(* Enumeration of bintrees by height in version 1. *)

(* Some proofs about "enumBT" function (see enumBT.ml) 
 *)

Set Implicit Arguments.
 
Require Import Arith.
Require Import Max.
Require Import MyTheoryList.
Import Wf_nat.
Require Import Omega.

Inductive bintree : Set := Leaf: bintree | Node: bintree -> bintree -> bintree.

Require Import SSLLoops.
Open Scope SSL_scope.

Fixpoint enumBT (St:Type) (n:nat) (f: bintree -> K St unit) {struct n} : (K St unit) :=
  match n with
  | O => f Leaf
  | (S p) => enumBT p (fun l => enumBT p (fun r => f (Node l r))   -; 
                                enumlt p (fun r => f (Node l r) -; 
                                                   f (Node r l)))
  end
with enumlt (St:Type) (n:nat) (f: bintree -> K St unit) {struct n} : (K St unit) :=
  match n with
  | O => skip 
  | (S p) => (enumBT p f) -; (enumlt p f)
  end.  


(** Proof methology about "enumBT" 
 * 
 * Assuming St: Type, and given two predicate
 * P Q: nat -> ((bintree -> K St unit) -> (K St unit)) -> Prop.
 * such that we want to prove "enumBT_P" and "enumlt_Q" below
 * 
 * We first prove "enumlt_Q_aux" below by structural induction over n.
 * Then, we prove "enumBT_P" by "nat_le_ind" induction below.
 * At last, "enum_lt_Q" is trivially derived from the two previous lemma.
 *
 * enumlt_Q_aux: forall (n:nat),
 *   (forall (m:nat), m < n -> P m (enumBT m)) -> Q n (enumlt n).
 * 
 * enumBT_P: forall (n:nat), P n (enumBT n).
 *
 * enumlt_Q: forall (n:nat), Q n (enumlt n).
 *
 *)

Lemma nat_le_ind: forall (P:nat -> Prop),
  (P O) 
   -> (forall (n:nat), (forall (m:nat), m <= n -> P m) -> P (S n))
     -> forall (n:nat), P n.
Proof.
  intros P H1 H2 n. pattern n. apply lt_wf_ind.
  clear n; intros n; case n; auto with arith.
Qed.

(** Monotonicity of enumBT **)
Lemma enumlt_monotonic_aux:
  forall (St: Type) (p1 p2: bintree -> K St unit) ,
    (forall (t:bintree) (st:St), [< st >] p1 t >==> p2 t)
     -> forall (n:nat), 
      (forall (m:nat) (st:St), m < n -> 
        [< st >] (enumBT m p1) >==> (enumBT m p2))  
             -> forall (st:St), ([< st >] (enumlt n p1) >==> (enumlt n p2)).
Proof.
  induction n; simpl; intros; auto with isl.
  SSL_monotony.
Qed.

Hint Resolve enumlt_monotonic_aux: bintree.

Lemma enumBT_monotonic:
  forall (St: Type) (n:nat) (p1 p2: bintree -> K St unit) ,
    (forall (t:bintree) (st:St), [< st >] (p1 t) >==> (p2 t))
     -> forall (st:St), [< st >] (enumBT n p1) >==> (enumBT n p2).
Proof.
  intros St n. elim n using nat_le_ind; simpl; intros; clear n; auto 10 with isl bintree arith.
Qed.

Hint Resolve enumBT_monotonic: bintree.

(** number of elements generated by enumBT **)

Require Import ZArith.
Open Scope Z_scope.

Fixpoint num2 (h: nat) {struct h}: Z*Z :=
  match h with
  | O => (1,0)   
  | (S h1) => 
         let (nh1,nlh1) := (num2 h1) in 
         let nlh := nh1 + nlh1 in
            (nh1*(nlh+nlh1),nlh)
  end.

Definition numBT (h:nat) : Z := fst (num2 h).
Definition numlt (h:nat) : Z := snd (num2 h).

Eval vm_compute in (numBT 4).
Eval vm_compute in (numBT 5).
Eval vm_compute in (numBT 8).

Definition incr (n:Z) : K Z unit :=
  bind (get Z) (fun x:Z => set (x+n)).

(** we first check the theorem  enum_incr with h=5 **) 
Goal [< 0 >] (enumBT 5 (fun t => incr 1)) >==> (incr (numBT 5)).
Proof.
  intros; apply refInEnv2wp; vm_compute.
  auto.
Qed.

Lemma incr_skip: forall (init:Z),
  ([< init >]  skip >==> (incr 0)).
Proof.
  SSL_simplify_all. 
  ring_simplify (init+0).
  auto.
Qed.

Lemma incr_seq: forall (n m init:Z), [< init >]  ((incr n) -; (incr m)) >==> (incr (n+m)).
Proof.
  intros; SSL_simplify_all. 
  intros; ring_simplify (init+(n+m)).
  auto.
Qed.

Hint Resolve incr_skip incr_seq: bintree.

Opaque Zmult.
Opaque Zplus.


Lemma enumlt_incr_gen_aux:
 forall (n:nat), 
      (forall (m:nat) (acc init:Z), (m < n)%nat -> 
         [< init >] (enumBT m (fun t => incr acc)) >==> (incr ((numBT m)*acc)))  
             -> forall (acc init:Z), [< init >] (enumlt n (fun t => incr acc)) >==> (incr ((numlt n)*acc)).
Proof.
  (* unfold numBT, numlt. *) induction n; simpl; intros.
  (* cas de base *)
  ring_simplify (0*acc).
  apply incr_skip.
  (* cas recursif *)
  unfold numlt; simpl; cutrewrite ((num2 n)=((numBT n), (numlt n))).
  simpl.
  rewrite  Zmult_plus_distr_l.
  SSL_trans.
  2: apply incr_seq.
  SSL_monotony.
  unfold numBT, numlt; case (num2 n); auto.
Qed.

Hint Resolve enumlt_incr_gen_aux: bintree.

Opaque Zmult.
Opaque Zplus.

Lemma  enum_incr_gen:
  forall (h:nat)  (acc init:Z), 
         [< init >] (enumBT h (fun t => incr acc))  >==> (incr ((numBT h)*acc)).
Proof.
  intros h. elim h using nat_le_ind; simpl.
  (* cas de base *)
  intros acc init.
  unfold numBT; simpl; ring_simplify (1*acc).
  simpl; auto with isl.
  (* cas recursif *)
  clear h; intros h IHh acc init; simpl.
  unfold numBT; simpl. cutrewrite ((num2 h)=((numBT h), (numlt h))).
  simpl.
  cutrewrite ((numBT h * (numBT h + numlt h + numlt h) * acc)=(numBT h * ((numBT h)*acc + numlt h * (acc + acc)))).
  SSL_trans.
  2: apply IHh; auto with arith.
  apply enumBT_monotonic.
  intros; SSL_trans.
  2: apply incr_seq.
  SSL_monotony.
  intros; SSL_trans.
  2: apply enumlt_incr_gen_aux; auto with bintree isl arith.
  auto with bintree.
  ring.
  unfold numBT, numlt; case (num2 h); auto.
Qed.

Theorem  enum_incr:
  forall (h:nat)  (init:Z), 
         [< init >] (enumBT h (fun t => incr 1)) >==> (incr (numBT h)).
Proof.
   intros h init; cutrewrite ((numBT h)=(numBT h)*1).  
   apply enum_incr_gen.
   ring.
Qed.

(** Some theory of binary trees **)
Close Scope Z_scope.


Fixpoint height (t:bintree) { struct t } : nat :=
  match t with
  | Leaf => 0
  | (Node l r) => S (max (height l) (height r))   
  end.

Inductive is_height: bintree -> nat -> Prop :=
 | Leaf_heightO: (is_height Leaf O)
 | Node_height_both: 
     forall (l r: bintree) (n:nat),
       (is_height l n) -> (is_height r n) -> (is_height (Node l r) (S n))
 | Node_height_left: 
     forall (l r: bintree) (n m:nat),
       (is_height l n) -> (is_height r m) -> m < n -> (is_height (Node l r) (S n))
 | Node_height_right: 
     forall (l r: bintree) (n m:nat),
       (is_height l n) -> (is_height r m) -> n < m -> (is_height (Node l r) (S m)).

Hint Resolve Leaf_heightO Node_height_both Node_height_left Node_height_right: bintree.

Lemma is_height_from_height:
  forall (t:bintree), (is_height t (height t)).
Proof.
  induction t; simpl; auto with bintree.
  case (lt_eq_lt_dec (height t1) (height t2)).
  intros H1; case H1.
  intros H2; rewrite max_r; auto with arith.
  apply Node_height_right with (n:=(height t1)); auto.
  intros H2; rewrite H2; rewrite max_r; auto with arith.
  apply Node_height_both; auto.
  rewrite <- H2; auto.
  intros H2; rewrite max_l; auto with arith.
  apply Node_height_left with (m:=(height t2)); auto.
Qed.

Lemma is_height_to_height:
  forall (t:bintree) (n:nat), (is_height t n) -> (height t)=n.
Proof.
  induction 1; simpl; subst; auto.
  rewrite IHis_height2.
  rewrite max_l; auto with arith.
  rewrite max_l; auto with arith.
  rewrite max_r; auto with arith.
Qed.

Lemma height_discr: 
  forall (t1 t2:bintree) (n m:nat), 
    (is_height t1 n) -> (is_height t2 m) -> n <> m -> t1 <> t2.
Proof.
  intros t1 t2 n m H1 H2 H3 H4.
  subst.
  rewrite <- (is_height_to_height H2) in H3.
  rewrite <- (is_height_to_height H1) in H3.
  tauto.
Qed. 
 
Hint Resolve is_height_to_height is_height_from_height: bintree.


Fixpoint eqbintree (t1 t2:bintree) { struct t1 }: bool :=
  match t1,t2 with
  | Leaf,Leaf => true
  | (Node l1 r1),(Node l2 r2) => Bool.andb (eqbintree l1 l2) (eqbintree r1 r2)
  | _,_ => false
  end. 

Lemma eqbintree_case: forall (t1 t2:bintree), (eqbintree t1 t2)=true \/ (eqbintree t1 t2)=false.
Proof.
  intros t1 t2; case (eqbintree t1 t2); auto.
Qed.

Lemma eq_eqbintree: forall (t1 t2:bintree), t1=t2 -> (eqbintree t1 t2)=true.
Proof.
  induction t1; simpl. 
  intros t2 H; inversion H; auto.
  intros t2 H; inversion H; auto. 
  rewrite IHt1_1; auto.
  rewrite IHt1_2; auto.
Qed.

Lemma eqbintree_eq: forall (t1 t2:bintree), (eqbintree t1 t2)=true -> t1=t2.
Proof.
  induction t1; simpl; intros t2; case t2; simpl; auto;
  try (intros; discriminate).
  intros b b0.
  case (eqbintree_case t1_1 b); intros H1; rewrite H1; simpl.
  rewrite (IHt1_1 b); auto.
  intros; rewrite (IHt1_2 b0); auto.
  intros; discriminate.
Qed.

Lemma noteq_eqbintree: forall (t1 t2:bintree), t1<>t2 -> (eqbintree t1 t2)=false.
Proof.
  intros t1 t2; case (eqbintree_case t1 t2); auto.
  intros H1; rewrite (eqbintree_eq _ _ H1).
  intros H2; case H2; auto.
Qed.

Lemma eqbintree_noteq: forall (t1 t2:bintree), (eqbintree t1 t2)=false -> t1 <> t2.
Proof.
  intros t1 t2 H1 H2.
  rewrite (eq_eqbintree H2) in H1.
  discriminate.
Qed.

(* preliminairies of enumBT_spec proof *)

Lemma enumlist_bintree:
  forall (n:nat) (l1 l2:list bintree),
    (enumlist (fun t => (height t) = n) l1)
     -> (enumlist (fun t => (height t) < n) l2)
       ->  (enumlist (fun t => (height t) = (S n)) 
              (fold_right (fun x1 m1 => (fold_right (fun x2 m2 => (Node x2 x1)::(Node x1 x2)::m2) nil l2
                              ++  fold_right (fun x2 m2 => (Node x1 x2)::m2) nil l1)
                              ++ m1) nil l1)).
Proof.
  intros n l1 l2 H1 H2.
  (** petit lemme 1 **)
  assert (X1: 
   (forall l : bintree,
       (enumlist (fun t => (exists r, (height r) = n /\ t=(Node l r)))
         (fold_right (fun r m2 => Node l r :: m2)
           nil l1)))).
  intros l; eapply fold_right_cons_enumlist with (f:=(Node l)).
  apply H1.
  simpl; auto.
  simpl; intros t1 t2 A B Z; inversion Z; auto.
  simpl; eauto.
  (** petit lemme 2 **)
  assert (X2: 
   (forall l : bintree, (height l)=n ->
       (enumlist (fun t => (exists r, (height r) < n /\ (t=(Node l r) \/ t=(Node r l))))
         (fold_right (fun r m2 => Node r l :: Node l r :: m2)
           nil l2)))).
  intros l0 XX; eapply fold_right_app_enumlist with (f:=fun r => Node r l0 :: Node l0 r :: nil).
  apply H2.
  firstorder eauto.
  simpl. intros t1 t2 A B Z. decompose [ex Logic.and] Z; clear Z.
  intuition subst.
  inversion H0; auto.
  inversion H3; auto.
  inversion H3; auto.
  inversion H; auto.
  simpl; intros; constructor 1.
  eauto 10 with lists.
  apply doublefree_hd.
  apply allS_not_in.
  apply allS_cons; auto with lists.
  intros Z; inversion Z. 
  subst.
  omega.
  eauto with lists.
  (** petit lemme 3 **)
  assert (X3: forall l : bintree, (height l)=n ->
    enumlist (fun t =>  
              (exists r, (height r) < n /\ (t=(Node l r) \/ t=(Node r l)))
           \/ (exists r, (height r) = n /\ t=(Node l r)))
         ((fold_right (fun r m2 => Node r l :: Node l r :: m2) nil l2)
         ++ (fold_right (fun r m2 => Node l r :: m2) nil l1))).
  intros l H; eapply enumlist_app.
  apply X2; auto.
  simpl; intros; firstorder (subst; auto).
  simpl; intros. firstorder subst.
  constructor 2.
  intros X; firstorder.
  inversion H3; subst; omega.
  inversion H3; subst; omega.
  eapply enumlist_compatible.
  2: apply X1.
  simpl; firstorder (subst; auto).
  intros X; firstorder.
  inversion H3; subst; omega.
  inversion H3; subst; omega.
  (** preuve du but initial **)
  clear X1 X2 H2.
  eapply fold_right_app_enumlist
    with (f:= fun (x1 : bintree) =>
      fold_right
         (fun (x2 : bintree) (m2 : list bintree) =>
          Node x2 x1 :: Node x1 x2 :: m2) nil l2 ++
       fold_right
         (fun (x2 : bintree) (m2 : list bintree) => Node x1 x2 :: m2) nil l1).
  apply H1.
  (* cas 1 *)
  clear H1; intros t H. assert (X:is_height t (S n)).
  rewrite <- H; auto with bintree.   
  inversion X.
  (* cas 1.1 *)   
  constructor 1 with (x:=l).
  intuition auto with bintree.
  eapply enumlist_cond_in. 
  apply X3; auto with bintree.
  simpl; firstorder eauto with bintree.  
  (* cas 1.2 *)
  constructor 1 with (x:=l).
  intuition auto with bintree.
  eapply enumlist_cond_in. 
  apply X3; auto with bintree.
  simpl. 
  rewrite <- (is_height_to_height H2) in H4.
  firstorder eauto with bintree. 
  (* cas 1.3 *)
  constructor 1 with (x:=r).
  intuition auto with bintree.
  eapply enumlist_cond_in. 
  apply X3; auto with bintree.
  simpl. 
  rewrite <- (is_height_to_height H1) in H4.
  firstorder eauto with bintree. 
  (*** cas 2 **) 
  clear H1; simpl; intros a b H3 H4 X; case X; clear X.
  intuition auto.
  case (enumlist_in_cond (X3 a H3) x); firstorder subst.
  (* 2.1 *)
  case (enumlist_in_cond (X3 b H4) (Node a x0)); firstorder subst.
  inversion H3; auto.
  inversion H3; subst. absurd (height x < height x); [ omega | auto ].
  inversion H3; auto.
  (* 2.2 *)  
  case (enumlist_in_cond (X3 b H4) (Node x0 a)); firstorder subst.
  inversion H3; subst. absurd (height x < height x); [ omega | auto ].
  inversion H3; auto.
  inversion H3; subst. rewrite H4 in H; absurd (height x < height x); [ omega | auto ].
  (* 2.3 *)
  case (enumlist_in_cond (X3 b H4) (Node a x0)); firstorder subst.
  inversion H3; auto.
  inversion H3; subst. absurd (height x < height x); [ omega | auto ].
  inversion H3; auto.
  (*** cas 3 ***) 
  clear H1; simpl; intros t H; constructor 1.
  (* cas 3.1 *)
  eapply allS_weaken.
  eapply enumlist_AllScond; eauto.
  simpl; firstorder subst; eauto with bintree.
  eapply is_height_to_height; eapply Node_height_both; auto with bintree.
  rewrite <- H0; auto with bintree.
  (* cas 3.2 *)
  eapply enumlist_doublefree; eauto.
Qed.

  
Fixpoint genBT (n:nat) {struct n} : (list bintree) :=
  match n with
  | 0 => Leaf::nil
  | (S n1) => 
      let l1 := (genBT n1) in
      let l2 := (genlt n1) in
        fold_right (fun x1 m1 => (fold_right (fun x2 m2 => (Node x2 x1)::(Node x1 x2)::m2) nil l2
                              ++  fold_right (fun x2 m2 => (Node x1 x2)::m2) nil l1)
                              ++ m1) nil l1
  end
with genlt (n:nat) {struct n} : (list bintree) :=
  match n with
  | 0 => nil 
  | (S n1) => (genlt n1) ++  (genBT n1)
  end.  

Lemma genlt_enumlist:
  forall (n:nat), 
    (forall (m:nat), m < n -> (enumlist (fun t => (height t) = m) (genBT m))) 
     -> (enumlist (fun t => (height t) < n) (genlt n)).
Proof.
  induction n; simpl.
    intros; apply enumlist_nil.
    intros x H1; inversion H1.
  (* cas rec *)
  intros H.
  eapply enumlist_app.
  eapply IHn.
  intros; eapply enumlist_compatible.
  2: apply H; auto with arith.    
  simpl; intuition auto.
  simpl; intros; omega.
  simpl; intros t H0; inversion H0; intuition auto.
  constructor 2. omega.
  eapply enumlist_compatible.
  2: apply H; auto with arith.
  simpl; intuition omega.
Qed.

Theorem genBT_enumlist:
  forall (n:nat), (enumlist (fun t => (height t)=n) (genBT n)).
Proof.
  intros n; elim n using nat_le_ind; simpl; clear n.
  (* cas de base *)
  apply enumlist_cons.
  intros t H; assert (X:is_height t 0).
  rewrite <- H; auto with bintree.
  inversion X; auto.
  auto with bintree.
  apply enumlist_nil.
  intros t H. assert (X:is_height t 0).
  intuition auto.
  rewrite <- H1; auto with bintree.
  inversion X; auto.
  intuition auto.
  (* cas rcursif *)  
  intros.
  apply enumlist_bintree.
  auto with arith.
  apply genlt_enumlist; auto with arith.
Qed.

(* end of preliminaries **)

Lemma enumlt_spec:
  forall (St:Type) (n:nat) (f: bintree -> K St unit), 
     (forall (m:nat) (st:St), m < n -> 
       [< st >](enumBT m f) >==> listRevIter (genBT m) f)%SSL
          ->  (forall (st:St), [< st >]
               (enumlt n f) >==> 
                  (listRevIter (genlt n) f))%SSL.     
Proof.
  induction n; simpl.
  auto with isl.
  intros f H st.
  SSL_trans.
  2: apply listRevIter_app_ref2.
  SSL_monotony.
Qed.

Hint Resolve enumlt_spec: bintree.

(* This proof does not use "refinement ideas", and could be translated in monads *)
Theorem enumBT_spec:
  forall (St:Type) (h:nat) (f: bintree -> K St unit) (st:St), 
          (refInEnv st (enumBT h f) (listRevIter (genBT h) f)).
Proof.
  intros St n. elim n using nat_le_ind; simpl; clear n.
  (* cas de base *)
  intros f st; simpl.
  SSL_simplify_right; auto with isl.
  (* cas recursif *)
  intros n IHn f st.
  SSL_trans.
  apply IHn. 
  auto with arith.
  SSL_trans.
  2: apply listRevIter_composeIn_app_ref1 
   with (f:=fun x1 => fold_right
         (fun (x2 : bintree) (m2 : list bintree) =>
          Node x2 x1 :: Node x1 x2 :: m2) nil (genlt n) ++
       fold_right
         (fun (x2 : bintree) (m2 : list bintree) => Node x1 x2 :: m2) nil
         (genBT n)).
  apply listRevIter_monotony.
  intros.
  SSL_trans.
  2: apply listRevIter_app_ref2.
  SSL_monotony.
  SSL_trans.
  2: apply listRevIter_composeIn_cons.
  auto with arith.
  intros; SSL_trans.
  apply enumlt_spec; auto with arith.
  SSL_trans.
  2: apply listRevIter_composeIn_app_ref1 with (f:=fun x2 => Node x2 x::Node x x2:: nil).
  simpl.
  apply listRevIter_monotony; intros.
  SSL_unassoc_right.
  SSL_simplify_right.
  auto with isl.
Qed.
