(* Primitive weakest preconditions
 * (i.e. those for which "Build_DS" is built at hand).
 *  S. Boulme ---  coq 8.0
 *)
(* begin hide *)
Set Implicit Arguments.

Require Export WP.
Require Export NonEmpty.
(* end hide *)

(** ** Module [FSL]: primitive operators of functional specifications

This module defines the primitive operators of functional
specifications.  These primitive operators are defined directly using
a pair of weakest-preconditions. In the following, the proofs that
these pairs satisfy the [is_wp_pair] condition are named with a suffix
``[_correct]''.  Others operators of FSL (like recursion operators)
are built using these primitive operators.

The primitive operators are split in two categories: those which are
primitive in the sense of the algebraic semantics: [val], [bind],
[any], [require] and [ensure].  And those which are defined as
primitive to increase the efficiency of wp-calculus: [abort], [magic]
and [or].  
*)

(** *** primitive operators of the algebraic semantics. *)

(** I first introduce [val] and [bind] operators. *)

(* begin hide *)
Lemma val_correct :
 forall (A : Type) (a : A), is_wp_pair (fun R => R a) (fun R => R a).
Proof.
  unfold is_wp_pair; firstorder eauto; subst; auto.
Qed.
(* end hide *)

Definition val (A : Type) (a:A) : DS A
 := Build_DS (sdemon:=fun R => R a) (angel:=fun R => R a) (val_correct a).
(*
Proof.
  refine (fun A a => Build_DS (sdemon:=fun R => R a) (angel:=fun R => R a) _). 
  apply val_correct; auto.
Defined.
*)

(* begin hide *)
Lemma val_impl : forall (A : Type) (a : A), impl (val a).
Proof.
  unfold impl; firstorder eauto.
Qed.
(* end hide *)

(* BIND *)

(* begin hide *)
Lemma bind_correct :
 forall (A B : Type) (s1 : DS A) (s2 : A -> DS B),
 is_wp_pair (fun R => sdemon s1 (fun a => sdemon (s2 a) R))
            (fun R => angel s1 (fun a => angel (s2 a) R)).
Proof.
  intros A B s1 s2; constructor 1.
  (* sdemon <-> *)
  constructor 1.
   (* sdemon -> 1 *)
    intros H; constructor 1.
    (* sdemon -> _ /\ *)
    eapply sdemon_weaken; eauto.
    simpl; eauto with wipi.
    (* sdemon -> /\ _ *)
    intros a H1; case (angel_unfold _ _ H1).
    intros a0 H2; lapply (sdemon_unfold _ _ a0 H); firstorder auto.
    apply (sdemon_unfold (s2 a0)); auto.
    (* -> sdemon *)
    intros. apply sdemon_fold; firstorder (eauto with wipi).
    apply sdemon_fold. 
    apply (sdemon_unfold s1 (fun a : A => pre (s2 a))); auto.
    intuition eauto with wipi.
   (* angel *)
   constructor 1.
   (* angel -> *)
   intros H; case (angel_unfold _ _ H); 
   intros a H0; case (angel_unfold (s2 a) R).
   firstorder auto.
   intros b H1; constructor 1 with (x:=b); firstorder auto .
   eauto with wipi.
   (* -> angel *)
   intros H; case H; clear H; intros b H.
   case (angel_unfold s1 (fun a : A => post (s2 a) b));
   intuition eauto with wipi.
Qed.
(* end hide *)

Definition bind (A B : Type) (s1:DS A) (s2:(A -> DS B)) : DS B
 := Build_DS 
       (sdemon:=fun R => sdemon s1 (fun a => sdemon (s2 a) R))
       (angel:=fun R => angel s1 (fun a => angel (s2 a) R))
       (bind_correct s1 s2).
(*
  refine
   (fun A B s1 s2 =>
    Build_DS (sdemon:=fun R => sdemon s1 (fun a => sdemon (s2 a) R))
      (angel:=fun R => angel s1 (fun a => angel (s2 a) R)) _).
  apply bind_correct; auto.
Defined.
*)

(** With the two following lemmas, I check that [bind] satisfies its expected [pre]/[post] semantics.
*)

Lemma bind_pre : forall (A B : Type) (s1 : DS A) (s2 : A -> DS B),
 pre (bind s1 s2) <-> 
   pre s1 /\ (forall a : A, post s1 a -> pre (s2 a)).
Proof.
  simpl in |- *; intros A B s1 s2; firstorder eauto with wipi.
  apply (sdemon_unfold s1 (fun a : A => pre (s2 a))); auto.
Qed.

Lemma bind_post : forall (A B : Type) (s1 : DS A) (s2 : A -> DS B) (b : B),
 (post (bind s1 s2) b) <-> 
   (exists a : A, (post s1 a) /\ (post (s2 a) b)).
Proof.
  simpl in |- *; intros A B s1 s2 b;
   case (DScorrect s1 (fun a => post (s2 a) b));
   intuition eauto.
Qed.

(* begin hide *)
Lemma bind_impl :
 forall (A B : Type) (s1 : DS A) (s2 : A -> DS B),
 impl s1 -> (forall a : A, (pre s1) -> impl (s2 a)) -> impl (bind s1 s2).
Proof.
  unfold impl in |- *; simpl in |- *; intros A B s1 s2 se1 se2 H;
   case (bind_pre s1 s2); simpl in |- *; intros H1 H2; 
   case se1.
  tauto.
  intros a H3; case (se2 a).
    firstorder eauto.
    firstorder eauto.
    intros b H4; constructor 1 with (extract := b).
    intros R; case (H3 (fun a => angel (s2 a) R)); case (H4 R); tauto.
Qed.

Lemma bind_strdet :
 forall (A B : Type) (s1 : DS A) (s2 : A -> DS B),
 strdet s1 -> (forall a : A, strdet (s2 a)) -> strdet (bind s1 s2).
Proof.
  unfold strdet in |- *; simpl in |- *; intros A B s1 s2 se1 se2 H;
   case (bind_pre s1 s2); simpl in |- *; intros H1 H2; 
   case se1.
  tauto.
  intros a H3; case (se2 a).
    assert (H4 : post s1 a).
     rewrite H3; auto. 
     intuition auto.
    intros b H4; constructor 1 with (DSwit := b).
    intros R; rewrite H3; auto.
Qed.
(* end hide *)

(** Refinement is monotonic with respect to [bind]
*)
Lemma bind_monotony : forall (A B : Type) (s1 s3 : DS A) (s2 s4 : A -> DS B),
 refines s1 s3 -> 
  (forall a : A, refines (s2 a) (s4 a)) -> 
   refines (bind s1 s2) (bind s3 s4).  
Proof.
  unfold refines in |- *; simpl in |- *;
   intros A B s1 s3 s2 s4 H1 H2 R.
  cut
   (forall R : B -> Prop,
    sdemon s3 (fun a => sdemon (s4 a) R) -> 
    sdemon s1 (fun a => sdemon (s2 a) R)).
  intros H; constructor 1; auto.
    intros X1 X2; case (H1 (fun a => angel (s4 a) R)); intros H3 H4;
     apply H4.
    eapply sdemon_weaken; eauto.
    eapply angel_weaken_pre; eauto.
    apply sdemon_weaken with (R1 := fun a => pre (s2 a)); auto.
    intros a H5 H6; case (H2 a R); intros H7 H8; apply H8; auto.
    case (DScorrect s3 (fun a => pre (s4 a))); intuition eauto. 
    apply H13; case (H1 (eq a)); intuition eauto.
  clear R; intros R H0;
   apply sdemon_weaken with (R1 := fun a => sdemon (s4 a) R).
    case (H1 (fun a => sdemon (s4 a) R)); auto.
    intros a X1; case (H2 a R); auto.
Qed.

Lemma bind_monotony_left: forall (A B:Type) (s1 s1':DS A) (s2: A -> DS B),
  (refines s1 s1') -> (refines (bind s1 s2) (bind s1' s2)).
Proof.
   intros; apply bind_monotony; auto with wprefines.
Qed.


(** This lemma corresponds to a rule for small-steps simplifications in refinement proofs,
see section %\ref{sec:smallsteps}%. A goal [refines (bind s1 s2) (bind s3 s4)] such 
[s1] refines [s3], can be simplified in [(pre s3 -> sdemon s1 (fun a : A => refines (s2 a) (s4 a)))].
If [s1] is sufficiently simple, this gives a simple formula.  
*)
Lemma bind_simpl_both_l : forall (A B : Type) (s1 s3 : DS A) (s2 s4 : A -> DS B),
 refines s1 s3 -> 
  (pre s3 -> sdemon s1 (fun a : A => refines (s2 a) (s4 a))) -> 
    refines (bind s1 s2) (bind s3 s4). 
Proof.
  simpl in |- *; intros; apply refines_fold.
  generalize (refines_unfold H); simpl in |- *; intros.
  assert (X : pre s3). eauto with wipi.
  apply sdemon_fold. eauto with wipi.
  intros a X0.
  case (H (eq a)); intuition auto.
  lapply (sdemon_unfold _ _ a H2); auto.
  intros X1; lapply (sdemon_unfold _ _ a H5); auto.
  intros X2; apply sdemon_fold.
  case (X2 (fun _ => True)); intuition auto.
  intros b X3; apply angel_fold with (a := a); auto.
  case (X2 (eq b)); intuition auto.
Qed.

(** The three following lemmas also correspond to various small-steps simplifications rules.
*)
Lemma bind_simpl_both : forall (A B : Type) (s1 s3 : DS A) (s2 s4 : A -> DS B),
 refines s1 s3 -> 
  (pre s1 -> sdemon s1 (fun a : A => refines (s2 a) (s4 a))) -> 
    refines (bind s1 s2) (bind s3 s4). 
Proof.
 intros; apply bind_simpl_both_l; auto.
 unfold refines in H; firstorder auto.
Qed.


Lemma bind_simpl_left : forall (A B : Type) (s1 : DS A) (s2 : A -> DS B) (s3 : DS B),
 sdemon s1 (fun a : A => refines (s2 a) s3) -> 
  refines (bind s1 s2) s3. 
Proof.
  simpl in |- *; intros; apply refines_fold.
  simpl in |- *; intros; eapply sdemon_weaken. eauto. 
  simpl in |- *; auto with wprefines.
Qed.

Lemma bind_simpl_right : forall (A B : Type) (s1 : DS A) (s2 : A -> DS B) (s3 : DS B),
 (pre s1 -> angel s1 (fun a => refines s3 (s2 a))) -> 
  refines s3 (bind s1 s2). 
Proof.
  simpl in |- *; intros A B s1 s2 s3 H; apply refines_fold. 
  simpl in |- *; intros H0. eauto with wipi.
  lapply H.
  intros H1; case (angel_unfold _ _ H1).
  intros a H3; case H3; intros H5 H6; simpl in |- *;
   lapply (refines_unfold H6).
  intros H4; eapply sdemon_weaken. eauto.
  simpl in |- *; intros b H7; apply angel_fold with (a := a); auto.
  apply (sdemon_unfold _ _ a H0); auto. 
  eapply sdemon_weaken; eauto.
Qed.

(* begin hide *)
Hint Resolve bind_monotony: wprefines.
(* end hide *)

(** I now prove the three monadic properties relating [bind] and [val]. Currently, I have not yet introduced 
the equivalence relation associated to [bind]. Thus, the three properties correspond to the six lemmas below.
*)
Lemma bind_val_l_ref1 : forall (A B : Type) (a : A) (s1 : A -> DS B),
 refines (bind (val a) s1) (s1 a).
Proof.
  eauto with wprefines wipi.
Qed.

Lemma bind_val_l_ref2 : forall (A B : Type) (a : A) (s1 : A -> DS B),
 refines (s1 a) (bind (val a) s1).
Proof.
  eauto with wprefines wipi. 
Qed.

Lemma bind_val_r_ref1 : forall (A : Type) (s1 : DS A),
 refines (bind s1 (fun a => val a)) s1.
Proof.
  eauto with wprefines wipi.   
Qed.

Lemma bind_val_r_ref2 : forall (A : Type) (s1 : DS A),
 refines s1 (bind s1 (fun a => val a)).
Proof.
  intros; apply refines_fold; simpl; eauto with wprefines wipi.   
Qed.

Lemma bind_assoc_ref1 : forall (A B C : Type) (s1 : DS A) (s2 : A -> DS B) (s3 : B -> DS C),
 refines (bind (bind s1 s2) s3)
   (bind s1 (fun a => bind (s2 a) s3)).
Proof.
  eauto with wprefines wipi.   
Qed.

Lemma bind_assoc_ref2 : forall (A B C : Type) (s1 : DS A) (s2 : A -> DS B) (s3 : B -> DS C),
 refines (bind s1 (fun a => bind (s2 a) s3))
   (bind (bind s1 s2) s3).
Proof.
  eauto with wprefines wipi.   
Qed.

(** bind & pre **)

Lemma bind_pre_intro :
 forall (A B: Type) (s1: DS A) (s2: A -> DS B) (s3: DS B),
 (pre s1 -> refines s3 (bind s1 s2)) -> refines s3 (bind s1 s2).
Proof.
  intros; apply refines_pre_intro; simpl.
  intros H1; apply H.
  eauto with wipi.
Qed.


(* REQUIRE *)

(* begin hide *)
Lemma require_correct :
 forall p : Prop, is_wp_pair (fun R => p /\ R tt) (fun R => R tt).
Proof.
  unfold is_wp_pair in |- *; firstorder eauto; subst; auto.
Qed.
(* end hide *)

(** Here is the definition of the [require] operator.
*)
Definition require (p:Prop) : DS unit
 := Build_DS (sdemon:=fun R => p /\ R tt) (angel:=fun R => R tt) (require_correct p).
(*
  refine
   (fun p =>
    Build_DS (sdemon:=fun R => p /\ R tt) (angel:=fun R => R tt) _).
  apply require_correct; auto.
Defined.
*)

(** With the two following lemmas, I check that [require] satisfies its expected [pre]/[post] semantics.
*)
Lemma require_pre : forall p : Prop,
  pre (require p) <-> p.
Proof.
  simpl in |- *; tauto.
Qed.

Lemma require_post : forall (p : Prop) (a : unit),
  (post (require p) a) <-> True.
Proof.
  simpl in |- *; intros H a; case a; tauto.
Qed.

(* begin hide *)
Lemma require_impl : forall p : Prop, impl (require p).
Proof.
  unfold impl in |- *; constructor 1 with (extract := tt); simpl in |- *;
   tauto.
Qed.

Lemma require_strdet : forall p : Prop, strdet (require p).
Proof.
  unfold strdet in |- *; constructor 1 with (DSwit := tt); simpl in |- *;
   tauto.
Qed.
(* end hide *)

(** The two following lemmas state respectively that [skip] refines any [require]-expression.
 and that [require p] refines [skip] if [p] holds.
*)
Lemma require_refined_tt : forall p : Prop,
 refines (val tt) (require p).
Proof.
  unfold refines in |- *; simpl in |- *; intuition auto.
Qed.  

Lemma require_refines_tt : forall p : Prop, 
  p -> refines (require p) (val tt).
Proof.
  unfold refines in |- *; simpl in |- *; intuition auto.
Qed.  

(* ASSERT *)

(* begin hide *)
Lemma ensure_correct : forall p : Prop, 
  is_wp_pair (fun R => p -> R tt) (fun R => p /\ R tt).
Proof.
  unfold is_wp_pair in |- *; firstorder eauto; subst; auto.
Qed.
(* end hide *)

(** Here is the definition of [ensure]:
*) 
Definition ensure (p:Prop) : DS unit
 := Build_DS (sdemon:=fun R => p -> R tt) (angel:=fun R => p /\ R tt) (ensure_correct p).

(*
  refine
   (fun p =>
    Build_DS (sdemon:=fun R => p -> R tt) (angel:=fun R => p /\ R tt) _). 
  apply ensure_correct; auto.
Defined.  
*)

(** With the two following lemmas, I check that [ensure] satisfies its expected [pre]/[post] semantics.
*)
Lemma ensure_pre : forall p : Prop, 
  pre (ensure p) <-> True.
Proof.
  simpl in |- *; tauto.
Qed.

Lemma ensure_post : forall (p : Prop) (a : unit), 
  post (ensure p) a  <-> p.
Proof.
  simpl in |- *; intros p a; case a; simpl in |- *; tauto.
Qed.

(** The two following lemmas state respectively that any [ensure]-expression refines [skip].
 and that [skip] refines [ensure p] if [p] holds.
*)
Lemma ensure_refines_tt : forall p : Prop, 
  refines (ensure p) (val tt).
Proof.
  unfold refines in |- *; simpl in |- *; intuition auto.
Qed.  

Lemma ensure_refined_tt : forall p : Prop, 
  p -> refines (val tt) (ensure p).
Proof.
  unfold refines in |- *; simpl in |- *; intuition auto.
Qed.  

(** The two following lemmas recall that refinement corresponds to a weakening on preconditions and
 a strengthening on postconditions.
*)
Lemma require_refines_require : forall p1 p2 : Prop, 
  (p2 -> p1) -> refines (require p1) (require p2).
Proof.
  intros; apply refines_fold; simpl in |- *; tauto.
Qed.

Lemma ensure_refines_ensure : forall p1 p2 : Prop, 
  (p1 -> p2) -> refines (ensure p1) (ensure p2).
Proof.
  intros; apply refines_fold; simpl in |- *; tauto.
Qed.


(** The four following lemmas are particular cases of [bind_simpl_right] and [bind_simpl_left].
*)
Lemma require_refined : forall (A : Type) (p : Prop) (s1 s2 : DS A),
 (p -> refines s1 s2) -> 
 refines s1 (bind (require p) (fun _ => s2)).
Proof.
  intros; apply bind_simpl_right; simpl; intuition auto with wprefines.
Qed.

Lemma ensure_refines : forall (A : Type) (p : Prop) (s1 s2 : DS A),
 (p -> refines s1 s2) -> 
 refines (bind (ensure p) (fun _ => s1)) s2.
Proof.
  intros; apply bind_simpl_left; simpl; intuition auto with wprefines.
Qed.

Lemma require_refines : forall (A : Type) (p : Prop) (s1 s2 : DS A),
 p -> refines s1 s2 -> 
 refines (bind (require p) (fun _ => s1)) s2.
Proof.
  intros; apply bind_simpl_left; simpl; intuition auto with wprefines.
Qed.

Lemma ensure_refined: forall (A : Type) (p : Prop) (s1 s2 : DS A),
 p -> refines s1 s2 -> 
 refines s1 (bind (ensure p) (fun _ => s2)).
Proof.
  intros; apply bind_simpl_right; simpl; intuition auto with wprefines.
Qed.

(* begin hide *)
Hint Resolve require_refines_require ensure_refines_ensure require_refines require_refined 
 ensure_refined ensure_refines bind_monotony: wprefines.
(* end hide *)

(* ANY *)

(* begin hide *)
Lemma any_correct :
 forall A : Type,
 is_wp_pair (fun R => forall a : A, R a) (fun R => exists a : A, R a).
Proof.
  unfold is_wp_pair in |- *; firstorder eauto.
Qed.
(* end hide *)

(** I now define [any].
*)  
Definition any (A : Type) : DS A
  := Build_DS 
      (sdemon:=fun R => forall a : A, R a)
      (angel:=fun R => exists a : A, R a)
      (any_correct (A:=A)).
(* 
  refine
   (fun A : Type =>
    Build_DS (sdemon:=fun R => forall a : A, R a)
      (angel:=fun R => exists a : A, R a) _).
  apply any_correct; auto.
Defined.
*)

(** With the two following lemmas, I check that [any] satisfies its expected [pre]/[post] semantics.
*)
Lemma any_pre : forall A : Type, 
  pre (any A) <-> True.
Proof.
  simpl in |- *; tauto.
Qed.

Lemma any_post : forall (A : Type) (a : A),
 (post (any A) a) <-> True.
Proof.
  simpl in |- *; intuition eauto.
Qed.

(** It is easy to prove the three following lemmas. In particular, the
second one is a particular case of [bind_simpl_right], whereas the
third one is a particular case of [bind_simpl_left].  *)
Lemma any_refined : forall (A : Type) (a : A), 
  refines (val a) (any A).
Proof.
  unfold refines in |- *; simpl in |- *; intuition eauto.
Qed.

Lemma any_refined_bind : forall (A B : Type) (s1 : A -> DS B) (s2 : DS B) (a : A),
 refines s2 (s1 a) -> refines s2 (bind (any A) s1).
Proof.
  intros; apply bind_simpl_right; simpl; eauto.
Qed.

Lemma any_refines : forall (A B : Type) (s1 : A -> DS B) (s2 : DS B),
 (forall a : A, refines (s1 a) s2) -> 
   refines (bind (any A) s1) s2. 
Proof.
 intros; apply bind_simpl_left; simpl; auto.
Qed.

(* begin hide *)
Hint Resolve any_refined any_refines any_refined_bind: wprefines.
(* end hide *)

(** With the properties proved above, it is easy to check that the
properties of the algebraic semantics are satisfied. Furthermore, we
have also proved that these definitions are equivalent to those of the
pre/post semantics.

*** Important properties of [any] operator.

I detail now some properties of [any].
*)

Lemma any_val_refined_unfold : forall (A B : Type) (f : A -> DS B) (s : DS A),
 refines s
   (bind (any A) (fun a => bind (f a) (fun _ => val a))) -> 
 (forall a : A, pre (f a)) -> sdemon s (fun a => angel (f a) (fun _ => True)).
Proof. 
  intros A B f s H1 H2; generalize (refines_unfold H1); simpl in |- *.
  intros H3; eapply sdemon_weaken; eauto. 
  simpl in |- *; intros a H4; case H4; intros a0 H5; case (angel_unfold _ _ H5).
  firstorder auto; subst; eauto with wipi.
Qed.

Lemma any_val_refined_fold : forall (A B : Type) (f : A -> DS B) (s : DS A),
 ((forall a : A, pre (f a)) -> sdemon s (fun a => angel (f a) (fun _ => True))) -> 
 refines s
   (bind (any A) (fun a => bind (f a) (fun _ => val a))).
Proof. 
  intros A B f s H; apply refines_fold; simpl in |- *.
  intros H1; intuition auto; apply sdemon_fold.
  eapply sdemon_weaken; eauto.
  intros a H3; constructor 1 with (x := a); intuition auto.
  lapply (sdemon_unfold _ _ a H0); eauto with wipi.
Qed.

Lemma any_on_non_empty_type_can_be_lifted : forall (A B C : Type) (s : DS A) (f : A -> B -> DS C),
 (NonEmpty B) -> 
 refines (bind s (fun a => bind (any B) (fun b => f a b)))
   (bind (any B) (fun b => bind s (fun a => f a b))).
Proof.
  intros A B C s f H; apply refines_fold; simpl in |- *.
  intros H'.
  apply
   sdemon_forall_lift2
    with
      (R := fun (a0 : B) (a : A) =>
            sdemon (f a a0)
              (fun a1 : C =>
               exists a2 : _,
                 angel s
                   (fun a3 : A => post (f a3 a2) a1))).
  auto.
  intros b;
   generalize
    (refines_unfold (refines_refl (bind s (fun a : A => f a b)))).
  intros H1;
   change
     (sdemon (bind s (fun a : A => f a b))
        (fun a : C =>
         exists a2 : _,
           post (bind s (fun a0 : A => f a0 a2)) a))
    in |- *.
 eapply sdemon_weaken. lapply H1. eauto. simpl in |- *; auto.
 simpl in |- *; eauto.
Qed. 

Lemma any_when_pre_can_be_lifted : forall (A B C : Type) (s : DS A) (f : A -> B -> DS C),
 pre s -> 
 refines (bind s (fun a => bind (any B) (fun b => f a b)))
   (bind (any B) (fun b => bind s (fun a => f a b))).
Proof.
  intros A B C s f H0; apply refines_fold; simpl in |- *.
  intros H.
  apply
   sdemon_forall_lift1
    with
      (R := fun (a0 : B) (a : A) =>
            sdemon (f a a0)
              (fun a1 : C =>
               exists a2 : _,
                 angel s
                   (fun a3 : A => post (f a3 a2) a1)));
   auto.
  intros b;
   generalize
    (refines_unfold (refines_refl (bind s (fun a : A => f a b)))).
  intros H1;
   change
     (sdemon (bind s (fun a : A => f a b))
        (fun a : C =>
         exists a2 : _,
           post (bind s (fun a0 : A => f a0 a2)) a))
    in |- *.
 eapply sdemon_weaken. lapply H1. eauto. simpl in |- *; auto.
 simpl in |- *; eauto.
Qed. 

(** *** primitive operators for ``efficiency'' 

Here, I define [abort], [magic] and [or] as primitive operators, because it allows to have
better wp-simplifications for these constructions.
*)

(* ABORT *)

(* begin hide *)
Lemma abort_correct :
 forall A : Type,
 is_wp_pair (fun _ : A -> Prop => False) (fun _ : A -> Prop => False).
Proof.
  unfold is_wp_pair in |- *; simpl in |- *; firstorder eauto.
Qed.
(* end hide *)

Definition abort (A : Type) : DS A
 := Build_DS (sdemon:=fun R => False) (angel:=fun R => False) (abort_correct (A:=A)).
(*
  refine
   (fun A =>
    Build_DS (sdemon:=fun R : A -> Prop => False) (angel:=fun R => False) _).
  apply abort_correct.
Defined.
*)

(* begin hide *)
Lemma abort_impl : forall A : Type, impl (abort A).
Proof.
  unfold impl in |- *; simpl in |- *; intros; tauto.
Qed.

Lemma abort_strdet : forall A : Type, strdet (abort A).
Proof.
  unfold strdet in |- *; simpl in |- *; intros; tauto.
Qed.
(* end hide *)

Lemma refines_abort :
 forall (A : Type) (s : DS A), refines s (abort A).
Proof.
  unfold refines in |- *; simpl in |- *; tauto.
Qed.

(* MAGIC *)

(* begin hide *)
Lemma magic_correct :
 forall A : Type,
 is_wp_pair (fun _ : A -> Prop => True) (fun _ : A -> Prop => False).
Proof.
  unfold is_wp_pair in |- *; simpl in |- *; firstorder eauto.
Qed.
(* end hide *)

Definition magic (A : Type) : DS A
 := Build_DS (sdemon:=fun R => True) (angel:=fun R => False) (magic_correct (A:=A)).
(*
  refine
   (fun A =>
    Build_DS (sdemon:=fun R : A -> Prop => True) (angel:=fun R => False) _).
  apply magic_correct.
Defined.
*)

(* Magic is never valid on non-empty types *)
(*
Lemma DSmagic_invalid: (A,B:Type)(DSvalid (magic A B)) <-> (A -> False).
Proof.
  Unfold DSvalid; Simpl; Tauto.
Qed.
*)

Lemma refines_magic :
 forall (A : Type) (s : DS A), refines (magic A) s.
Proof.
  unfold refines in |- *; simpl in |- *; tauto.
Qed.

(* begin hide *)
Hint Resolve val_impl require_impl abort_impl bind_impl: wpimpl.
(* end hide *)

(* OR *)
                                                                                          
(* begin hide *)
Lemma or_correct: forall (A:Type) (s1 s2:(DS A)),
  (is_wp_pair (fun R => (sdemon s1 R) /\ (sdemon s2 R))
              (fun R => (angel s1 R) \/ (angel s2 R))).
Proof.
  intros A s1 s2 R; generalize (DScorrect s1) (DScorrect s2);  
  unfold is_wp_pair; intros H1 H2; case (H1 R); case (H2 R); intuition (eauto 20);
  decompose [ex and or] H8; intuition eauto.
Qed.
(* end hide *)

Definition or (A:Type) (s1 s2:(DS A)) : (DS A)
:= Build_DS (sdemon:=(fun R => (sdemon s1 R) /\ (sdemon s2 R)))
            (angel:=(fun R => (angel s1 R) \/ (angel s2 R)))
            (or_correct s1 s2).
(*  
  refine (fun A s1 s2 => (Build_DS (sdemon:=(fun R => (sdemon s1 R) /\ (sdemon s2 R)))
                                     (angel:=(fun R => (angel s1 R) \/ (angel s2 R))) _)).
  apply or_correct; auto.
Defined.
*)

Lemma or_pre: forall (A:Type) (s1 s2:(DS A)), (pre (or s1 s2)) <-> (pre s1) /\ (pre s2).
Proof.
  simpl; intuition auto.
Qed.
                                                                                          
Lemma or_post: forall (A:Type) (s1 s2:(DS A)) (a:A),
 (post (or s1 s2) a) <-> (post s1 a) \/ (post s2 a).
Proof.
  simpl; intuition auto.
Qed.
             
                                                                            
(* Proprit fondamentale du "OR" vis-a-vis de "REFINES" *)
Lemma refines_or_intro_l: forall (A:Type) (s1 s2 s:(DS A)),
  (refines s s1) -> (refines s (or s1 s2)).
Proof.
  intros A s1 s2 s H; eapply refines_trans. apply H. unfold refines; simpl; tauto.
Qed.

Lemma refines_or_intro_r: forall (A:Type) (s1 s2 s:(DS A)),
  (refines s s2) -> (refines s (or s1 s2)).
Proof.
  intros A s1 s2 s H; eapply refines_trans. apply H. unfold refines; simpl; tauto.
Qed.
                                                                                          
Lemma refines_or_elim: forall (A:Type) (s1 s2 s:(DS A)),
  (refines s1 s) -> (refines s2 s) -> (refines (or s1 s2) s).
Proof.
  unfold refines; simpl; intros A s1 s2 s H H0 R.
  case (H R); case (H0 R); tauto.
Qed.

(* begin hide *)
Hint Resolve refines_or_intro_l refines_or_intro_r refines_or_elim: wprefines.
(* end hide *)

(** Here is the usual definition of [or] using [bind] and [any]. 
  It allows to inherits properties of [bind], but has less powerful wp.
 *)
Definition or_x  (A : Type) (s1 s2:DS A) := 
  (bind (any bool) (fun b => if b then s1 else s2)).

Lemma refines_or_x: forall (A : Type) (s1 s2:DS A),
 (refines (or s1 s2) (or_x s1 s2)).
Proof.
  intros A s1 s2; unfold or_x; apply refines_or_elim. 
  apply any_refined_bind with (a:=true); auto with wprefines.
  apply any_refined_bind with (a:=false); auto with wprefines.
Qed.

Lemma refines_x_or: forall (A : Type) (s1 s2:DS A),
 (refines (or_x s1 s2) (or s1 s2)).
Proof.
  intros A s1 s2; unfold or_x; apply any_refines; intros b;
  case b; auto with wprefines. 
Qed.

Lemma or_monotony :
 forall (A : Type) (s1 s2 s3 s4 : DS A),
 refines s1 s3 -> 
 refines s2 s4 -> refines (or s1 s2) (or s3 s4).
Proof.
  auto with wprefines.
Qed.

(* begin hide *)
Hint Resolve refines_or_x refines_x_or: wprefines. 
(* end hide *)

(*** Assoc rules ***)
                                                                                
(* rem: 3 consquences des proprits fdales du OR *)
Lemma refines_or_assoc_l :
 forall (A : Type) (s1 s2 s3 : DS A),
 refines (or (or s1 s2) s3) (or s1 (or s2 s3)).
Proof.
  auto with wprefines.
Qed.
                                                                                
Lemma refines_or_assoc_r :
 forall (A : Type) (s1 s2 s3 : DS A),
 refines (or s1 (or s2 s3)) (or (or s1 s2) s3).
Proof.
  auto with wprefines.
Qed.
                                                                                
Lemma refines_or_sym :
 forall (A : Type) (s1 s2 : DS A),
 refines (or s1 s2) (or s2 s1).
Proof.
  auto with wprefines.
Qed.

Lemma bind_or_distr_r_1:
  forall (A B:Type) (s1 s2: DS A) (s: A -> (DS B)),
   (refines (bind (or s1 s2) s) (or (bind s1 s) (bind s2 s))).
Proof.
  intros; apply refines_trans with (s2:=(bind (or_x s1 s2) s)); auto with wprefines.
  unfold or_x; eapply refines_trans.
  apply bind_assoc_ref1.
  apply any_refines; intros b; case b; auto with wprefines.
Qed.

Lemma bind_or_distr_r_2:
  forall (A B:Type) (s1 s2: DS A) (s: A -> (DS B)),
   (refines (or (bind s1 s) (bind s2 s)) (bind (or s1 s2) s)).
Proof.
  intros; apply refines_or_elim; auto with wprefines.
Qed.

Lemma bind_or_distr_l_1:
 forall (A B:Type) (s: DS A) (s1 s2: A -> (DS B)),
   (refines (bind s (fun a => (or (s1 a) (s2 a)))) (or (bind s s1) (bind s s2))).
Proof.
  intros; apply refines_trans with (s2:=(bind s (fun a => (or_x (s1 a) (s2 a)))));
  auto with wprefines.
  unfold or_x; eapply refines_trans.
  apply any_on_non_empty_type_can_be_lifted with 
    (f:=(fun a (b:bool) => if b then s1 a else s2 a));
  auto with nonempty.
  apply any_refines; intros b; case b; auto with wprefines.
Qed.

Lemma bind_or_distr_l_2:
 forall (A B:Type) (s: DS A) (s1 s2: A -> (DS B)),
   (refines (or (bind s s1) (bind s s2)) (bind s (fun a => (or (s1 a) (s2 a))))).
Proof.
  intros; apply refines_or_elim; auto with wprefines.
Qed.

(* begin hide *)

(**  AND **)

Definition and (A: Type) (s1 s2:DS A): DS A
 := (bind (require ((pre s1) \/ (pre s2)))
       (fun _ => bind (any A)
          (fun a => bind (ensure (((pre s1) -> (post s1 a)) /\ 
                                  ((pre s2) -> (post s2 a))))
                          (fun _ => (val a))))).

Lemma and_pre: forall (A:Type) (s1 s2:(DS A)),
  (pre (and s1 s2)) <-> ((pre s1) \/ (pre s2)).
Proof.
  intros; simpl; firstorder auto.
Qed.

Lemma and_post: forall (A:Type) (s1 s2: (DS A)) (a:A),
    (post (and s1 s2) a) <-> (((pre s1) -> (post s1 a)) 
                               /\ ((pre s2) -> (post s2 a))).
Proof.
  intros; simpl; firstorder eauto.
  subst; auto.
  subst; auto.
Qed.

Lemma and_elim_l: forall (A:Type) (s1 s2 s:DS A),
  (refines s1 s) -> (refines (and s1 s2) s).
Proof.
  intros A s1 s2 s H H0; 
  case (refines_prepost s1 s);   
  intros H1 H2; clear H2; apply refines_fold.
  simpl.
  intuition auto.
Qed.

Lemma and_elim_r: forall (A:Type) (s1 s2 s:DS A),
  (refines s2 s) -> (refines (and s1 s2) s).
Proof.
  intros A s1 s2 s H H0; 
  case (refines_prepost s2 s);   
  intros H1 H2; clear H2; apply refines_fold.
  simpl; firstorder auto.
Qed.

Lemma and_intro: forall (A:Type) (s1 s2 s:DS A),
  (refines s s1) -> (refines s s2) -> (refines s (and s1 s2)).
Proof.
  intros A s1 s2 s H1 H2;
  case (refines_prepost s s1);
  case (refines_prepost s s2);
  intros H3 H4 H5 H6.
  clear H4 H6; case (refines_prepost s (and s1 s2)).
  intros H4 H6; clear H4.
  apply H6; clear H6.
  intros H4; simpl in H4.
  constructor 1.
  intuition auto.
  intros a H; case (and_post s1 s2 a).
  intros H6 H7; clear H6.
  intuition auto.
Qed.

Hint Resolve and_elim_l and_elim_r and_intro: wprefines.
  
Lemma and_monotony :
 forall (A : Type) (s1 s2 s3 s4 : DS A),
 refines s1 s3 -> 
 refines s2 s4 -> refines (and s1 s2) (and s3 s4).
Proof.
  intros; auto with wprefines.
Qed.

Hint Resolve and_monotony: wprefines. 

Lemma and_ensure_ref1: forall (P Q: Prop),
  (refines (and (ensure P) (ensure Q)) (ensure (P /\ Q))).
Proof.
  intros; apply refines_fold; simpl.
  intuition auto.
Qed.

Lemma and_ensure_ref2: forall (P Q: Prop),
  (refines (ensure (P /\ Q)) (and (ensure P) (ensure Q))).
Proof.
  intros; apply and_intro;
  apply refines_fold; simpl; intuition auto.
Qed.

Lemma refines_and_assoc_l :
 forall (A : Type) (s1 s2 s3 : DS A),
 refines (and (and s1 s2) s3) (and s1 (and s2 s3)).
Proof.
  intros; auto with wprefines.
Qed.
                                                                                
Lemma refines_and_assoc_r :
 forall (A : Type) (s1 s2 s3 : DS A),
 refines (and s1 (and s2 s3)) (and (and s1 s2) s3).
Proof.
  intros; auto with wprefines.
Qed.
                                                                                
Lemma refines_and_sym :
 forall (A : Type) (s1 s2 : DS A),
 refines (and s1 s2) (and s2 s1).
Proof.
  intros; auto with wprefines.
Qed.

Lemma bind_and_distr_r_1:
  forall (A B:Type) (s1 s2: DS A) (s: A -> (DS B)),
   (refines (bind (and s1 s2) s) (and (bind s1 s) (bind s2 s))).
Proof.
 intros; apply and_intro;
 apply bind_monotony; auto with wprefines.
Qed.

Lemma bind_and_distr_l_1:
 forall (A B:Type) (s: DS A) (s1 s2: A -> (DS B)),
   (refines (bind s (fun a => (and (s1 a) (s2 a)))) (and (bind s s1) (bind s s2))).
Proof.
 intros; apply and_intro;
 apply bind_monotony; auto with wprefines.
Qed.

(* FALSE PROPOSITION: bind_and_distr_l_2
 forall (A B:Type) (s: DS A) (s1 s2: A -> (DS B)),
   (pre (bind s s1)) -> (pre (bind s s2)) ->   
    (refines (and (bind s s1) (bind s s2)) 
             (bind s (fun a => (and (s1 a) (s2 a)))) ).
*)

Lemma bind_and_distr_l_2_cx1:
  (refines (val 1) (and (bind (any nat) (fun n => bind (ensure (n=1)) (fun _ => (val n))))
                        (bind (any nat) (fun n => bind (ensure (n=0)) (fun _ => (val (S n))))))).
Proof.
  intros; apply and_intro; 
  apply bind_simpl_right; simpl.
  intros; constructor 1 with (x:=1).
  apply bind_simpl_right; simpl; intros.
  intuition auto with wprefines.  
  intros; constructor 1 with (x:=0).
  apply bind_simpl_right; simpl; intros.
  intuition auto with wprefines.  
Qed. 

Lemma bind_and_distr_l_2_cx2:
  (refines (bind (any nat) (fun n => and (bind (ensure (n=1)) (fun _ => (val n)))
                                         (bind (ensure (n=0)) (fun _ => (val (S n))))))
           (magic nat)).
Proof.
  intros; apply refines_fold.
  simpl.
  intuition auto.
  subst.
  discriminate H2.
Qed.

(* sdemon_split:
  forall (A:Type) (s1 s2: DS A) (R: A -> Prop),
    (sdemon (and s1 s2) R) -> (sdemon s1 R) \/ (sdemon s2 R).
Proof.
  simpl; intros A s1 s2 R H.
  case H; clear H; intros H1 H2.
  case H1; [constructor 1 | constructor 2]; 
  apply sdemon_fold; intuition auto.
Qed.
*)

(* FALSE PROPOSITION bind_and_distr_r_2:
  forall (A B:Type) (s1 s2: DS A) (s: A -> (DS B)),
   (refines (and (bind s1 s) (bind s2 s)) (bind (and s1 s2) s)).
*)
  
Lemma bind_and_distr_r_2_cx1:
  (refines (val 0) 
           (and (bind (val 0) (fun x => bind (require (x=0)) (fun _ => (val x))))
                (bind (val 1) (fun x => bind (require (x=0)) (fun _ => (val x)))))
           ).
Proof.
  intros; apply refines_fold; simpl.
  firstorder auto.
Qed.
          
Lemma bind_and_distr_r_2_cx2:
  (refines (bind (and  (val 0) (val 1))
                (fun x => bind (require (x=0)) (fun _ => (val x))))
           (magic nat)).
Proof.
  intros; apply refines_fold; simpl.
  firstorder auto.
  subst.
  discriminate H1.
Qed.

(** SYNC **)

Definition sync (A B:Type) (s :A -> (DS B)) := 
  (bind (require (exists a:A, (pre (s a))))
      (fun _ => bind (any B) 
           (fun b => bind (ensure (forall (a:A), (pre (s a)) -> (post (s a) b)))
                 (fun _ => (val b))))).

Lemma sync_pre: forall (A B:Type) (s: A -> (DS B)),
  (pre (sync s)) <-> (exists a:A, (pre (s a))).
Proof.
  intros; simpl; firstorder auto.
Qed.

Lemma sync_post:  forall (A B:Type) (s: A -> (DS B)) (b:B),
    (post (sync s) b) <-> forall (a:A), (pre (s a)) -> (post (s a) b).
Proof.
  intros; simpl; firstorder eauto.
  subst; auto.
Qed.

Lemma sync_elim: forall (A B:Type) (s1: A -> (DS B)) (s2:(DS B)) (a:A),
  (refines (s1 a) s2) -> (refines (sync s1) s2).
Proof.
  intros A B s1 s2 a H. 
  case (refines_prepost (s1 a) s2);   
  intros H1 H2; clear H2; apply refines_fold.
  simpl.
  intuition eauto.
Qed.

Lemma sync_intro: forall (A B:Type) (s1:(DS B)) (s2: A -> (DS B)),
  (forall (a:A), (refines s1 (s2 a))) -> (refines s1 (sync s2)).
Proof.
  intros A B s1 s2 H. 
  case (refines_prepost s1 (sync s2)).
  intros H0 H1; apply H1; clear H0 H1.
  intros H0; case (sync_pre s2).
  intros H1 H2; case H1; clear H1 H2; auto.
  intros a; case (refines_prepost s1 (s2 a)).
  intros H1 H2; clear H2.
  simpl; constructor 1.
  firstorder auto.
  intros b H3; constructor 1 with (x:=b).
  firstorder auto.
Qed.

Hint Resolve sync_elim sync_intro: wprefines2.
  
Lemma sync_monotony :
 forall (A B: Type) (s1 s2: A -> DS B),
 (forall (a:A), refines (s1 a) (s2 a)) -> 
    refines (sync s1) (sync s2).
Proof.
  intros; eauto with wprefines2.
Qed.

Hint Resolve sync_monotony: wprefines2. 


Lemma sync_elimX: forall (A B C:Type) (s1: A -> (DS B)) 
(s2: B -> (DS C)) (s3: DS C) (a:A),
  (refines (bind (s1 a) s2) s3) -> (refines (bind (sync s1) s2) s3).
Proof.
  intros A B C s1 s2 s3 a H. 
  eapply refines_trans.
  2: apply H.
  eauto with wprefines wprefines2.
Qed.

Lemma sync_introX: forall (A B C:Type) (s1: DS C) (s2: DS B) (s3: B -> DS C),
  (A -> refines s1 (bind s2 s3)) -> (refines s1 (bind (sync (fun _:A => s2)) s3)).
Proof.
  intros A B C s1 s2 s3 H. 
  case (refines_prepost s1 (bind (sync (fun _:A => s2)) s3)).
  intros H0 H1; apply H1; clear H0 H1.
  simpl; intuition auto.
  case H1; clear H1.
  intros H0 H1.
  case (refines_prepost s1 (bind s2 s3)).
  intros H3 H4; generalize (H3 (H H0)); clear H3 H4.
  simpl.
  intros X; lapply X; clear X.
  intuition auto.
  apply sdemon_fold; auto.
  case H1; clear H1.
  intros H' H1.
  case (refines_prepost s1 (bind s2 s3)).
  intros H3 H4; generalize (H3 (H H')); clear H3 H4.
  simpl.
  intros X; lapply X; clear X.
  clear H.
  intuition auto.
  case (angel_unfold _ _ (H4 a H0)).
  clear H4 H2; firstorder auto.
  apply sdemon_fold; auto.
Qed.

Lemma bind_sync_distr_r_1:
  forall (A B C:Type) (s1: A->DS B) (s2: B -> (DS C)),
   (refines (bind (sync s1) s2) (sync (fun x => bind (s1 x) s2))).
Proof.
 intros; apply sync_intro.
 intros; apply bind_monotony; eauto with wprefines wprefines2.
Qed.

Lemma bind_sync_distr_l_1:
 forall (A B C:Type) (s1: DS A) (s2: A -> B -> (DS C)),
   (refines (bind s1 (fun a => sync (s2 a))) (sync (fun b => (bind s1 (fun a => s2 a b))))).
Proof.
 intros; apply sync_intro.
 intros; apply bind_monotony; eauto with wprefines wprefines2.
Qed.

(*
Lemma guard_conjunc:
  forall (A:Type) (g1 g2 g3: A -> Prop) (sp: forall (B:Type), (A -> DS B) -> (DS B)),
    (forall (x:A), (g1 x) -> (g2 x) -> (g3 x)) ->
      (forall (B:Type) (i j: A -> (DS B)), 
          (forall (x:A), refines (i x) (j x)) -> (refines (sp B i) (sp B j))) ->
       (forall (B:Type) (i: A -> (DS B)), 
         refines (sp B i) (bind (any A) (fun x => bind (ensure (g1 x)) (fun _ => i x)))) -> 
       (forall (B:Type) (i: A -> (DS B)), 
         refines (sp B i) (bind (any A) (fun x => bind (ensure (g2 x)) (fun _ => i x)))) -> 
       forall (B:Type) (i: A -> (DS B)), 
         (forall (x y:A)(b:B), (post (i x) b) -> (post (i y) b) -> x=y) -> 
         refines (sp B i) (bind (any A) (fun x => bind (ensure (g3 x)) (fun _ => i x))).
Proof.
  intros A g1 g2 g3 sp H H0 H1 H2 B i H3.
  apply refines_pre_intro.
  simpl.
  intros H4.
  apply refines_trans with 
   (s2:=
     (sp B (fun x => (bind (require (g3 x)) (fun _ => i x))))). 
  apply H0; auto with wprefines.

  apply refines_fold; simpl.
  generalize (refines_unfold (H1 B (fun x => (bind (require (g3 x)) (fun _ => i x)))))
             (refines_unfold (H2 B (fun x => (bind (require (g3 x)) (fun _ => i x))))); simpl.
  clear H2 H3; intros H2 H3 H5.
  lapply H3. lapply H2.
  clear H2 H3; intros H2 H3.
  apply sdemon_fold.
  eauto with wipi.
  intros a H5.
  case (sdemon_unfold _ _ _ H2 H5).
  intros x Hx.
  case (sdemon_unfold _ _ _ H3 H5).
  intros y Hy.
  lapply (H1 x y a); intuition auto.
  subst.
  constructor 1 with (x:=y).
  intuition auto.
  intuition auto.
  intuition auto.

Qed.
*)

Extraction Inline val_impl bind_impl require_impl any require ensure abort abort_impl impl.
Extract Inductive unit => "unit" [ "()" ].

Unset Implicit Arguments.
(* end hide *)
