Set Implicit Arguments.

Require Export FSL.
Require Export WfT.
Require Export Iter.
Require Export FSLTarski.

Section Fix.

Variable A B: Type.

Variable F: (A -> (DS B)) -> (A -> (DS B)).
Variable R: A -> A -> Prop.

Hypothesis F_monotony:
  forall (f g: A -> (DS B)), 
   (forall (x:A), (refines (f x) (g x)))
      -> forall (x:A), (refines (F f x) (F g x)).

Definition call_decreasing :=
  forall (f:A -> (DS B)) (x:A), (refines (F f x) 
                                       (F (fun y =>
                                               (bind (ensure (R y x))
                                                   (fun _ => (f y))))
                                          x)).

Hypothesis Rwf : (well_founded R).

Fixpoint iterWF (x:A) (a: Acc R x) {struct a} : (DS B)
  := (F  (fun y => (bind (any (R y x))
              (fun h => (iterWF (Acc_inv a y h))))) x).

Lemma iterWF_non_acc_dep:
 forall (x:A) (r s: Acc R x),
    (refines (iterWF r) (iterWF s)).  
Proof.
   intros x r; elim r using Acc_inv_dep; simpl; clear r x.
   intros x H1 H2.
   intros s; generalize H1 H2; clear H1 H2; case s; simpl.
   auto with wprefines.
Qed.

Hint Resolve iterWF_non_acc_dep: wprefines.

Definition fixF (x:A) := (iterWF (Rwf x)).

Lemma iterWF_refines_F:
  forall (x:A) (r: Acc R x),
      (refines (iterWF r) (F fixF x)).
Proof.
  intros x r; elim r using Acc_inv_dep; clear r x; simpl.
  intros x H1 H2.
  apply F_monotony.
  intros y; apply any_refines; intros H3.
  unfold fixF; apply iterWF_non_acc_dep.
Qed.

Lemma fixF_postfix:
   forall (x:A), (refines (fixF x) (F fixF x)).
Proof.
  unfold fixF; intros; apply iterWF_refines_F.
Qed.

Lemma fixF_mostmagic:
  forall (fixO: A -> (DS B)), 
    (forall (x:A), (refines (F fixO x) (fixO x)))
     -> forall (x:A), (refines (fixF x) (fixO x)).
Proof.
  intros fixO H x; unfold fixF.
  generalize (Rwf x); intros r; elim r using Acc_inv_dep; clear r x; simpl.
  intros x H1 H2.
  eapply refines_trans.
  2: apply H.
  apply F_monotony.
  intros y.
  apply any_refines; intros H3; auto.
Qed.

Hypothesis F_decreasing: call_decreasing.

Lemma fixF_prefix:
   forall (x:A), (refines (F fixF x) (fixF x)).
Proof.
  unfold fixF. unfold call_decreasing in F_decreasing.
  intros x; eapply refines_trans.
  apply F_decreasing.
  case (Rwf x); simpl.
  intros x0 H; apply F_monotony.
  intros y; apply bind_simpl_left; simpl.  
  intros H1; apply any_refined_bind with (a:=H1).
  apply iterWF_non_acc_dep.
Qed. 

Lemma iterfixF_prefix:
  forall (n:nat) (x:A), 
     (refines (iter n F fixF x) (fixF x)). 
Proof.
  induction n; simpl; auto with wprefines.
  intros x; eapply refines_trans.
  2: apply fixF_prefix.
  auto.
Qed.

Section FixFInv.

Variable inv: (A -> (DS B)) -> A -> Type.

Hypothesis inv_compat: forall (x:A) (s1 s2: A -> (DS B)),
  (inv s1 x) -> (refines (s1 x) (s2 x)) -> (refines (s2 x) (s1 x)) -> 
    (inv s2 x).

Section FixFrect.

Hypothesis F_preserves_inv: forall (f: A -> (DS B)) (x:A), 
  (forall (y:A), (R y x) -> (inv f y)) 
    -> inv (F f) x.

Theorem fixF_rect:
   forall (x:A), inv fixF x.
Proof.
  intros x; elim (Rwf x) using Acc_rect; clear x.
  simpl; intros x H1 H2.
  eapply inv_compat.
  2: apply fixF_prefix.
  apply F_preserves_inv.
  auto.
  apply fixF_postfix.
Qed.

End FixFrect.

Lemma inv_fixF_require:
  (forall (x:A) (Q: A -> Prop) (s:A -> (DS B)), 
      ((Q x) -> (inv s x)) -> (inv (fun y => (bind (require (Q y)) (fun _ => (s y)))) x))
   -> (forall (f: A -> (DS B)), (forall (x:A), (inv f x))
             -> forall (x:A), inv (F f) x)
   -> forall (x:A), inv fixF x.
Proof.
  intros inv_require inv_F x; apply fixF_rect; clear x; auto.
  intros f x H; apply inv_compat with (s1:=(F (fun y =>
                                               (bind (require (R y x))
                                                   (fun _ => (f y)))))); auto.
  (* apply inv_F.
  intros y; apply inv_require with (Q:=(fun y => (R y x))); auto. *)
  eapply refines_trans.
  unfold call_decreasing in F_decreasing; apply  F_decreasing.
  auto with wprefines.
  auto with wprefines.
Qed.

Lemma inv_fixF_ensure:
   (forall (x:A) (Q: A -> Prop) (s: A -> (DS B)), 
      ((Q x) -> (inv s x)) -> (inv (fun y => (bind (ensure (Q y)) (fun _ => (s y)))) x))
   -> (forall (f: A -> (DS B)), (forall (x:A), (inv f x))
             -> forall (x:A), inv (F f) x)
   -> forall (x:A), inv fixF x.
Proof.
  intros inv_ensure inv_F x; apply fixF_rect; clear x; auto.
  intros f x H; apply inv_compat with (s1:=(F (fun y =>
                                               (bind (ensure (R y x))
                                                   (fun _ => (f y)))))); auto.
  (* apply inv_F.
  intros y; apply inv_ensure with (Q:=(fun y => (R y x))); auto. *)
  eapply refines_trans.
  unfold call_decreasing in F_decreasing; apply  F_decreasing.
  auto with wprefines.
Qed.

End FixFInv.


Lemma fixF_mostabort:   forall (fixO: A -> (DS B)), 
    (forall (x:A), (refines (fixO x) (F fixO x)))
     -> forall (x:A), (refines (fixO x) (fixF x)).
Proof.
  intros fixO H x. pattern fixF, x; apply inv_fixF_require; clear x;  auto with wprefines.
  intros; apply refines_trans with (s2:=(s1 x)); auto.
  intros; eapply refines_trans.
  apply H. auto.
Qed.

(* another proof ! *)

Lemma fixF_mostmagic2:   forall (fixO: A -> (DS B)), 
    (forall (x:A), (refines (F fixO x) (fixO x)))
     -> forall (x:A), (refines (fixF x) (fixO x)).
Proof.
  intros fixO H x; pattern fixF, x; apply inv_fixF_ensure; clear x; auto with wprefines.
  intros; apply refines_trans with (s2:=(s1 x)); auto.
  intros; eapply refines_trans.
  2: apply H. auto.
Qed.


(* link with "sfixF" *)
Lemma sfixF_refines_fixF:
 forall x, refines (sfixF F x) (fixF x).
Proof.
  intros; apply sfixF_mostmagic.
  apply fixF_prefix.
Qed.

Lemma fixF_refines_sfixF:
 forall x, refines (fixF x) (sfixF F x).
Proof.
  intros; apply fixF_mostmagic.
  apply sfixF_prefix; auto.
Qed.

Lemma sfixF_mostabort:forall (fixO: A -> (DS B)), 
    (forall (x:A), (refines (fixO x) (F fixO x)))
     -> forall (x:A), (refines (fixO x) (sfixF F x)).
Proof.
  intros; eapply refines_trans.
  apply fixF_mostabort; auto.
  apply fixF_refines_sfixF.
Qed.

Section Determinitic_Fixpoints.


(** fixF preserves determinism *)
Variable F_impl: 
 forall (f: A -> (DS B)), (forall (x:A), (impl (f x))) -> forall (x:A), impl (F f x).


Lemma fixF_impl: forall (x:A), impl (fixF x).
Proof.
(* In comments below, the original proof. Here, I "inline" the proof
   of "inv_fixF_require" because coq extraction (in v8.1) does not
   inline the "inv_compat" proof !!  
*)
 
(*
  intros x; pattern fixF, x. apply inv_fixF_require ; auto with wprefines.
  intros; eapply impl_compat; eauto.  
  intros; apply bind_impl; simpl; intuition.
*)

  intros x; pattern fixF, x; apply fixF_rect; clear x; auto.
  intros; eapply impl_compat; eauto.
  intros f x H.
  lapply (F_impl (fun y => bind (require (R y x)) (fun _ => f y))).
  intros H1; eapply impl_compat.
  3: apply (H1 x).
  unfold call_decreasing in F_decreasing.
  eapply refines_trans.
  apply F_decreasing.
  auto with wprefines.
  auto with wprefines.
  intros; apply bind_impl; simpl; intuition.
Qed.

Lemma sfixF_impl: forall (x:A), impl (sfixF F x).
Proof.
  intros; eapply impl_compat.
  apply fixF_refines_sfixF.
  apply sfixF_refines_fixF.
  apply fixF_impl.
Qed.

(* We can extract the fixpoint into a function *)
Definition dfix: forall (x:A), (pre (sfixF F x)) -> B 
 := fun (x:A) (H:pre (sfixF F x)) => extract (sfixF_impl H).
  
(* This fixpoint satisfies the (extracted) fixpoint equation *)
Lemma fix_dfix: forall (x:A) (H:pre (sfixF F x)),
  (extract (F_impl _ sfixF_impl _ (refines_weakpre (sfixF_prefix F F_monotony x) H)))=(dfix H).
Proof.
  intros x H; unfold dfix; apply (extract_unicity (sfixF_prefix F F_monotony x)).
Qed.

End Determinitic_Fixpoints.
End Fix.


Section Monotony.

Variable A B: Type.

Variable F1 F2: (A -> (DS B)) -> (A -> (DS B)).
Variable R1 R2: A -> A -> Prop.

Hypothesis F1_monotony:
  forall (f g:A -> (DS B)), 
   (forall (x:A), (refines (f x) (g x))) -> 
      forall (x:A), (refines (F1 f x) (F1 g x)).

Hypothesis F2_monotony:
  forall (f g:A -> (DS B)), 
   (forall (x:A), (refines (f x) (g x))) -> 
      forall (x:A), (refines (F2 f x) (F2 g x)).

Hypothesis R1wf : (well_founded R1).
Hypothesis R2wf : (well_founded R2).

Hypothesis F1_refines_F2:
  forall (f: A -> (DS B)) (x:A), (refines (F1 f x) (F2 f x)).

Hypothesis F2_decreses: (call_decreasing F2 R2).

Lemma fixF1_refines_fixF2:  forall (x:A), (refines (fixF F1 R1wf x) (fixF F2 R2wf x)).
Proof.
  intros; apply fixF_mostmagic; auto.
  intros; eapply refines_trans.
  2: apply fixF_prefix; auto.
  auto.
Qed.

End Monotony.


Section Morph.

Variable A1 A2 B1 B2: Type.

Variable F1: (A1 -> (DS B1)) -> (A1 -> (DS B1)).
Variable F2: (A2 -> (DS B2)) -> (A2 -> (DS B2)).
Variable G: (A1 -> (DS B1)) -> (A2 -> (DS B2)).

Hypothesis F1_monotony:
  forall (f g:A1 -> (DS B1)), 
   (forall (x:A1), (refines (f x) (g x))) -> 
      forall (x:A1), (refines (F1 f x) (F1 g x)).

Hypothesis F2_monotony:
  forall (f g:A2 -> (DS B2)), 
   (forall (x:A2), (refines (f x) (g x))) -> 
      forall (x:A2), (refines (F2 f x) (F2 g x)).

Hypothesis G_monotony:
  forall (f g:A1 -> (DS B1)), 
  (forall (x:A1), (refines (f x) (g x))) -> 
      forall (x:A2), (refines (G f x) (G g x)).
 
Variable R1: A1 -> A1 -> Prop.
Hypothesis R1wf : (well_founded R1).

Variable R2: A2 -> A2 -> Prop.
Hypothesis R2wf : (well_founded R2).

Lemma fix2_refines_G_fix1:
 (call_decreasing F1 R1)
  -> (forall (f:A1 -> (DS B1)) (x:A2), (refines (F2 (G f) x) (G (F1 f) x)))
    -> forall (x:A2), (refines (fixF F2 R2wf x) (G (fixF F1 R1wf) x)).
Proof.
  intros F1_decreses F2G_refines_GF1; apply fixF_mostmagic; auto.
  intros; eapply refines_trans.
  apply F2G_refines_GF1.
  apply G_monotony.
  intros; apply fixF_prefix; auto.
Qed.

Lemma G_fix1_refines_fix2:
  (call_decreasing F2 R2) -> 
     (forall (f:A1 -> (DS B1)) (x:A2), (refines (G (F1 f) x) (F2 (G f) x))) 
      -> forall (x:A2), (refines (G (fixF F1 R1wf) x) (fixF F2 R2wf x)).
Proof.
   intros F2_decreasing GF1_refines_F2G x; pattern (fixF F2 R2wf), x.
   apply inv_fixF_require; clear x; auto with wprefines.
   intros; eapply refines_trans; eauto.
   intros; apply refines_trans with (s2:= (G (F1 (fixF F1 R1wf)) x)).
   apply G_monotony; intros;  apply fixF_postfix; auto.
   eapply refines_trans; auto.
Qed.

End Morph.

Ltac sfixF_mabort :=
  match goal with
    |- (refines ?t (sfixF _ ?x)) => 
          let t1:=(lambda x t) in
            eapply sfixF_mostabort with (fixO:=t1)
            || idtac "Tactic too weak: try something else..."
    | _ => idtac "User Error: goal does not match"
    end.

Extraction Inline Acc_rect fixF_rect (* inv_fixF_require *) fixF_impl sfixF_impl.

Unset Implicit Arguments.
