Set Implicit Arguments.

Require Export FSL.
Require Import Tarski.

Definition glb (A:Type) (P:(DS A)->Prop) : (DS A) 
 := (sync (fun sp => bind (require (P sp)) (fun _ => sp))).

Lemma glb_refines:
  forall (A:Type) (P:(DS A)->Prop) (sp:(DS A)),
    (P sp) -> (refines (glb P) sp).
Proof.
  intros A P sp H; unfold glb.
  apply sync_elim with (a:=sp).
  apply bind_simpl_left; simpl.
  intuition auto with wprefines.    
Qed.

Lemma glb_refined:
  forall (A:Type) (P:(DS A)->Prop) (sp:(DS A)),
    (forall (sp1:DS A), P sp1 -> (refines sp sp1)) 
        -> (refines sp (glb P)).
Proof.
  intros A P sp H; unfold glb.
  apply sync_intro.
  intros; apply bind_simpl_right; simpl.
  intuition auto.
Qed.

Lemma FSL_inf_semilattice:
  forall (A:Type), (inf_semilattice (refines (A:=A)) (glb (A:=A))).
Proof.
  intros; constructor 1; constructor 1.
  unfold Relation_Definitions.reflexive; auto with wprefines.
  unfold Relation_Definitions.transitive. apply refines_trans.
  apply glb_refines.
  apply glb_refined.
Qed.

Hint Resolve FSL_inf_semilattice: lattice.

Section Most_Magic.

Variable A: Type.

Section SFixDS_Def.

Variable F: (DS A) -> (DS A).

Hypothesis F_monotonic:
  forall (sp1 sp2:(DS A)),
    (refines sp1 sp2) -> (refines (F sp1) (F sp2)).

Definition sfixDS: (DS A) :=
  (sfix (refines (A:=A)) (glb (A:=A)) F).

Lemma sfixDS_mostmagic:
  forall (sp:(DS A)),
    (refines (F sp) sp)
      -> (refines sfixDS sp).
Proof.
  unfold sfixDS. intros; apply sfix_smallest with (R:=refines (A:=A)) ;
  auto with lattice.
Qed.
  
Lemma sfixDS_prefix:
  (refines (F sfixDS) sfixDS).
Proof.
  unfold sfixDS. intros; apply sfix_prefix with (R:=refines (A:=A)) ;
  auto with lattice.
Qed.

Lemma sfixDS_postfix:
  (refines sfixDS (F sfixDS)).
Proof.
  unfold sfixDS. intros; apply sfix_postfix with (R:=refines (A:=A)) ;
  auto with lattice.
Qed.

Lemma iterFsfixDS_refines_sfixDS:
  forall (n:nat), (refines (Iter.iter n F sfixDS) sfixDS).
Proof.
  unfold sfixDS. intros; apply iterFsfix_R_sfix with (R:=refines (A:=A)) ;
  auto with lattice.  
Qed.
                                                                                         
Lemma sfixDS_refines_iterFsfixDS:
  forall (n:nat), (refines sfixDS (Iter.iter n F sfixDS)).
Proof.
  unfold sfixDS. intros; apply sfix_R_iterFsfix with (R:=refines (A:=A)) ;
  auto with lattice.  
Qed.

End SFixDS_Def.

Section SFix_monotony.
                                                                                         
Variable F1 F2: (DS A) -> (DS A).
(* Hypothesis F1_monotony: forall (x y:DS A), (refines x y) -> (refines (F1 x) (F1 y)). *)
Hypothesis F2_monotony: forall (x y:DS A), (refines x y) -> (refines (F2 x) (F2 y)).
                                                                                         
Lemma sfixDS_monotony:
  (forall (x:DS A), (refines (F1 x) (F2 x))) -> (refines (sfixDS F1) (sfixDS F2)).
Proof.
  unfold sfixDS. intros; apply sfix_monotony with (R:=refines (A:=A)) ;
  auto with lattice.  
Qed.
                                                                                         
End SFix_monotony.

End Most_Magic.

Section SFixDS_Morph.
                                                                                         
Variable A1 A2: Type.
                                                                                        
Variable F1: DS A1 -> DS A1.
Hypothesis F1_monotony: forall (x y:DS A1), (refines x y) -> (refines (F1 x) (F1 y)).
                                                                                         
Variable F2: DS A2 -> DS A2.
(*Hypothesis F2_monotony: forall (x y:DS A2), (refines x y) -> (refines (F2 x) (F2 y)).*)
                                                                                         
Variable G: DS A1 -> DS A2.
Hypothesis G_monotony: forall (x y:DS A1), (refines x y) -> (refines (G x) (G y)).
                                                                                         
Lemma sfixDS2_refines_G_sfixDS1:
 (forall (x: DS A1), (refines (F2 (G x)) (G (F1 x))))
    -> (refines (sfixDS F2) (G (sfixDS F1))).
Proof.
  unfold sfixDS. intros; 
  apply sfix2_R_G_sfix1 
     with (R1:=refines (A:=A1)) (R2:=(refines (A:=A2)));
  auto with lattice.  
Qed.

End SFixDS_Morph.
 
Definition lub (A:Type) (P:(DS A)->Prop) : (DS A) 
 := (bind (any (DS A)) (fun sp =>
          bind (ensure (P sp)) (fun _ => sp))).

Lemma lub_refined:
  forall (A:Type) (P:(DS A)->Prop) (sp:(DS A)),
    (P sp) -> (refines sp (lub P)).
Proof.
  intros A P sp H; unfold lub.
  apply bind_simpl_right; simpl.
  intros; constructor 1 with (x:=sp).
  apply bind_simpl_right; simpl.
  intuition auto with wprefines.    
Qed.

Lemma lub_refines:
  forall (A:Type) (P:(DS A)->Prop) (sp:(DS A)),
    (forall (sp1:DS A), P sp1 -> (refines sp1 sp)) 
        -> (refines (lub P) sp).
Proof.
  intros A P sp H; unfold lub.
  apply bind_simpl_left; simpl.
  intros; apply bind_simpl_left; simpl.
  intuition auto.
Qed.

Lemma FSL_sup_semilattice:
  forall (A:Type), (sup_semilattice (refines (A:=A)) (lub (A:=A))).
Proof.
  intros; constructor 1; constructor 1.
  unfold Relation_Definitions.reflexive; auto with wprefines.
  unfold Relation_Definitions.transitive. apply refines_trans.
  apply lub_refined.
  apply lub_refines.
Qed.

Hint Resolve FSL_sup_semilattice: lattice.

Section Most_Abort.

Variable A: Type.

Section BFixDS_Def.

Variable F: (DS A) -> (DS A).

Hypothesis F_monotonic:
  forall (sp1 sp2:(DS A)),
    (refines sp1 sp2) -> (refines (F sp1) (F sp2)).

Definition bfixDS: (DS A) :=
  (bfix (refines (A:=A)) (lub (A:=A)) F).

Lemma bfixDS_mostabort:
  forall (sp:(DS A)),
    (refines sp (F sp))
      -> (refines sp bfixDS).
Proof.
  unfold bfixDS. intros; apply bfix_biggest with (R:=refines (A:=A)) ;
  auto with lattice.
Qed.
  
Lemma bfixDS_prefix:
  (refines (F bfixDS) bfixDS).
Proof.
  unfold bfixDS. intros; apply bfix_prefix with (R:=refines (A:=A)) ;
  auto with lattice.
Qed.

Lemma bfixDS_postfix:
  (refines bfixDS (F bfixDS)).
Proof.
  unfold bfixDS. intros; apply bfix_postfix with (R:=refines (A:=A)) ;
  auto with lattice.
Qed.

Lemma iterFbfixDS_refines_bfixDS:
  forall (n:nat), (refines (Iter.iter n F bfixDS) bfixDS).
Proof.
  unfold bfixDS. intros; apply iterFbfix_R_bfix with (R:=refines (A:=A)) ;
  auto with lattice.  
Qed.
                                                                                         
Lemma bfixDS_refines_iterFbfixDS:
  forall (n:nat), (refines bfixDS (Iter.iter n F bfixDS)).
Proof.
  unfold bfixDS. intros; apply bfix_R_iterFbfix with (R:=refines (A:=A)) ;
  auto with lattice.  
Qed.

End BFixDS_Def.

Section BFix_monotony.
                                                                                         
Variable F1 F2: (DS A) -> (DS A).
Hypothesis F1_monotony: forall (x y:DS A), (refines x y) -> (refines (F1 x) (F1 y)).
(*Hypothesis F2_monotony: forall (x y:DS A), (refines x y) -> (refines (F2 x) (F2 y)).*)
                                                                                         
Lemma bfixDS_monotony:
  (forall (x:DS A), (refines (F1 x) (F2 x))) -> (refines (bfixDS F1) (bfixDS F2)).
Proof.
  unfold bfixDS. intros; apply bfix_monotony with (R:=refines (A:=A)) ;
  auto with lattice.  
Qed.
                                                                                         
End BFix_monotony.

End Most_Abort.

Section BFixDS_Morph.
                                                                                         
Variable A1 A2: Type.
                                                                                        
Variable F1: DS A1 -> DS A1.
Hypothesis F1_monotony: forall (x y:DS A1), (refines x y) -> (refines (F1 x) (F1 y)).
                                                                                         
Variable F2: DS A2 -> DS A2.
(*Hypothesis F2_monotony: forall (x y:DS A2), (refines x y) -> (refines (F2 x) (F2 y)).*)
                                                                                         
Variable G: DS A1 -> DS A2.
Hypothesis G_monotony: forall (x y:DS A1), (refines x y) -> (refines (G x) (G y)).
                                                                                         
Lemma G_bfixDS1_refines_bfixDS2:
 (forall (x: DS A1), (refines (G (F1 x))) (F2 (G x)))
    -> (refines (G (bfixDS F1)) (bfixDS F2)).
Proof.
  unfold bfixDS. intros; 
  apply G_bfix1_R_bfix2 
     with (R1:=refines (A:=A1)) (R2:=(refines (A:=A2)));
  auto with lattice.  
Qed.

End BFixDS_Morph.
 
Section SFixF.

Variable A B: Type.

Definition refinesF (f1 f2: A -> DS B) := 
  forall (x:A), refines (f1 x) (f2 x).

Definition glbF (P:(A->DS B)->Prop) (x:A) : (DS B) 
 := (sync (fun sp => bind (require (P sp)) (fun _ => sp x))).

Lemma glbF_refines:
  forall (P:(A->DS B)->Prop) (sp:A -> DS B),
    (P sp) -> (refinesF (glbF P) sp).
Proof.
  intros P sp H x; unfold glbF.
  apply sync_elim with (a:=sp).
  apply bind_simpl_left; simpl.
  intuition auto with wprefines.    
Qed.

Lemma glbF_refined:
  forall (P:(A -> DS B)->Prop) (sp:A->DS B),
    (forall (sp1:A -> DS B), P sp1 -> (refinesF sp sp1)) 
        -> (refinesF sp (glbF P)).
Proof.
  unfold refinesF; intros P sp H x; unfold glbF.
  apply sync_intro.
  intros; apply bind_simpl_right; simpl.
  intuition auto.
Qed.

Lemma FSL_F_inf_semilattice: inf_semilattice refinesF glbF.
Proof.
  unfold refinesF, glbF; intros; constructor 1; constructor 1.
  unfold Relation_Definitions.reflexive; auto with wprefines.
  unfold Relation_Definitions.transitive.  
  intros; eapply refines_trans; eauto.
  apply glbF_refines.
  apply glbF_refined.
Qed.

Hint Resolve FSL_F_inf_semilattice: lattice.

Section SFixF_Def.

Variable F: (A->DS B) -> (A -> DS B).

Hypothesis F_monotonic:
  forall (sp1 sp2:A->DS B),
    (forall x, refines (sp1 x) (sp2 x)) -> 
    (forall x, refines (F sp1 x) (F sp2 x)).

Definition sfixF: A-> (DS B) :=
  (sfix (refinesF) (glbF) F).

Lemma sfixF_mostmagic:
  forall (fixO:A -> (DS B)),
    (forall x, refines (F fixO x) (fixO x))
      -> forall x, refines (sfixF x) (fixO x).
Proof.
  intros sp H. 
  change (refinesF sfixF sp). 
  unfold sfixF; apply sfix_smallest with (R:=refinesF).
  auto with lattice.
  unfold refinesF; auto.
Qed.
  
Lemma sfixF_prefix:
  forall x, refines (F sfixF x) (sfixF x).
Proof.
  change (refinesF (F sfixF) sfixF).
  unfold sfixF.
  apply sfix_prefix with (R:=refinesF) ;
  auto with lattice.
Qed.

Lemma sfixF_postfix:
  forall x, refines (sfixF x) (F sfixF x).
Proof.
  change (refinesF sfixF (F sfixF)).
  unfold sfixF; apply sfix_postfix with (R:=refinesF) ;
  auto with lattice.
Qed.

Require Export Iter.

Lemma itersfixF_prefix:
  forall (n:nat) (x:A), 
     (refines (iter n F sfixF x) (sfixF x)). 
Proof.
  induction n; simpl; auto with wprefines.
  intros x; eapply refines_trans.
  2: apply sfixF_prefix.
  auto.
Qed.

Lemma itersfixF_postfix:
  forall (n:nat) (x:A), 
     (refines (sfixF x) (iter n F sfixF x)). 
Proof.
  induction n; simpl; auto with wprefines.
  intros x; eapply refines_trans.
  apply sfixF_postfix.
  auto.
Qed.

End SFixF_Def.

Section SFixF_monotony.
                                                                                         
Variable F1 F2: (A -> (DS B)) -> (A -> (DS B)).
(*Hypothesis F1_monotony: forall (f g: A->DS B), (forall x, refines (f x) (g x)) -> (forall x, refines (F1 f x) (F1 g x)).*)
Hypothesis F2_monotony: forall (f g: A->DS B), (forall x, refines (f x) (g x)) -> (forall x, refines (F2 f x) (F2 g x)).
                                                                                         
Lemma sfixF_monotony:
  (forall (f:A->DS B) (x:A), refines (F1 f x) (F2 f x)) -> forall x,refines (sfixF F1 x) (sfixF F2 x).
Proof.
  intros H; change (refinesF (sfixF F1) (sfixF F2)).
  unfold sfixF.  apply sfix_monotony with (R:=refinesF) ; auto with lattice.  
Qed.
                                                                                         
End SFixF_monotony.

End SFixF.

Section SFixF_Morph.

Hint Resolve FSL_F_inf_semilattice: lattice.
                                                                             
Variable A1 A2 B1 B2: Type.
                                                                                        
Variable F1: (A1 -> DS B1) -> A1 -> (DS B1).
Hypothesis F1_monotony: forall (f g:A1 -> DS B1), (forall x, refines (f x) (g x)) -> forall x, refines (F1 f x) (F1 g x).
                                                                                         
Variable F2: (A2 -> DS B2) -> A2 -> (DS B2).
(*Hypothesis F2_monotony: forall (f g:A2 -> DS B2), (forall x, refines (f x) (g x)) -> forall x, refines (F2 f x) (F2 g x).*)
                                                                                         
Variable G: (A1 -> DS B1) -> A2 -> (DS B2).
Hypothesis G_monotony: forall (f g:A1 -> DS B1), (forall x, refines (f x) (g x)) -> forall x, refines (G f x) (G g x).
                                                                                         
Lemma sfixF2_refines_G_sfixF1:
 (forall (f: A1 -> DS B1) (x:A2), refines (F2 (G f) x) (G (F1 f) x))
    -> forall x, refines (sfixF F2 x) (G (sfixF F1) x).
Proof.
  intros H; change (refinesF (sfixF F2) (G (sfixF F1))). 
  unfold sfixF; 
  apply sfix2_R_G_sfix1 
     with (R1:=refinesF (A:=A1) (B:=B1)) (R2:=refinesF (A:=A2) (B:=B2));
  auto with lattice.  
Qed.

End SFixF_Morph.

Section BFixF.

Variable A B: Type.

Definition lubF (P:(A -> DS B)->Prop) (x:A): (DS B)
 := bind (any (A -> DS B)) (fun sp =>
          bind (ensure (P sp)) (fun _ => sp x)).

Lemma lubF_refined:
  forall (P:(A->DS B)->Prop) (sp:A -> DS B),
    (P sp) -> (refinesF sp (lubF P)).
Proof.
  intros P sp H x; unfold lubF.
  apply bind_simpl_right; simpl.
  intros; constructor 1 with (x:=sp).
  apply bind_simpl_right; simpl.
  intuition auto with wprefines.    
Qed.

Lemma lubF_refines:
  forall (P:(A -> DS B)->Prop) (sp:A->DS B),
    (forall (sp1:A -> DS B), P sp1 -> (refinesF sp1 sp)) 
        -> (refinesF (lubF P) sp).
Proof.
  unfold refinesF; intros P sp H x; unfold lubF.
  apply bind_simpl_left; simpl.
  intros; apply bind_simpl_left; simpl.
  intuition auto.
Qed.

Lemma FSL_F_sup_semilattice: sup_semilattice (refinesF (A:=A) (B:=B)) lubF.
Proof.
  unfold refinesF, lubF; intros; constructor 1; constructor 1.
  unfold Relation_Definitions.reflexive; auto with wprefines.
  unfold Relation_Definitions.transitive.  
  intros; eapply refines_trans; eauto.
  apply lubF_refined.
  apply lubF_refines.
Qed.

Hint Resolve FSL_F_sup_semilattice: lattice.


Section BFixF_Def.

Variable F: (A->DS B) -> (A -> DS B).

Hypothesis F_monotonic:
  forall (sp1 sp2:A->DS B),
    (forall x, refines (sp1 x) (sp2 x)) -> 
    (forall x, refines (F sp1 x) (F sp2 x)).

Definition bfixF: A-> (DS B) :=
  (bfix (refinesF (A:=A) (B:=B)) (lubF) F).

Lemma bfixF_mostabort:
  forall (sp:A -> (DS B)),
    (forall x, refines (sp x) (F sp x))
      -> forall x, refines (sp x) (bfixF x).
Proof.
  intros sp H. 
  change (refinesF sp bfixF). 
  unfold bfixF; apply bfix_biggest with (R:=refinesF (A:=A) (B:=B)).
  auto with lattice.
  unfold refinesF; auto.
Qed.
  
Lemma bfixF_prefix:
  forall x, refines (F bfixF x) (bfixF x).
Proof.
  change (refinesF (F bfixF) bfixF).
  unfold bfixF.
  apply bfix_prefix with (R:=refinesF (A:=A) (B:=B)) ;
  auto with lattice.
Qed.

Lemma bfixF_postfix:
  forall x, refines (bfixF x) (F bfixF x).
Proof.
  change (refinesF bfixF (F bfixF)).
  unfold bfixF; apply bfix_postfix with (R:=refinesF (A:=A) (B:=B)) ;
  auto with lattice.
Qed.

End BFixF_Def.

Section BFixF_monotony.
                                                                                         
Variable F1 F2: (A -> (DS B)) -> (A -> (DS B)).
Hypothesis F1_monotony: forall (f g: A->DS B), (forall x, refines (f x) (g x)) -> (forall x, refines (F1 f x) (F1 g x)).
(*Hypothesis F2_monotony: forall (f g: A->DS B), (forall x, refines (f x) (g x)) -> (forall x, refines (F2 f x) (F2 g x)).*)
                                                                                         
Lemma bfixF_monotony:
  (forall (f:A->DS B) (x:A), refines (F1 f x) (F2 f x)) -> forall x,refines (bfixF F1 x) (bfixF F2 x).
Proof.
  intros H; change (refinesF (bfixF F1) (bfixF F2)).
  unfold bfixF.  apply bfix_monotony with (R:=refinesF (A:=A) (B:=B)) ; auto with lattice.  
Qed.
                                                                                         
End BFixF_monotony.

End BFixF.

Section BFixF_Morph.

Hint Resolve FSL_F_sup_semilattice: lattice.
                                                                             
Variable A1 A2 B1 B2: Type.
                                                                                        
Variable F1: (A1 -> DS B1) -> A1 -> (DS B1).
Hypothesis F1_monotony: forall (f g:A1 -> DS B1), (forall x, refines (f x) (g x)) -> forall x, refines (F1 f x) (F1 g x).
                                                                                         
Variable F2: (A2 -> DS B2) -> A2 -> (DS B2).
(*Hypothesis F2_monotony: forall (f g:A2 -> DS B2), (forall x, refines (f x) (g x)) -> forall x, refines (F2 f x) (F2 g x).*)
                                                                                         
Variable G: (A1 -> DS B1) -> A2 -> (DS B2).
Hypothesis G_monotony: forall (f g:A1 -> DS B1), (forall x, refines (f x) (g x)) -> forall x, refines (G f x) (G g x).
                                                                                         
Lemma G_sfixF1_refines_bfixF2:
 (forall (f: A1 -> DS B1) (x:A2), refines (G (F1 f) x) (F2 (G f) x))
    -> forall x, refines (G (bfixF F1) x) (bfixF F2 x) .
Proof.
  intros H; change (refinesF (G (bfixF F1)) (bfixF F2)). 
  unfold bfixF; 
  apply G_bfix1_R_bfix2 
     with (R1:=refinesF (A:=A1) (B:=B1)) (R2:=refinesF (A:=A2) (B:=B2));
  auto with lattice.  
Qed.

End BFixF_Morph.

(* abstraction of a term in an other term *)
Ltac lambda v t :=
  match t with
  | ?f v => f
  | ?f ?x v => (f x)
  | _ => match (eval pattern v in t) with
            ?f v => f
            end
end.

Ltac sfixF_mmagic :=
  match goal with
    |- (refines (sfixF _ ?x) ?t) => 
          let t1:=(lambda x t) in
            apply sfixF_mostmagic with (fixO:=t1)
            || idtac "Tactic too weak: try something else..."
    | _ => idtac "User Error: goal does not match"
    end.
