(** ** Module [ISL]: specifications with a local state.

This module defines operators for a Dijsktra specification language
with state and exception handling. Monad operators are extended to
deal with local states.
*)

(* begin hide *)
Set Implicit Arguments.

Require Export FSL.
Require Export Place.
(* end hide *)

(** The type of specification is also parametrized by [St] the local
type of states and [A] the type of the result. This definition is the
standard implementation of monad with a state.
*)

Definition DS (St A:Type) := St -> WP.DS (prodT St A).  

(** This definition uses [WP.DS] the [DS] type defined in module [WP].
Assuming that [st] is a state and [sp] a specification, the term
[(pre (sp st))] is the precondition of [sp] on the initial state [st].
The same remarks applies for [(sdemon (sp st))] and [(angel (sp st))].
*)
 
(* begin hide *)
Delimit Scope ISL_scope with ISL. 
Bind Scope ISL_scope with DS.
(* end hide *)

(** I first define [refInEnv] the relation of refinement in a
environment, using [refines] relation of [WP]. 
*)

Definition refInEnv  (A St: Type) (st:St) (sp1 sp2 : DS St A) : Prop :=
  WP.refines (sp1 st) (sp2 st).

(** For a given [st], [(refInEnv st)] is reflexive and transitive.
*)
Lemma refInEnv_refl :
 forall (A St: Type) (st: St) (sp : DS St A), refInEnv st sp sp.
Proof.
  unfold refInEnv in |- *; auto with wprefines.
Qed.

Lemma refInEnv_trans :
 forall (A St: Type) (st:St) (sp1 sp2 sp3 : DS St A),
 refInEnv st sp1 sp2 -> refInEnv st sp2 sp3 -> refInEnv st sp1 sp3.
Proof.
  unfold refInEnv in |- *; eauto with wprefines.
Qed.

(** This lemma is the proof rule stating how a refinement goal
can be simplified by a wp-computation.
*)
Lemma refInEnv2wp :
 forall (A St: Type) (st:St) (sp1 sp2 : DS St A),
 (pre (sp2 st) -> 
  sdemon (sp1 st) (fun a => post (sp2 st) a)) -> 
 refInEnv st sp1 sp2.
Proof.
  unfold refInEnv in |- *; auto with wprefines.
Qed.

(** The following lemma is the converse of the preceding one.
*)
Lemma refInEnv_from_wp :
 forall (A St: Type) (st:St) (sp1 sp2 : DS St A),
 refInEnv st sp1 sp2 -> 
 pre (sp2 st) -> 
 sdemon (sp1 st) (fun a => post (sp2 st) a).
Proof.
  unfold refInEnv in |- *; intros A St st sp1 sp2 H H1;
   apply (WP.refines_unfold H); auto.
Qed.

(** In a refinement proof, 
the precondition of the refined specification can still be assumed.
*)
Lemma refInEnv_pre_intro:
  forall (A St: Type) (st:St) (sp1 sp2 : DS St A),
   ((pre (sp2 st)) -> refInEnv st sp1 sp2) -> refInEnv st sp1 sp2.
Proof.
  intros; unfold refInEnv; apply refines_pre_intro; auto.
Qed.

(** Below, I define [val] and [raise] operators which
corresponds to their homonym in the algebraic semantics.
*)

Definition val (A St: Type) (a : A) : DS St A :=
  fun st => FSL.val (pairT st a).

(** I extend here [set] and [get] operators of monad, with a place
parameter.  Hence [set] and [get] are close to the usual constructions
of programming language, allowing to affect a reference or to derefence.
*)

Definition set (A St :Type) (v:(Place St A)) (x : A) : DS St unit :=
  fun st => FSL.val (pairT (write v st x) tt).

Definition get (A St:Type) (v:(Place St A)) : DS St A := 
  fun st => FSL.val (pairT st (read v st)).

(** The major change with respect to the usual monadic operators is [bind].
*)
Definition bind (A B St1 St2: Type) (v:(Place St1 St2)) (m : DS St2 A) (f : A -> DS St1 B) : 
  DS St1 B :=
  fun st =>
  FSL.bind (m (read v st))
    (fun p => let (st',a):=p in f a (write v st st')).

Lemma bind_monotony :
 forall (A B St1 St2: Type) (v1 v2:(Place St1 St2)) (sp1 sp2 : DS St2 A) (f1 f2 : A -> DS St1 B) (st : St1),
 (eqPl v1 v2) -> 
 refInEnv (read v2 st) sp1 sp2 -> 
 (forall (st2 : St2) (x : A), refInEnv (write v2 st st2) (f1 x) (f2 x)) -> 
 refInEnv st (bind v1 sp1 f1) (bind v2 sp2 f2).
Proof.
  unfold refInEnv, bind in |- *; intros.
  apply FSL.bind_monotony.
  rewrite eqPl_read with (v2:=v2); auto with wprefines.
  intros p; case p; simpl in |- *;
  intros; rewrite eqPl_write with (v2:=v2); 
  auto with wprefines.
Qed.

(* begin hide *)  
Lemma bind_monotonyx :
 forall (A B St1 St2: Type) (v:(Place St1 St2)) (sp1 sp2 : DS St2 A) (f1 f2 : A -> DS St1 B) (st : St1),
 refInEnv (read v st) sp1 sp2 -> 
 (forall (st2 : St2) (x : A), refInEnv (write v st st2) (f1 x) (f2 x)) -> 
 refInEnv st (bind v sp1 f1) (bind v sp2 f2).
Proof.
  intros; apply bind_monotony; auto with place.
Qed.

Hint Extern 0 (WP.refines _ _) => refine (FSL.bind_assoc_ref1 _ _ _): wprefines.
Hint Resolve refInEnv_refl bind_monotony: isl.
(* end hide *)


(** Here is a notation for IF-Then-Else *)
Notation "'If' [ v ]  pb 'Then' sp1 'Else' sp2" :=
  (bind v pb (bool_rect (fun _ => (DS _ _)) sp1 sp2))
  (at level 20, v, pb, sp1, sp2 at level 90): ISL_scope.

Lemma if_monotony :
 forall (A St1 St2 : Type) (v: (Place St1 St2)) (spb1 spb2 : DS St2 bool) (sp1 sp2 sp3 sp4 : DS St1 A) (st : St1),
 refInEnv (read v st) spb1 spb2 -> 
 (forall st2 : St2, refInEnv (write v st st2) sp1 sp2) -> 
 (forall st2 : St2, refInEnv (write v st st2) sp3 sp4) -> 
 refInEnv st
   (If[ v] spb1 Then sp1 Else sp3)%ISL
   (If[ v] spb2 Then sp2 Else sp4)%ISL.
Proof.
  simpl in |- *; intros; apply bind_monotonyx; auto with isl.
  intros st0 x; case x; auto with isl.
Qed.

(** [bind] is associative. It is expressed by the two following lemmas *)
Lemma bind_assoc_ref1 :
 forall (A B C St1 St2 St3: Type) (v1: (Place St2 St1)) (sp1 : DS St1 A) 
  (v2: (Place St3 St2)) (sp2 : A -> DS St2 B) (sp3 : B -> DS St3 C) (st : St3),
 refInEnv st (bind v2 (bind v1 sp1 sp2) sp3) 
   (bind (v2#v1) sp1 (fun a => bind v2 (sp2 a) sp3)).
Proof.
  unfold bind, refInEnv in |- *.
  intros A B C St1 St2 St3 v1 sp1 v2 sp2 sp3 st.
  eapply WP.refines_trans.
  apply bind_assoc_ref1.
  apply FSL.bind_monotony; auto with wprefines.
  intros a; case a; case a; simpl; intros.
  intros; apply FSL.bind_monotony; auto with wprefines; 
  autorewrite with placeeq; auto with wprefines. 
  intros b; case b; case b; simpl; intros.
  autorewrite with placeeq; auto with wprefines. 
Qed.

(* begin hide *)
Hint Extern 0 (WP.refines _ _) => refine (FSL.bind_assoc_ref2 _ _ _): wprefines.
(* end hide *)

Lemma bind_assoc_ref2 :
 forall (A B C St1 St2 St3: Type) (v1: (Place St2 St1)) (sp1 : DS St1 A) 
  (v2: (Place St3 St2)) (sp2 : A -> DS St2 B) (sp3 : B -> DS St3 C) (st : St3),
   refInEnv st (bind (v2#v1) sp1 (fun a => bind v2 (sp2 a) sp3))
               (bind v2 (bind v1 sp1 sp2) sp3).
Proof.
  unfold bind, refInEnv in |- *.
  intros A B C St1 St2 St3 v1 sp1 v2 sp2 sp3 st.
  eapply WP.refines_trans.
  2: apply bind_assoc_ref2.
  apply FSL.bind_monotony; auto with wprefines.
  intros a; case a; case a; simpl; intros.
  intros; apply FSL.bind_monotony; auto with wprefines; 
  autorewrite with placeeq; auto with wprefines. 
  intros b; case b; case b; simpl; intros.
  autorewrite with placeeq; auto with wprefines. 
Qed.

(* begin hide *)
Lemma bind_changeV_left:
 forall (A B St1 St2: Type) (v1 v2:(Place St1 St2)) (sp1 : DS St2 A) (sp2 : A -> DS St1 B) (sp3: (DS St1 B)) 
   (st : St1),
     (eqPl v1 v2) -> 
       (refInEnv st (bind v1 sp1 sp2) sp3)
       -> (refInEnv st (bind v2 sp1 sp2) sp3).
Proof.
  intros; apply refInEnv_trans with (sp2:=(bind v1 sp1 sp2)); auto with isl place.
Qed.
 
Lemma bind_changeV_right:
 forall (A B St1 St2: Type) (v1 v2:(Place St1 St2)) (sp1 : DS St2 A) (sp2 : A -> DS St1 B) (sp3: (DS St1 B)) 
   (st : St1),
     (eqPl v1 v2) -> 
       (refInEnv st sp3 (bind v1 sp1 sp2))
       -> (refInEnv st sp3 (bind v2 sp1 sp2)).
Proof.
  intros; apply refInEnv_trans with (sp2:=(bind v1 sp1 sp2)); auto with isl place.
Qed.

Lemma bind_val_r_ref1:
  forall (St A: Type) (sp1 sp2: DS St A) (sp3:A -> (DS St A)) (st : St),
    (refInEnv st sp1 sp2)
      -> (forall (a:A) (st:St), (refInEnv st (sp3 a) (val a)))   
         ->  (refInEnv st (bind allPl sp1 sp3) sp2).
Proof.
  intros St A sp1 sp2 sp3 st H1 H2; 
  apply refInEnv_trans with (sp2:=(bind allPl sp1 (fun a => (val a)))).
  apply bind_monotonyx; simpl; auto with isl.
  clear H2; generalize (refInEnv_from_wp H1); intros H2.
  apply refInEnv2wp; simpl.
  intros H3; eapply sdemon_weaken.
  apply H2; auto.
  intros b; case b; simpl; auto.
Qed. 

Lemma bind_val_r_ref2:
  forall (St A: Type) (sp1 sp2: DS St A) (sp3:A -> (DS St A)) (st : St),
    (refInEnv st sp2 sp1)
      -> (forall (a:A) (st:St), (refInEnv st (val a) (sp3 a)))   
         ->  (refInEnv st sp2 (bind allPl sp1 sp3)).
Proof.
  intros A St sp1 sp2 sp3 st H1 H2; 
  apply refInEnv_trans with (sp2:=(bind allPl sp1 (fun a => (val a)))).
  clear H2; generalize (refInEnv_from_wp H1); intros H2.
  apply refInEnv2wp; simpl.
  intros H3; eapply sdemon_weaken.
  apply H2.
  eauto with wipi.
  simpl; intros b H4.
  eapply angel_weaken.
  apply H4.
  intros b0 H5; subst.
  case b0; simpl; auto.  
  apply bind_monotonyx; auto with isl.
Qed. 

Lemma bind_left_unassoc :
 forall (A B C St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (sp1 : DS St3 A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind (v1#v2) sp1 (fun a => bind v1 (sp2 a) sp3)) sp4 -> 
 refInEnv st (bind v1 (bind v2 sp1 sp2) sp3) sp4.
Proof.
  intros; refine (refInEnv_trans (bind_assoc_ref1 _ _ _ _ _ _) _); auto.
Qed.

Lemma bind_left_unassoc_allPl1 :
 forall (A B C St1 St3: Type) (v2:(Place St1 St3)) (sp1 : DS St3 A) (sp2 : A -> DS St1 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind v2 sp1 (fun a => bind allPl (sp2 a) sp3)) sp4 -> 
 refInEnv st (bind allPl (bind v2 sp1 sp2) sp3) sp4.
Proof.
  intros; apply bind_left_unassoc.
  eapply bind_changeV_left. 2:eauto. auto with place.
Qed.

Lemma bind_left_unassoc_allPl2 :
 forall (A B C St1 St2: Type) (v1:(Place St1 St2)) (sp1 : DS St2 A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind v1 sp1 (fun a => bind v1 (sp2 a) sp3)) sp4 -> 
 refInEnv st (bind v1 (bind allPl sp1 sp2) sp3) sp4.
Proof.
  intros; apply bind_left_unassoc.
  eapply bind_changeV_left. 2:eauto. auto with place.  
Qed.

Lemma bind_left_unassoc_nonePl2 :
 forall (A B C St1 St2: Type) (v1:(Place St1 St2)) (v2:(Place St2 unit)) (sp1 : DS unit A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind nonePl sp1 (fun a => bind v1 (sp2 a) sp3)) sp4 -> 
 refInEnv st (bind v1 (bind v2 sp1 sp2) sp3) sp4.
Proof.
  intros; apply bind_left_unassoc.
  eapply bind_changeV_left. 2:eauto. auto with place.  
Qed.

Lemma bind_left_assoc :
 forall (A B C St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (v:(Place St1 St3)) (sp1 : DS St3 A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 (eqPl v (v1#v2)) -> 
 refInEnv st (bind v1 (bind v2 sp1 sp2) sp3) sp4 -> 
 refInEnv st (bind v sp1 (fun a => bind v1 (sp2 a) sp3)) sp4.
Proof.
  intros; apply bind_changeV_left with (v1:=(v1#v2)); auto with place.
  auto with isl.
  refine (refInEnv_trans (bind_assoc_ref2 _ _ _ _ _ _) _); auto.
Qed.

Lemma bind_left_assoc_allPl1 :
 forall (A B C St1 St3: Type) (v2:(Place St1 St3)) (sp1 : DS St3 A) (sp2 : A -> DS St1 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind allPl (bind v2 sp1 sp2) sp3) sp4 -> 
 refInEnv st (bind v2 sp1 (fun a => bind allPl (sp2 a) sp3)) sp4.
Proof.
  intros; eapply bind_left_assoc.
  2: eauto.
  auto with place.
Qed.

Lemma bind_left_assoc_allPl2 :
 forall (A B C St1 St2: Type) (v1:(Place St1 St2)) (sp1 : DS St2 A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind v1 (bind allPl sp1 sp2) sp3) sp4 -> 
 refInEnv st (bind v1 sp1 (fun a => bind v1 (sp2 a) sp3)) sp4.
Proof.
  intros; eapply bind_left_assoc.
  2: eauto.
  auto with place.
Qed.

Lemma bind_left_assoc_nonePl2 :
 forall (A B C St1 St2: Type) (v1:(Place St1 St2)) (v2:(Place St1 unit)) (sp1 : DS unit A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind v1 (bind nonePl sp1 sp2) sp3) sp4 -> 
 refInEnv st (bind v2 sp1 (fun a => bind v1 (sp2 a) sp3)) sp4.
Proof.
  intros; eapply bind_left_assoc.
  2: eauto.
  auto with place.
Qed.

Lemma bind_right_unassoc :
 forall (A B C St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (sp1 : DS St3 A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind (v1#v2) sp1 (fun a => bind v1 (sp2 a) sp3)) -> 
 refInEnv st sp4 (bind v1 (bind v2 sp1 sp2) sp3).
Proof.
  intros; refine (refInEnv_trans _ (bind_assoc_ref2 _ _ _ _ _ _)); auto.
Qed.

Lemma bind_right_unassoc_allPl1 :
 forall (A B C St1 St3: Type) (v2:(Place St1 St3)) (sp1 : DS St3 A) (sp2 : A -> DS St1 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind v2 sp1 (fun a => bind allPl (sp2 a) sp3)) -> 
 refInEnv st sp4 (bind allPl (bind v2 sp1 sp2) sp3).
Proof.
  intros; apply bind_right_unassoc.
  eapply bind_changeV_right. 2:eauto. auto with place.  
Qed.

Lemma bind_right_unassoc_allPl2 :
 forall (A B C St1 St2: Type) (v1:(Place St1 St2)) (sp1 : DS St2 A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind v1 sp1 (fun a => bind v1 (sp2 a) sp3)) -> 
 refInEnv st sp4 (bind v1 (bind allPl sp1 sp2) sp3).
Proof.
  intros; apply bind_right_unassoc.
  eapply bind_changeV_right. 2:eauto. auto with place.  
Qed.

Lemma bind_right_unassoc_nonePl2 :
 forall (A B C St1 St2: Type) (v1:(Place St1 St2)) (v2:(Place St2 unit)) (sp1 : DS unit A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind nonePl sp1 (fun a => bind v1 (sp2 a) sp3)) -> 
 refInEnv st sp4 (bind v1 (bind v2 sp1 sp2) sp3).
Proof.
  intros; apply bind_right_unassoc.
  eapply bind_changeV_right. 2:eauto. auto with place.  
Qed.

Lemma bind_right_assoc :
 forall (A B C St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (v:(Place St1 St3)) (sp1 : DS St3 A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 (eqPl v (v1#v2)) -> 
 refInEnv st sp4 (bind v1 (bind v2 sp1 sp2) sp3) -> 
 refInEnv st sp4 (bind v sp1 (fun a => bind v1 (sp2 a) sp3)).
Proof.
  intros; apply bind_changeV_right with (v1:=(v1#v2)); auto with place.
  refine (refInEnv_trans _ (bind_assoc_ref1 _ _ _ _ _ _)); auto.
Qed.

Lemma bind_right_assoc_allPl1 :
 forall (A B C St1 St3: Type) (v2:(Place St1 St3)) (sp1 : DS St3 A) (sp2 : A -> DS St1 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind allPl (bind v2 sp1 sp2) sp3) -> 
 refInEnv st sp4 (bind v2 sp1 (fun a => bind allPl (sp2 a) sp3)).
Proof.
 intros; eapply bind_right_assoc.
 2: eauto.
 auto with place.
Qed.

Lemma bind_right_assoc_allPl2 :
 forall (A B C St1 St2: Type) (v1:(Place St1 St2)) (sp1 : DS St2 A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind v1 (bind allPl sp1 sp2) sp3) -> 
 refInEnv st sp4 (bind v1 sp1 (fun a => bind v1 (sp2 a) sp3)).
Proof.
 intros; eapply bind_right_assoc.
 2: eauto.
 auto with place.
Qed.

Lemma bind_right_assoc_nonePl2 :
 forall (A B C St1 St2: Type) (v1:(Place St1 St2)) (v2:(Place St1 unit)) (sp1 : DS unit A) (sp2 : A -> DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind v1 (bind nonePl sp1 sp2) sp3) -> 
 refInEnv st sp4 (bind v2 sp1 (fun a => bind v1 (sp2 a) sp3)).
Proof.
 intros; eapply bind_right_assoc.
 2: eauto.
 auto with place.
Qed.
(* end hide *)

(** Now, let us see "Proof by step-by-step execution" lemmas 
*) 

Lemma bind_simpl_left :
 forall (A B St1 St2: Type) (v:(Place St1 St2)) (sp1 : DS St2 A) (sp2 : A -> DS St1 B) (sp3 : DS St1 B) (st : St1),
 sdemon (sp1 (read v st))
   (fun p => let (st',a):=p in refInEnv st (bind allPl (set v st') (fun _ => sp2 a)) sp3) 
    -> refInEnv st (bind v sp1 sp2) sp3. 
Proof.
  simpl in |- *; unfold refInEnv, bind in |- *; intros.
  refine (FSL.bind_simpl_left _ _ _ _); auto.
  simpl in |- *; intros; eapply sdemon_weaken. eauto.
  simpl in |- *; intros p; case p; case p; simpl in |- *; auto with wprefines.
Qed.

(** Here is a "pure" version of the preceding lemma (when "v" is nonePl) *)

Lemma bind_simplPure_left :
 forall (A B St: Type) (v:(Place St unit)) (sp1 : DS unit A) (sp2 : A -> DS St B) (sp3 : DS St B) (st : St),
 sdemon (sp1 tt)
   (fun p => refInEnv st (sp2 (sndT p)) sp3) -> refInEnv st (bind v sp1 sp2) sp3. 
Proof.
  simpl in |- *; intros; apply bind_changeV_left with (v1:=nonePl (A:=St)); auto with place.
  unfold refInEnv, bind in * |- *; intros.
  refine (FSL.bind_simpl_left _ _ _ _); auto.
  simpl in |- *; intros; eapply sdemon_weaken. eauto.
  simpl in |- *; intros p; case p; simpl in |- *; auto with wprefines.
Qed.

Lemma bind_simpl_right :
 forall (A B St1 St2: Type) (v:(Place St1 St2)) (sp1 : DS St2 A) (sp2 : A -> DS St1 B) (sp3 : DS St1 B) (st : St1),
 (pre (sp1 (read v st)) -> 
  angel (sp1 (read v st))
    (fun p => let (st',a):=p in
         refInEnv (write v st st') (bind allPl (set allPl st) (fun _ => sp3)) (sp2 a)
     )) -> refInEnv st sp3 (bind v sp1 sp2). 
Proof.
  simpl in |- *; unfold refInEnv, bind in |- *; intros.
  refine (FSL.bind_simpl_right _ _ _ _); auto.
  simpl in |- *; intros; eapply angel_weaken. eauto.
  simpl in |- *; intros p; case p; simpl in |- *; auto with wprefines.
Qed.

(* begin hide *)
Lemma bind_simplPure_right :
 forall (A B St: Type) (v:(Place St unit)) (sp1 : DS unit A) (sp2 : A -> DS St B) (sp3 : DS St B) (st : St),
 (pre (sp1 tt) -> 
  angel (sp1 tt)
    (fun p => refInEnv st sp3 (sp2 (sndT p)))) -> refInEnv st sp3 (bind v sp1 sp2). 
Proof.
  intros; apply bind_changeV_right with (v1:=nonePl (A:=St)); auto with place.
  unfold refInEnv, bind in * |- *; intros.
  refine (FSL.bind_simpl_right _ _ _ _); auto.
  simpl in |- *; intros; eapply angel_weaken. eauto.
  simpl in |- *; intros p; case p; simpl in |- *; auto with wprefines.
Qed.
(* end hide *)

Lemma bind_simpl_both :
 forall (A B St1 St2: Type) (v1 v2:(Place St1 St2)) (sp1 sp3 : DS St2 A) (sp2 sp4 : A -> DS St1 B) (st : St1),
 (eqPl v1 v2) -> 
 refInEnv (read v1 st) sp1 sp3 -> 
  (pre (sp1 (read v1 st)) -> 
   sdemon (sp1 (read v1 st))
     (fun p => let (st',a):=p in 
                 refInEnv (write v1 st st') (sp2 a) (sp4 a)))
   -> refInEnv st (bind v1 sp1 sp2) (bind v2 sp3 sp4).
Proof.
  simpl in |- *; unfold refInEnv, bind in |- *; intros.
  refine (FSL.bind_simpl_both _ _ _ _); auto.
  rewrite eqPl_read with (v1:=v2) (v2:=v1); auto with place.
  intros; eapply sdemon_weaken. eapply (H1 H2).
  simpl in |- *; intros p; case p; simpl in |- *; 
  intros; rewrite eqPl_write with (v1:=v2) (v2:=v1); auto with place wprefines.
Qed. 

(* begin hide *)
Lemma seq_left_assoc :
 forall (B C St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (v:(Place St1 St3)) (sp1 : DS St3 unit) (sp2 : DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 (eqPl v (v1#v2)) -> 
 refInEnv st (bind v1 (bind v2 sp1 (fun _: unit => sp2)) sp3) sp4 -> 
 refInEnv st (bind v sp1 (fun _: unit => bind v1 sp2 sp3)) sp4.
Proof.
  intros; apply bind_left_assoc with (v2:=v2) (sp2:=(fun _:unit => sp2));
  auto.
Qed.

Lemma seq_left_assoc_allPl1 :
 forall (B C St1 St3: Type) (v2:(Place St1 St3)) (sp1 : DS St3 unit) (sp2 : DS St1 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind allPl (bind v2 sp1 (fun _: unit => sp2)) sp3) sp4 -> 
 refInEnv st (bind v2 sp1 (fun _: unit => bind allPl sp2 sp3)) sp4.
Proof.
  intros; eapply seq_left_assoc.
  2: eauto.
  auto with place.
Qed.

Lemma seq_left_assoc_allPl2 :
 forall (B C St1 St2: Type) (v1:(Place St1 St2)) (sp1 : DS St2 unit) (sp2 : DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind v1 (bind allPl sp1 (fun _: unit => sp2)) sp3) sp4 -> 
 refInEnv st (bind v1 sp1 (fun _: unit => bind v1 sp2 sp3)) sp4.
Proof.
  intros; eapply seq_left_assoc.
  2: eauto.
  auto with place.
Qed.

Lemma seq_left_assoc_nonePl2 :
 forall (B C St1 St2: Type) (v1:(Place St1 St2)) (v2:(Place St1 unit)) (sp1 : DS unit unit) (sp2 : DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind v1 (bind nonePl sp1 (fun _: unit => sp2)) sp3) sp4 -> 
 refInEnv st (bind v2 sp1 (fun _: unit => bind v1 sp2 sp3)) sp4.
Proof.
  intros; eapply seq_left_assoc.
  2: eauto.
  auto with place.
Qed.

Lemma seq_right_assoc :
 forall (B C St1 St2 St3 : Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (v:(Place St1 St3)) (sp1 : DS St3 unit) (sp2 : DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 (eqPl v (v1#v2)) -> 
 refInEnv st sp4 (bind v1 (bind v2 sp1 (fun _: unit => sp2)) sp3) -> 
 refInEnv st sp4 (bind v sp1 (fun _: unit => bind v1 sp2 sp3)).
Proof.
  intros; apply bind_right_assoc with (v2:=v2) (sp2:=(fun _:unit => sp2)); auto.
Qed.

Lemma seq_right_assoc_allPl1 :
 forall (B C St1 St3 : Type) (v2:(Place St1 St3)) (sp1 : DS St3 unit) (sp2 : DS St1 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind allPl (bind v2 sp1 (fun _: unit => sp2)) sp3) -> 
 refInEnv st sp4 (bind v2 sp1 (fun _: unit => bind allPl sp2 sp3)).
Proof.
 intros; eapply seq_right_assoc.
 2: eauto.
 auto with place.
Qed.

Lemma seq_right_assoc_allPl2 :
 forall (B C St1 St2: Type) (v1:(Place St1 St2)) (sp1 : DS St2 unit) (sp2 : DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind v1 (bind allPl sp1 (fun _: unit => sp2)) sp3) -> 
 refInEnv st sp4 (bind v1 sp1 (fun _: unit => bind v1 sp2 sp3)).
Proof.
 intros; eapply seq_right_assoc.
 2: eauto.
 auto with place.
Qed.


Lemma seq_right_assoc_nonePl2 :
 forall (B C St1 St2: Type) (v1:(Place St1 St2)) (v2:(Place St1 unit)) (sp1 : DS unit unit) (sp2 : DS St2 B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind v1 (bind nonePl sp1 (fun _: unit => sp2)) sp3) -> 
 refInEnv st sp4 (bind v2 sp1 (fun _: unit => bind v1 sp2 sp3)).
Proof.
 intros; eapply seq_right_assoc.
 2: eauto.
 auto with place.
Qed.


Lemma bind_enter_left :
 forall (A B St1 St2: Type) (v:(Place St1 St2)) (sp1 sp2:DS St2 A) (sp3 : A -> DS St1 B) (sp4: DS St1 B) (st : St1),
 refInEnv (read v st) sp1 sp2 -> 
 refInEnv st (bind v sp2 sp3) sp4 -> refInEnv st (bind v sp1 sp3) sp4.
Proof.
  intros; eapply refInEnv_trans.
  2: eauto.
  auto with isl place.
Qed.

Lemma bind_enter_right :
 forall (A B St1 St2: Type) (v:(Place St1 St2)) (sp1 sp2: DS St2 A) (sp3 : A -> DS St1 B) (sp4: DS St1 B) (st : St1),
 refInEnv (read v st) sp2 sp1 -> 
 refInEnv st sp4 (bind v sp2 sp3) -> refInEnv st sp4 (bind v sp1 sp3).
Proof.
  intros; eapply refInEnv_trans. 
  eauto.
  auto with isl place.
Qed.

(*
Lemma bind_monotony_left :
 forall (A B : Type) (sp1 : DS A) (f1 f2 : A -> DS B) (sp2 : DS B) (st : St),
 (forall (x : A) (st : St), refInEnv st (f1 x) (f2 x)) -> 
 refInEnv st (bind sp1 f2) sp2 -> refInEnv st (bind sp1 f1) sp2.
Proof.
  intros A B sp1 f1 f2 sp2 st H1 H2.
  apply refInEnv_trans with (sp2 := bind sp1 f2); auto with isl.  
Qed.
 
Lemma bind_monotony_right :
 forall (A B : Type) (sp1 : DS A) (f1 f2 : A -> DS B) (sp2 : DS B) (st : St),
 (forall (x : A) (st : St), refInEnv st (f2 x) (f1 x)) -> 
 refInEnv st sp2 (bind sp1 f2) -> refInEnv st sp2 (bind sp1 f1).
Proof.
  intros A B sp1 f1 f2 sp2 st H1 H2.
  apply refInEnv_trans with (sp2 := bind sp1 f2); auto with isl.  
Qed.
*)
(* end hide *)

(** Here are a few lemmas to simplify "set". There are symmetric lemmas for the right side of refines relation.
 *)

Lemma set_skip_left :
 forall (A St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (sp1 sp2 : DS St1 A) (st : St1) (x:St3),
 x=(read (v1#v2) st) -> refInEnv st sp1 sp2 -> refInEnv st (bind v1 (set v2 x) (fun _ => sp1)) sp2.
Proof.
  intros A St1 St2 St3 v1 v2 sp1 sp2 st x H1 H; subst; apply refInEnv2wp.
   generalize (refInEnv_from_wp H); simpl in |- *; 
   autorewrite with placeeq; auto.
Qed. 

(* begin hide *)
Lemma set_skip_right :
 forall (A St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (sp1 sp2 : DS St1 A) (st : St1) (x:St3),
 x=(read (v1#v2) st) -> refInEnv st sp1 sp2 -> refInEnv st sp1 (bind v1 (set v2 x) (fun _ => sp2)).
Proof.
  intros A St1 St2 St3 v1 v2 sp1 sp2 st x H1 H; subst; apply refInEnv2wp.
   generalize (refInEnv_from_wp H); simpl in |- *; 
   autorewrite with placeeq; auto.
Qed. 
(* end hide *)

Lemma set_swap_leftright:
 forall (A St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (sp1 sp2 : DS St1 A) (st : St1) (x:St3),
   refInEnv (write (v1#v2) st x) sp1 (bind allPl (set allPl st) (fun _ => sp2)) 
     -> refInEnv st (bind v1 (set v2 x) (fun _ => sp1)) sp2.
Proof.
   intros A St1 St2 St3 v1 v2 sp1 sp2 st x H; apply refInEnv_trans 
    with (sp2:=(bind v1 (set v2 x) (fun _ => (bind allPl (set allPl st) (fun _ => sp2)))));
   auto with isl.
   apply seq_left_assoc with (v2:=v1); auto with place.
   apply bind_simpl_left; simpl.
   apply set_skip_left; [ simpl; reflexivity | idtac ]. 
   auto with isl.
Qed.

(* begin hide *)
Lemma set_swap_rightleft:
 forall (A St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (sp1 sp2 : DS St1 A) (st : St1) (x:St3),
  refInEnv (write (v1#v2) st x) (bind allPl (set allPl st) (fun _ => sp1)) sp2 
     -> refInEnv st sp1 (bind v1 (set v2 x) (fun _ => sp2)).
Proof.
  intros A St1 St2 St3 v1 v2 sp1 sp2 st x H; apply refInEnv_trans 
    with (sp2:=(bind v1 (set v2 x) (fun _ => (bind allPl (set allPl st) (fun _ => sp1)))));
   auto with isl.
   apply seq_right_assoc with (v2:=v1); auto with place.
   apply bind_simpl_right; simpl.
   intros; apply set_skip_left; [ simpl; reflexivity | idtac ]. 
   auto with isl.
Qed.
(* end hide *)

Lemma set_simplify_left :
 forall (A St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (sp1 sp2 : DS St1 A) (st : St1) (x:St3),
 refInEnv st (bind v1 (set v2 x) (fun _ => sp1)) sp1 -> 
 refInEnv st sp1 sp2 -> 
 refInEnv st (bind v1 (set v2 x) (fun _ => sp1)) sp2.
Proof.
  intros; apply refInEnv_trans with (sp2 := sp1); auto.
Qed.

(* begin hide *)
Lemma set_simplify_right :
 forall (A St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (sp1 sp2 : DS St1 A) (st : St1) (x:St3),
 refInEnv st sp2 (bind v1 (set v2 x) (fun _ => sp2)) -> 
 refInEnv st sp1 sp2 -> 
 refInEnv st sp1 (bind v1 (set v2 x) (fun _ => sp2)).
Proof.
  intros; apply refInEnv_trans with (sp2 := sp2); auto.
Qed.


Lemma seq_monotony_left :
 forall (B St1 St2: Type) (v:(Place St1 St2)) (sp1 : DS St2 unit) (sp2 sp3 sp4 : DS St1 B) (st : St1),
 (forall st2 : St2, refInEnv (write v st st2) sp2 sp3) -> 
 refInEnv st (bind v sp1 (fun _ : unit => sp3)) sp4 -> 
 refInEnv st (bind v sp1 (fun _ : unit => sp2)) sp4.
Proof.
  intros; apply refInEnv_trans with (sp2 := bind v sp1 (fun _ : unit => sp3));
  auto with isl place.  
Qed.

(*
Lemma seq_simplify_left :
 forall (B St1 St2: Type) (v:(Place St1 St2)) (sp1 : DS St2 unit) (sp2 sp3 : DS St1 B) (st : St1),
 (sdemon (sp1 (read v st))
   (fun p =>
    match sndT p with
    | Union_left a =>
        refInEnv (write v st (fstT p)) sp2 (bind allPl (set allPl st) (fun _ => sp3))
    | Union_right e =>
        refInEnv (write v st (fstT p)) (raise _ e) (bind allPl (set allPl st) (fun _ => sp3))
    end)) -> 
  refInEnv st (bind v sp1 (fun _ : unit => sp2)) sp3.
Proof.
  simpl in |- *; unfold refInEnv, bind in |- *; intros.
  refine (FSL.bind_simpl_left _ _ _ _); auto.
  simpl in |- *; intros; eapply sdemon_weaken. eauto.
  simpl in |- *; intros p; case p; intros stp rp; case rp; simpl in |- *; auto with wprefines.
Qed.
*)

Lemma seq_monotony_right :
 forall (B St1 St2: Type) (v:(Place St1 St2)) (sp1 : DS St2 unit) (sp2 sp3 sp4 : DS St1 B) (st : St1),
 (forall st2 : St2, refInEnv (write v st st2) sp2 sp3) -> 
 refInEnv st sp4 (bind v sp1 (fun _ : unit => sp2)) -> 
 refInEnv st sp4 (bind v sp1 (fun _ : unit => sp3)).
Proof.
  intros; apply refInEnv_trans with (sp2 := bind v sp1 (fun _ : unit => sp2));
  auto with isl place.  
Qed.


Lemma seq_left_assoc_nonePl1 :
 forall (B C St1 St3 : Type) (v1:(Place St1 unit)) (v2:(Place St1 St3)) (sp1 : DS St3 unit) (sp2 : DS unit B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st (bind allPl (bind v2 sp1 (fun _: unit => (bind nonePl sp2 (fun a => val a)))) sp3) sp4 -> 
 refInEnv st (bind v2 sp1 (fun _: unit => bind v1 sp2 sp3)) sp4.
Proof.
  intros; eapply refInEnv_trans.
  2:eauto.
  apply bind_right_unassoc_allPl1.
  apply bind_monotonyx; auto with isl.
  intros; apply bind_right_unassoc_allPl1.
  apply bind_monotony; auto with place isl.
  intros; apply bind_simpl_right; simpl.
  intros; apply set_skip_left; simpl; auto with isl.
Qed.

Lemma seq_right_assoc_nonePl1 :
 forall (B C St1 St3 : Type) (v1:(Place St1 unit)) (v2:(Place St1 St3)) (sp1 : DS St3 unit) (sp2 : DS unit B) 
   (sp3 : B -> DS St1 C) (sp4 : DS St1 C) (st : St1),
 refInEnv st sp4 (bind allPl (bind v2 sp1 (fun _: unit => (bind nonePl sp2 (fun a => val a)))) sp3) -> 
 refInEnv st sp4 (bind v2 sp1 (fun _: unit => bind v1 sp2 sp3)).
Proof.
  intros; eapply refInEnv_trans.
  eauto.
  apply bind_left_unassoc_allPl1.
  apply bind_monotonyx; auto with isl.
  intros; apply bind_left_unassoc_allPl1.
  apply bind_monotony; auto with place isl.
  intros; apply bind_simpl_left; simpl.
  intros; apply set_skip_left; simpl; auto with isl.
Qed.

Hint Resolve bind_val_r_ref1 bind_val_r_ref2: isl.

Lemma bind_liftV_unassoc_left :
 forall (A B St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (v:(Place St1 St3)) 
  (sp1 : DS St3 A) (sp2 : A -> DS St1 B) (sp3 : DS St1 B) (st : St1),
 (eqPl v (v1#v2)) -> 
 refInEnv st (bind v1 (bind v2 sp1 (fun a => (val a))) sp2) sp3 -> 
 refInEnv st (bind v sp1 sp2) sp3.
Proof.
  intros; eapply refInEnv_trans.
  2: apply H0.
  apply bind_right_unassoc.
  apply bind_monotony; auto with isl place. 
  simpl; intros; apply bind_simpl_right; simpl.
  intros; apply set_skip_left; simpl; auto with isl.
  autorewrite with placeeq; auto.
Qed.

Lemma bind_liftV_unassoc_right :
 forall (A B St1 St2 St3: Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) (v:(Place St1 St3)) 
  (sp1 : DS St3 A) (sp2 : A -> DS St1 B) (sp3 : DS St1 B) (st : St1),
 (eqPl v (v1#v2)) -> 
 refInEnv st sp3 (bind v1 (bind v2 sp1 (fun a => (val a))) sp2) -> 
 refInEnv st sp3 (bind v sp1 sp2).
Proof.
  intros; eapply refInEnv_trans.
  apply H0.
  apply bind_left_unassoc.
  apply bind_monotony; auto with isl place. 
  simpl; intros; apply bind_simpl_left; simpl.
  intros; apply set_skip_left; simpl; auto with isl.
Qed.

Lemma set_permut_sumPl_right:
  forall (A B St St1 St2:Type) (v1:(Place St St1)) (v2:(Place St St2)) 
    (u: sumPl v1 v2) (sp:(DS St B)) (sp1: DS St1 A) (sp2: A->(DS St B)) (st:St) (x:St2),
    (refInEnv st
      sp (bind v1 sp1 (fun a => (bind allPl (set v2 x) (fun _ => (sp2 a))))))
    -> 
    (refInEnv st
       sp (bind allPl (set v2 x) (fun _ => (bind v1 sp1 sp2)))).
Proof.
  intros; eapply refInEnv_trans.
  apply H.
 clear H.
 apply refInEnv2wp; simpl.
 generalize (refInEnv_from_wp (refInEnv_refl (write v2 st x) (bind v1 sp1 sp2))).
 simpl.
  rewrite <- (mkEnv_readLR u st).
 autorewrite with placeeq.
  intros X H1.
 eapply sdemon_weaken.
 apply X.
 eapply sdemon_weaken.
 apply H1.
 intros a; case a; simpl; auto.
 intros a; case a; simpl.
 intros s b; rewrite <- (mkEnv_readLR u st).
 autorewrite with placeeq.
 intros.
 rewrite <- (mkEnv_readLR u st).
 rewrite writeL_mkEnv.
 rewrite writeR_mkEnv.
 eapply sdemon_weaken.
 apply H.
 simpl; intros.
 autorewrite with placeeq.
 eapply angel_weaken.
 apply H0.
 intros d; case d; simpl.
 intros s0 a1; auto.
Qed.

Lemma set_permut_sumPl_left:
  forall (A B St St1 St2:Type) (v1:(Place St St1)) (v2:(Place St St2)) 
    (u: sumPl v1 v2) (sp: DS St B) (sp1: DS St2 A) (sp2: A->(DS St B)) (st:St) (x:St1),
    (refInEnv st
      sp (bind v2 sp1 (fun a => (bind allPl (set v1 x) (fun _ => (sp2 a)))))) ->
    (refInEnv st
      sp (bind allPl (set v1 x) (fun _ => (bind v2 sp1 sp2)))).
Proof.
  intros.
  refine (set_permut_sumPl_right (swapsumPl u) _ _).
  auto.
Qed.
(* end hide *)

(** Here is the lift operator from state transformer monad. It allows to define [any], [require], ...
directly from FSL. *)
Definition lift (St A: Type) (wp : WP.DS A) : DS St A :=
  fun st => FSL.bind wp (fun a => FSL.val (pairT st a)). 

Lemma bind_lift_changeV_equiv: forall (A B St1 St2 St3:Type) 
 (v1:(Place St1 St2)) (v2:(Place St1 St3)) (wp:(WP.DS A)) (sp:A -> (DS St1 B)) (st:St1),
  (refInEnv st (bind v1 (lift wp) sp) (bind v2 (lift wp) sp)).
Proof.
  intros; generalize (refInEnv_from_wp (refInEnv_refl st (bind allPl (lift wp) sp))).
  simpl; intros; apply refInEnv2wp.
  simpl; intros; eapply sdemon_weaken. apply H.
  eapply sdemon_weaken. apply H0.
  simpl; autorewrite with placeeq; auto.
  simpl; autorewrite with placeeq; auto.
Qed. 

(** [lift] and [set] commutate. *)
Lemma set_permut_lift_left:
  forall (A B St1 St2 St3:Type) (v1:(Place St1 St2)) (v2:(Place St2 St3)) 
    (wp: WP.DS A) (sp1: A -> (DS St1 B)) (sp2: (DS St1 B)) (st:St1) (x:St3),
    (refInEnv st
      (bind nonePl (lift wp) (fun a => (bind v1 (set v2 x) (fun _ => (sp1 a))))) sp2)
   ->    
    (refInEnv st 
      (bind v1 (set v2 x) (fun _ => (bind nonePl (lift wp) sp1))) sp2).
Proof.
  intros; eapply refInEnv_trans. 
  2:eauto.
 apply refInEnv2wp; simpl.
 intros H1.
 generalize (refInEnv_from_wp (refInEnv_refl (write (v1#v2) st x) (bind nonePl (lift wp) sp1))).
 simpl.
 auto.
Qed.

Definition abort (St A: Type) : DS St A := lift (FSL.abort A).

Definition magic (St A : Type) : DS St A := lift (FSL.magic A).

Definition require (St: Type) (p : Prop) : DS St unit := lift (FSL.require p).

Definition ensure (St: Type) (p : Prop) : DS St unit := lift (FSL.ensure p).

Definition any (St A: Type) : DS St A := lift (FSL.any A).

Lemma any_permut_bind :
 forall (A B C St1 St2: Type) (v:(Place St1 St2)) (sp : DS St2 A) (f : A -> B -> DS St1 C) 
 (st : St1),
 pre (sp (read v st)) \/ (NonEmpty B) -> 
 refInEnv st (bind v sp (fun a => bind nonePl (any B) (fun b => f a b)))
             (bind nonePl (any B) (fun b => bind v sp (fun a => f a b))).
Proof.
  simpl in |- *;
   unfold abort, any, refInEnv, bind, lift in |- *;
   intros A B C St1 St2 v sp f st H; apply WP.refines_fold; 
   simpl in |- *.
  intros H1; apply sdemon_fold.
  case H; auto.
  intros H2; case H2; intros x; generalize (H1 x); eauto with wipi.
  intros p; case p; simpl in |- *; clear p; simpl in |- *.
  intros st' a H2 b. lapply
  (WP.refines_unfold (refInEnv_refl st (bind v sp (fun a : A => f a b)))).
  intros H3;
   lapply (sdemon_unfold (sp (read v st)) _ (pairT st' a) H3); 
   auto.
  simpl in |- *; intros; eapply sdemon_weaken. eauto. simpl in |- *; eauto.
  simpl in |- *; auto.
Qed.

(* begin hide *)
Definition or (St A:Type) (sp1 sp2:(DS St A)) : (DS St A)
 := (fun st => (or (sp1 st) (sp2 st))).

Lemma refInEnv_or_intro_l: forall (St A:Type) (sp1 sp2 sp:(DS St A)) (st:St),
  (refInEnv st sp sp1) -> (refInEnv st sp (or sp1 sp2)).
Proof.
  unfold refInEnv, or; auto with wprefines.
Qed.

Lemma refInEnv_or_intro_r: forall (St A:Type) (sp1 sp2 sp:(DS St A)) (st:St),
  (refInEnv st sp sp2) -> (refInEnv st sp (or sp1 sp2)).
Proof.
  unfold refInEnv, or; auto with wprefines.
Qed.

Lemma refInEnv_or_elim: forall (St A:Type) (sp1 sp2 sp:(DS St A)) (st:St),
  (refInEnv st sp1 sp) -> (refInEnv st sp2 sp) -> 
    (refInEnv st (or sp1 sp2) sp).
Proof.
  unfold refInEnv, or; auto with wprefines.
Qed.

Hint Resolve refInEnv_or_intro_l refInEnv_or_intro_r refInEnv_or_elim
 set_skip_left set_skip_right: isl.

Definition or_x  (St A : Type) (sp1 sp2:DS St A) :=
  (bind nonePl (any bool) (fun b => if b then sp1 else sp2)).
                                                                                                        
Lemma refInEnv_or_x: forall (St A : Type) (sp1 sp2:DS St A) (st:St),
 (refInEnv st (or sp1 sp2) (or_x sp1 sp2)).
Proof.
  intros; unfold or_x; apply refInEnv_or_elim.
  apply bind_simpl_right; simpl.
  intros H; clear H.
  constructor 1 with (x:=true); auto with isl.
  apply bind_simpl_right; simpl.
  intros H; clear H.
  constructor 1 with (x:=false); auto with isl.
Qed.
                                                                                                        
Lemma refines_x_or: forall (St A : Type) (sp1 sp2:DS St A) (st:St),
 (refInEnv st (or_x sp1 sp2) (or sp1 sp2)).
Proof.
  intros; unfold or_x; apply bind_simpl_left; simpl.
  intros b; case b; auto with isl.
Qed.

Hint Resolve refInEnv_or_x refines_x_or: isl.

Lemma refInEnv_or_bind_assoc2_1:
 forall (St1 St2 A B:Type) (v:Place St1 St2) (sp: DS St2 A) (sp1 sp2: A -> (DS St1 B)) (st:St1),
   (refInEnv st (bind v sp (fun a => (or (sp1 a) (sp2 a)))) (or (bind v sp sp1) (bind v sp sp2))).
Proof.
  intros; apply refInEnv_trans with (sp2:=(bind v sp (fun a => (or_x (sp1 a) (sp2 a)))));
  auto with isl place.
  unfold or_x; eapply refInEnv_trans.
  apply any_permut_bind with
    (f:=(fun a (b:bool) => if b then sp1 a else sp2 a)).
  auto with nonempty.
  apply bind_simpl_left; simpl ; intros b; case b; 
  auto with isl.
Qed.

Lemma if_or_ref1:
  forall (St1 St2 A:Type) (v:Place St1 St2) (spb: DS St2 bool) (sp1 sp2: DS St1 A) (st: St1),
     (refInEnv st (If[v] spb Then sp1 Else sp2)%ISL
                 (or (bind v spb (fun b => bind nonePl (ensure (b=true)) (fun _ => sp1)))
                     (bind v spb (fun b => bind nonePl (ensure (b=false)) (fun _ => sp2))))).
Proof.
  intros.
  apply refInEnv_trans with 
    (sp2:=(bind v spb (fun b => (or (bind nonePl (ensure (b=true)) (fun _ => sp1))
                                    (bind nonePl (ensure (b=false)) (fun _ => sp2)))))).
  apply bind_monotony; simpl; auto with isl place.
  intros st0 b; case b; simpl.
  apply refInEnv_or_intro_l; apply bind_simpl_right; simpl;
  intuition auto with isl.
  apply refInEnv_or_intro_r; apply bind_simpl_right; simpl;
  intuition auto with isl.
  apply refInEnv_or_bind_assoc2_1 with
    (sp1:=(fun b : bool => bind nonePl (ensure (b = true)) (fun _ : unit => sp1)))
    (sp2:=(fun b : bool => bind nonePl (ensure (b = false)) (fun _ : unit => sp2))).
Qed.

Lemma if_elim:
 forall (St1 St2 A:Type) (v:Place St1 St2) (spb: DS St2 bool) (sp1 sp2 sp: DS St1 A) (st: St1),
   (refInEnv st (bind v spb (fun b => (bind nonePl (ensure (b=true)) (fun _ => sp1)))) sp) -> 
   (refInEnv st (bind v spb (fun b => (bind nonePl (ensure (b=false)) (fun _ => sp2)))) sp) -> 
   (refInEnv st (If[v] spb Then sp1 Else sp2)%ISL sp).
Proof.
  intros; eapply refInEnv_trans.
  apply if_or_ref1.
  auto with isl.
Qed.

Definition sync (St A B:Type) (sp: A ->(DS St B)) : (DS St B) 
 := fun st => (FSL.sync (fun x:A => (sp x st))).
 
Lemma sync_elim: forall (A B St:Type) (sp1: A -> (DS St B)) (sp2:(DS St B)) (a:A) (st:St),
  (refInEnv st (sp1 a) sp2) -> (refInEnv st (sync sp1) sp2).
Proof.
  unfold refInEnv, sync; intros; eapply sync_elim; eauto.
Qed.

Lemma sync_intro: forall (A B St:Type) (sp1:(DS St B)) (sp2: A -> (DS St B)) (st:St),
  (forall (a:A), (refInEnv st sp1 (sp2 a))) -> (refInEnv st sp1 (sync sp2)).
Proof.
  unfold refInEnv, sync; auto with wprefines2.
Qed.

Hint Resolve sync_elim sync_intro: isl2.
                                                                                
Lemma sync_monotony :
 forall (A B St: Type) (s1 s2: A -> DS St B) (st:St),
 (forall (a:A), refInEnv st (s1 a) (s2 a)) ->
    refInEnv st (sync s1) (sync s2).
Proof.
  intros; eauto with isl2.
Qed.
                                                                                
Hint Resolve sync_monotony: isl2.
 
Lemma bind_sync_distr_r_1:
  forall (A B C St St0:Type) (v:Place St St0) (s1: A->DS St0 B) (s2: B -> (DS St C)) (st:St),
   (refInEnv st (bind v (sync s1) s2) (sync (fun x => bind v (s1 x) s2))).
Proof.
 intros; apply sync_intro.
 intros; apply bind_monotony; eauto with isl isl2 place.
Qed.
                                                                                
Lemma bind_sync_distr_l_1:
 forall (A B C St St0:Type) (v:Place St St0) (s1: DS St0 A) (s2: A -> B -> (DS St C)) (st:St),
   (refInEnv st (bind v s1 (fun a => sync (s2 a))) 
                (sync (fun b => (bind v s1 (fun a => s2 a b))))).
Proof.
 intros; apply sync_intro.
 intros; apply bind_monotony; eauto with isl isl2 place.
Qed.
 
Unset Implicit Arguments.
(* end hide *)
