(* begin hide *)
Set Implicit Arguments.

Require Export ISLBlock.
(* end hide *)

(** ** Module [AbsPost]: Hoare Specifications

In this module, I define operations to handle specifications a la Hoare in ISL. 

A Hoare specification is a black-box specification: a specification in
which the result or the final state is only determined by some
predicates. I propose here two kind of Hoare specifications: Hoare
specifications that may update the state, and purely functional
Hoare specifications. Of course, the second ones are a special case of
the first ones. As in section %\ref{subsec:extdsl}%, Hoare
specifications are decomposed into a precondition and an abstract
postcondition.

*** Hoare specifications without exception raising.

With respect to section %\ref{subsec:extdsl}%, I now consider that a
such Hoare specification may work only on a place of the current
state. 
*)
Section UPD_ABSPOST.

(** I assume [St] the type of global state, [St2] the type of local state. [v] is the place on which works the abstract postcondition. 
[A] is the type of the returned result. [q] is the postcondition relating the final state and 
the result.
*)
Variable St St2: Type.

Variable v: (Place St St2).

Variable A: Type.

Variable q: St2 -> A -> Prop.

(**  The abstract postcondition of [v] and [q] is defined as follows:
*)
Definition upd_abspost: (DS St A)
  := (Bind result <-/ (Any A) In
              [v](Bind new_st <-/ (Any St2) In 
                   (allPl :<- new_st) -;  
                   (Ensure (q new_st result))) -;
                 (Val result))%ISL.

(** In the following lemmas, I assume an initial state [old_st].
*)
Variable old_st: St.

(** Here is a rule stating how to prove that a specification [pt] refines
an abstract postcondition using wp-calculus. This rule assume that the
current state [tmp_st] has been obtained from [old_st] only by
modifications in [v]. 
Hence, the current goal is equivalent to 
[(refInEnv old_st (allPl :<- tmp_st -; pt) upd_abspost)]
This rule generalizes the usual one which assumes [tmp_st=old_st]. 
This generalization is useful in proofs using small-steps simplifications, because goals of the form
``[refInEnv old_st pt0 upd_abspost]'' can be simplified before to use 
this rule.
*)
Lemma updabspost2wp:
  forall (pt:(DS St A)) (tmp_st: St),
  tmp_st=(write v old_st (read v tmp_st)) -> 
  (sdemon (pt tmp_st)
   (fun p => let (new_st,a):=p in 
      new_st = (write v old_st (read v new_st)) 
      /\ (q (read v new_st) a)))      
    -> (refInEnv tmp_st pt upd_abspost).
Proof.
  intros sp tmp_st H1 H2; rewrite H1 in H2; rewrite H1; clear H1; unfold upd_abspost.
  apply refInEnv2wp; simpl.
  intros; eapply sdemon_weaken.
  apply H2.
  intros a; case a; simpl.
  intros st0 a0 H3; case H3; intros H4 H5;
  constructor 1 with (x:=a0).
  constructor 1 with (x:=(read v st0)).
  autorewrite with placeeq.
  rewrite <- H4.
  intuition auto.
Qed.

(** This lemma corresponds to a ``refinement in a context'' rule. In a
goal [(refInEnv old_st sp1 sp2)] such [sp1] satifies a postcondition
[q], it is possible to strengthen [sp1] with a clause ensuring that
the result and the final state of [sp1] satisfy [q].  
*)
Lemma intro_updabspost:
  forall (sp1 sp2:(DS St A)),
   (refInEnv old_st sp1 upd_abspost)
      -> (refInEnv old_st (Bind a <- sp1 In
                          (Bind new_st <- !allPl In 
                           (Ensure ((q (read v new_st) a) 
                                    /\ new_st=(write v old_st (read v new_st))))) -;
                         (Val a))
                      sp2)%ISL       
  -> (refInEnv old_st sp1 sp2)%ISL.
Proof.
  intros sp1 sp2 H1 H. eapply refInEnv_trans.
  2:apply H.
  clear H; unfold refInEnv.
  apply refines_fold.
  simpl; intros H3.
  lapply (refines_unfold H1).
  intros H4; apply sdemon_fold.
  eapply sdemon_weaken.
  apply H4.
  auto.
  intros a H5.
  apply angel_fold with (a:=a); intuition auto.
  case (sdemon_unfold _ _ a H4 H5).
  simpl.
  intros x H6; decompose [ex Logic.and] H6.
  subst. simpl.  autorewrite with placeeq. intuition auto.
  simpl; intuition auto. 
Qed.

(** This rule is a generalization of the preceding one.
*)
Lemma intro_bind_updabspost:
  forall (B:Type) (sp1:(DS St A)) (sp2: A -> (DS St B)) (sp3: (DS St B)),
    (refInEnv old_st sp1 upd_abspost)
        -> (refInEnv old_st (Bind a <- sp1 In
                             (Bind new_st <- ! allPl In
                               (Ensure ((q (read v new_st) a) 
                                /\ new_st=(write v old_st (read v new_st))))) -;
                           (sp2 a))
                       sp3)%ISL
  -> (refInEnv old_st (bind allPl sp1 sp2) sp3)%ISL.
Proof.
  intros B sp1 sp2 sp3 H1 H3.
  eapply refInEnv_trans.
  2:apply H3.
  apply refInEnv_trans with
   (sp2:=(Bind b <- (Bind a <- sp1 In (Bind new_st <- !allPl In 
                           Ensure q (read v new_st) a /\ new_st = write v old_st (read v new_st)) -;
                           (Val a)) In (sp2 b))%ISL).
   ISL_monotony.
   eapply intro_updabspost.
   apply H1.
   auto with isl.
   ISL_unassoc_left.
   ISL_monotony.
   intros; ISL_simplify_left; simpl.
   intros; ISL_simplify_right; simpl.
   intuition auto.
   auto with isl.
Qed.

End UPD_ABSPOST.

(** *** Pure Hoare specifications.

Given a type [A] and a predicate [q:A -> Prop], I now define
[pure_abspost q] as a specification equivalent to [(upd_abspost _
nonePl (fun _ => q)))].  However, it is interesting to simplify the
corresponding specification.

*)

Section PURE_ABSPOST.

Variable St A: Type.

Variable q: A -> Prop.

Definition pure_abspost: (DS St A) 
  := (Bind result <-/ (Any A) In  
            (Ensure (q result)) /;
                  (Val result))%ISL.

(** The rules of the preceding section can be specialized to this case. I do not give the details.%\\%
[...]
*)
(* begin hide *)
Variable old_st: St.

Lemma pureabspost2wp:
  forall (sp:(DS St A)),  
  (sdemon (sp old_st)
   (fun p => let (new_st,a):=p in
               new_st=(old_st) /\ (q a)))      
    -> (refInEnv old_st sp pure_abspost).
Proof.
  intros sp H2; subst; unfold pure_abspost.
  apply refInEnv2wp; simpl.
  intros; eapply sdemon_weaken.
  apply H2.
  intros a; case a; simpl.
  intros st0 a0 H3; case H3; intros H4 H5;
  constructor 1 with (x:=a0).
  rewrite <- H4.
  intuition auto.
Qed.

Lemma pure_abspost_exc_abspost_ref1:
  (refInEnv old_st pure_abspost (upd_abspost nonePl (fun _ => q))).
Proof.
  apply updabspost2wp with (old_st:=old_st);
  simpl; auto.
Qed.

Lemma pure_abspost_exc_abspost_ref2:
  (refInEnv old_st (upd_abspost nonePl (fun _ => q)) pure_abspost).
Proof.
  apply pureabspost2wp;
  simpl; auto.
Qed.

Lemma intro_pureabspost:
  forall (sp1 sp2:(DS St A)),
   (refInEnv old_st sp1 pure_abspost)
      -> (refInEnv old_st (Bind a <- sp1 In
                       ((Bind new_st <- !allPl In 
                          Ensure ((q a) /\ new_st=old_st)) -;
                         (Val a)))
                      sp2)%ISL       
  -> (refInEnv old_st sp1 sp2)%ISL.
Proof.
  intros; apply intro_updabspost with (v:=(nonePl (A:=St))) (q:=(fun _:unit => q));
  simpl; auto.
  ISL_trans. 
  eauto.
  apply pure_abspost_exc_abspost_ref1.
Qed.

Lemma intro_bind_pureabspost:
  forall (B:Type) (sp1:(DS St A)) (sp2: A -> (DS St B)) (sp3: (DS St B)),
      (refInEnv old_st sp1 pure_abspost)
       -> (refInEnv old_st (Bind a <- sp1 In
                          ((Bind new_st <- !allPl In
                            Ensure ((q a) /\ new_st=old_st)) -;
                           (sp2 a)))
                       sp3)%ISL
  -> (refInEnv old_st (bind allPl sp1 sp2) sp3)%ISL.
Proof.
  intros; apply intro_bind_updabspost with (v:=(nonePl (A:=St))) (q:=(fun _:unit => q));
  simpl; auto.
  ISL_trans. 
  eauto.
  apply pure_abspost_exc_abspost_ref1.
Qed.
(* end hide *)

End PURE_ABSPOST.

(* begin hide *)
Hint Resolve pure_abspost_exc_abspost_ref1 pure_abspost_exc_abspost_ref2: islblock.
(* end hide *)

(** *** Rules for Hoare specifications.
*)

Lemma updabspost_monotony: forall (St St2 A:Type) (v1 v2:(Place St St2)) (q1 q2: St2 -> A -> Prop) (st:St),
  (eqPl v1 v2) -> 
   (forall (st2:St2) (a:A), (q1 st2 a) -> (q2 st2 a)) -> 
    (refInEnv st (upd_abspost v1 q1) (upd_abspost v2 q2)).
Proof.
  intros; eapply updabspost2wp with (old_st:=st). 
  autorewrite with placeeq; auto.
  simpl.
  intros; rewrite eqPl_write with (v1:=v1) (v2:=v2); auto.
  autorewrite with placeeq.   
  auto.
Qed.

Lemma updabspost_split: forall (St St2 A:Type) (v:(Place St St2)) (q1 q2: St2 -> A -> Prop) 
  (st:St) (sp:(DS St A)),
   (refInEnv st sp (upd_abspost v q1))
     -> (refInEnv st sp (upd_abspost v q2))
      ->
    (refInEnv st sp (upd_abspost v (fun (stf:St2) (r:A) => (q1 stf r)  /\ (q2 stf r)))).
Proof.
  intros St St2 A v q1 q2 st sp H1 H2.
  eapply intro_updabspost with (v:=v).
  apply H2.
  ISL_enter_left.
  apply H1.
  intros; eapply updabspost2wp with (old_st:=st). 
  autorewrite with placeeq; auto.
  simpl.
  intros x stf;  autorewrite with placeeq.   
  intuition auto.
Qed.

Lemma updabspost_splitforall: forall (St St2 A B:Type) (v:(Place St St2)) 
  (p: B -> Prop) (q: B -> St2 -> A -> Prop) 
  (st:St) (sp:(DS St A)),
   (exists x:B, (p x)) -> 
   (forall (x:B), (p x) -> (refInEnv st sp (upd_abspost v (q x)))) ->
    (refInEnv st sp (upd_abspost v (fun (stf:St2) (r:A) => forall (x:B), (p x) -> (q x stf r)))).
Proof.
  intros.
  eapply updabspost2wp with (old_st:=st). 
  autorewrite with placeeq; auto.
  case H; clear H.
  intros x H; apply sdemon_fold.  
  lapply (refines_unfold (H0 x H)); simpl; intros; eauto with wipi.
  intros a; case a; clear a.
  intros s a H1; constructor 1.
  lapply (refines_unfold (H0 x H)); simpl; auto.
  intros H2; lapply (sdemon_unfold _ _ (pairT s a) H2); auto.
  clear H2; intros H2.
  decompose [ex Logic.and] H2.
  inversion H5. 
  autorewrite with placeeq.
  auto.
  intros y Hy; lapply (refines_unfold (H0 y Hy)); simpl; auto.
  intros H2; lapply (sdemon_unfold _ _ (pairT s a) H2); auto.
  clear H2; intros H2.
  decompose [ex Logic.and] H2.
  inversion H5. 
  autorewrite with placeeq.
  auto.
Qed.

Lemma embed_updabspost_ref1: forall (St St2 St3 A:Type) (v1:(Place St St2)) (v2:(Place St2 St3)) 
 (q: St3 -> A -> Prop) (st:St),
  (refInEnv st (Embed [v1] (upd_abspost v2 q))%ISL (upd_abspost (v1#v2) q)).
Proof.
  intros; eapply updabspost2wp with (old_st:=st).
  autorewrite with placeeq; auto.
  simpl.
  intros; autorewrite with placeeq.
  auto.
Qed.

Lemma bind_updabspost_ref1:
 forall (St St2 St3 A B:Type) (v1:(Place St St2)) (v2:(Place St2 St3)) 
  (q: St3 -> A -> Prop) (st:St) (sp: A -> (DS St B)),
  (refInEnv st  
      (bind v1 (upd_abspost v2 q) sp)
      (bind allPl (upd_abspost (v1#v2) q) sp)).
Proof.
  intros; apply bind_liftV_unassoc_left with (v1:=(allPl (A:=St))) (v2:=v1).
  auto with place.
  ISL_monotony.
  ISL_trans.
  2:apply embed_updabspost_ref1.
  auto with isl.
Qed.

(* begin hide *)
Lemma embed_pureabspost_ref1: forall (St St2 A:Type) (v:(Place St St2)) (q: A -> Prop) (st:St),
  (refInEnv st (Embed [v] (pure_abspost q))%ISL 
               (pure_abspost q)).
Proof.
  intros; apply pureabspost2wp.
  simpl; intros; autorewrite with placeeq; auto.  
Qed.

Lemma bind_pureabspost_ref1:
 forall (St St2 A B:Type) (v:(Place St St2)) (q: A -> Prop) (st:St) (sp: A -> (DS St B)),
  (refInEnv st  
      (bind v (pure_abspost q) sp)
      (bind allPl (pure_abspost q) sp)).
Proof.
  intros; apply bind_liftV_unassoc_left with (v1:=(allPl (A:=St))) (v2:=v).
  auto with place.
  ISL_monotony.
  ISL_trans.
  2:apply embed_pureabspost_ref1.
  auto with isl.
Qed.
(* end hide *)

Lemma embed_updabspost_ref2: forall (St St2 St3 A:Type) (v1:(Place St St2)) (v2:(Place St2 St3)) 
 (q: St3 -> A -> Prop) (st:St),
  (refInEnv st (upd_abspost (v1#v2) q) (Embed [v1] (upd_abspost v2 q))%ISL).
Proof.
  intros; unfold upd_abspost. 
  ISL_unassoc_right.
  ISL_monotony.
  intros; ISL_Trivial.
Qed.

Lemma bind_updabspost_ref2:
 forall (St St2 St3 A B:Type) (v1:(Place St St2)) (v2:(Place St2 St3)) 
  (q: St3 -> A -> Prop) (st:St) (sp: A -> (DS St B)),
  (refInEnv st  
      (bind allPl (upd_abspost (v1#v2) q) sp)
      (bind v1 (upd_abspost v2 q) sp)).
Proof.
  intros; apply bind_liftV_unassoc_right with (v1:=(allPl (A:=St))) (v2:=v1).
  auto with place.
  ISL_monotony.
  ISL_trans.
  apply embed_updabspost_ref2.
  auto with isl.
Qed.

Lemma embed_pureabspost_ref2: forall (St St2 A:Type) (v:(Place St St2)) (q: A -> Prop) (st:St),
  (refInEnv st (pure_abspost q)
               (Embed [v] (pure_abspost q))%ISL).
Proof.
  intros; unfold pure_abspost. 
  ISL_unassoc_right.
  ISL_monotony.
  intros; ISL_simplify_all; simpl; autorewrite with placeeq; auto.
Qed.

(* begin hide *)
Lemma bind_pureabspost_ref2:
 forall (St St2 A B:Type) (v:(Place St St2)) (q: A -> Prop) (st:St) (sp: A -> (DS St B)),
  (refInEnv st  
      (bind allPl (pure_abspost q) sp)
      (bind v (pure_abspost q) sp)).
Proof.
  intros; apply bind_liftV_unassoc_right with (v1:=(allPl (A:=St))) (v2:=v).
  auto with place.
  ISL_monotony.
  ISL_trans.
  apply embed_pureabspost_ref2.
  eauto with isl.
Qed.
(* end hide *)

Lemma updabspost_liftPl_ref1: forall (St St2 St3 A:Type) (v1:(Place St St2)) (v2:(Place St2 St3)) 
 (q: St3 -> A -> Prop) (tmp_st old_st:St) (st:St2),
  tmp_st=(write v1 old_st st) 
  -> (refInEnv tmp_st 
       (upd_abspost v1 (fun (e':St2) (x:A) => e'=(write v2 st (read v2 e')) 
                           /\ (q (read v2 e') x)))  
       (upd_abspost (v1#v2) q)).
Proof.
  intros; subst. 
  eapply updabspost2wp with (old_st:=(write v1 old_st st)).
  simpl; autorewrite with placeeq; auto.
  simpl.
  intros; autorewrite with placeeq.
  intuition auto.
  rewrite <- H0.
  auto.
Qed.

Lemma updabspost_liftPl_ref2: forall (St St2 St3 A:Type) (v1:(Place St St2)) (v2:(Place St2 St3)) 
 (q: St3 -> A -> Prop) (tmp_st old_st:St) (st:St2),
  tmp_st=(write v1 old_st st) 
  -> (refInEnv tmp_st 
       (upd_abspost (v1#v2) q)
       (upd_abspost v1 (fun (e':St2) (x:A) => e'=(write v2 st (read v2 e')) 
                           /\ (q (read v2 e') x)))).
Proof.
  intros; subst. 
  eapply updabspost2wp with (old_st:=old_st).
  simpl; autorewrite with placeeq; auto.
  simpl.
  intros; autorewrite with placeeq.
  intuition auto.
Qed.

(*
Lemma block_updabspost_ref1: 
 forall (St St1 St2 St3 A:Type) (v1:Place St St1) (bkv:Place St St2)
   (u:sumPl v1 bkv) (v2:Place St2 St3) (q: St3 -> A -> Prop) (st:St),
    (refInEnv st 
      (upd_abspost allPl (fun (e':St) (x:A) => 
             e'=(write bkv e' (write v2 (read bkv st) (read (bkv#v2) e'))) 
                 /\ (q (read (bkv#v2) e') x)))
    (  (Bind a <-[bkv](upd_abspost v2 q) In 
        (Bind x' <-/ (Any St1) In v1 :<- x') -; (Val a)))%ISL.
Proof.
  intros.
  unfold upd_abspost at 1; repeat (intros; ISL_simplify_left).
  simpl; intuition auto.  
  ISL_simplify_right.
  intros; ISL_simplify_right; intros.
  constructor 1 with (x:=a).
  constructor 1 with (x:=(read v2 (read bkv a0))).
  intuition auto.
  ISL_simplify_right.
  intros; constructor 1 with (x:=(read v1 a0)).
  ISL_simplify_left.
  rewrite <- (mkEnv_readLR u st).
  autorewrite with placeeq.
  rewrite <- H0.
  ISL_Trivial.  
Qed.

(* begin hide *)
Lemma block_pureabspost_ref1: forall (St St1 St2 A:Type) 
   (v1:Place St St1) (bkv:Place St St2) (u:sumPl v1 bkv) 
   (q: A -> Prop) (st:St),
    (refInEnv st
      (upd_abspost _ allPl (fun (e':St) (x:A) => e'=(write bkv e' (read bkv st)) /\ (q x))) 
      (Bind a <-[bkv](pure_abspost q) In 
        (Bind x' <-/ (Any St1) In v1 :<- x') -; (Val a)))%ISL.
Proof.
  intros. ISL_trans.
  ISL_trans.
  2: apply block_updabspost_ref1 with (v1:=v1) (bkv:=bkv) 
    (v2:=(nonePl (A:=St2))) (q:=(fun (_:unit) => q)) (st:=st).
  simpl; auto with isl.
  auto.
  ISL_monotony.
  auto with islblock.
Qed.
*)

Unset Implicit Arguments.
(* end hide *)
