(* Below, a little "hack" to deal with a coqdoc bug *)
(** printing [(*]*) %{\tt[}%*)
(** printing ] %{\tt]}%*)
(* begin hide *)
Require Export ISL.
(* end hide *)

(** ** Module [ISLfriendly]: Macros and tactics in the prover 
*)

Hint Resolve refInEnv_refl bind_val_r_ref1 bind_val_r_ref2: isl.
Hint Resolve bind_monotony if_monotony : isl.
Hint Resolve set_skip_left set_skip_right: isl.

(** *** Simple tactics to simplify [refInEnv] proofs using wp-calculus 
Below, tactics [ISL_simplify_all] and [ISL_Trivial] apply wp to the whole goal. *)
Ltac ISL_simplify_all := intros; apply refInEnv2wp; simpl.
Ltac ISL_Trivial := abstract (intros; apply refInEnv2wp; vm_compute; intuition eauto).
(** Tactic [ISL_pre_intro] introduces the precondition of the right-side of the goal into hypothesis *)
Ltac ISL_pre_intro := apply refInEnv_pre_intro; simpl.
(** Tactic [ISL_simplify_both] corresponds to [bind_simpl_both] lemma *)
Ltac ISL_simplify_both :=
  apply bind_simpl_both; simpl in |- *; auto with isl place.
(** Roughly speaking, [ISL_simplify_left] corresponds to
[bind_simpl_left] lemma.  But, when the first part of the left
expression is a set expression, it tries also to eliminate it.  If it
fails, then it uses [set_swap_left_right] to move the set expression
on the right side of the goal.  
*)
Ltac ISL_skip_set := apply set_skip_left ; [(*]*) simpl; reflexivity | idtac ].
Ltac ISL_simplify_left :=
  (apply bind_simplPure_left || ISL_skip_set ||
    (apply set_simplify_left; [(*]*) ISL_Trivial; fail | idtac ] ) ||
      (apply set_swap_leftright; 
         try (apply set_simplify_right; [(*]*) ISL_Trivial; fail | idtac ] )) ||
        (apply bind_simpl_left; simpl in |- *;
          try ISL_skip_set)); simpl.
(** Tactic [ISL_simplify_right] is quite symmetric to [ISL_simplify_left], except that,
when the right expression is a set expression, it is never moved on the left.
*)
Ltac ISL_simplify_right :=
  (apply bind_simplPure_right ||
  (apply set_simplify_right ; [(*]*) ISL_Trivial; fail | idtac ] ) ||
      (apply bind_simpl_right; simpl in |- *;
        try (let z:= fresh "H" in (intro z; ISL_skip_set; generalize z; clear z)))); simpl.

(** *** Tactics using monotony to decompose the goal *)
Ltac ISL_enter_left := eapply bind_enter_left; simpl.
Ltac ISL_enter_right := eapply bind_enter_right; simpl.
Ltac ISL_monotony_left :=
  eapply seq_monotony_left; simpl (* || eapply bind_monotony_left *).
Ltac ISL_monotony_right :=
  eapply seq_monotony_right; simpl (* || eapply bind_monotony_right *).
Ltac ISL_monotony :=
  (apply if_monotony || apply bind_monotony); simpl; auto with isl place.
(* begin hide *)
Ltac ISL_trans := eapply refInEnv_trans.
(* end hide *)

(** *** Tactics changing the goal using associativity *)
Ltac ISL_unassoc_left := 
  apply bind_left_unassoc_allPl1 || apply bind_left_unassoc_allPl2 || 
  apply bind_left_unassoc_nonePl2 || apply bind_left_unassoc.
Ltac ISL_unassoc_right :=
  apply bind_right_unassoc_allPl1 || apply bind_right_unassoc_allPl2 || 
  apply bind_right_unassoc_nonePl2 || apply bind_right_unassoc.
Ltac ISL_assoc_left := 
  apply seq_left_assoc_allPl1 || apply seq_left_assoc_allPl2 || 
  apply seq_left_assoc_nonePl2 || apply seq_left_assoc_nonePl1 ||
  apply bind_left_assoc_allPl1 || apply bind_left_assoc_allPl2 || 
  apply bind_left_assoc_nonePl2.
Ltac ISL_assoc_right := 
  apply seq_right_assoc_allPl1 || apply seq_right_assoc_allPl2 || 
  apply seq_right_assoc_nonePl2 || apply seq_right_assoc_nonePl1 ||
  apply bind_right_assoc_allPl1 || apply bind_right_assoc_allPl2 || 
  apply bind_right_assoc_nonePl2.


(** *** Syntactic Sugar *)

Delimit Scope ISL_scope with ISL. 
Bind Scope ISL_scope with DS.

Notation "'Val' v" := (val v) (at level 90): ISL_scope.
Notation "'Abort'" := (abort _) (at level 90): ISL_scope.
Notation "v  :<-  x" := (set v x) (at level 90, right associativity): ISL_scope.
Notation "! v" := (get v) (at level 90): ISL_scope.
Notation "'Any' A" := (any A) (at level 90): ISL_scope.
Notation "'Req' p" := (require p) (at level 90): ISL_scope.
Notation "'Ensure' p" := (ensure p) (at level 90): ISL_scope.
Notation "'Ext' x => f" := (val (fun x => (val f))) (at level 90, f at level 0): ISL_scope. 
Notation "'Skip'" := (val tt) (at level 90): ISL_scope.

(* Notations for binders *)

Notation "'If' [ v ]  pb 'Then' sp1 'Else' sp2" :=
  (bind v pb (bool_rect (fun _ => (DS _ _)) sp1 sp2))
  (at level 20, v, pb, sp1, sp2 at level 90): ISL_scope.

Notation "'If' pb 'Then' sp1 'Else' sp2" :=
  (bind allPl pb (bool_rect (fun _ => (DS _ _)) sp1 sp2))
  (at level 20, pb, sp1, sp2 at level 90): ISL_scope.

Notation "'If' / pb 'Then' sp1 'Else' sp2" :=
  (bind nonePl pb (bool_rect (fun _ => (DS _ _)) sp1 sp2))
  (at level 20, pb, sp1, sp2 at level 90): ISL_scope.

Notation "'Bind' x <-[ v ] sp1 'In' sp2" := (bind v sp1 (fun x => sp2))
  (at level 95, v, sp1 at level 90, right associativity): ISL_scope.
Notation "'Bind' x <- sp1 'In' sp2" := (bind allPl sp1 (fun x => sp2))
  (at level 95, sp1 at level 90, right associativity): ISL_scope.
Notation "'Bind' x <-/ sp1 'In' sp2" := (bind nonePl sp1 (fun x => sp2))
  (at level 95, sp1 at level 90, right associativity): ISL_scope.

Notation "[ v ] sp1  -;  sp2" := (bind v sp1 (fun _ : unit => sp2))
  (at level 95, v, sp1 at level 90, right associativity): ISL_scope.
Notation "sp1  -;  sp2" := (bind allPl sp1 (fun _ : unit => sp2))
  (at level 95, right associativity): ISL_scope.
Notation "sp1  /;  sp2" := (bind nonePl sp1 (fun _ : unit => sp2))
  (at level 95, right associativity): ISL_scope.

(* A REVOIR *)
Notation "f  @-  arg" := (bind allPl arg (fun x => bind allPl f (fun g => g x)))
  (at level 110, left associativity): ISL_scope. 

(* begin hide *)
Notation " [<  st  >]  sp1  >==>  sp2" := (refInEnv st sp1 sp2) 
 (at level 200): ISL_scope.
(* end hide *)

(* A coqdoc bug again... *)
(** [Notation "]%{\tt[<}%[ st ]%{\tt>]}%[  sp1  >==>  sp2" := (refInEnv st sp1 sp2)] *)
(** %~~%[ (at level 200): ISL_scope.] *)

