(* ATTENTION,
   Les commentaires de ce fichier sont interpretes par "coqdoc"
   pour generer un fichier tex, inclus dans le rapport. 
*)
(* begin hide *)
Set Implicit Arguments.

Require Import WfT.
(* end hide *)

(** ** Total correctness in presence of non-determinism %\label{subsec:starRel}%

The purpose of this section is to sketch how I formalize the total
correctness theorem. As usual in presence of non-determinism, the
operational semantics is a small-steps semantics. In other word, my
semantics is based on a binary relation [leadsto] which describes
one-step of computation.  [leadsto] relates two couples formed of a
program and a state.  Hence, [leadsto (p,st) (p',st')] indicates that
in the state [st], [p] can be computed in [p'], the resulting state
being [st'].  It is non-deterministic in the sense that for a given
[(p,st)], there are maybe several couple [(p',st')] such the relation
holds.

A strategy of computation can be defined as a sequence of couple,
successively in relation via [leadsto]. Given of a couple [(p,st)],
stating that all strategies of computations terminate can be
expressed as [(Acc (fun y x => leadsto x y) (p,st))].

In order to express arbitrary but finite number of steps of
computations, we define the binary relation [leadstostar] as the
reflexive and transitive closure of [leadsto]. 

In the following, I give the definition [(starRel R)] of the
reflexive and transitive closure of a relation [R]. I prove that if
I can decide whether some element is minimal for [R], and if for a
given [x], [(Acc R x)] then there exists a minimal element [y] such
[(starRel R y x)].

For [(fun y x => leadsto x y)], the minimal element are couple
[(p,st)] such [p] is a value. And of course, being a value is a
decidable property of programs. In other word, the total correctness
can be formalized as the two properties:
  - [leadstostar] induces a special case of refinement.
  - [(Acc (fun y x => leadsto x y) (p,st))] holds when [p] can be
interpreted as a DSL specification [ps] such [(pre ps st)].

I prove now the announced property: accessibility implies the
existence of a path to reach a minimal element if minimality is
decidable. First, I give an inductive definition of [(starRel R)].
*)
Inductive starRel (A: Type) (R: A -> A -> Prop) : A -> A -> Prop :=
  | starRel_refl: forall (x:A), (starRel R x x)
  | starRel_next: 
     forall (x y z:A), (R x y) -> (starRel R y z) -> (starRel R x z). 

(** The three following lemmas shows that [(starRel R)] is the smallest relation
that is reflexive, transitive and that includes [R]*)

Lemma starRel_trans:
  forall (A:Type) (R: A -> A -> Prop) (x y z:A),
   (starRel R x y) -> (starRel R y z) -> (starRel R x z).
Proof.
  intros A R x y z H; generalize z; elim H; clear x y z H.
  auto.
  intros; eapply starRel_next; eauto.
Qed.

Lemma starRel_includes:
  forall (A:Type) (R: A -> A -> Prop) (x y:A), (R x y) -> (starRel R x y).
Proof.
  intros; eapply starRel_next.
  eauto.
  apply starRel_refl.
Qed. 

Lemma starRel_smallest:
 forall (A:Type) (R P: A -> A -> Prop), 
   (forall x:A, (P x x))
   -> (forall (x y:A), (R x y) -> (P x y))
   -> (forall (x y z:A), (P x y) -> (P y z) -> (P x z))
    -> forall (x y:A), (starRel R x y) -> (P x y).
Proof.
  induction 4; auto.
  eapply H1.
  apply H0; eauto.
  auto.  
Qed.
  
(** This small lemma shows that permutation of arguments is preserved 
via [starRel] (usually [leadsto] relation is written in the other way).
*)
Lemma starRel_swap:
  forall (A:Type) (R: A -> A -> Prop) (x y:A),
    (starRel R x y) -> (starRel (fun y x => R x y) y x).
Proof.
  induction 1.
  apply starRel_refl.
  eapply starRel_trans.
  eauto.
  apply starRel_includes; auto.
Qed.     
  

(** Here is the announced property of [Acc]. *)
Theorem Acc_findmin:
 forall (A:Type) (R: A -> A -> Prop) 
  (R_min_dec: forall (x:A), { forall (y:A), ~(R y x) } + { exists y:A, (R y x) }) 
   (x:A), (Acc R x) -> (exists y:A, (starRel R y x) /\ forall (z:A), ~(R z y)).
Proof.
  induction 2.
  case (R_min_dec x).
  intros H1; constructor 1 with (x:=x).
  intuition auto.
  apply starRel_refl.
  intros H1.
  case H1.
  intros x0 H2; case (H0 x0); auto.
  intros x1 H3; constructor 1 with (x:=x1).
  intuition auto.
  eapply starRel_trans.
  eapply H4.
  apply starRel_includes; auto.
Qed.

(* begin hide *)
Require Export ISL.
Require Export ISLWhile.
Require Import Iter.

Section ISLprogs.

(* end hide *)
(** ** A non-deterministic programming language and its semantics

In this subsection, I define the abstract syntax of a
non-deterministic expression language.  The role of this syntax is
mainly to give a support to express non-deterministic reduction rules
on a sub-language of DSL and the associated notion of
strong-termination.  This sub-language contains all basic constructions
except [ensure], and a [any] restricted to non-empty types. Moreover,
I explicitly introduce a [while] operator to express non-terminating
iteration. In particular, this language of expressions contains a
[require] operator which allows to express partial expressions like
division by zero or access out of the bounds of an array (see the
semantics below).  Non-determinism is also a useful feature to give a
semantics to ``real'' programming language aspects, such:
non-initialized variables, pseudo-random computations or undetermined
order of evaluation in some subexpressions.  It also allows to give a
very abstract view of interactions with environment, like keyboard
inputs.

I interpret this programming language for ISL, a state-handling DSL
formally defined chapter %\ref{chap:prover}%. The main particularity
of this DSL, is that it allows to deal with a local state. But I do
not use this feature here, and use it as a standard DSL. At this stage, 
the reader need only to know that the
type of an ISL specification is of the form [(DS St A)] where 
[St] is the type of its local state and [A]
the type of the result.
*)

(** Thus, I first assume [St] the type of the global state, and 
    interpret programs as ISL specifications in [(DS St)].  
*)

Variable St: Type.

(** I define the abstract syntax of the expression language by the
[prog] inductive type below.

A main characteristic of this abstract syntax is that it is %{\em
very}% abstract: 
  - the type of expressions [prog] is parametrized by the type of the
  result.  Hence, it is a typed syntax. Furthermore, I interpret values
  as %\coq{}% values (and types as %\coq{}% types).
  - variables and bindings (in the [bind] construction) are simulated
  by %\coq{}% functions. This avoids to explicitly redefine a notion of variable 
  substitution.
  - the definition of [bind] introduced what I call a 
  %{\em capture of structure}% by analogy with a capture of variable. 
  Indeed in an expression ``[bind p1 p2]'', [p2] is a function depending
  on the result of [p1]. But the syntactic structure of [p2] also depends 
  on this result. Hence, [p2] can be any %\coq{}% function, performing
  pattern-matching or structural recursion. For instance, even if the
  syntax does not introduce an ``if-then-else'', I can define it as%\\%
 [bind cond (fun b => if b then pt else pe)] where [cond]
  is of type [(prog bool)] and [pt], [pe] are of type [(prog A)] for
  any type [A]. Hence, the computational meaning of this syntax is
  defined by the operational semantics I give below, but also
  implicitly, by %\coq{}% reduction. However, this rules does not
  interfere with the result given here, since at our level, %\coq{}%
  reduction is deterministic and terminating. Let me remark however
  that by using a more complicated definition, I could avoid this
  capture of structure. But it does not seem necessary here. On the
  contrary, this capture of structure enriches the meaning of this
  language.
  
%\noindent% Here is the inductive definition

*)
 
Inductive prog: Type -> Type :=
| val: forall (A:Type), A -> (prog A)
| bind: forall (A B:Type), (prog A) -> (A -> (prog B)) -> (prog B)
| require: Prop -> (prog unit)
| any: forall (A:Type), (NonEmpty A) -> (prog A)
| get: (prog St)
| set: St -> (prog unit)
| while: (prog bool) -> (prog unit) -> (prog unit).

(** Let me note that this syntax allows only %{\em first-order}%
expressions.  Indeed, a high-order operator such [lambda] below is
%{\em not correct}% because it provokes a universe inconsistency.
[[
Definition lambda (A B:Type) (f:(A -> (prog B))): prog (A -> (prog B)) := (val f).
]]
It is possible to extend the syntax in order to support high-order expressions, but
it requires to introduce an abstract syntax for types: it thus a bit
more complex and I do not consider that here.%\\%

I now introduce the small-steps (operational) semantics for %{\em prog}%.  
As there is no syntax of ``if-then-else'' in the programming
language, I use the above trick to express one computation step of
[while]. Hence, I define 
[[
Definition unfold_while (cond:(prog bool)) (body:(prog unit)): (prog unit) 
 :=(bind cond (fun b => if b then (bind body (fun _ => (while cond body))) else (val tt))).
]]
*)
(* begin hide *)

(* un bug etrange de "coqdoc -latex" si on visualise ca !!! *)

Definition unfold_while (cond:(prog bool)) (body:(prog unit)): (prog unit) 
 :=(bind cond (fun b => if b then (bind body (fun _ => (while cond body))) else (val tt))).

(* end hide *)

(** The predicate [leadsto] defines one computation step on the
language.  %\footnote{As explained above, this computation-step may
contain arbitrary \coq{} computations.}% A computation transforms the
expression involved but also the global state.  More precisely, if [b]
is a boolean [(leadsto p st p' st' b)] means that it is possible to
computate a step from [p] in the state [st] to [p'] in the state
[st']. The boolean [b] indicates whether the computation has aborted:
if it is [true], a [require] clause has been violated.  Because of
[any] and [require], this semantics is non-deterministic.  In
particular, a [require] expression may not abort even if its clause
has been violated.  %\footnote{Alternatively, I could also force a
[require] expression to abort if the clause is not satisfied. The
correctness theorem would be preserved (only a few intermediate lemmas
need to be adjusted).  However, this semantics seems less interesting
than the one presented here since the computations may stuck on a
non-decidable property.}% 
*)

Inductive leadsto: forall (A:Type), (prog A) -> St -> (prog A) -> St -> bool -> Prop :=
| leadsto_bind:  
   forall (A B:Type) (p1a p1b:(prog A)) (p2: A -> (prog B)) (st1 st2:St) (ab:bool), 
      (leadsto p1a st1 p1b st2 ab) 
         -> (leadsto (bind p1a p2) st1 (bind p1b p2) st2 ab)
| leadsto_bind_val: 
   forall (A B:Type) (a:A) (p:A -> (prog B)) (st:St), 
                    (leadsto (bind (val a) p) st (p a) st false)
| leadsto_abort:
   forall (P:Prop) (st:St), 
      ~P -> (leadsto (require P) st (require P) st true)
| leadsto_require: 
   forall (P:Prop) (st:St), (leadsto (require P) st (val tt) st false)
| leadsto_any: 
   forall (A:Type) (An:NonEmpty A) (a:A) (st:St), (leadsto (any An) st (val a) st false)
| leadsto_get: 
   forall (st:St), (leadsto get st (val st) st false)
| leadsto_set: 
   forall (st1 st2:St), (leadsto (set st2) st1 (val tt) st2 false)
| leadsto_while: 
   forall (cond:(prog bool)) (body:(prog unit)) (st:St),
     (leadsto (while cond body) st (unfold_while cond body) st false).

(** Here are some explanations about this rules:
  - [leadsto_bind] are [leadsto_bind_val] are the two possible reduction rules 
for a [bind] expression. [leadsto_bind] propagates the reduction applicable on the left subexpression. 
  - [leadsto_abort] and [leadsto_require] give the semantics of [require].
When, the required property is not satisfied, the program may (but is not obliged to) abort. 
If the property is satisfied, the only applicable rule is [leadsto_require], and [require] reduces to [skip].
  - [leadsto_any] means that [any] reduces to any value [a] of the (non-empty) type, 
without changing the state.
  - [leadsto_get] and [leadsto_set] mean respectively that [get]
returns the current state without modifying it whereas [set] erases
the value of the state with its parameter.
  - [leadsto_while] means that [while] simply reduces to [unfold_while] 
without changing the state.

I have proved some trivial but fundamental properties of values: 
``to be a value'' is a decidable property of programs, 
and values are the only irreducible expressions via [leadsto]. Let me remark, that this last assertion
holds because [any] are applied to non-empty types, and there is no [ensure] construct in the 
language. In other word, we have the following lemma:
[[
Lemma leadsto_values:
  forall (A:Type) (p:(prog A)) (st:St),
     { (exists a:A, p=(val a)) /\ forall (p':(prog A)) (st':St), ~(leadsto p st p' st' false)} + 
              { exists p':(prog A), exists st':St, (leadsto p st p' st' false)}.
]]
*)

(* begin hide *)

Hint Resolve leadsto_bind_val leadsto_abort leadsto_bind leadsto_require
  leadsto_any leadsto_get leadsto_set leadsto_while: islprog.

(* end hide *)

(** I now define [leadstostar] as kind of reflexive and transitive closure of [leadsto], using an inductive definition.
Actually [leadstostar] is not exactly a reflexive and transitive closure, because in case of abortion, 
the reduction stops. 
*)

Inductive leadstostar (A:Type):  (prog A) -> St -> (prog A) -> St -> bool -> Prop :=
 | leadstostar_refl: forall (p:(prog A)) (st:St), (leadstostar p st p st false)
 | leadstostar_next_false: forall (p1 p2 p3: (prog A)) (st1 st2 st3:St) (b:bool),
     (leadsto p1 st1 p2 st2 false) -> 
       (leadstostar p2 st2 p3 st3 b) -> (leadstostar p1 st1 p3 st3 b)
 | leadstostar_next_true: forall (p1 p2: (prog A)) (st1 st2:St),
     (leadsto p1 st1 p2 st2 true) -> (leadstostar p1 st1 p2 st2 true).

(** Actually, restricted on the non-aborting cases, [leadstostar] is
the reflexive and transitive closure of [leadsto] as stated by the
following lemma:
[[
Lemma leadstostar_starRel:
  forall (A:Type) (p1 p2:(prog A)) (st1 st2:St),
    (leadstostar p1 st1 p2 st2 false) <-> 
        (starRel  
             (fun (y x:(prod (prog A) St)) =>
                      leadsto (fst x) (snd x) (fst y) (snd y) false)
             (pair p2 st2) (pair p1 st1)).
]]

Thus, the picture of the theorem sketched section
%\ref{subsec:starRel}% is recovered when considering non-aborting
computations. As we will see below, under the hypotheses of the
correctness theorem, there are only non-aborting computations. 
*)

(* begin hide *)
Hint Resolve leadstostar_refl leadstostar_next_false leadstostar_next_true: islprog.

(* end hide *)
(** ** Statement of the total-correctness theorem  
I now introduce a predicate [semDS] relating a program [p] to its ISL
interpretation [ps].  The important point is that I interpret the
[while] operator as [whileWF], because I am here interested in %{\em
total}% correctness. Other operators are directly interpreting by
their corresponding one in ISL.  Let me remark that [semDS] relation
is not a function because in the case of the [while] operator, several
interpretations are possible, depending on the invariant.

Let me now explain some syntactic details about specifications in ISL
(see chapter %\ref{chap:prover}% for more details):
  - [(Val a)%ISL], [(Req P)%ISL] and [(Any A)%ISL] 
are respectively the ISL notations for the usual [(val a)],
[(require P)] and [(any A)].
  - [(ISL.bind allPl pf1 pf2)] is the ISL notation for [(bind pf1 pf2)]. 
The term [allPl] means here that [pf1] works on the global state [St].
  - [(!allPl)%ISL] and [(allPl :<- st)%ISL] are respectively reading and writing in [allPl] 
the global state. Thus, they are respectively the ISL notation for [get] and [set st].
*)

Inductive semDS: forall (A:Type), (prog A) -> (DS St A) -> Prop :=
| val_proof: forall (A:Type) (a:A), (semDS (val a) (Val a)%ISL)
| bind_proof: forall (A B:Type) (p1:(prog A)) (p2: A -> (prog B)) 
   (pf1:(DS St A)) (pf2: A -> (DS St B)),
     (semDS p1 pf1) -> (forall (x:A),(semDS (p2 x) (pf2 x)))
       -> (semDS (bind p1 p2) (ISL.bind allPl pf1 pf2))
| require_proof: forall (P:Prop), (semDS (require P) (Req P)%ISL)
| any_proof: forall (A:Type) (An:NonEmpty A), (semDS (any An) (Any A)%ISL)
| get_proof: (semDS get (!allPl)%ISL)
| set_proof: forall (st:St), (semDS (set st) (allPl :<- st)%ISL)
| while_proof: forall (cond:(prog bool)) (body:(prog unit)) 
   (inv: St -> Prop) (R: St -> St -> Prop)
   (condpf: (DS St bool)) (bodypf: (DS St unit)),
     (semDS cond condpf)
      -> (semDS body bodypf)
        -> (semDS (while cond body) (whileWF condpf bodypf inv R)).
(* begin hide *)
Hint Resolve val_proof bind_proof require_proof any_proof get_proof set_proof while_proof: islprog.
(* end hide *)

(** Hence, I can now formally write the statement of the theorem of this section. 
In this statement, [seq] of ISL is simply denoted by the infix symbol ``[-;]''.
[[
Theorem semDS_correct:
  forall (A:Type) (p:(prog A)) (ps:(DS St A)) (st:St),
   (semDS p ps)
     -> (pre (ps st))
        ->  (forall (p':(prog A)) (st':St) (b:bool),
              (leadstostar p st p' st' b) -> 
                 b=false
                 /\ (exists ps':(DS St A), 
                       (semDS p' ps') /\ (refInEnv st ((allPl :<- st') -; ps')  ps))%ISL)
           /\ (Acc (fun (y x:(prod (prog A) St)) =>
                      leadsto (fst x) (snd x) (fst y) (snd y) false)
                    (pair p st)).   
]]

This theorem states that if the initial state satisfies the precondition of the initial
expression, then computations can not abort. We thus recover the
picture described section %\ref{subsec:starRel}%.  Hence, the theorem
is mainly divided in two parts. First, the left side of the
conjunction in the conclusion corresponds to the preservation of the
refinement semantics along arbitrary [leadstostar] reductions. And
second, the right side of the conjunction corresponds to termination
of an arbitrary sequence of [leadsto] reductions. In the following, I
name the left-side as the %{\em preservation property}% and the
right-side as the %{\em termination property}%.

I give the intermediary steps of the proof in the next subsection.
Mainly, I first need to prove the preservation, because this property
is used in the proof of the termination.  Indeed, in this last proof,
I need to prove that the state remains in the precondition of the
computed expression, along reductions. This property is a direct
consequence of the preservation property.

** Main steps to prove this result 
I now explain some details about the %\coq{}% proof of this result,
because this proof has an interest by the %\coq{}% technique
involved. Indeed, the preceding result has been proved in the ground
theory of %\coq{}% without any variant of Streicher's %{\em K}% axiom!
It is quite surprising, because there are many type dependencies in
the preceding inductive definitions. The inversion of these kind of
inductive definitions requires usually this axiom. For instance, with
the previous definition of [prog], a simple statement such 
[(val a)=(val b) -> a=b] is not provable without %{\em K}% axiom.

In this proof, I have simply redefine [leadsto], [leadstostar] and
[semDS] using a continuation passing style. I still use inductive
types, but the continuation style give me a better inversion.  This
approach has thus two advantages. First, it avoids to add an axiom.
And more important (because, it is known that this axiom is consistent
with %\coq{}% theory), it avoids the tedious proofs I have when each
inversion require to invoke [inj_pair2] (a consequence of %{\em K}%
axiom). Actually, by using such a continuation passing style, we
recover some kinds of weakest-precondition calculi. 

Actually, the continuation technique used here is quite similar to the
introduction of ``evaluation contexts'' usually associated with such
operational semantics.
*)

(** *** Step 1: Reformulating [leadsto], [leadstostar] and [semDS] using angelic WP.
 
Hence, I first begin by defining a kind of angelic
weakest-precondition associated to [leadsto]. Informally, given an
expression [p] of type [(prog A)] and an initial state [st], and a
predicate [R] of type [((prog A) -> St -> Prop)], I define [areduce p st R] as the
weakest-precondition ensuring that there exists [p'] and [st'] such
[(leadsto p st p' st' false)] and [(R p' st')]. I do not consider here aborting computations, 
since they can not happen under the hypotheses of the correctness theorem. Formally, here is the
definition of [areduce]:
*)

Inductive areduce: forall (A:Type), (prog A) -> St -> ((prog A) -> St -> Prop) -> Prop :=
| areduce_bind_val: 
  forall (A B:Type) (a:A) (p:A -> (prog B)) (st:St) (R: (prog B) -> St -> Prop) ,
      (R (p a) st) -> (areduce (bind (val a) p) st R) 
| areduce_bind_nonval: 
   forall (A B:Type) (p1:(prog A)) (p2: A -> (prog B)) (st:St) (R: (prog B) -> St -> Prop), 
      (areduce p1 st (fun p' st' => (R (bind p' p2) st')))
        ->  (areduce (bind p1 p2) st R)
| areduce_require: forall (P:Prop) (st:St) (R:(prog unit) -> St -> Prop),
    (R (val tt) st) -> (areduce (require P) st R)
| areduce_any: 
    forall (A:Type) (An:NonEmpty A) (a:A) (st:St) (R:(prog A) -> St -> Prop), 
       (R (val a) st) -> (areduce (any An) st R)
| areduce_get: 
    forall (st:St) (R:(prog St) -> St -> Prop), 
        (R (val st) st) -> (areduce get st R)
| areduce_set: 
    forall (st1 st2:St) (R:(prog unit) -> St -> Prop),
       (R (val tt) st2) -> (areduce (set st2) st1 R)
| areduce_while:
    forall (st:St) (cond: (prog bool)) (body: (prog unit)) (R:(prog unit) -> St -> Prop), 
     (R (unfold_while cond body) st)
      -> (areduce (while cond body) st R).
(* begin hide *)
Hint Resolve areduce_bind_val areduce_bind_nonval areduce_require areduce_any areduce_get areduce_set
 areduce_while: islprog.
(* end hide *)

(** Actually, [areduce p st R] could be alternatively defined as a
fixpoint. Defining it as an inductive type provides an elimination
scheme that simplifies some proofs: indeed, its case analysis pattern is
different from [prog].  I will now construct a predicate [areduceinv]
exactly equivalent to [areduce], but such inversion of [areduce]
corresponds to a reduction step on [areduceinv].  But first, let me
introduce a predicate [is_val] indicating whether a program is a value
or not.  *)

Definition is_val (A:Type) (p:(prog A)) : Prop :=
  match p with
  | (val _ a) => True
  | _ => False
  end.

(** Of course, [is_val] is trivally a decidable predicate. *)

Lemma is_val_dec: forall (A:Type) (p:(prog A)), { (is_val p) } + { ~(is_val p) }.
Proof.
  intros A p; case p; simpl; intuition auto.
Defined.

(** Given a program [p: (prog A)], a predicate [R: A -> Prop] and a
proof [H:(is_val p)], I build the weakest-condition [(v_inv p R H)]
such [R] holds for the underlying value of [p]. Hence, [v_inv] is an
inverting function for [is_val] predicate. *) 

Definition v_inv (A:Type) (p:(prog A)): (A -> Prop) -> (is_val p) -> Prop
 := match p in (prog A) return (A -> Prop) -> (is_val p) -> Prop with
    | (val A a) => fun R H => (R a)
    |   x => fun R absurd => match absurd return Prop with end
    end.

(** [v_inv] satisfies the following principle: *)

Lemma v_inv_ind: forall (A:Type) (P: (prog A) -> Prop) (R: A -> Prop),
  (forall (a:A), (R a) -> (P (val a)))
   -> forall (p:(prog A)) (H:(is_val p)), (v_inv p R H) -> (P p).
Proof.
  intros A P R HR p; generalize P R HR; case p; simpl; 
  intuition auto.
Qed.

(* begin hide *)
Lemma is_val_inv: forall (A:Type) (p:(prog A)),
  (is_val p) -> exists a:A, p=(val a).
Proof.
  intros A p; case p; simpl; intuition eauto.
Qed.
(* end hide *)

(** This allows now to define [areduceinv]. In the [bind] case, I proceed by a
case analysis on [(is_val_dec p1)]. Alternatively, I could also
directly proceed by case analysis on [p1]. But, in proofs, it is more convenient
to invert using  [is_val_dec]. Indeed, this introduces only two
cases, instead of seven cases (six cases, being exactly treated in the
same way).
*)
Definition areduceinv (A:Type) (p:(prog A)) (st:St): ((prog A) -> St -> Prop) -> Prop :=
  match p in (prog A) return ((prog A) -> St -> Prop) -> Prop with
  | (val A a) => fun R => False
  | (bind A B p1 p2) => fun R => 
     match (is_val_dec p1) with
     | (left H) => (v_inv _ (fun a => (R (p2 a) st)) H)
     | _ => areduce p1 st (fun p' st' => R (bind p' p2) st')
     end
  | require P => fun R => (R (val tt) st)
  | (any A p) => fun R => exists a:A,(R (val a) st)
  | get => fun R => (R (val st) st)
  | (set v) => fun R => (R (val tt) v)
  | (while cond body) => fun R => (R (unfold_while cond body) st)
  end.

(** As announced, [areduceinv] and [areduce] are equivalent. The
first lemma below, is proved by induction on [areduce], while the
second is proved by induction on [p]. *) 
Lemma areduce_areduceinv: 
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop),
   (areduce p st R) -> (areduceinv p st R).
Proof.
 induction 1; try (simpl; eauto; fail).
 generalize p2 H IHareduce; case p1; clear A p1 p2 H IHareduce; simpl; auto.
 intros A a p2 H; inversion H.
Qed.

Lemma areduceinv_areduce: 
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop),
   (areduceinv p st R) -> (areduce p st R).
Proof.
  induction p; try (simpl; eauto with islprog; fail).
  (* cas val *) simpl; tauto.
  (* cas bind *) 
  intros st R; simpl; case (is_val_dec p);
  simpl; auto with islprog.  
  intros H0 X; pattern p; elim X using v_inv_ind.
  auto with islprog.
  (* cas any *)
  intros st R H; case H; eauto with islprog.
Qed.  
(* begin hide *)
Hint Resolve areduceinv_areduce areduce_areduceinv: islprog.
(* end hide *)

(** The following lemmas prove that [~(is_val p)] is equivalent
to [forall (st:St), (areduce p st (fun _ _ => True))]. In this case, 
[p] is said %{\em reducible}%. Let me first remark, that without the non-emptiness condition
on [any] programs, the first lemma below would be false. Second, the second lemma is more 
general than the right side of the equivalence. But, this trivial property is crucial in the 
following. It gives a way to generalize a reducibility hypothesis into a nice inductive property
(see [checkDS_preservation_aux] below).
*)
Lemma non_val_reducible: forall (A:Type) (p:(prog A)), 
  ~(is_val p) -> forall (st:St), (areduce p st (fun _ _ => True)).
Proof.
  induction p; simpl; auto with islprog.
  tauto.
  intros H0 st; generalize IHp p0 H; case p;
  clear A p IHp p0 H; simpl; auto with islprog.
  case n. intros a X st; apply areduce_any with (a:=a).
  auto.
Qed.
(* begin hide *)
Hint Resolve non_val_reducible: islprog.
(* end hide *)

Lemma areduce_nonval: forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop),
  (areduce p st R) -> ~(is_val p).
Proof.
  intros A p st R H; case H; simpl; auto.
Qed.

(** Now, I prove an important property of [areduce]. It is an angelic
weakest-preconditions, and thus, it admits an angelic normal form. I prove
this lemma on [areduceinv] instead of [areduce]. The proof is made by
induction on [p]. Let me remark, that both sides of the equivalence
must be proved simultaneously.
*) 
Lemma areduce_angelwp: 
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop), 
   (areduceinv p st R) <-> 
    exists p':(prog A), exists st':St, 
      ((areduceinv p st (fun p0 st0 => p0=p' /\ st0=st'))  /\ (R p' st')).
Proof.
  induction p.
  (* val *)
  simpl.
  firstorder auto.
  (* bind *)
  intros st R; clear H.
    case (IHp st (fun p' st' => (R (bind p' p0) st'))).
  intros H H0.
  simpl; case (is_val_dec p).  
   (* val *)
   generalize p0; case p; simpl; intuition eauto.
   abstract firstorder subst; eauto.
   (* non_val *)
   intros; intuition auto.
     (* -> *)
     decompose [ex Logic.and] (H (areduce_areduceinv H1)).
     constructor 1 with (x:=(bind x p0)).
     constructor 1 with (x:=x0); intuition auto with islprog.
     case (IHp st (fun p' st' => bind p' p0 = bind x p0 /\ st' = x0)).
     eauto 15 with islprog.
     (* <- *)
     decompose [ex Logic.and] H1.
     apply areduceinv_areduce.
     apply H0.
     case (IHp st (fun p' st' => bind p' p0 = x /\ st' = x0)).
     intros X1 X2.
     decompose [ex Logic.and] (X1 (areduce_areduceinv H2)) ;
     subst; eauto.
  (* require *)
  simpl; firstorder auto.
  subst; auto.
  (* any *)
  simpl.
  firstorder eauto.
  subst; eauto.
  (* get *)
  simpl; firstorder auto.
  subst; auto.
  (* set *)
  simpl; firstorder auto.
  subst; auto.
  (* while *)
  simpl.
  intuition eauto.
  decompose [ex Logic.and] H.
  subst; auto.
Qed.

(* begin hide *)
Lemma areduce_unfold:
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop), 
   (areduce p st R) -> 
    exists p':(prog A), exists st':St, 
      ((areduce p st (fun p0 st0 => p0=p' /\ st0=st'))  /\ (R p' st')).
Proof.
  intros A p st R; case (areduce_angelwp p st R).
  intros H H0 H1;  decompose [ex Logic.and] (H (areduce_areduceinv H1)).
  eauto with islprog.
Qed.

Lemma areduce_fold:
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop) (p':(prog A)) (st':St), 
  (areduce p st (fun p0 st0 => p0=p' /\ st0=st')) -> (R p' st') -> (areduce p st R).
Proof.
  intros A p st R; case (areduce_angelwp p st R).
  intros H H0 H1;  firstorder eauto with islprog.
Qed.
(* end hide *)

(** A standard consequence of such a normal form property, is the weakening rule.
*)
Lemma areduce_weaken:  
 forall (A:Type) (p:(prog A)) (st:St) (P Q:(prog A) -> St -> Prop), 
   (areduce p st P) -> 
     (forall p' st', (P p' st') -> (Q p' st')) -> 
        (areduce p st Q).
Proof.
  intros A p st P Q H1 H2.
  decompose [ex Logic.and] (areduce_unfold H1).
  clear H1; eapply areduce_fold; eauto.
Qed.

(** As [areduce] is an angelic wp for [leadsto],
 I now define [areducestar] an angelic wp for [leadstostar].
*)
Definition areducestar (A:Type) (R:(prog A) -> St -> Prop) (p:(prog A)) (st:St)
 := exists n:nat, (iter n (fun R p st => (areduce p st R)) R p st).

(** The three following lemmas are trivially derived from [areducestar] definition.
*)
Lemma areducestar_refl: forall (A:Type) (R:(prog A) -> St -> Prop) (p:(prog A)) (st:St),
  (R p st) -> (areducestar R p st).
Proof.
  unfold areducestar ; constructor 1 with (x:=O); simpl; auto.
Qed.

Lemma areduce_areducestar: forall (A:Type) (R:(prog A) -> St -> Prop) (p:(prog A)) (st:St),
  (areduce p st R) -> (areducestar R p st).
Proof.
  unfold areducestar ; constructor 1 with (x:=(1)); simpl; auto.
Qed.

Lemma areduce_areducestar_trans:
 forall (A:Type) (R:(prog A) -> St -> Prop) (p:(prog A)) (st:St),
  (areduce p st (areducestar R)) -> (areducestar R p st).
Proof.
  intros A R p st H; decompose [ex Logic.and] (areduce_unfold H).
  case H2.
  intros n H3.
  constructor 1 with (x:=(S n)).
  simpl; eapply areduce_fold; eauto.
Qed.

(* begin hide *)


Lemma areducestar_trans: forall (A:Type) (R:(prog A) -> St -> Prop) (p:(prog A)) (st:St),
  (areducestar (areducestar R) p st) -> (areducestar R p st).
Proof.
  intros A R p st H; case H; clear H.
  intros n; generalize R p st; elim n; clear n R p st; simpl; auto.
  intros n HR R p st H.
  apply areduce_areducestar_trans.
  eapply areduce_weaken; eauto.
Qed.

Hint Resolve areducestar_refl areduce_areducestar areducestar_trans: islprog.
(* end hide *)

(** The two following lemmas correspond to the normal form property of
angelic weakest-preconditions. Here, the property has been split in two lemmas, because 
the two sides of the equivalence can be proved independently. In both cases, the proof is done
by induction on the [n] of the [areducestar] in hypothesis.
*) 
Lemma areducestar_unfold: forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop), 
   (areducestar R p st) -> 
    exists p':(prog A), exists st':St, 
      ((areducestar (fun p0 st0 => p0=p' /\ st0=st') p st)  /\ (R p' st')).
Proof.
  intros A p st R H; case H; clear H.
  intros n; generalize p st R; elim n; clear n p st R; simpl.
  (* cas 0 *)
  intros.
  constructor 1 with (x:=p).
  constructor 1 with (x:=st).
  eauto with islprog.
  (* cas (S n) *)
  intros n HR p st R H; decompose [ex Logic.and] (areduce_unfold H); clear H.
  decompose [ex Logic.and] (HR _ _ _ H2).
  constructor 1 with (x:=x1).
  constructor 1 with (x:=x2).
  intuition auto.
  apply areduce_areducestar_trans.
  eapply areduce_weaken; eauto.
  simpl; intuition auto; subst.
  auto.
Qed.

Lemma areducestar_fold:
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop) (p':(prog A)) (st':St), 
  (areducestar (fun p0 st0 => p0=p' /\ st0=st') p st) 
    -> (R p' st') -> (areducestar R p st).
Proof.
  intros A p st R p' st' H; case H; clear H.
  intros n; generalize p st R p' st'; elim n; clear n p st R p' st'; simpl.
  (* cas 0 *)
  intros; intuition auto; subst.
  auto with islprog.
  (* cas (S n) *)
  intros n HR p st R p' st' H1 H2.
  decompose [ex Logic.and] (areduce_unfold H1); clear H1.
  apply areduce_areducestar_trans.
  eapply areduce_weaken; eauto.
  simpl; intuition auto; subst.
  eauto.
Qed.

(** Of, course, I also derive a weakening lemma from the two preceding lemmas.
*)
Lemma areducestar_weaken:  
 forall (A:Type) (p:(prog A)) (st:St) (P Q:(prog A) -> St -> Prop), 
   (areducestar P p st) -> 
     (forall p' st', (P p' st') -> (Q p' st')) -> 
        (areducestar Q p st).
Proof.
  intros A p st P Q H1 H2.
  decompose [ex Logic.and] (areducestar_unfold H1).
  clear H1; eapply areducestar_fold; eauto.
Qed.

(* begin hide *)
(*
Lemma areducestar_val: forall (A:Type) (a:A) (st:St) (R: (prog A) ->St ->Prop),
  (areducestar R (val a) st) -> (R (val a) st).
Proof.
  intros A a st R H; case H.
  intros n; case n; clear n; simpl; auto.
  intros n X; case (areduce_nonval X).
  simpl; auto.
Qed.
*)
(* end hide *)

(** I now introduce a decomposition lemma to prove a property such
[(areducestar R (bind p1 p2) st)].  In particular, I want to reduce the
proof of such a property, to a proof of [(areducestar R' p1 st)] for
some [R'] that is to find. At a first step, I prove the decomposition lemma 
for [R'] being [(fun p' st' => (areducestar R (bind p' p2) st'))].
The proof is done by induction on the [n] of the [areducestar] hypothesis.
*)
Lemma areducestar_bind:
  forall (A B:Type) (p1:(prog A)) (p2:A -> (prog B)) (st:St)
    (R: (prog B) -> St -> Prop), 
   (areducestar (fun p' st' => (areducestar R (bind p' p2) st')) p1 st)
      -> (areducestar R (bind p1 p2) st).
Proof.
  intros A B p1 p2 st R H; case H; clear H.
  intros n; generalize p1 p2 st R; elim n; clear p1 p2 st R n; simpl; auto.
  intros n HR p1 p2 st R H.
  decompose [ex Logic.and] (areduce_unfold H); clear H.
  apply areduce_areducestar_trans.
  apply areduce_bind_nonval.
  eapply areduce_weaken; eauto.
  simpl; intros; intuition auto; subst.
  auto.
Qed.

(** Then I prove the decomposition lemma for [R'] being 
[(fun p' st' => (exists a:A, p'=(val a) /\ (areducestar R (p2 a) st')))].
The proof is simply a weakening on the preceding lemma.
*)
Lemma areducestar_bind_val:
  forall (A B:Type) (p1:(prog A)) (p2:A -> (prog B)) (st:St)
    (R: (prog B) -> St -> Prop), 
   (areducestar (fun p' st' => 
        (exists a:A, p'=(val a) /\ (areducestar R (p2 a) st'))) p1 st)
     -> (areducestar R (bind p1 p2) st).
Proof.
  intros; apply areducestar_bind.
  eapply areducestar_weaken; eauto.
  simpl; intros p' st' H1.
  decompose [ex Logic.and] H1; subst.
  apply areduce_areducestar_trans;
  auto with islprog.
Qed.

(* begin hide *)
(*
Lemma areducestar_disj:
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop),
   (areducestar R p st) -> (R p st) \/ (areduce p st (areducestar R)).
Proof.
  intros A p st R H; case H.
  intros n; case n; clear n; simpl; auto.
  intros n H1; constructor 2.
  eapply areduce_weaken.
  eauto.
  intros;
  constructor 1 with (x:=n).
  simpl; auto.  
Qed.
*)
(* end hide *)

(** I now define [checkDS] predicate, which is the angelic wp
corresponding to [semDS]. On the contrary of [areduce], I define it by
a fixpoint. Indeed, the corresponding inductive predicate is not
correct, because of a non-positive occurrence in the [bind]
case. Furthermore, the corresponding elimination scheme, would not
provide a case analysis different from the one of [prog].
*)
Fixpoint checkDS (A:Type) (p:(prog A)) {struct p}: ((DS St A) -> Prop) -> Prop :=
  match p in (prog A) return ((DS St A) -> Prop) -> Prop with
  | (val A a) => fun R => (R (Val a)%ISL)
  | (bind A B p1 p2) => fun R => 
      (exists pf2: A -> (DS St B),
        (checkDS p1 (fun pf1 => 
          (R (ISL.bind allPl pf1 pf2))
           /\ forall (a:A), (checkDS (p2 a) (fun pf2' => pf2'=(pf2 a))))))
  | require P => fun R => (R (Req P)%ISL)
  | (any A An) => fun R => (R (Any A)%ISL)
  | get => fun R => (R (!allPl)%ISL)
  | set v => fun R => (R (allPl :<- v)%ISL)
  | (while cond body) => 
       fun R => (checkDS cond (fun c => (checkDS body (fun b => 
          exists inv, exists W, (R (whileWF c b inv W))))))
  end.

(** Hence, [checkDS] satisfies a normal form property for angelic weakest-precondition. 
As usual, I derive from this lemma a weakening lemma.
*)
Lemma checkDS_angelwp: 
 forall (A:Type) (p:(prog A)) (R:(DS St A) -> Prop), 
   (checkDS p R) <-> 
    (exists ps:(DS St A), (checkDS p (fun ps0 => ps0=ps)) /\ (R ps)).
Proof.
  induction p; simpl.
  (* val *) firstorder (try subst; eauto).
  (* bind *) intros R.  constructor 1.
    (*  -> *)
    intros H1; case H1; clear H1.
    intros pf2 H1.
    case (IHp 
      (fun pf1 : DS St A =>
        R (ISL.bind allPl pf1 pf2) /\
        (forall a : A,
         checkDS (p0 a) (fun pf2' : DS St B => pf2' = pf2 a)))).
    intros IHp1 IHp2; clear IHp2.
    decompose [ex Logic.and] (IHp1 H1); clear IHp1 H1.
    constructor 1 with (x:=(ISL.bind allPl x pf2));
    intuition auto.
    constructor 1 with (x:=pf2).
    case (IHp (fun pf1 : DS St A =>
      ISL.bind allPl pf1 pf2 = ISL.bind allPl x pf2 /\
      (forall a : A, checkDS (p0 a) (fun pf2' : DS St B => pf2' = pf2 a)))).
    intros IHp1 IHp2; apply IHp2; clear IHp2 IHp1.
    intuition eauto.
    (* <- *)
    intros H1; decompose [ex Logic.and] H1; clear H1.
    constructor 1 with (x:=x0).
    case (IHp (fun pf1 : DS St A =>
        ISL.bind allPl pf1 x0 = x /\
        (forall a : A, checkDS (p0 a) (fun pf2' : DS St B => pf2' = x0 a)))).
    intros IHp1 IHp2; decompose [ex Logic.and] (IHp1 H0); clear IHp1 IHp2 H0.
    subst; case (IHp (fun pf1 : DS St A =>
              (R (ISL.bind allPl pf1 x0)) /\
               (forall a : A, checkDS (p0 a) (fun pf2' : DS St B => pf2' = x0 a)))).
    intros IHp1 IHp2; apply IHp2; clear IHp1 IHp2.
    eauto.
 (* require *) firstorder (try subst; eauto).
 (* any *) firstorder (try subst; eauto).
 (* get *) firstorder (try subst; eauto).
 (* set *) firstorder (try subst; eauto).
 (* while *) intros R.  constructor 1.
   (* sens -> *)
   intros H1. 
   case (IHp1 (fun c : DS St bool =>
      checkDS p2
          (fun b : DS St unit =>
           exists inv : St -> Prop,
             (exists W : St -> St -> Prop, R (whileWF c b inv W)))));
   intros IH1 IH2.
   decompose [ex Logic.and] (IH1 H1); clear H1 IH1 IH2.
   case (IHp2 (fun b : DS St unit =>
        exists inv : St -> Prop,
          (exists W : St -> St -> Prop, R (whileWF x b inv W)))).
   intros IH1 IH2.
   decompose [ex Logic.and] (IH1 H2); clear H2 IH1 IH2.
   constructor 1 with (x:=(whileWF x x0 x1 x2)).
   intuition auto.
   case (IHp1  (fun c : DS St bool =>
     checkDS p2
       (fun b : DS St unit =>
         exists inv : St -> Prop,
           (exists W : St -> St -> Prop, whileWF c b inv W = whileWF x x0 x1 x2))));
   intros IH1 IH2.
   apply IH2; clear IH1 IH2.
   constructor 1 with (x:=x).
   intuition auto.
   case (IHp2 (fun b : DS St unit =>
     exists inv : St -> Prop,
      (exists W : St -> St -> Prop, whileWF x b inv W = whileWF x x0 x1 x2)));
   intros IH1 IH2; apply IH2; clear IH1 IH2.
   constructor 1 with (x:=x0).
   intuition eauto.  
   (* sens <- *)
    intros H1; decompose [ex Logic.and] H1; clear H1.
    case (IHp1 (fun c : DS St bool =>
     checkDS p2
       (fun b : DS St unit =>
        exists inv : St -> Prop,
          (exists W : St -> St -> Prop, R (whileWF c b inv W)))));
    intros IH1 IH2; apply IH2; clear IH1 IH2.
    case (IHp1 (fun c : DS St bool =>
        checkDS p2
          (fun b : DS St unit =>
           exists inv : St -> Prop,
             (exists W : St -> St -> Prop, whileWF c b inv W = x)))); 
   intros IH1 IH2; decompose [ex Logic.and] (IH1 H0); clear H0 IH1 IH2.
   constructor 1 with (x:=x0).    
   intuition auto.
   case (IHp2 (fun b : DS St unit =>
        exists inv : St -> Prop,
          (exists W : St -> St -> Prop, whileWF x0 b inv W = x)));
   intros IH1 IH2; decompose [ex Logic.and] (IH1 H3); clear H3 IH1 IH2.
   subst.
   case (IHp2 (fun b : DS St unit =>
   exists inv : St -> Prop,
     (exists W : St -> St -> Prop, R (whileWF x0 b inv W))));
   intros IH1 IH2; apply IH2; clear IH1 IH2.
   eauto.
Qed.

(* begin hide *)
Lemma checkDS_unfold:
 forall (A:Type) (p:(prog A)) (R:(DS St A) -> Prop), 
   (checkDS p R) -> 
    (exists ps:(DS St A), (checkDS p (fun ps0 => ps0=ps)) /\ (R ps)).
Proof.
  intros A p R H; case (checkDS_angelwp p R); auto.
Qed.

Lemma checkDS_fold:
 forall (A:Type) (p:(prog A)) (R:(DS St A) -> Prop) (ps:(DS St A)), 
  (checkDS p (fun ps0 => ps0=ps)) -> (R ps) -> 
     (checkDS p R).
Proof.
  intros A p R H1 H2; case (checkDS_angelwp p R).
  firstorder eauto.  
Qed.

Lemma checkDS_weaken:  
 forall (A:Type) (p:(prog A)) (P Q:(DS St A) -> Prop), 
   (checkDS p P) -> 
     (forall ps, (P ps) -> (Q ps)) -> 
        (checkDS p Q).
Proof.
  intros A p P Q H1 H2.
  decompose [ex Logic.and] (checkDS_unfold _ _ H1).
  clear H1; eapply checkDS_fold; eauto.
Qed.
(* end hide *)

(** *** Step 2: proving a reformulation of the preservation property.

I now want to prove the first part of the theorem: the preservation of 
the refinement semantics along arbitrary [leadstostar] reduction.
In order to express the notion of %{\em
arbitrary reductions}%, I introduce a demonic weakest-precondition
for [leadstostar], called [dreducestar]. I first define a demonic
weakest-precondition for [leadsto], called [dreduce]. Construction of
[dreduce] and [dreducestar] is dual to [areduce] and [areducestar]
respectively. Hence, the only differences between [areduce] and
[dreduce] are the treatment of [any] and [require].  For [any],
[dreduce] imposes that the postcondition holds for %{\em all}% values
of the type.  For [require], [dreduce] requires a proof of the
required property.
*)
Inductive dreduce: forall (A:Type), (prog A) -> St -> ((prog A) -> St -> Prop) -> Prop :=
| dreduce_bind_val: 
  forall (A B:Type) (a:A) (p:A -> (prog B)) (st:St) (R: (prog B) -> St -> Prop) ,
      (R (p a) st) -> (dreduce (bind (val a) p) st R) 
| dreduce_bind_nonval: 
   forall (A B:Type) (p1:(prog A)) (p2: A -> (prog B)) (st:St) (R: (prog B) -> St -> Prop), 
      (dreduce p1 st (fun p' st' => (R (bind p' p2) st')))
        ->  (dreduce (bind p1 p2) st R)
| dreduce_require: forall (P:Prop) (st:St) (R:(prog unit) -> St -> Prop),
    P -> (R (val tt) st) -> (dreduce (require P) st R)
| dreduce_any: 
    forall (A:Type) (An:NonEmpty A) (st:St) (R:(prog A) -> St -> Prop), 
       (forall (a:A), (R (val a) st)) -> (dreduce (any An) st R)
| dreduce_get: 
    forall (st:St) (R:(prog St) -> St -> Prop), 
        (R (val st) st) -> (dreduce get st R)
| dreduce_set: 
    forall (st1 st2:St) (R:(prog unit) -> St -> Prop),
       (R (val tt) st2) -> (dreduce (set st2) st1 R)
| dreduce_while:
    forall (st:St) (cond: (prog bool)) (body: (prog unit)) (R:(prog unit) -> St -> Prop), 
     (R (unfold_while cond body) st)
      -> (dreduce (while cond body) st R).
(* begin hide *)
Hint Resolve dreduce_bind_val dreduce_bind_nonval dreduce_require dreduce_any dreduce_get 
dreduce_set dreduce_while: islprog.
(* end hide *)

(** As for [areduce], I define a [dreduceinv] predicate equivalent to [dreduce], 
and such inversion of
[dreduce] corresponds to reduction of [dreduceinv].
*)
Definition dreduceinv (A:Type) (p:(prog A)) (st:St): ((prog A) -> St -> Prop) -> Prop :=
  match p in (prog A) return ((prog A) -> St -> Prop) -> Prop with
  | (val A a) => fun R => False
  | (bind A B p1 p2) => fun R => 
     match (is_val_dec p1) with
     | (left H) => (v_inv _ (fun a => (R (p2 a) st)) H)
     | _ => dreduce p1 st (fun p' st' => R (bind p' p2) st')
     end
  | require P => fun R => P /\ (R (val tt) st)
  | (any A An) => fun R => forall a:A,(R (val a) st)
  | get => fun R => (R (val st) st)
  | (set v) => fun R => (R (val tt) v)
  | (while cond body) => fun R => (R (unfold_while cond body) st)
  end.

(* begin hide *)
Lemma dreduce_dreduceinv: 
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop),
   (dreduce p st R) -> (dreduceinv p st R).
Proof.
 induction 1; try (simpl; eauto; fail).
 generalize p2 H IHdreduce; case p1; clear A p1 p2 H IHdreduce; simpl; auto.
 intros A a p2 H; inversion H.
Qed.

Lemma dreduceinv_dreduce: 
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop),
   (dreduceinv p st R) -> (dreduce p st R).
Proof.
  induction p; try (simpl; intuition eauto with islprog; fail).
  (* cas bind *) 
  intros st R; generalize IHp p0 H; 
  case p; clear IHp p0 H A p;
  simpl; auto with islprog.  
Qed.  

Hint Resolve dreduceinv_dreduce dreduce_dreduceinv: islprog.
(* end hide *)

(** I also prove that [dreduce] satisfies the normal form property
corresponding to demonic weakest-precondition.  As usual, this normal
form involves the corresponding angelic weakest-precondition, which is
here [areduce].
*) 
Lemma dreduce_demonwp: 
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop), 
   (dreduceinv p st R) <-> 
    (dreduceinv p st (fun _ _ => True)) /\ 
    (forall (p':(prog A)) (st':St), 
       (areduceinv p st (fun p0 st0 => p0=p' /\ st0=st') -> (R p' st'))).
Proof.
  induction p; try (simpl; intuition subst; auto; fail).
  (* bind: preuve a revoir, cf. areduce_angelwp *)
  intros st R; clear H.
    case (IHp st (fun p' st' => (R (bind p' p0) st'))).
  intros H H0.
  simpl; case (is_val_dec p).  
  (* val *)
  generalize p0; case p; simpl; intuition eauto.
  abstract firstorder subst; eauto.
  (* non_val *)
  intros; intuition auto with islprog.
    (* -> 1 *)
     decompose [Logic.and] (H (dreduce_dreduceinv H0));
     intuition auto with islprog.
    (* -> 2 *)
    decompose [ex Logic.and] (areduce_unfold H2); subst;
    decompose [Logic.and] (H (dreduce_dreduceinv H0));
    intuition auto with islprog.
    (* <- *)
    decompose [ex Logic.and] H;
    apply dreduceinv_dreduce;
    apply H1; auto with islprog.
    intros; apply H3;
    generalize (areduceinv_areduce _ _ _ H0).
    intros; eapply areduce_weaken; eauto;
    simpl; intuition subst; auto.
    simpl; intuition subst; auto.
    simpl; intuition subst; auto.
  (* any *)
  simpl; firstorder subst; eauto.
Qed.

(* begin hide *)
Lemma dreduce_unfold:
 forall (A:Type) (p p':(prog A)) (st st':St) (R:(prog A) -> St -> Prop), 
   (dreduce p st R) -> (areduce p st (fun p0 st0 => p0=p' /\ st0=st'))  -> (R p' st').
Proof.
  intros A p p' st st' R H1 H2; case (dreduce_demonwp p st R).
  intros H H0;  decompose [Logic.and] (H (dreduce_dreduceinv H1)).
  firstorder auto with islprog.
Qed.

Lemma dreduce_fold:
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop),
  (dreduce p st (fun _ _ => True)) 
  -> (forall (p':(prog A)) (st':St), 
        (areduce p st (fun p0 st0 => p0=p' /\ st0=st')) -> (R p' st'))
   -> (dreduce p st R).
Proof.
  intros A p st R; case (dreduce_demonwp p st R).
  firstorder auto with islprog.
Qed.
(* end hide *)

(** This normal form property also implies a weakening property.
*)
Lemma dreduce_weaken:  
 forall (A:Type) (p:(prog A)) (st:St) (P Q:(prog A) -> St -> Prop), 
   (dreduce p st P) -> 
     (forall p' st', (P p' st') -> (Q p' st')) -> 
        (dreduce p st Q).
Proof.
  intros A p st P Q H1 H2.
  case (dreduce_demonwp p st P);
  intros H H0; decompose [Logic.and] (H (dreduce_dreduceinv H1)).
  apply dreduce_fold.
  firstorder auto with islprog.
  intros; apply H2; auto with islprog.
Qed.

(** The property [(dreduce p st (fun _ _ => True))] implies that [p] is not a value. 
But the converse is not true, since [(deduce (require False) st (fun _ _ => True))] is false.
But as implied by [checkDS_preservation_aux] below, if [p] is reducible and [st] is in the precondition of 
[p], then [(deduce p st (fun _ _ => True))] holds.
*) 
Lemma dreduce_not_isval: forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop),
  (dreduce p st R) -> ~(is_val p).
Proof.
  intros A p st R H; case H; simpl; auto.
Qed.

(** Now, I express a weak version of the preservation:
``refinement is preserved by an arbitrary [leadsto] reduction''.
More formally, the following lemma states that given [p], [st] and [ps] satisfying the hypotheses of 
the correctness theorem, then if [p] is reducible then for all p' and st' such [(leadsto p st p' st' false)]
then [(bind (set st') (fun _ => p'))] can be interpreted as a specification refining [ps]. 
In the proof of this lemma, the hypothesis ``[p] %{\em reducible}%'' is generalized as [(areduce p st R)] 
for some [R]. Hence, the proof is performed by induction on this hypothesis. 
It presents no particular difficulty, but the possibility to ``invert'' [dreduce] easily is very helpful 
here.
*)
Lemma checkDS_preservation_aux:
 forall (A:Type) (p:(prog A)) (st:St) (ps:(DS St A)),
 ~(is_val p) -> 
   (checkDS p (fun ps0 => ps0=ps)) 
        -> (pre (ps st))
         -> (dreduce p st (fun p' st' =>
               (checkDS p' (fun ps'' => 
                  (refInEnv st ((allPl :<- st') -; ps'') ps)%ISL)))).
Proof.
  intros A p st ps H; generalize (non_val_reducible _ H st) ps.
  generalize (fun (_ : prog A) (_ : St) => True); clear H ps.
  induction 1.
  (* bind_val *) 
  simpl; intros ps H0 X; apply dreduce_bind_val.
  decompose [ex Logic.and] H0; clear H0.
  subst.  eapply checkDS_fold; eauto.
  ISL_simplify_all; simpl; intuition eauto with wipi.
  (* bind_non_val *) 
  simpl; intros ps H0 X; apply dreduce_bind_nonval.
  case H0; clear H0.
  intros pf2 H0.
  decompose [ex Logic.and] (checkDS_unfold _ _ H0).
  clear H0; subst.
  lapply (IHareduce _ H2).
  intros H3; eapply dreduce_weaken; eauto.
  clear H3; simpl; intros p' st' H3.
  constructor 1 with (x:=pf2).
  eapply checkDS_weaken; eauto.
  simpl; intuition auto.
  ISL_assoc_left.
  ISL_monotony.
  simpl in X; eauto with wipi.
  (* require *)  
  simpl; intros; subst; apply dreduceinv_dreduce; simpl.
  simpl in H1; intuition auto.
  ISL_Trivial.
  (* any *)  
  simpl; intros; subst; apply dreduceinv_dreduce; simpl.
  ISL_Trivial.
  (* get *)
  simpl; intros; subst; apply dreduceinv_dreduce; simpl.
  ISL_Trivial.
  (* set *)
  simpl; intros; subst; apply dreduceinv_dreduce; simpl.
  ISL_Trivial.
  (* while *)
  simpl; intros; subst; apply dreduceinv_dreduce; simpl.
  decompose [ex Logic.and] (checkDS_unfold _ _ H0).
  decompose [ex Logic.and] (checkDS_unfold _ _ H4); clear H4; subst.   
  constructor 1 with 
    (x:=(fun (a:bool) => (if a then (x0 -; (whileWF x x0 x1 x2))
                               else Skip)%ISL)).
  eapply checkDS_weaken.
  apply H3.
  simpl; intros; subst; intuition auto.
  ISL_simplify_left. 
  ISL_trans.
  2: apply whileWF_unfold_left.
  ISL_monotony.
  case a; simpl; auto.
  constructor 1 with (x:=(fun (_:unit) => (whileWF x x0 x1 x2))).
  eapply checkDS_weaken; eauto.
  simpl; intros; subst; intuition auto.
Qed.



(**  I now define [dreducestar]. The main idea in this definition, 
     is that the postcondition [R] must be preserved by all iteration of [dreduce], 
     but only while programs are reducible.  
     Indeed, if I have not introduced this restriction, then
     [dreducestar] would never be provable.
*)
Definition dreducestar (A:Type) (R:(prog A) -> St -> Prop) (p:(prog A)) (st:St)
 := forall n:nat, (iter n (fun R p st => ~(is_val p) -> (dreduce p st R)) R p st).

(* begin hide *)
Lemma dreducestar_corefl: forall (A:Type) (R:(prog A) -> St -> Prop) (p:(prog A)) (st:St),
  (dreducestar R p st) -> (R p st).
Proof.
  intros A R p st H; exact (H 0).
Qed.

Lemma dreducestar_dreduce: forall (A:Type) (R:(prog A) -> St -> Prop) (p:(prog A)) (st:St),
  (dreducestar R p st) -> ~(is_val p) -> (dreduce p st R).
Proof.
  intros A R p st H; exact (H (1)).
Qed.

Lemma dreducestar_cotrans_dreduce: forall (A:Type) (R:(prog A) -> St -> Prop) (p:(prog A)) (st:St),
  (dreducestar R p st) -> ~(is_val p) -> (dreduce p st (dreducestar R)).
Proof.
  intros A R p st H H0; apply dreduce_fold.
  generalize (H (1)); simpl.
  intuition auto; eapply dreduce_weaken; eauto.
  intros p' st' H1 n.
  generalize (H (S n)); simpl.
  intuition auto. apply (dreduce_unfold H3 H1).
Qed.

Lemma dreducestar_cotrans: forall (A:Type) (R:(prog A) -> St -> Prop) (p:(prog A)) (st:St),
  (dreducestar R p st) -> (dreducestar (dreducestar R) p st).
Proof.
  intros A R p st H; 
  intros n; generalize R p st H; elim n; clear n R p st H; simpl; auto.
  intros n HR R p st H H0.
  generalize (dreducestar_cotrans_dreduce H H0).
  intros; eapply dreduce_weaken; eauto.
Qed.

Hint Resolve areduce_nonval dreducestar_dreduce dreducestar_cotrans: islprog.
(* end hide *)

(** The basic property of [dreducestar] is to be the demonic weakest precondition associated 
to [areducestar], as stated by the two following lemmas. The first lemma is proved by induction on the [n] 
of the [areducestar] hypothesis, while the second lemma is proved by induction on the [n] of the [dreducestar] 
in the goal. In both lemmas, I simply apply the corresponding relations between [dreduce] and [areduce].
*)
Lemma dreducestar_unfold:
 forall (A:Type) (p p':(prog A)) (st st':St) (R:(prog A) -> St -> Prop), 
   (dreducestar R p st) -> (areducestar (fun p0 st0 => p0=p' /\ st0=st') p st) 
      -> (R p' st').
Proof.
  intros A p p' st st' R H1 H2; case H2; clear H2.
  intros n; generalize p p' st st' R H1; elim n; clear n p p' st st' R H1; simpl.
  (* cas 0 *)
  intros; intuition auto; subst.
  apply dreducestar_corefl; auto.
  (* cas (S n) *)
  intros n HR p p' st st' R H1 H2.
  decompose [ex Logic.and] (areduce_unfold H2); clear H2.
  eapply HR.
  2: eauto.
  lapply (dreducestar_cotrans_dreduce H1); eauto with islprog.
  intros; eapply dreduce_unfold; eauto with islprog.
Qed.

Lemma dreducestar_fold:
 forall (A:Type) (p:(prog A)) (st:St) (R:(prog A) -> St -> Prop),
  (dreducestar (fun p' _ => True) p st) 
  -> (forall (p':(prog A)) (st':St), 
        (areducestar (fun p0 st0 => p0=p' /\ st0=st') p st) -> (R p' st'))
   -> (dreducestar R p st).
Proof.
  intros A p st R H1 H2 n.
  generalize p st R H1 H2; elim n; clear n p st R H1 H2; simpl;
  intuition auto with islprog.
  (* cas (S n) *)
  apply dreduce_fold; auto with islprog.
  intros; apply H.
  eapply dreducestar_unfold; eauto with islprog.
  intros; apply H2.
  apply areduce_areducestar_trans.
  eapply areduce_fold; eauto.
Qed.

(* begin hide *)
Lemma dreducestar_weaken:  
 forall (A:Type) (p:(prog A)) (st:St) (P Q:(prog A) -> St -> Prop), 
   (dreducestar P p st) -> 
     (forall p' st', (P p' st') -> (Q p' st')) -> 
        (dreducestar Q p st).
Proof.
  intros A p st P Q H1 H2; eapply dreducestar_fold.
  intros n; generalize p st P H1; elim n; clear p st P Q H1 H2 n; simpl; auto.
  intros n HR p st P H1 H2.
  generalize (dreducestar_cotrans_dreduce H1 H2).
  intros; eapply dreduce_weaken; eauto.
  intros; apply H2.
  eapply dreducestar_unfold; eauto.
Qed.

Lemma iter_n_weaken: forall (n:nat) (A:Type) (p:(prog A)) (st:St) (P Q:(prog A) -> St -> Prop),
  (iter n (fun R p st => ~(is_val p) -> (dreduce p st R)) P p st)
    -> (forall p' st', (P p' st') -> (Q p' st')) -> 
          (iter n (fun R p st => ~(is_val p) -> (dreduce p st R)) Q p st).
Proof.
  induction n; simpl; auto.
  intros; eapply dreduce_weaken; eauto.
Qed.
(* end hide *)

(** Below is the reformulation of the preservation property that I now 
prove. The proof is a simple induction on the [n] of the [dreducestar].
The recursive case is a direct consequence of %\\%[checkDS_preservation_aux].
*)
Lemma checkDS_preservation:
 forall (A:Type) (p:(prog A)) (st:St) (ps:(DS St A)),
      (checkDS p (fun ps0 => ps0=ps)) -> (pre (ps st))
        -> (dreducestar (fun p' st' =>
               (checkDS p' (fun ps'' =>
                  (refInEnv st ((allPl :<- st') -; ps'') ps)%ISL))) p st).
Proof.
  intros A p st ps H X n; generalize p st ps H X; elim n; clear p st ps H X n; simpl.
  (* cas de base *)
  intros; eapply checkDS_weaken; eauto.
  simpl; intros; subst; auto with isl.
  (* (S n) *)
  intros n HR p st ps H1 X H2.
  generalize (checkDS_preservation_aux _ _ _ H2 H1 X).
  intros; eapply dreduce_weaken; eauto.
  simpl; intros.
  decompose [ex Logic.and] (checkDS_unfold _ _ H0).
  lapply (HR _ st' _ H4).
  intros; eapply iter_n_weaken; eauto.
  simpl; intros.
  eapply checkDS_weaken; eauto.
  simpl; intros; ISL_trans; eauto.
  ISL_simplify_right; auto with islprog.
  unfold refInEnv, refines in H5.
  abstract firstorder auto.
Qed.

(** The lemma below is direct consequence of [checkDS_preservation]. It shows 
that if the initial state satisfies the precondition of the initial expression, 
then in any reduction, the derived state satisfies the precondition of the 
corresponding derived expression.
*)
Lemma checkDS_pre_preservation:
 forall (A:Type) (p:(prog A)) (st:St), 
    (checkDS p (fun ps => pre (ps st)))
           -> (dreducestar 
                  (fun p' st' => (checkDS p' (fun ps' => pre (ps' st'))))
                  p st).
Proof.
  intros A p st H.
  decompose [ex Logic.and] (checkDS_unfold _ _ H). 
  generalize (checkDS_preservation p st _ H1 H2).
  intros H3; eapply dreducestar_weaken.
  apply H3.
  simpl; intros p' st' H0; eapply checkDS_weaken.
  apply H0.
  simpl; intros ps H4.
  unfold refInEnv, refines in H4.
  abstract firstorder auto.
Qed.

(** *** Step 3: proving a reformulation of the termination property
I first introduce the relation [reduce] on which I prove the
accessibility property. This is a special case of [areduce].
*)
Definition reduce (A: Type) (y x: (prod (prog A) St)) 
  := (areduce (fst x) (snd x) (fun p' st' => p'=(fst y) /\ st'=(snd y))).

(** I first prove the accessibility values. It is trivial because, 
values have no predecessor for [reduce].
*)
Lemma Acc_val:
  forall (A:Type) (a:A) (st:St),
    (Acc (reduce (A:=A)) (pair (val a) st)).
Proof.
  intros A a st; constructor 1.
  intros y H0; case (areduce_nonval H0 I).
Qed.

(* begin hide *)
Hint Resolve Acc_val: islprog.
(* end hide *)

(** I then prove that [bind] preserves accessibility of [reduce].
Indeed, intuitively, if the left subexpression of the [bind] is accessible, 
it will always be reduced to a value. And then, the accessibility of the [bind] 
is deduced from the accessibility of the right subexpression. 
This property is thus proved by induction on the accessibility of left subexpression
(or more formally the hypothesis [(Acc (reduce (A:=A)) x)] in the following lemma). 
*)
Lemma Acc_bind_preserv:
  forall (A:Type) (x:(prod (prog A) St)),
    (Acc (reduce (A:=A)) x) ->
     forall (B:Type) (p2: A -> (prog B)),
      (forall (a:A) (st':St), 
         (areducestar (fun p' st0 => p'=(val a) /\ st0=st') (fst x) (snd x))
            -> (Acc (reduce (A:=B)) (pair (p2 a) st')))
         -> (Acc (reduce (A:=B)) (pair (bind (fst x) p2) (snd x))).
Proof.
  induction 1.
  intros B p2 H1.
  constructor 1.
  unfold reduce at 1; simpl; intros.
  generalize (areduce_areduceinv H2); simpl; clear H2.
  case (is_val_dec (fst x)).
  (* bind_val *)
  intros i X.
  generalize H1; clear H1.
  pattern (fst x); elim X using v_inv_ind; clear i X.
  case y; simpl; intuition subst; auto.
  apply H2.
  constructor 1 with (x:=0).
  simpl; auto.
  (* bind_nonval *)
  intros H2 H3.
  decompose [ex] (areduce_unfold H3); clear H3.
  case H5; case y; simpl; clear y H5.
  intuition subst.
  apply H0 with (y:=(pair x0 s)); auto.
  simpl; intros; apply H1.
  apply areduce_areducestar_trans.
  eapply areduce_weaken. 
  eauto.
  simpl.
  intuition subst; auto.
Qed.

(** If a state [st] satisfies the precondition of an expression [bind p1 p2], then it satisfies 
also trivially the precondition of [p1]. *)
Lemma checkDS_pre_bind_1:
 forall (A B:Type) (p1:(prog A)) (p2:A -> (prog B)) (st:St),
    (checkDS (bind p1 p2) (fun ps => pre (ps st)))
    -> (checkDS p1 (fun ps => pre (ps st))).
Proof.
  simpl; intros.
  decompose [ex Logic.and] H; clear H.
  eapply checkDS_weaken.
  apply H0.
  simpl; clear H0.
  intuition auto.
  eauto with wipi.
Qed.

(** If a state [st] satisfies the precondition of an expression [bind
p1 p2], then if for any value [a] and state [st'], such [p1] in the
state [st] can be reduced in [a] and [st'], then [st'] is in the
precondition of [(p2 a)].  This lemma is a consequence of
[checkDS_preservation].  
*)
Lemma checkDS_pre_bind_2:
 forall (A B:Type) (p1:(prog A)) (p2:A -> (prog B)) (st:St) (a:A),
    (checkDS (bind p1 p2) (fun ps => pre (ps st)))
    -> (dreducestar 
         (fun p' st' => 
            p'=(val a) -> (checkDS (p2 a) (fun ps => pre (ps st'))))
         p1 st).
Proof.
  simpl; intros.
  decompose [ex Logic.and] H; clear H.
  decompose [ex Logic.and] (checkDS_unfold _ _ H0); clear H0. 
  lapply (checkDS_preservation p1 st _ H1).
  intros H2; eapply dreducestar_weaken.
  apply H2.
  simpl; clear H2; intros p' st' H0 H4; subst. 
  eapply checkDS_weaken.
  apply H3.
  simpl; intros ps H4; subst.
  simpl in H0.
  unfold refInEnv, refines in H0.
  case (H0 (fun a : prod St A =>
       pre match a with
           | pair stx a0 => x a0 stx
           end)).
  simpl; intros X1 X2; clear H0 X2.
  auto.
  eauto with wipi.
Qed.

(** I now prove the reformulation of the termination theorem.  It
is proved by induction on [p]. The cases [val], [require], [get] and
[set] corresponds to the [Acc_val] lemma. The [bind] case uses the
property [checkDS_pre_bind_1] and [checkDS_pre_bind_2] in order to
discharge the hypotheses of the induction hypotheses and then apply
[Acc_bind_preserv]. The [while] case is the most technical: it uses
the preconditions of [whileWF]. In particular, it performs an
induction over the variant and applies [Acc_bind_preserv] in order to
prove that the unfolding of [while] terminates.  
*)
Lemma checkDS_termination:
 forall (A:Type) (p:(prog A)) (st:St), 
    (checkDS p (fun ps => pre (ps st)))
           -> (Acc (reduce (A:=A)) (pair p st)).
Proof.
  induction p; auto with islprog.
  (* bind *)
  intros st H0.
  apply Acc_bind_preserv with (x:=(pair p st)).
  apply IHp.
  eapply checkDS_pre_bind_1.
  eauto.
  simpl; intros; apply H.
  apply (dreducestar_unfold (checkDS_pre_bind_2 _ a H0) H1).
  auto.
  (* require  *)
  intros st H.
  constructor 1.
  unfold reduce at 1; simpl; intros y H0.
  generalize (areduce_areduceinv H0); simpl; clear H0.
  case y; simpl; intros; intuition subst; auto with islprog.
  (* any *)
  intros st H.
  constructor 1.
  unfold reduce at 1; simpl; intros y H0.
  generalize (areduce_areduceinv H0); simpl; clear H0.
  intro H0; case H0; clear H0; intros a.
  case y; simpl; intros; intuition subst; auto with islprog.
  (* get *)
  intros st H.
  constructor 1.
  unfold reduce at 1; simpl; intros y H0.
  generalize (areduce_areduceinv H0); simpl; clear H0.
  case y; simpl; intros; intuition subst; auto with islprog.
  (* set *)
  intros st H.
  constructor 1.
  unfold reduce at 1; simpl; intros y H0.
  generalize (areduce_areduceinv H0); simpl; clear H0.
  case y; simpl; intros; intuition subst; auto with islprog.
  (* while *)
  intros st H.
  simpl in H.
  case (checkDS_unfold _ _ H); clear H.
  intros c H; case H; clear H; intros H1 H2.
  case (checkDS_unfold _ _ H2); clear H2.
  intros b H; case H; clear H; intros H2 H; case H; clear H.
  intros inv H; case H; clear H; intros R H.
  decompose [Logic.and] H.
  clear H3 H; unfold well_founded in H5.
  generalize H7; elim (H5 st); clear H7 H5 st.
  intros st H3 Hacc H5.
  constructor 1; intros y; case y.
  unfold reduce at 1; simpl.
  intros p0 st0 H6.
  generalize (areduce_areduceinv H6); simpl; clear H6.
  unfold unfold_while; simpl; intuition subst.
  assert (IFRAD: 
            (checkDS 
              (bind p1
               (fun b1 =>
                  if b1 then p2 else val tt))
              (fun ps => ps=(If c Then b Else Skip)%ISL))).
  (* preuve de IFRAD *)
    simpl; 
    constructor 1 with
     (x:=(bool_rect (fun _ => (DS _ _)) b (Skip)%ISL)).
    eapply checkDS_weaken.
    eauto.
    simpl; intuition subst; auto.
    case a0; simpl; auto.
  (* *)
  assert (IFPRE:
            (checkDS 
              (bind p1
               (fun b1 =>
                  if b1 then p2 else val tt))
              (fun ps => pre (ps st0)))).
  (* preuve de IFPRE *)
    eapply checkDS_weaken.
    apply IFRAD.
    simpl; intros; subst. 
    eapply inv_preservation_pre; eauto.
  (* *)
  apply Acc_bind_preserv with (x:=(pair p1 st0)).
  apply IHp1.
  eapply checkDS_pre_bind_1.
  apply IFPRE.
  simpl; intros a0 st'; case a0; auto with islprog.
  (*** cas then ***)
  intros; apply Acc_bind_preserv with (x:=(pair p2 st')).
  apply IHp2.
  apply (dreducestar_unfold (checkDS_pre_bind_2 _ true IFPRE) H).
  auto.
  simpl; intros a1 st1 H6.
  apply Hacc.
  (* variant *)
  generalize (H0 st0 H5); clear H0.
  intros H0.
  lapply (checkDS_preservation p1 st0 c H1).
  intros H8; lapply (dreducestar_unfold (p':=(val true)) (st':=st') H8); clear H8.
  simpl.
  intros H8;
  assert (H9:([< st0 >] allPl :<- st' -; b >==> upd_abspost allPl (fun (new_st : St) (_ : unit) => R new_st st0))%ISL).
  ISL_trans; eauto.
  ISL_enter_right.
  eauto.
  ISL_simplify_right.
  intros; ISL_simplify_left.
  ISL_simplify_both.
  intros; ISL_simplify_right.
  intuition auto with isl.
  clear H8.
  lapply (checkDS_preservation (bind (set st') (fun _:unit => p2))
    st0 (allPl :<- st' -; b)%ISL).
  intros H8; lapply H8; simpl; clear H8.
  intros H8; lapply (dreducestar_unfold (p':=(val a1)) (st':=st1) H8); clear H8.
  simpl; case a1.
  intros H8.
  lapply (refInEnv_from_wp (refInEnv_trans H8 H9)).
  simpl.
  intros H10; decompose [ex Logic.and] H10.
  inversion_clear H12.
  auto.
  simpl; auto.
  apply areducestar_bind_val.
  apply areduce_areducestar.
  apply areduce_set.
  constructor 1 with (x:=tt).
  intuition auto.
  unfold refInEnv, refines in H9. simpl in H9.
  abstract firstorder auto.
  simpl.
  constructor 1 with (x:=(fun _:unit => b)).
  intuition auto.
  auto.
  unfold refInEnv, refines in H0. 
  case (H0 (fun _ => True)).
  clear H0; simpl; intros X1 X2; clear X2.
  intuition eauto with wipi.
  (* invariant *)
  generalize (H4 st0 H5); clear H4.
  intros H7.
  lapply (checkDS_preservation (bind p1 (fun b => if b then p2 else (val tt)))
              st0 (If c Then b Else Skip)%ISL).
  intros H8; lapply H8; clear H8.
  intros H8; lapply (dreducestar_unfold (p':=(val a1)) (st':=st1) H8); clear H8.
  simpl; intros H8.
  assert (H9: ([< st0 >] allPl :<- st1 >==> (upd_abspost allPl (fun (new_st : St) (_ : unit) => inv new_st)))%ISL).
  ISL_trans.
  2: apply H7.
  ISL_trans.
  2: apply H8.
  case a1; ISL_Trivial.
  lapply (refInEnv_from_wp H9).
  simpl; intro H10. decompose [ex Logic.and] H10.
  inversion_clear H12.
  auto.
  simpl; auto.
  apply areducestar_bind_val.
  eapply areducestar_weaken.
  eauto.
  simpl; intuition subst.
  constructor 1 with (x:=true).
  intuition auto.
  unfold refInEnv, refines in H7. 
  abstract firstorder auto.
  auto. 
Qed.

  
(** *** Last step: proving the first formulation of the theorem  

 I have now to prove the correspondence between [semDS] and [checkDS] on one hand, 
 and [leadstostar] and [areducestar] on the other hand. 
 The two following lemmas correspond to this first part. The first is simply 
 proved by induction on [semDS]. And the second is proved by induction on [p].
*)
Lemma semDS_checkDS: forall (A:Type) (p:(prog A)) (ps:(DS St A)),
   (semDS p ps) -> (checkDS p (fun ps' => ps'=ps)).
Proof.
  intros A p ps H; elim H; simpl; intuition eauto.
  constructor 1 with (x:=pf2).
  eapply checkDS_weaken; eauto.
  simpl; intros; subst; intuition auto.
  eapply checkDS_weaken; eauto.
  simpl; intros; eapply checkDS_weaken; eauto.
  simpl; intros; subst; eauto.
Qed.

Lemma checkDS_semDS: forall (A:Type) (p:(prog A)) (ps:(DS St A)),
   (checkDS p (fun ps' => ps'=ps)) -> (semDS p ps).
Proof.
  induction p; simpl; intros; subst; eauto with islprog.
  (* bind *)
  decompose [ex Logic.and] H0; clear H0.
  decompose [ex Logic.and] (checkDS_unfold _ _ H1).
  subst; auto with islprog.
  (* while *)
  decompose [ex Logic.and] (checkDS_unfold _ _ H).
  decompose [ex Logic.and] (checkDS_unfold _ _ H2).
  subst; auto with islprog.
Qed.


(** I now establish a correspondence between [leadsto] and [areduce].
The first lemma below is proved by induction on [leadsto] and the
second by induction on [areduce].  *)
Lemma leadsto_areduce:
  forall (A:Type) (p1 p2:(prog A)) (st1 st2:St),
    (leadsto p1 st1 p2 st2 false) 
        -> (areduce p1 st1 (fun p' st' => p'=p2 /\ st'=st2)).
Proof.
  intros A p1 p2 st1 st2 H;
  assert (H1: (exists b:bool,(leadsto p1 st1 p2 st2 b) /\ b=false)).
  eauto.
  case H1; intros b H2; case H2; clear H1 H2 H.
  intros H; elim H; simpl; firstorder eauto with islprog.
  (* bind *)
  apply areduce_bind_nonval.
  intros; eapply areduce_weaken; eauto.
  simpl; intros; intuition subst; auto.
  discriminate H1.
Qed.  

Lemma areduce_leadsto:
  forall (A:Type) (p1 p2:(prog A)) (st1 st2:St),
    (areduce p1 st1 (fun p' st' => p'=p2 /\ st'=st2))
      -> (leadsto p1 st1 p2 st2 false).
Proof.
  intros A p1 p2 st1 st2 H; generalize p2 st1 st2 (areduce_areduceinv H);
  elim p1; clear A p1 p2 st1 st2 H; 
  try (simpl; firstorder subst; eauto with islprog; fail).
  (* bind *)
  intros A B p1a HR1 p2 HR2 p1b st1 st2;
  simpl; case (is_val_dec p1a).
  intros H X; elim X using v_inv_ind.
  intuition subst; auto with islprog.
  intros H1 H2; 
  decompose [ex Logic.and] (areduce_unfold H2).
  subst; auto with islprog.
Qed.  

(** This equivalence is extended to [leadstostar] and [areducestar]. The first lemma below is proved
by induction on [leadstostar] and the second is proved by induction on the [n] in the [areducestar] 
hypothesis. 
*)
Lemma leadstostar_areducestar:
  forall (A:Type) (p1 p2:(prog A)) (st1 st2:St),
    (leadstostar p1 st1 p2 st2 false)
        -> (areducestar (fun p' st' => p'=p2 /\ st'=st2) p1 st1).
Proof.
  intros A p1 p2 st1 st2 H;
  assert (H1: (exists b:bool,(leadstostar p1 st1 p2 st2 b) /\ b=false)).
  eauto.
  case H1; intros b H2; case H2; clear H1 H2 H.
  intros H; elim H; simpl; auto with islprog.
  intros; apply areduce_areducestar_trans.
  generalize (leadsto_areduce H0).
  intros; eapply areduce_weaken; eauto.
  simpl; intros; intuition subst; auto.
  intros; discriminate H1.
Qed.

Lemma areducestar_leadstostar:
  forall (A:Type) (p1 p2:(prog A)) (st1 st2:St),
    (areducestar (fun p' st' => p'=p2 /\ st'=st2) p1 st1)
      -> (leadstostar p1 st1 p2 st2 false).
Proof.
  intros A p1 p2 st1 st2 H; case H.
  intros n; generalize p1 p2 st1 st2; elim n; clear p1 p2 st1 st2 H n; simpl.
  intuition subst; auto with islprog.
  intros n HR p1 p2 st1 st2 H.
  decompose [ex Logic.and] (areduce_unfold H).
  apply leadstostar_next_false with (p2:=x) (st2:=x0).
  apply areduce_leadsto; auto.
  auto.
Qed.   

(** I prove that a [leadsto] computation can not abort when the initial state is in the precondition of the 
initial expression. Then, I extend this result to [leadstostar] computations. 
*)
Lemma leadsto_nonabort:
   forall (A:Type) (p1 p2:(prog A)) (st1 st2:St) (b:bool),
    (leadsto p1 st1 p2 st2 b) -> (checkDS p1 (fun ps => pre (ps st1)))
        -> b=false.
Proof.
  induction 1; try (simpl; auto; firstorder auto).
  apply IHleadsto.
  eapply checkDS_weaken; eauto.
  simpl; intros.
  intuition eauto with wipi. 
Qed.

Lemma leadstostar_nonabort:
   forall (A:Type) (p1 p2:(prog A)) (st1 st2:St) (b:bool),
    (leadstostar p1 st1 p2 st2 b) -> (checkDS p1 (fun ps => pre (ps st1)))
        -> b=false.
Proof.
  induction 1; auto with islprog.
  intros H1; apply IHleadstostar.
  generalize (leadsto_areduce H).
  intros H2.
  apply (dreducestar_unfold (checkDS_pre_preservation _ _ H1) (areduce_areducestar H2)).
  intros H2; apply (leadsto_nonabort H H2).
Qed.


(* begin hide *)
Lemma leadstostar_starRel:
  forall (A:Type) (p1 p2:(prog A)) (st1 st2:St),
    (leadstostar p1 st1 p2 st2 false) <-> 
        (starRel  
             (fun (y x:(prod (prog A) St)) =>
                      leadsto (fst x) (snd x) (fst y) (snd y) false)
             (pair p2 st2) (pair p1 st1)).
Proof.
  intros A p1 p2 st1 st2. constructor 1.
  intros H.
  assert (H1: (exists b:bool,(leadstostar p1 st1 p2 st2 b) /\ b=false)).
  eauto.
  case H1; intros b H2; case H2; clear H1 H2 H.
  intros H; elim H; simpl; auto with islprog.
  intros; apply starRel_refl.  
  intros; eapply starRel_trans; simpl; eauto.
  apply starRel_includes; simpl; auto.
  intros.  discriminate H1.
  intro H.
  change (leadstostar 
   (fst (pair p1 st1)) 
   (snd (pair p1 st1))
   (fst (pair p2 st2))
   (snd (pair p2 st2))
   false).
  generalize (pair p1 st1) (pair p2 st2) (starRel_swap H).
  clear H st2 st1 p2 p1.
  intros p p0 H; pattern p, p0.
  elim H; eauto with islprog.
Qed.

Lemma leadsto_values:
  forall (A:Type) (p:(prog A)) (st:St),
     { (exists a:A, p=(val a)) /\ forall (p':(prog A)) (st':St), ~(leadsto p st p' st' false)} + 
              { exists p':(prog A), exists st':St, (leadsto p st p' st' false)}.
Proof.
  intros A p st; case (is_val_dec p).
  intros H; constructor 1.
  case (is_val_inv _ H).
  intros a H'; subst.
  intuition eauto.
  inversion_clear H0.
  intros H; constructor 2.
  decompose [ex Logic.and] (areduce_unfold (non_val_reducible _ H st)).
  constructor 1 with (x:=x).
  constructor 1 with (x:=x0).
  apply areduce_leadsto; auto.
Qed.
(* end hide *)

(** At last, I prove the theorem. I deduce directly the left side
from [checkDS_preservation] and the preceding correspondences. Let me
remark that [dreducestar_unfold] and [areducestar_unfold] allows to
establish how [dreduce] can be expressed in terms of
[leadstostar]. The right side is directly deduced from
[checkDS_termination]. 
*)
Theorem semDS_correct:
  forall (A:Type) (p:(prog A)) (ps:(DS St A)) (st:St),
   (semDS p ps)
     -> (pre (ps st))
        ->  (forall (p':(prog A)) (st':St) (b:bool),
              (leadstostar p st p' st' b) -> 
                 b=false
                 /\ (exists ps':(DS St A), 
                       (semDS p' ps') /\ (refInEnv st ((allPl :<- st') -; ps')  ps))%ISL)
           /\ (Acc (fun (y x:(prod (prog A) St)) =>
                      leadsto (fst x) (snd x) (fst y) (snd y) false)
                    (pair p st)).   
Proof.
  intros A p ps st H1 H2; constructor 1.
  (* autoreduction *)
  intros p' st' b H3.
  lapply (leadstostar_nonabort H3).
  firstorder subst; auto.
  decompose [ex Logic.and] (checkDS_unfold _ _
   (dreducestar_unfold 
     (checkDS_preservation p st ps (semDS_checkDS H1) H2)
     (leadstostar_areducestar H3))).
  constructor 1 with (x:=x). 
  intuition auto.
  apply checkDS_semDS; auto.
  apply checkDS_fold with (ps:=ps).
  apply semDS_checkDS; auto.
  apply H2.
  (* termination *)
  apply Acc_enforce with (R1:=(reduce (A:=A))).
  apply checkDS_termination; auto.
  eapply checkDS_fold.
  2: apply H2.
  apply semDS_checkDS; auto.
  unfold reduce; intuition auto; apply leadsto_areduce; auto.
Qed.


(** The technique involved in this theorem shows that CPS techniques
and more precisely weakest-preconditions can be useful to invert
complex inductive predicates. However a simpler (but weaker) version
of this theorem could be probably be stated using a big-step semantics
instead of a small-step one, using coinduction to express
non-termination as exposed in \cite{Leroy-coindsem}.
*)

(* begin hide *)
End ISLprogs.

Unset Implicit Arguments.
(* end hide *)
