(* begin hide *)
Set Implicit Arguments.
 
Require Export ISLAbsPost.
Require Export ZBoolAux.
Require Export Array.
(* end hide *)

(** First, I give ISL functions to access and store values in a array. 
I interpret here arrays as functions. *)

Section array.

Variable A: Type.

Record StArray : Type := { 
  length: Z ;
  abs_array: (AbsArray A) 
}.
 
Identity Coercion AbsArray2Fun: AbsArray >-> Funclass.
Coercion abs_array: StArray >-> AbsArray.

Definition array: (Place StArray (AbsArray A)).
 refine (Build_Place
            abs_array
            (fun e x => (Build_StArray (length e) x)) _ _ _);
 abstract (intros e; case e; simpl; auto).
Defined.

(** Let us note that [access] function (and similarly [store] and [swap]) below, 
the array is passed ``by variable'' (and not ``by copy''). This possible here, because
places are first class values in our model. And user proofs will not involve aliases problem 
(e.g. user will not have to reason about equality or inequality of
places) because these functions have a single place argument. *)

 
Definition access (Env: Type) (t:(Place Env StArray)) (i:Z) : (DS Env A) :=
   (Bind tv <- !t In
    If / (Val ((Zle_bool 0 i)%Z && (Zlt_bool i (length tv))))%bool 
    Then (Val (tv i))
    Else Abort)%ISL.

Lemma access_ok:
  forall (Env: Type) (t:(Place Env StArray)) (i:Z) (e:Env),
   (0 <= i < (length (read t e)))%Z -> 
     ([< e >] (access t i) >==> (Val ((read t e) i)))%ISL.
Proof.
  intros Env t i e H; unfold access.
  ISL_simplify_left.
  rewrite (Zle_le_bool 0 i); try abstract omega.
  rewrite (Zlt_lt_bool i (length (read t e))); try abstract omega.
  simpl; ISL_simplify_left; auto with isl.
Qed.
  
Definition store (Env: Type) (t:(Place Env StArray)) (i:Z) (a:A): (DS Env unit) :=
   (Bind tv <- !t In
    If / (Val ((Zle_bool 0 i)%Z && (Zlt_bool i (length tv))))%bool 
    Then 
      t#array :<- (fun j => (if (Zeq_bool i j) then a else (tv j)))
    Else 
      Abort)%ISL.

Lemma store_ok:
  forall (Env: Type) (t:(Place Env StArray)) (i:Z) (a:A) (e:Env),
   (0 <= i < (length (read t e)))%Z -> 
     ([< e >] (store t i a) >==> 
        t#array :<- (fun j => (if (Zeq_bool i j) then a else (read t e j))))%ISL.
Proof.
  intros Env t i a e H; unfold store.
  ISL_simplify_left.
  rewrite (Zle_le_bool 0 i); try abstract omega.
  rewrite (Zlt_lt_bool i (length (read t e))); try abstract omega.
  simpl; ISL_simplify_left; auto with isl.
Qed.

Definition swap (Env: Type) (t:(Place Env StArray)) (i j:Z): (DS Env unit) :=
  (Bind ti <- (access t i) In
    Bind tj <- (access t j) In
      (store t i tj) -; (store t j ti))%ISL.

Lemma swap_ok:
  forall (Env: Type) (t:(Place Env StArray)) (i j:Z) (e:Env), 
    (0 <= i < (length (read t e)))%Z -> (0 <= j < (length (read t e)))%Z
     -> ([< e >] (swap t i j) >==> 
          (upd_abspost (t#array)
          (fun new_t result => (Exchange (read t e) new_t i j))))%ISL.
Proof.
  unfold swap.
  intros Env t i j e H1 H2.
  ISL_enter_left; [ apply access_ok; auto | ISL_simplify_left].
  ISL_enter_left; [ apply access_ok; auto | ISL_simplify_left].
  ISL_enter_left. apply store_ok; auto. ISL_simplify_left. 
  simpl.
  ISL_trans. apply store_ok. autorewrite with placeeq; simpl; auto.
  ISL_simplify_right.
  intros. eapply updabspost2wp with (v:=t#array) (* *) (old_st:=e) (* *).
  autorewrite with placeeq; eauto.
  simpl; autorewrite with placeeq; simpl.
  intuition auto.
  constructor 1.
  rewrite (Zeq_eq_bool j j); auto.
  generalize (Zeq_cases j i); case (Zeq_bool j i).
  intros; subst; auto.
  rewrite (Zeq_eq_bool i i); auto.
  intros k Ha Hb; rewrite (Zdiff_eq_bool j k); auto.
  rewrite (Zdiff_eq_bool i k); auto.
Qed.

Lemma swap_permutation:
  forall (Env: Type) (t:(Place Env StArray)) (i j:Z) (e:Env), 
    (0 <= i < (length (read t e)))%Z -> (0 <= j < (length (read t e)))%Z
    -> ([< e >] (swap t i j) >==> 
       (upd_abspost (t#array)
          (fun new_t result => 
             (permutation (read t e) new_t 0 ((length (read t e))-1)))))%ISL.
Proof.
  intros; ISL_trans.
  apply swap_ok; auto.
  apply updabspost2wp with (old_st:=e) (v:=t#array).
  autorewrite with placeeq; auto.
  simpl; intros; autorewrite with placeeq; simpl.
  intuition auto.
  eapply Exchange_Permutation.  
  3: eauto.
  abstract omega.
  abstract omega.
Qed.

End array.
(* begin hide *)

Require Export ISLWhile.
Import Zwf.
Import WfT.

(* end hide *)

(** Maximum sort is then decomposed in two functions: [maximum] which
computes the index of the maximum in a slice of an array, and [maxisort] that sorts
the slice of an array. The array is a global variable shared by these
two functions. 
The slice is represented by two indexes [a] (the first
index of the slice) and [b] (the second index of the slice) which are
parameters of these functions.  *)


Section MaximumDef.

(** Function [maximum] introduces two local variables [ind] (an index, 
such that successive indexes accessed by [maximum] lies in the interval [a]..[ind]), and
[res] (the index of the maximum in [a]..[ind]).
*)
Record place_maximum: Type := {
  res: Z;
  ind: Z 
}.

(** Place [pL] corresponds to the pair of this two local variables
and place [pR] is the global array variable.  
*)
 
Definition maximum (a b:Z) : (DS (StArray Z) Z) :=
 (block (Build_place_maximum a a)
    (Bind init_t <- !pR In
    (whileWF
(** %{\hspace*{7ex}\tt(*}% condition to continue %{\tt*)}% *)
      (Bind s <- !pL In (Val (Zlt_bool (ind s) b)))
(** %{\hspace*{7ex}\tt(*}% body %{\tt*)}% *)
      (Bind s <- !pL In
         let nind := ((ind s)+1)%Z in
          let r := (res s) in
            If (Bind tr <- (access pR r) In
                Bind tn <- (access pR nind) In
                 (Val (Zlt_bool tr tn)%Z))
            Then
              pL :<- (Build_place_maximum nind nind)
            Else
              pL :<- (Build_place_maximum (res s) nind))
(** %{\hspace*{7ex}\tt(*}% invariant %{\tt*)}% *)
      (fun e => 
        let r:=(res (read pL e)) in
        let i:=(ind (read pL e)) in
        let t:=(read pR e) in
          (a <= i <= b)%Z /\  (a <= r <= b)%Z /\ (Maximize t (t r) a i) /\ t=init_t)
(** %{\hspace*{7ex}\tt(*}% variant %{\tt*)}% *)
      (fun x y => (Zwf_up b (ind (read pL x)) (ind (read pL y))))
    ) -;
    Bind s <- !pL In (Val (res s)))
   )%ISL.

Lemma maximum_spec:
  forall (a b:Z) (t:(StArray Z)),
    ((0 <= a <= b)%Z /\ (b < (length t))%Z) -> 
       ([< t >] (maximum a b) 
                >==> (pure_abspost (fun r => ((a <= r <= b)%Z 
                                             /\ (Maximize t (t r) a b%Z))))
       )%ISL.
Proof.
  intros a b t H; unfold maximum.
  apply block_enter_left.
  ISL_unassoc_left.
  ISL_simplify_left.
  ISL_unassoc_left.
  ISL_enter_left.
  apply whileWF_spec.
  ISL_simplify_left; intuition (auto; try abstract omega).
  (* preservation de l'invariant *)
    unfold inv_preservation.
    intros s; case s; intros vl m; case vl; intros r i; simpl; clear vl s; intros; ISL_simplify_left.
    intuition auto; subst.
    generalize (Zlt_cases i b); case (Zlt_bool i b); simpl.
    (* corps de la boucle *)
      intros; ISL_simplify_left.
      ISL_unassoc_left.
      ISL_enter_left; [ apply access_ok; simpl; abstract omega | ISL_simplify_left].
      ISL_unassoc_left; ISL_enter_left; [ apply access_ok; simpl; abstract omega | ISL_simplify_left].
      generalize (Zlt_cases (t r) (t (i+1)%Z)); case (Zlt_bool (t r) (t (i+1)%Z)); simpl;
      intros; ISL_simplify_left; apply updabspost2wp with (old_st:=(pairT (Build_place_maximum r i) t)); 
      simpl; auto; intuition (auto; try abstract omega); constructor 1; 
      try abstract omega; intros; ring_simplify (i+1-1)%Z; auto.
      (* cas du then *)
      eapply Maximize_Zle. eauto.
      try abstract omega.
    (* sortie de la boucle *)
    intros; apply updabspost2wp with (old_st:=pairT (Build_place_maximum r i) t); simpl; auto; intuition (auto; try abstract omega).
  (* decroissance du variant *)
    unfold variant_decreasing.
    intros s; case s; intros vl m; case vl; intros r i; simpl; clear vl s; intros; ISL_simplify_left.
    intuition auto; subst.
    ISL_simplify_left.
    intros X; generalize (Zlt_bool_lt _ _ X); intros X0.
    ISL_simplify_left. 
    ISL_unassoc_left.
    ISL_enter_left; [ apply access_ok; simpl; abstract omega | ISL_simplify_left].
    ISL_unassoc_left; ISL_enter_left; [ apply access_ok; simpl; abstract omega | ISL_simplify_left].
    unfold Zwf_up; generalize (Zlt_cases (t r) (t (i+1)%Z)); case (Zlt_bool (t r) (t (i+1)%Z)); 
    simpl; intros; ISL_simplify_left; apply updabspost2wp with (old_st:=(pairT (Build_place_maximum r i) t));
    simpl; auto; intuition (auto; try abstract omega).
  (* well_foundedT *)
  apply wf_inverse_image with (f:=(fun x: prodT place_maximum (StArray Z) => (ind (fstT x)))).
  apply Zwf_up_well_founded.
  (* initialisation de l'invariant *)
  constructor 1; try abstract omega.
  intros X; absurd (a < a)%Z.
  abstract omega.
  auto.
  (* post-condition generale *)
  intros; ISL_trans.
  2: apply embed_pureabspost_ref2.
  simpl; apply pureabspost2wp with (old_st:=pairT (Build_place_maximum a a) t); simpl; auto.
  intuition auto.
  rewrite <- H9; case a0; simpl; auto.
  rewrite <- H9.
  generalize (Zlt_bool_ge _ _ H4).
  intros; assert (b=(ind (fstT a0))).
  abstract omega.
  subst; auto.
Qed.

End MaximumDef.

Section MaxisortDef.

(** Function [maxisort] introduces a local variable represented as
[pL] place below. This local variable describes the slice [!pL]..[b]
which is sorted at each step of the loop. *)


Definition maxisort (a b:Z) : (DS (StArray Z) unit) :=
 (block b
    (Bind init_t <- !pR In
    (whileWF
(** %{\hspace*{7ex}\tt(*}% condition to continue %{\tt*)}% *)
      (Bind i <- !pL In (Val (Zle_bool a i)))
(** %{\hspace*{7ex}\tt(*}% body %{\tt*)}% *)
      (Bind i <- !pL In
       Bind r <-[pR] (maximum a i) In
         (swap pR i r) -;
         (pL :<- (i-1)%Z))
(** %{\hspace*{7ex}\tt(*}% invariant %{\tt*)}% *)
      (fun e => 
        let i:=(read pL e) in
        let t:=(read pR e) in
             ((a-1) <= i <= b)%Z 
          /\ ((i < b)%Z -> Maximize t (t (i+1))%Z a i) 
          /\ (length t)=(length init_t) 
          /\ (Sorted t (i+1)%Z b) 
          /\ (permutation t init_t a b)
       )
(** %{\hspace*{7ex}\tt(*}% variant %{\tt*)}% *)
      (fun x y => (Zwf a (read pL x) (read pL y)))
    )))%ISL.

Lemma maxisort_spec:
  forall (a b:Z) (t:(StArray Z)),
    ((0 <= a <= b)%Z /\ (b < (length t))%Z) -> 
       ([< t >] (maxisort a b) >==> 
         (upd_abspost (array Z) 
           (fun new_t _ => ((Sorted new_t a b) /\ (permutation new_t t a b)))))%ISL.
Proof.
  intros a b t H; unfold maxisort.
  apply block_enter_left; simpl.
  ISL_unassoc_left.
  ISL_simplify_left.
  ISL_enter_left.
  apply whileWF_spec.
  ISL_simplify_left. intuition (auto; try abstract omega).
  (* preservation de l'invariant *)
    unfold inv_preservation.
    intros s; case s; intros i new_t; simpl; clear s; intros; ISL_simplify_left.
    intuition auto.
    generalize (Zle_cases a i); case (Zle_bool a i); simpl.
    (* corps de la boucle *)
     intros; ISL_simplify_left.
     ISL_enter_left.
     eapply maximum_spec.
     simpl; abstract omega.
     ISL_trans.
     eapply bind_pureabspost_ref1. 
     ISL_simplify_left.
     intros. ISL_simplify_left.
     ISL_enter_left.
     apply swap_ok; simpl; try abstract omega.
     eapply updabspost2wp with (old_st:=(pairT i new_t)); simpl; eauto; 
     intuition (auto; try abstract omega).
     (* Maximize *)
     ring_simplify (i-1+1)%Z.
     assert (C1: (Exchange new_t a2 i a0)); auto.
     case C1; clear C1. intros C1 C2 C3.
     apply Maximize_ext2; intros i0 C6. rewrite <- C2.
     case (Z_eq_dec i0 a0).
       intros C7; subst.
       rewrite <- C1.
       eapply Maximize_ext1.
       eapply H11.
       abstract omega.
     intros C7; rewrite <- C3; try abstract omega.
       eapply Maximize_ext1.
       eapply H11.
       abstract omega.
     (* Sorted *)
     ring_simplify (i-1+1)%Z.
     assert (C1: (Exchange new_t a2 i a0)); auto.
     case C1; clear C1. intros C1 C2 C3.
     apply Sorted_ext2.
     intros k C4. rewrite <- (C3 (k+1)%Z); try abstract omega.
     case (Z_eq_dec i k).
       intros; subst.
       rewrite <- C2.
       eapply Maximize_ext1.
       apply H3.
       abstract omega.
       abstract omega.
     intros C5; rewrite <- C3; try abstract omega.
       eapply Sorted_ext1.
       apply H6.
       abstract omega.          
       abstract omega.
     (* Permutation *)
     eapply permutation_trans.
     2: apply H8.
     apply permutation_sym.
     eapply Exchange_Permutation.  
     3: eauto.
     abstract omega.
     abstract omega.
    (* sortie de la boucle *)
    intros; eapply updabspost2wp with (old_st:=pairT i new_t); 
    simpl; eauto; intuition (auto; try abstract omega).
  (* decroissance du variant *)
    unfold variant_decreasing.
    intros s; case s; intros i new_t; simpl; clear s; intros; ISL_simplify_left.
    intuition auto.
    ISL_simplify_left.
    intros; ISL_simplify_left.
    ISL_enter_left.
     eapply maximum_spec.
     generalize (Zle_bool_le _ _ H7).
     simpl. abstract omega.
     ISL_trans.
     eapply bind_pureabspost_ref1. 
     ISL_simplify_left.
     intros. ISL_simplify_left.
     ISL_enter_left.
     apply swap_ok; simpl; try abstract omega.
     eapply updabspost2wp with (old_st:=(pairT i new_t)); simpl; eauto; 
     intuition (auto; try abstract omega).
    (* fin simplification *)
    unfold Zwf; abstract omega.
  (* well_foundedT *)
  apply wf_inverse_image with (f:=(fstT (A:=Z) (B:=(StArray Z)))) .
  apply Zwf_well_founded.
  (* post-condition *)
  absurd (b<b)%Z.
  try abstract omega.
  auto.
  apply Sorted_ext2.
  intros i X; try abstract omega.
  apply permutation_eqext.
  apply eqext_refl.
  ISL_trans.
  2: apply embed_updabspost_ref2.
  eapply updabspost2wp with (old_st:=pairT b t).
  case t; simpl; auto.
  simpl.
  rewrite <- H5.
  intuition auto.
  case (sndT a0); simpl; auto.
  eapply Sorted_union.
  2: apply H7.
  generalize (Zle_bool_gt _ _ H4).
  intros; assert ((fstT a0)=(a-1)%Z).
  abstract omega.
  rewrite H10.
  ring_simplify (a-1+1)%Z.
  apply Sorted_ext2.
  intros i X; try abstract omega.
Qed.

End MaxisortDef.
