(* The game of nim *)

Set Implicit Arguments.

Require Export FSLTarski.
Require Export FSLAbsPost.
Require Export ZBoolAux.

Open Scope Z_scope.

Definition specPlayer (n: Z) : DS Z :=
  bind (require (n > 0)) 
          (fun _ => absPost (fun m => 1 <= m <= 3 /\ m <= n)).

Definition unfold_game (players: bool -> Z -> DS Z) (game: bool*Z -> (DS bool)) (pa:bool*Z)
 := let (turn,n) := pa in
      if (Zeq_bool n 0) 
      then 
         val (negb turn)
      else
         bind (players turn n) (fun m => game (negb turn, n-m)). 

Lemma unfold_game_monotonic1:
  forall (players: bool -> Z -> DS Z) (game1 game2: bool*Z -> DS bool),
  (forall pa, refines (game1 pa) (game2 pa))
   -> (forall pa, refines (unfold_game players game1 pa) (unfold_game players game2 pa)).
Proof.
  unfold unfold_game; intros.
  case pa.
  intros b n.
  case (Zeq_bool n 0); auto with wprefines.
Qed.

Lemma unfold_game_monotonic2:
  forall (pl1 pl2: bool -> Z -> DS Z),
   (forall b x, refines (pl1 b x) (pl2 b x)) -> 
   forall (game: bool*Z -> DS bool) (pa: bool*Z),
     refines (unfold_game pl1 game pa) (unfold_game pl2 game pa).
Proof.
  intros; unfold unfold_game.
  case pa; clear pa; intros start n; simpl.
  case (Zeq_bool n 0); case start; auto with wprefines.
Qed.

Hint Resolve unfold_game_monotonic1 unfold_game_monotonic2: playmatches.

Definition game (players: bool -> Z -> DS Z)  (start:bool) (n:Z) : (DS bool)  :=
  bind (require ((forall b x, refines (players b x) (specPlayer x)) /\ 
                  n >= 0))
         (fun _ => sfixF (unfold_game players) (start,n)).

Lemma game_monotonic:
  forall (pl1 pl2: bool -> Z -> DS Z),
   (forall b x, refines (pl1 b x) (pl2 b x)) -> 
     forall (turn: bool) (n:Z), 
         refines (game pl1 turn n) (game pl2 turn n).
Proof.
  intros pl1 pl2 Hpl turn n; unfold game.
  apply require_refined; intros H.
  apply require_refines; intuition ((try eapply refines_trans; eauto; fail) || idtac).
  apply sfixF_monotony;  auto with playmatches.
Qed.

Lemma game_symmetry:
  forall (players: bool -> Z -> DS Z) (turn: bool) (n:Z), 
         refines (game players turn n) 
                    (bind (game (fun b => players (negb b)) (negb turn) n) (fun b => val (negb b))).
Proof.
  intros players turn n; unfold game.
  eapply refines_trans.
  2: apply bind_assoc_ref2.
  apply bind_simpl_right; simpl. 
  intuition. 
  apply require_refines.
  intuition.
  generalize (H (negb b) x).
  case b; simpl; auto.
  apply sfixF2_refines_G_sfixF1 with 
    (G:=fun (f: bool*Z -> DS bool) (p:bool*Z)=> 
            let (turn,n) := p in 
            bind (f (negb turn,n)) (fun b => val (negb b)))
    (x:=(turn,n)); auto with playmatches.
  intros f g H3 x; case x; auto with wprefines.
  clear H2 turn n; intros f x; case x; clear x; intros turn n.
  unfold unfold_game.
  case (Zeq_bool n 0); simpl.
  apply bind_simpl_right; simpl; case turn; simpl; auto with wprefines.
  case turn; simpl; apply bind_assoc_ref2.
Qed.

Lemma game_dont_abort:
  forall (players: bool -> Z -> DS Z) (turn: bool) (n:Z),
     (forall b n,  refines (players b n) (specPlayer n)) ->
     n >= 0 -> 
         refines (game players turn n) 
                    (any bool).
Proof.
  intros pl turn n Hpl Hn.
  eapply refines_trans.
  apply game_monotonic with (pl2:=fun b:bool => specPlayer); auto with wprefines.
  unfold game.
  apply require_refines; intuition auto with wprefines.
  eapply refines_trans.
  eapply sfixF_mostmagic with (fixO:=fun (p:bool*Z) =>
    let (turn,n):=p in 
    bind (require (n>=0))
           (fun _ => any bool)).
  clear Hn Hpl pl turn n; intros x; case x; intros turn n; clear x; simpl.
  apply require_refined; intuition.
  generalize (Zeq_cases n 0); case (Zeq_bool n 0).
  auto with wprefines.
  unfold specPlayer; case turn; intros; apply bind_simpl_left; simpl; intuition.
  simpl; intuition.
Qed.

Definition winPlayer (n: Z) : DS Z := val (Zmax 1 (n mod 4)).

Lemma winPlayer_iscorrect:
  forall (n:Z),  refines (winPlayer n) (specPlayer n).
Proof.
  intros; unfold specPlayer; apply bind_simpl_right; simpl.
  intuition auto.
  apply abspost2wp; simpl.
  intuition.
  apply Zle_max_l.
  apply Zmax_case.
  omega.
  generalize (Z_mod_lt n 4); omega.
  apply Zmax_case.
  omega.
  assert (X: n mod 4 = n - 4 * (Zdiv n 4)).
  generalize (Z_div_mod_eq n 4); omega.
  rewrite X.
  generalize (Z_div_ge0 n 4); omega.
Qed.

Hint Resolve winPlayer_iscorrect: playmatches.

Definition mk_pair (A:Type) (x y: A) (b:bool): A :=
  if b then x else y.

Definition winning_game (turn:bool) (n:Z) : DS bool
  := if (Zeq_bool (n mod 4) 0)
       then
          (if turn then (any bool) else (val true))
        else
          (if turn then (val true) else (any bool)).

Theorem winPlayer_game:
   forall (otherplayer: Z -> DS Z) (turn:bool) (n:Z),
      (forall x, refines (otherplayer x) (specPlayer x)) ->
      n >= 0  ->
        refines (game (mk_pair winPlayer otherplayer) turn n)
                (winning_game turn n).
Proof.
  intros op turn n Hop Hn.
  eapply refines_trans.
  apply game_monotonic with (pl2:=(mk_pair winPlayer specPlayer)).
  intros b; unfold mk_pair; case b; simpl; auto with wprefines.
  clear Hop op; unfold game.
  apply require_refines.
  intuition.
  unfold mk_pair; case b; simpl; auto with playmatches wprefines.
  eapply refines_trans.
  eapply sfixF_mostmagic with (fixO:=fun (p:bool*Z) =>
    let (turn,n):=p in
    bind (require (n>=0))
           (fun _ => winning_game turn n)).
  (** 1. coeur de la preuve **)
  unfold unfold_game, winning_game; intros x; case x; clear x Hn n.
  intros b n; apply require_refined; intros H.
  generalize (Zeq_cases n 0); case (Zeq_bool n 0).
  (* 1.1 cas de base *)
  intros; subst; case b; simpl; auto with wprefines.
  (* 1.2 cas recursif *)
  intros H0; case b; simpl.
  (* 1.2.1 cas: winPlayer turn *)
  unfold winPlayer.
  generalize (Zeq_cases (n mod 4) 0); case (Zeq_bool (n mod 4) 0).
  (* 1.2.1.1 winPlayer stucks *)
  intros H1; rewrite H1; unfold Zmax; simpl.
  apply bind_simpl_left; simpl.
  apply bind_simpl_left; simpl.
  intuition.
  case (Zeq_bool ((n - 1) mod 4) 0); auto with wprefines.
  (* 1.2.1.2 winPlayer wins *)
  intros H1; cutrewrite ((Zmax 1 (n mod 4))=(n mod 4)).
  apply bind_simpl_left; simpl.
  apply bind_simpl_left; simpl.
  cutrewrite ((n - (n mod 4))=4*(n / 4)).
  intuition.
  generalize (Z_div_ge0 n 4); omega.
  cutrewrite (4 * (n/4) = 0 + ((n/4) * 4)).
  rewrite Z_mod_plus; simpl; auto with wprefines.
  omega.
  ring.
  generalize (Z_div_mod_eq n 4); omega.
  apply (Zmax_case_strong 1 (n mod 4)); intros H2; [
    generalize (Z_mod_lt n 4); omega |
    auto ].
  (* 1.2.2 cas: specPlayer turn *)
  unfold specPlayer.
  apply bind_simpl_left; simpl.
  intuition.
  apply bind_simpl_left; simpl.
  intuition.
  generalize (Zeq_cases (n mod 4) 0); case (Zeq_bool (n mod 4) 0);
  generalize (Zeq_cases ((n-a) mod 4) 0); case (Zeq_bool ((n-a) mod 4) 0);
  auto with wprefines.
  (* 1 <= a <= 3  /\  n mod 4 =0  -> (n-a) mod 4 <> 0 *)
  intros H2 H5. assert (H6: 4 > 0).
  omega.
  absurd ((n-a) mod 4<>0); auto.
  clear H2.
  rewrite (Z_div_mod_eq n 4); auto.
  rewrite H5.
  cutrewrite ((4 * (n / 4) + 0 - a) = (-a + (n / 4)*4)).
  rewrite Z_mod_plus; auto.
  intros H2.
  absurd (a mod 4 = 0).
  intros H7.
  generalize (Z_div_mod_eq a 4 H6).
  intuition.
  generalize (Z_mod_zero_opp _ _ H6 H2).
  cutrewrite (- -a =a ); omega.
  ring.
  (***** fin ****)
  intuition.
Qed.

Lemma winPlayer_against_winPlayer:  forall (turn:bool) (n:Z), 
    n >= 0  ->
    refines (game (fun _ => winPlayer) turn n)
            (val (eqb (Zeq_bool (n mod 4) 0) (negb turn))).
Proof.
  intros turn n Hn.   generalize (Zeq_cases (n mod 4) 0); case (Zeq_bool (n mod 4) 0); case turn; simpl;
  intros H;
  match goal with
  |- refines ?a (val false) => 
         eapply refines_trans; [
               apply game_symmetry
             | eapply refines_trans; [
                 apply bind_monotony with (s4:=fun b : bool => val (negb b)); [
                    eapply refines_trans; 
                      [ idtac | apply winPlayer_game; try (intros; apply winPlayer_iscorrect) ] ;
                      [ apply game_monotonic; intros b; case b; simpl; auto with wprefines | auto ] 
                  | auto with wprefines]
                | idtac ]]
  | |- refines ?a (val true) => 
      eapply refines_trans; [ 
                    eapply refines_trans; 
                      [ idtac | apply winPlayer_game; try (intros; apply winPlayer_iscorrect) ] ;
                      [ apply game_monotonic; intros b; case b; simpl; auto with wprefines | auto ] 
                   | idtac ]
  end ;
  unfold winning_game; (rewrite (Zeq_eq_bool _ _ H) || rewrite (Zdiff_eq_bool _ _ H)); simpl; 
  apply refines_fold; simpl; auto.
Qed.

Require Export FSLFixF.
Require Import Zwf.
Import WfT.

Definition unfold_game_trm (players: bool -> Z -> DS Z) (game: bool*Z -> (DS bool)) (pa:bool*Z)
 := bind (require ((snd pa)>=0)) (fun _ => unfold_game players game pa).

Definition Rgame := (fun  (p1 p2: bool*Z) => Zwf 0 (snd p1) (snd p2)).

Lemma Rgame_well_founded: (well_founded Rgame).
Proof.
   unfold Rgame.
   apply wf_inverse_image with (f:=fun (p:bool*Z) => (snd p)).
   apply Zwf_well_founded.
Qed.


Lemma unfold_game_decreasing: 
  forall (players: bool -> Z -> DS Z),
     (forall b (n:Z),  refines (players b n) (specPlayer n)) ->
     call_decreasing (unfold_game_trm players)  
                            (fun  (p1 p2: bool*Z) => Zwf 0 (snd p1) (snd p2)).
Proof.
  unfold call_decreasing, unfold_game_trm.
  intros players H f x; case x; intros b n; clear x; simpl.
  apply bind_simpl_both.
  auto with wprefines.
  simpl; intuition.
  generalize (Zeq_cases n 0); case (Zeq_bool n 0); simpl; intuition.
  eapply intro_bind_abspost_l.
     eapply refines_trans; eauto;
     unfold specPlayer; apply require_refines;  [ omega | auto with wprefines].
  unfold Zwf; intuition.
Qed.

Definition game_trm (players: bool -> Z -> DS Z)  (turn:bool) (n:Z) : (DS bool)  :=
  bind (require (forall b x, refines (players b x) (specPlayer x)))
       (fun _ => fixF (unfold_game_trm players) Rgame_well_founded (turn,n)).

Lemma game_trm_game:
  forall (players: bool -> Z -> DS Z) (turn: bool) (n:Z),
     refines (game_trm players turn n) 
             (game players turn n).
Proof.
  unfold game_trm, game.
  intros; apply require_refined.
  simpl; intuition.
  apply require_refines; simpl; intuition.
  eapply refines_trans.
  apply fixF_mostmagic with (fixO := fun p: bool*Z => bind (require ((snd p)>=0)) (fun _ => sfixF (unfold_game players) p)).
  unfold unfold_game_trm; intuition.
  intros x ; case x; clear H1 turn n; intros turn n.
  unfold unfold_game_trm; simpl.
  apply bind_simpl_both; simpl; intuition.
  eapply refines_trans.
  2: (apply sfixF_prefix; auto with playmatches).
  unfold unfold_game at 2; generalize (Zeq_cases n 0); case (Zeq_bool n 0); simpl; intuition.
  eapply intro_bind_abspost_l.
      eapply refines_trans; eauto;
      unfold specPlayer; apply require_refines;  [ omega | auto with wprefines].
  unfold Zwf; intuition.
  intuition.
Qed.

Lemma game_trm_aux:
  forall (players: bool -> Z -> DS Z) (turn: bool) (n:Z),
         refines (game players turn n)
                 (game_trm players turn n).
Proof.
  unfold game_trm.
  intros; apply require_refined.
  simpl; intuition.
  apply fixF_mostabort with 
    (fixO := fun p => game players (fst p) (snd p))
    (x:=(turn,n)).
  unfold unfold_game_trm; intuition.
  apply unfold_game_decreasing; auto.
  unfold unfold_game_trm. intros x; clear turn n.
  unfold game; apply bind_monotony.
  apply refines_fold; simpl; intuition.
  intros; eapply refines_trans.
  eapply sfixF_postfix.
  intuition.
  case x; clear x a; intros; apply unfold_game_monotonic1.
  intuition.
Qed.
