(* play with matches *)

Set Implicit Arguments.

Require Export FSLTarski.
Require Export FSLAbsPost.
Require Export ZBoolAux.

Open Scope Z_scope.

Definition specPlayer (n: Z) : DS Z :=
  bind (require (n > 0)) 
          (fun _ => absPost (fun m => 1 <= m <= 3 /\ m <= n)).

Definition unfold_game (pt pf: Z -> DS Z) (game: bool*Z -> (DS bool)) (pa:bool*Z)
 := let (b,n) := pa in
      if (Zeq_bool n 0) 
      then 
         val (negb b)
      else
         bind (if b then (pt n) else (pf n)) (fun m => game (negb b, n-m)). 

Lemma unfold_game_monotonic1:
  forall (pt pf: Z -> DS Z) (game1 game2: bool*Z -> DS bool),
  (forall pa, refines (game1 pa) (game2 pa))
   -> (forall pa, refines (unfold_game pt pf game1 pa) (unfold_game pt pf game2 pa)).
Proof.
  unfold unfold_game; intros.
  case pa.
  intros b n.
  case (Zeq_bool n 0); auto with wprefines.
Qed.

Lemma unfold_game_monotonic2:
  forall (pt1 pf1 pt2 pf2: Z -> DS Z),
   (forall x, refines (pt1 x) (pt2 x)) -> 
   (forall x, refines (pf1 x) (pf2 x)) ->
   forall (game: bool*Z -> DS bool) (pa: bool*Z),
     refines (unfold_game pt1 pf1 game pa) (unfold_game pt2 pf2 game pa).
Proof.
  intros; unfold unfold_game.
  case pa; clear pa; intros start n; simpl.
  case (Zeq_bool n 0); case start; auto with wprefines.
Qed.

Hint Resolve unfold_game_monotonic1 unfold_game_monotonic2: playmatches.

Definition game (pt pf: Z -> DS Z)  (start:bool) (n:Z) : (DS bool)  :=
  bind (require ((forall x, refines (pt x) (specPlayer x)) /\ 
                      (forall x, refines (pf x) (specPlayer x)) /\
                      n >= 0))
         (fun _ => sfixF (unfold_game pt pf) (start,n)).

Lemma game_monotonic:
  forall (pt1 pf1 pt2 pf2: Z -> DS Z),
   (forall x, refines (pt1 x) (pt2 x)) -> 
   (forall x, refines (pf1 x) (pf2 x)) ->
     forall (start: bool) (n:Z), 
         refines (game pt1 pf1 start n) (game pt2 pf2 start n).
Proof.
  intros pt1 pf1 pt2 pf2 Ht Hf start n; unfold game.
  apply require_refined; intros H.
  apply require_refines; intuition ((try eapply refines_trans; eauto; fail) || idtac).
  apply sfixF_monotony;  auto with playmatches.
Qed.

Lemma game_symmetry:
  forall (pt pf: Z -> DS Z) (start: bool) (n:Z), 
         refines (game pt pf start n) 
                    (bind (game pf pt (negb start) n) (fun b => val (negb b))).
Proof.
  intros pt pf start n; unfold game.
  eapply refines_trans.
  2: apply bind_assoc_ref2.
  apply bind_simpl_right; simpl. 
  intuition. 
  apply require_refines; intuition.
  apply sfixF2_refines_G_sfixF1 with 
    (G:=fun (f: bool*Z -> DS bool) (p:bool*Z)=> 
            let (start,n) := p in 
            bind (f (negb start,n)) (fun b => val (negb b)))
    (x:=(start,n)); auto with playmatches.
  intros f g H2 x; case x; auto with wprefines.
  clear H3 start n; intros f x; case x; clear x; intros start n.
  unfold unfold_game.
  case (Zeq_bool n 0); simpl.
  apply bind_simpl_right; simpl; case start; simpl; auto with wprefines.
  case start; simpl; apply bind_assoc_ref2.
Qed.

Lemma game_dont_abort:
  forall (pt pf: Z -> DS Z) (start: bool) (n:Z),
     (forall (n:Z),  refines (pt n) (specPlayer n)) ->
     (forall (n:Z),  refines (pf n) (specPlayer n)) ->
     n >= 0 -> 
         refines (game pt pf start n) 
                    (any bool).
Proof.
  intros pt pf start n Ht Hf Hn.
  eapply refines_trans.
  apply game_monotonic with (pt2:=specPlayer) (pf2:=specPlayer); auto with wprefines.
  unfold game.
  apply require_refines; intuition auto with wprefines.
  eapply refines_trans.
  eapply sfixF_mostmagic with (sp:=fun (p:bool*Z) =>
    let (start,n):=p in 
    bind (require (n>=0))
           (fun _ => any bool)).
  clear Hn Hf Ht pt pf start n; intros x; case x; intros start n; clear x; simpl.
  apply require_refined; intuition.
  generalize (Zeq_cases n 0); case (Zeq_bool n 0).
  auto with wprefines.
  unfold specPlayer; case start; intros; apply bind_simpl_left; simpl; intuition.
  simpl; intuition.
Qed.

Definition myplayer (n: Z) : DS Z := val (Zmax 1 (n mod 4)).

Lemma myplayer_iscorrect:
  forall (n:Z),  refines (myplayer n) (specPlayer n).
Proof.
  intros; unfold specPlayer; apply bind_simpl_right; simpl.
  intuition auto.
  apply abspost2wp; simpl.
  intuition.
  apply Zle_max_l.
  apply Zmax_case.
  omega.
  generalize (Z_mod_lt n 4); omega.
  apply Zmax_case.
  omega.
  assert (X: n mod 4 = n - 4 * (Zdiv n 4)).
  generalize (Z_div_mod_eq n 4); omega.
  rewrite X.
  generalize (Z_div_ge0 n 4); omega.
Qed.

Hint Resolve myplayer_iscorrect: playmatches.

Definition winning_game (start:bool) (n:Z) : DS bool 
  := if (Zeq_bool (n mod 4) 0) 
       then
          (if start then (any bool) else (val true))
        else
          (if start then (val true) else (any bool)).

Theorem myplayer_game:
   forall (otherplayer: Z -> DS Z) (start:bool) (n:Z), 
      (forall x, refines (otherplayer x) (specPlayer x)) ->
      n >= 0  ->
        refines (game myplayer otherplayer start n)
                   (winning_game start n). 
Proof.
  intros op start n Hop Hn.
  eapply refines_trans.
  apply game_monotonic with (pt2:=myplayer) (pf2:=specPlayer); auto with wprefines.
  unfold game.
  apply require_refines; auto with playmatches wprefines.
  eapply refines_trans.
  eapply sfixF_mostmagic with (sp:=fun (p:bool*Z) =>
    let (start,n):=p in 
    bind (require (n>=0))
           (fun _ => winning_game start n)).
  unfold unfold_game, winning_game; intros x; case x; clear x Hn n.
  (** 1. coeur de la preuve **)
  intros b n; apply require_refined; intros H.  
  generalize (Zeq_cases n 0); case (Zeq_bool n 0).
  (* 1.1 cas de base *) 
  intros; subst; case b; simpl; auto with wprefines.
  (* 1.2 cas recursif *)
  intros H0; case b; simpl.
  (* 1.2.1 cas: myplayer starts *)
  unfold myplayer.
  generalize (Zeq_cases (n mod 4) 0); case (Zeq_bool (n mod 4) 0).
  (* 1.2.1.1 myplayer stucks *)
  intros H1; rewrite H1; unfold Zmax; simpl.
  apply bind_simpl_left; simpl.
  apply bind_simpl_left; simpl.
  intuition.
  case (Zeq_bool ((n - 1) mod 4) 0); auto with wprefines.
  (* 1.2.1.2 myplayer wins *)
  intros H1; cutrewrite ((Zmax 1 (n mod 4))=(n mod 4)).
  apply bind_simpl_left; simpl.
  apply bind_simpl_left; simpl.
  cutrewrite ((n - (n mod 4))=4*(n / 4)).
  intuition.
  generalize (Z_div_ge0 n 4); omega.
  cutrewrite (4 * (n/4) = 0 + ((n/4) * 4)).
  rewrite Z_mod_plus; simpl; auto with wprefines.
  omega.
  ring.
  generalize (Z_div_mod_eq n 4); omega.
  apply (Zmax_case_strong 1 (n mod 4)); intros H2; [
    generalize (Z_mod_lt n 4); omega |
    auto ].
  (* 1.2.2 cas: specPlayer starts *)
  unfold specPlayer.
  apply bind_simpl_left; simpl.
  intuition.
  apply bind_simpl_left; simpl.
  intuition.
  generalize (Zeq_cases (n mod 4) 0); case (Zeq_bool (n mod 4) 0);
  generalize (Zeq_cases ((n-a) mod 4) 0); case (Zeq_bool ((n-a) mod 4) 0);
  auto with wprefines.
  (* 1 <= a <= 3  /\  n mod 4 =0  -> (n-a) mod 4 <> 0 *)
  intros H2 H5. assert (H6: 4 > 0).
  omega.
  absurd ((n-a) mod 4<>0); auto.
  clear H2.
  rewrite (Z_div_mod_eq n 4); auto.
  rewrite H5.
  cutrewrite ((4 * (n / 4) + 0 - a) = (-a + (n / 4)*4)).
  rewrite Z_mod_plus; auto.
  intros H2.
  absurd (a mod 4 = 0).
  intros H7.
  generalize (Z_div_mod_eq a 4 H6).
  intuition.
  generalize (Z_mod_zero_opp _ _ H6 H2).
  cutrewrite (- -a =a ); omega.
  ring.
  (***** fin ****)
  intuition.
Qed.

Lemma  myplayer_against_myplayer:  forall (start:bool) (n:Z), 
    n >= 0  ->
    refines (game myplayer myplayer start n)
               (val (eqb (Zeq_bool (n mod 4) 0) (negb start))).
Proof.
  intros start n Hn;   generalize (Zeq_cases (n mod 4) 0); case (Zeq_bool (n mod 4) 0); case start; simpl;
  intros H; match goal with
  |- refines ?a (val false) => 
         eapply refines_trans; [
             apply game_symmetry
             | eapply refines_trans;  [
                apply bind_monotony with (s4:=fun b : bool => val (negb b)); [
                       apply myplayer_game; auto with playmatches
                     | auto with wprefines ]
                | idtac ]]
  | |- refines ?a (val true) => 
       eapply refines_trans; [ apply myplayer_game; auto with playmatches | idtac ]
  end;
  unfold winning_game; (rewrite (Zeq_eq_bool _ _ H) || rewrite (Zdiff_eq_bool _ _ H)); simpl; 
  apply refines_fold; simpl; auto.
Qed.

Require Export FSLFixF.
Require Import Zwf.
Import WfT.

Definition unfold_game_trm (pt pf: Z -> DS Z) (game: bool*Z -> (DS bool)) (pa:bool*Z)
 := bind (require ((snd pa)>=0)) (fun _ => unfold_game pt pf game pa).

Definition Rgame := (fun  (p1 p2: bool*Z) => Zwf 0 (snd p1) (snd p2)).

Lemma Rgame_well_founded: (well_foundedT Rgame).
Proof.
   unfold Rgame.
   apply wfT_lift with (f:=fun (p:bool*Z) => (snd p)).
   apply wf2wfT.
   apply Zwf_well_founded.
Qed.


Lemma unfold_game_decreasing: 
  forall (pt pf: Z -> DS Z),
     (forall (n:Z),  refines (pt n) (specPlayer n)) ->
     (forall (n:Z),  refines (pf n) (specPlayer n)) ->
     call_decreasing (unfold_game_trm pt pf)  
                            (fun  (p1 p2: bool*Z) => Zwf 0 (snd p1) (snd p2)).
Proof.
  unfold call_decreasing, unfold_game_trm.
  intros pt pf H1 H2 f x; case x; intros b n; clear x; simpl.
  apply bind_simpl_both.
  auto with wprefines.
  simpl; intuition.
  generalize (Zeq_cases n 0); case (Zeq_bool n 0); simpl; intuition.
  eapply intro_bind_abspost_l.
    case b; eapply refines_trans; eauto.
      unfold specPlayer; apply require_refines;  [ omega | auto with wprefines].
      unfold specPlayer; apply require_refines;  [ omega | auto with wprefines].
  unfold Zwf; intuition.
Qed.

Definition game_trm (pt pf: Z -> DS Z)  (start:bool) (n:Z) : (DS bool)  :=
  bind (require ((forall x, refines (pt x) (specPlayer x)) /\ 
                      (forall x, refines (pf x) (specPlayer x))))
         (fun _ => fixF (unfold_game_trm pt pf) Rgame_well_founded (start,n)).

Lemma game_trm_game:
  forall (pt pf: Z -> DS Z) (start: bool) (n:Z),
     refines (game_trm pt pf start n) 
                (game pt pf start n).
Proof.
  unfold game_trm, game.
  intros; apply require_refined.
  simpl; intuition.
  apply require_refines; simpl; intuition.
  eapply refines_trans.
  apply fixF_mostmagic with (fixO := fun p: bool*Z => bind (require ((snd p)>=0)) (fun _ => sfixF (unfold_game pt pf) p)).
  unfold unfold_game_trm; intuition.
  intros x ; case x; clear H2 start n; intros start n.
  unfold unfold_game_trm; simpl.
  apply bind_simpl_both; simpl; intuition.
  eapply refines_trans.
  2: (apply sfixF_prefix; auto with playmatches).
  unfold unfold_game at 2; generalize (Zeq_cases n 0); case (Zeq_bool n 0); simpl; intuition.
  eapply intro_bind_abspost_l.
    case start; eapply refines_trans; eauto.
      unfold specPlayer; apply require_refines;  [ omega | auto with wprefines].
      unfold specPlayer; apply require_refines;  [ omega | auto with wprefines].
  unfold Zwf; intuition.
  intuition.
Qed.

Lemma game_trm_aux:
  forall (pt pf: Z -> DS Z) (start: bool) (n:Z),
         refines (game pt pf start n)
                    (game_trm pt pf start n).
Proof.
  unfold game_trm.
  intros; apply require_refined.
  simpl; intuition.
  apply fixF_mostabort with 
    (fixO := fun p => game pt pf (fst p) (snd p))
    (x:=(start,n)).
  unfold unfold_game_trm; intuition.
  apply unfold_game_decreasing; auto.
  unfold unfold_game_trm. intros x; clear start n.
  unfold game; apply bind_monotony.
  apply refines_fold; simpl; intuition.
  intros; eapply refines_trans.
  eapply sfixF_postfix.
  intuition.
  case x; clear x a; intros; apply unfold_game_monotonic1.
  intuition.
Qed.
