(* begin hide *)
Set Implicit Arguments.
(* end hide *)
 
(** ** Module [Place]: Abstract management of ``states'' %\label{subsec:Place}%

This module describes a model of state conversions using places.
Informally, a place is typically a mutable variable. And a set of
place is also a place. Given a type [Env] of environments and a type
[A], [(Place Env A)] is the type of places of type [A] in the
environment of type [Env]. In typical use of this module, [A] and
[Env] are uples.

*** Definition and examples of places 

In this definition, places are defined by two functions [read] and
[write] that allow respectively to read the value of the place in
the environment, and to write a new value for the place in the
environment, creating thus a new environment (a functional denotation
of places). The relation between [read] and [write] are described by
some invariant properties.
*)

Record Place (Env A: Type) : Type := {
  read: Env -> A ;
  write: Env -> A -> Env ;
  read_write: forall (e:Env) (x:A), (read (write e x))=x ;
  write_read: forall (e:Env), (write e (read e))=e ;
  write_write: forall (e:Env) (x y:A), (write (write e x) y)=(write e y)
}. 

(* begin hide *)
Hint Rewrite -> read_write write_write write_read: placeeq.
(* end hide *)

(** I first introduce [allPl], the place designing the whole environment. 
The [read] function is identity and the [write] function is [fun _ x => x]. 
The %\coq{}% definition of this function is interactive.
*)
Definition allPl (A: Type): (Place A A).
 refine (fun A =>
         (Build_Place 
            (fun x => x)
            (fun _ x => x)
            _
            _  
            _
            ));
 abstract auto.
Defined.
Implicit Arguments allPl [A].

(** I now introduce [nonePl], the empty place. 
The [read] function is [fun x => tt] and the [write] function is [fun x _ => x]. 
*)
Definition nonePl (A: Type): (Place A unit).
 refine (fun A =>
         (Build_Place 
            (fun x => tt)
            (fun x _ => x)
            _
            _  
            _
            ));
 abstract (auto || intros B x; case x; auto).
Defined.
Implicit Arguments nonePl [A].

(** I now introduced the composition of two places, which returns a place.
- [read] function is [fun e => (read v2 (read v1 e))] 
- [write] function is [fun e a => (write v1 e (write v2 (read v1 e) a))].
*)
Definition liftPl (Env A B: Type) (v1:(Place Env A)) (v2:(Place A B)): (Place Env B).
 refine (fun Env A B v1 v2 => 
          (Build_Place 
            (fun e => (read v2 (read v1 e)))
            (fun e a => (write v1 e (write v2 (read v1 e) a)))
            _
            _  
            _
            )); 
 abstract (intros; autorewrite with placeeq using (try reflexivity)).
Defined.

(** The composition of two places [v1] and [v2] is simply noted [v1 # v2].
*)
Notation "v1 # v2" := (liftPl v1 v2) (at level 20, right associativity).

(** The left member of a pair is a place.  [read] is simplify the
first projection. [write] is [fun e x => (pairT x (sndT e))].  
*)
Definition pL (A B:Type): (Place (A*B) A).
 refine (fun A B => 
          (Build_Place 
            (fstT (A:=A) (B:=B))
            (fun e x => (pairT x (sndT e)))
            _
            _  
            _
            ));
 abstract (intros e; case e; simpl; auto).
Defined.
Implicit Arguments pL [A B].

(** Symetrically, the right member of pair is also a place.
*)
Definition pR (A B:Type): (Place (A*B) B).
 refine (fun A B => 
          (Build_Place 
            (sndT (A:=A) (B:=B))
            (fun e x => (pairT (fstT e) x))
            _
            _  
            _
            ));
 abstract (intros e; case e; simpl; auto).
Defined.
Implicit Arguments pR [A B].

(** The permutation of a pair is also a place. 
[read] is [(fun p => (pairT (sndT p) (fstT p)))] and 
[write] is [(fun _ p => (pairT (sndT p) (fstT p)))].
*)
Definition pairSPl (A B: Type): (Place (prodT A B) (prodT B A)).
  refine (fun A B =>
            (Build_Place
              (fun p => (pairT (sndT p) (fstT p)))
              (fun _ p => (pairT (sndT p) (fstT p)))
              _
              _
              _
             ));
  abstract ( (simpl; auto; fail) || (intros e p; case p; simpl; auto; fail) 
    || (intros e; case e; simpl; auto)).
Qed.

(** *** Equality of places 

I first prove that [write] function of places is injective on its rightmost argument.
*)  
Lemma write_right_inj: forall (Env A:Type) (v:(Place Env A)) (e:Env) (x y:A),
  (write v e x)=(write v e y) -> x=y.
Proof.
  intros Env A v e x y H.
  rewrite <- (read_write v e x).
  rewrite H; autorewrite with placeeq.
  auto.
Qed.

(** I now introduce an equality relation between places. Two places are equals if and only if
their respective [write] functions are extensionally equals.
*)
Definition eqPl (Env A:Type) (v1 v2:(Place Env A)) : Prop := 
  forall (e:Env) (x:A), (write v1 e x)=(write v2 e x).

(* begin hide *)
Hint Unfold eqPl: place.
(* end hide *)

Lemma eqPl_refl: forall (Env A:Type) (v:(Place Env A)), (eqPl v v).
Proof.
  auto with place.
Qed.

Lemma eqPl_sym: forall (Env A:Type) (v1 v2:(Place Env A)), (eqPl v1 v2) -> (eqPl v2 v1).
Proof.
  auto with place.
Qed.

Lemma eqPl_trans: forall (Env A:Type) (v1 v2 v3:(Place Env A)),
  (eqPl v1 v2) -> (eqPl v2 v3) -> (eqPl v1 v3).
Proof.
  unfold eqPl; intros; eapply trans_eq; auto.
Qed.
 
(** [read] functions of equal places are extensionally equals. 
*)
Lemma eqPl_read: forall (Env A:Type) (v1 v2:(Place Env A)) (e:Env),
  (eqPl v1 v2) -> (read v1 e)=(read v2 e).
Proof.
  unfold eqPl; intros Env A v1 v2 e H.
  replace (read v2 e) with (read v2 (write v2 e (read v1 e))).
  autorewrite with placeeq; auto.
  rewrite <- (H e (read v1 e)).
  autorewrite with placeeq; auto.
Qed.

(* begin hide *)
Lemma eqPl_write: forall (Env A:Type) (v1 v2:(Place Env A)) (e:Env) (x:A),
  (eqPl v1 v2) -> (write v1 e x)=(write v2 e x).
Proof.
  unfold eqPl; auto.
Qed.
(* end hide *)


(** [liftPl] is compatible with this equality *)
Lemma liftPl_compat: forall (Env A B:Type) (v1a v2a:(Place Env A)) (v1b v2b:(Place A B)),
 (eqPl v1a v2a) -> (eqPl v1b v2b) -> (eqPl (v1a#v1b) (v2a#v2b)).
Proof.
  intros; unfold eqPl.
  simpl; intros. 
  rewrite eqPl_write with (v1:=v1a) (v2:=v2a); auto.
  rewrite eqPl_write with (v1:=v1b) (v2:=v2b); auto.
  rewrite eqPl_read with (v1:=v1a) (v2:=v2a); auto.
Qed.

(** [liftPl] is associative with this equality *)
Lemma liftPl_assoc: forall (Env A B C:Type) 
  (v1:(Place Env A)) (v2:(Place A B)) (v3:(Place B C)),
    (eqPl ((v1#v2)#v3) (v1#v2#v3)).
Proof.
  auto with place.
Qed.

(** [allPl] is a neutral element for [liftPl] *)
Lemma liftPl_allPl_left: forall (Env A:Type) (v:(Place Env A)), 
  (eqPl (allPl#v) v).
Proof.
  auto with place.
Qed.  

Lemma liftPl_allPl_right: forall (Env A:Type) (v:(Place Env A)),
  (eqPl (v#allPl) v).
Proof.
  auto with place.
Qed.  

(** [nonePl] is the unique place of type [unit]. In other words, all place of type 
[unit] are equals. Let us remark that this property does not hold with [allPl] because
[(pairSPl A A)] could for instance have the same type than [(allPl (A:=(prodT A A)))].
*)
Lemma nonePl_uniq_aux: forall (Env:Type) (v:(Place Env unit)),
  (eqPl v nonePl).
Proof.
  unfold eqPl; intros; simpl.
  case x.
  generalize (write_read v e).
  case (read v e).
  auto.
Qed.

Lemma nonePl_uniq: forall (Env:Type) (v1 v2:(Place Env unit)),
  (eqPl v1 v2).
Proof.
  intros; eapply eqPl_trans.
  apply nonePl_uniq_aux.
  apply eqPl_sym; apply nonePl_uniq_aux.
Qed.

(* begin hide *)
(* FAUX:  
Lemma allPl_uniq: forall (Env:Type) (v:(Place Env Env)),
  (eqPl v allPl).
Contre_ex:
  read (x,y:A*A) := (y,x)
  write (e:A*A) (x,y: A*A) := (y,x)
*)

Hint Resolve liftPl_assoc liftPl_allPl_left liftPl_allPl_right: place.
Hint Resolve nonePl_uniq eqPl_refl eqPl_sym liftPl_compat: place.

(* end hide *)

(** *** Bipartition of a state into two places.

We now define a bipartition of a state. Given a state [Env], two
places [vL] and [vR] of respective type [(Place Env A)] and [(Place Env B)], 
[(sumPl vL VR)] is the type of ``proofs'' that [vL] and [vR] form a
bipartition of [Env]. These proofs introduce a function [mkEnv] allowing
to construct a new value of type [Env] from a value of type [A] and a
value of type [B]. Actually, [(sumPl vL VR)] implies that [Env] is 
isomorphic to [(prodT A B)].  This is used in module [ISLBlock].
*)

Record sumPl (Env A B: Type) (vL:(Place Env A)) (vR:(Place Env B)) : Type :=
 { 
   mkEnv: A -> B -> Env ;
   mkEnv2writeL: forall (e:Env) (x:A), (mkEnv x (read vR e))=(write vL e x)  ;
   mkEnv2writeR: forall (e:Env) (x:B), (mkEnv (read vL e) x)=(write vR e x)  
 }.

(* begin hide *)
Hint Rewrite -> mkEnv2writeL mkEnv2writeR: placeeq.
(* end hide *)

(** The three following lemmas shows that [(mkEnv u)] is like the
constructor of a pair, where as [(read vL)] and [(read vR)] are the
corresponding destructors.
*)
Lemma mkEnv_readLR: forall (Env A B:Type) (vL:Place Env A) (vR:Place Env B) 
    (u:sumPl vL vR) (e:Env), 
     (mkEnv u (read vL e) (read vR e))=e.
Proof.
  intros; autorewrite with placeeq.
  auto.
Qed.

Lemma readL_mkEnv: forall (Env A B:Type) (vL:Place Env A) (vR:Place Env B) 
    (u:sumPl vL vR) (x:A) (y:B), (read vL (mkEnv u x y))=x.
Proof.
  intros; rewrite <- (read_write vR (mkEnv u x y) y).
  autorewrite with placeeq.
  auto.
Qed.

Lemma readR_mkEnv: forall (Env A B:Type) (vL:Place Env A) (vR:Place Env B) 
    (u:sumPl vL vR) (x:A) (y:B), (read vR (mkEnv u x y))=y.
Proof.
  intros; rewrite <- (read_write vL (mkEnv u x y) x).
  autorewrite with placeeq.
  auto.
Qed.

(* begin hide *)
Hint Rewrite -> readL_mkEnv readR_mkEnv: placeeq.
(* end hide *)

(** Hence, by lemma [mkEnv_readLR], we could write elements of [Env] in a ``normal form'' [(mkEnv u x y)].
The two following lemmas, show how [(write vL)] and [(write vR)] can be defined using this normal form.
*)
 
Lemma writeL_mkEnv: forall (Env A B:Type) (vL:Place Env A) (vR:Place Env B) 
   (u:sumPl vL vR) (x1 x2:A) (y:B),
       (write vL (mkEnv u x1 y) x2)=(mkEnv u x2 y).
Proof.
  intros; rewrite <- (mkEnv2writeL u (mkEnv u x1 y) x2).
  autorewrite with placeeq.
  auto.
Qed.

Lemma writeR_mkEnv: forall (Env A B:Type) (vL:Place Env A) (vR:Place Env B)
   (u:sumPl vL vR) (x:A) (y1 y2:B),
       (write vR (mkEnv u x y1) y2)=(mkEnv u x y2).
Proof.
  intros; rewrite <- (mkEnv2writeR u (mkEnv u x y1) y2).
  autorewrite with placeeq.
  auto.
Qed.

(* begin hide *)
Hint Rewrite -> writeL_mkEnv writeR_mkEnv: placeeq.
(* end hide *)

(** Of course, the two places [(pL A B)] and [(pR A B)] form a bipartition of [(prodT A B)]. 
*)
Definition pairsumPl (A B:Type): (sumPl (pL (A:=A) (B:=B)) pR).
 refine (fun A B => 
          (Build_sumPl _ _
            (fun x y => (pairT x y))
            _
            _  
            ));
 abstract (intros e; case e; simpl; auto).
Defined.

(* begin hide *)
Definition trivPl (A:Type): (sumPl (nonePl (A:=A)) allPl).
 refine (fun A => 
          (Build_sumPl _ _
            (fun _ x => x)
            _
            _  
            ));
 abstract (intros e; simpl; auto).
Defined.

Definition swapsumPl (Env A B:Type) (vL:Place Env A) (vR:Place Env B) (u:sumPl vL vR)
  : (sumPl vR vL).
  refine (fun Env A B vL vR u =>
            (Build_sumPl _ _ 
               (fun x y => mkEnv u y x)
                _
                _));
  abstract (intros; autorewrite with placeeq; auto).
Defined.

Definition pl2prod (Env A B:Type) (vL:Place Env A) (vR:(Place Env B)) (e:Env): (prodT A B) 
 := (pairT (read vL e) (read vR e)).

Definition prod2pl (Env A B:Type) (vL:Place Env A) (vR:(Place Env B)) (u:(sumPl vL vR)) 
 (p:(prodT A B)): Env 
 := (mkEnv u (fstT p) (sndT p)).

Lemma pl2prod2pl_id: forall (Env A B:Type) (vL:Place Env A) (vR:(Place Env B)) (u:(sumPl vL vR))
  (e:Env), (prod2pl u (pl2prod vL vR e))=e.
Proof.
  unfold prod2pl; simpl.
  intros; autorewrite with placeeq.
  auto.
Qed.

Lemma prod2pl2prod_id: forall (Env A B:Type) (vL:Place Env A) (vR:(Place Env B)) (u:(sumPl vL vR))
  (p:prodT A B), (pl2prod vL vR (prod2pl u p))=p.
Proof.
  unfold pl2prod, prod2pl; intros; case p; simpl.
  intros; autorewrite with placeeq.
  auto.
Qed.
(* end hide *)

(*
Record extPl : Type :=
  { extEnv: Type -> Type ;
    locT: Type ;
    vloc: forall ext, Place (extEnv ext) locT ;
    vext: forall ext, Place (extEnv ext) ext ;
    sext: forall ext, sumPl (vloc ext) (vext ext)
   }.  

Definition eext: extPl := (Build_extPl (fun ext => ext) _ _ trivPl).

Definition xext (B: extPl) (A: Type): extPl.
  intros; refine (Build_extPl (fun ext => extEnv B (A*ext)) 
                              (locT:=extEnv B A)
                              _
                              (fun ext => (vext B A*ext)#pR)
                              _
                 ).
  
*)


(* begin hide *)
(* PROPRIETE NON-DEMONTRABLE ??? 

Lemma interfWrite: forall (Env A B:Type) (v1:(Place Env A)) (v2:(Place Env B)),
  forall (e1 e2:Env) (x:A) (y:B),
    (read v1 (write v2 (write v1 e1 x) y))=(read v1 (write v2 (write v1 e2 x) y)).
Proof.
  intros.

Qed.
*)

(* inclusion de place. ne marche pas bien a cause du pb ci-dessus...

Definition containsPl (Env A B:Type) (v1:(Place Env A)) (v2:(Place Env B)): Type 
  := forall (e:Env) (x:B), (write v1 e (read v1 (write v2 e x)))=(write v2 e x).   

Lemma containsPl_write1:
 forall (Env A B:Type) (v1:(Place Env A)) (v2:(Place Env B)) (e:Env) (e':A) (x:B),
   (containsPl v1 v2) -> (write v1 e (read v1 (write v2 (write v1 e e') x)))=(write v2 (write v1 e e') x).
Proof.
 intros Env A B v1 v2 e e' x H; unfold containsPl in H.
 eapply trans_eq.
 2: apply H.
 autorewrite with placeeq.
 auto.
Qed.

Definition subPl (Env A B:Type) (v1:(Place Env A)) (v2:(Place Env B)): 
   Env -> (containsPl v1 v2) -> (Place A B).
  refine (fun Env A B v1 v2 e h =>
             Build_Place (fun e' => (read v2 (write v1 e e')))
                       (fun e' x => (read v1 (write v2 (write v1 e e') x)))
                       _ _ _);
  abstract (intros; try (rewrite containsPl_write1) ; autorewrite with placeeq; auto).
Defined.

Lemma subPl_eqPl: forall (Env A B:Type) (v1a v1b:(Place Env A)) (v2a v2b:(Place Env B)) (ea eb:Env)
  (ha:(containsPl v1a v2a)) (hb:(containsPl v1b v2b)),
  (eqPl v1a v1b) -> (eqPl v2a v2b) -> (eqPl (subPl ea ha) (subPl eb hb)).
Proof.
  intros; unfold eqPl; simpl.
  intros e x. 
  rewrite eqPl_write with (v1:=v1a) (v2:=v1b); auto.
  rewrite eqPl_write with (v1:=v2a) (v2:=v2b); auto.
  rewrite eqPl_read with (v1:=v1a) (v2:=v1b); auto.
  assert 
  ((write v1b (write v2b (write v1b eb e) x) (read v1b (write v2b (write v1b eb e) x))
    = (write v1b (write v2b (write v1b eb e) x) (read v1b (write v2b (write v1b ea e) x))))).
  autorewrite with placeeq.
Qed.
*)

(* version 1. Pas tres bon ?

(** inclusion de place **)

Record inclPl (Env A B:Type) (v1:(Place Env A)) (v2:(Place Env B)): Type := {
  iWitV: (Place B A) ;
  iWitVproof: (eqPl v1 (v2#iWitV)) 
}.

Definition nonePl_included (Env A: Type) (v:(Place Env A)): (inclPl nonePl v).
  refine (fun Env A v => Build_inclPl (iWitV:=nonePl) _).
  abstract auto with place.
Defined.  

Definition allPl_includes (Env A: Type) (v:(Place Env A)): (inclPl v allPl).
  refine (fun Env A v => Build_inclPl (iWitV:=v) _).
  abstract auto with place.
Defined.  

Definition eqPl_inclPl (Env A: Type) (v1:(Place Env A)) (v2:(Place Env A)): (eqPl v1 v2) -> (inclPl v1 v2).
  refine (fun Env A v1 v2 h => Build_inclPl (iWitV:=allPl) _).
  abstract auto with place.
Defined.  

Definition inclPl_trans (Env A B C: Type) (v1:(Place Env A)) (v2:(Place Env B)) (v3:(Place Env C)): 
  (inclPl v1 v2) -> (inclPl v2 v3) -> (inclPl v1 v3).
  refine (fun Env A B C v1 v2 v3 i1 i2 => Build_inclPl (iWitV:=(iWitV i2)#(iWitV i1)) _).
  abstract 
   (case i1; case i2; simpl;
    intros; eapply eqPl_trans; [ eauto | idtac ];
    eapply eqPl_trans; [ idtac | apply liftPl_assoc] ;
    auto with place).
Defined.  

Definition liftPl_inclPl (Env A B C D: Type) (v1a:(Place Env A)) (v1b:(Place A B)) (v2a:(Place Env C)) (v2b:(Place C D))
 (i1:(inclPl v1a v2a)): (inclPl ((iWitV i1)#v1b) v2b) -> (inclPl (v1a#v1b) (v2a#v2b)).
  refine (fun Env A B C D v1a v1b v2a v2b i1 i2 => Build_inclPl (iWitV:=(iWitV i2)) _).
  abstract 
   (generalize i2; case i1; clear i1 i2; simpl;
    intros va H1 i2; case i2; simpl; clear i2;
    intros vb H2;
    apply eqPl_trans with (v2:=v2a#v2b#vb); auto with place;
    apply eqPl_trans with (v2:=v2a#va#v1b); auto with place;
    apply eqPl_trans with (v2:=(v2a#va)#v1b); auto with place).
Defined.
*)

Unset Implicit Arguments.
(* end hide *)
