(** ** Module [SSL]: specifications with a global state.

This module defines operators for a Dijsktra specification language
with state and exception handling. Monad operators are extended to
deal with local states.
*)

(* begin hide *)
Set Implicit Arguments.

Require Export FSL.
(* end hide *)

(** The type of specification is also parametrized by [St] the global
type of states.
*)

Definition K (St A:Type) := St -> WP.DS (prodT St A).  

(** This definition uses [WP.DS] the [DS] type defined in module [WP].
Assuming that [st] is a state and [sp] a specification, the term
[(pre (sp st))] is the precondition of [sp] on the initial state [st].
The same remarks applies for [(sdemon (sp st))] and [(angel (sp st))].
*)
 
(* begin hide *)
Delimit Scope SSL_scope with SSL. 
Bind Scope SSL_scope with K.
(* end hide *)

(** I first define [refInEnv] the relation of refinement in a
environment, using [refines] relation of [WP]. 
*)

Definition refInEnv  (A St: Type) (st:St) (sp1 sp2 : K St A) : Prop :=
  WP.refines (sp1 st) (sp2 st).

(** For a given [st], [(refInEnv st)] is reflexive and transitive.
*)
Lemma refInEnv_refl :
 forall (A St: Type) (st: St) (sp : K St A), refInEnv st sp sp.
Proof.
  unfold refInEnv in |- *; auto with wprefines.
Qed.

Lemma refInEnv_trans :
 forall (A St: Type) (st:St) (sp1 sp2 sp3 : K St A),
 refInEnv st sp1 sp2 -> refInEnv st sp2 sp3 -> refInEnv st sp1 sp3.
Proof.
  unfold refInEnv in |- *; eauto with wprefines.
Qed.

(** This lemma is the proof rule stating how a refinement goal
can be simplified by a wp-computation.
*)
Lemma refInEnv2wp :
 forall (A St: Type) (st:St) (sp1 sp2 : K St A),
 (pre (sp2 st) -> 
  sdemon (sp1 st) (fun a => post (sp2 st) a)) -> 
 refInEnv st sp1 sp2.
Proof.
  unfold refInEnv in |- *; auto with wprefines.
Qed.

(** The following lemma is the converse of the preceding one.
*)
Lemma refInEnv_from_wp :
 forall (A St: Type) (st:St) (sp1 sp2 : K St A),
 refInEnv st sp1 sp2 -> 
 pre (sp2 st) -> 
 sdemon (sp1 st) (fun a => post (sp2 st) a).
Proof.
  unfold refInEnv in |- *; intros A St st sp1 sp2 H H1;
   apply (WP.refines_unfold H); auto.
Qed.

(** In a refinement proof, 
the precondition of the refined specification can still be assumed.
*)
Lemma refInEnv_pre_intro:
  forall (A St: Type) (st:St) (sp1 sp2 : K St A),
   ((pre (sp2 st)) -> refInEnv st sp1 sp2) -> refInEnv st sp1 sp2.
Proof.
  intros; unfold refInEnv; apply refines_pre_intro; auto.
Qed.

(** Below, I define [val] and [raise] operators which
corresponds to their homonym in the algebraic semantics.
*)

Definition val (A St: Type) (a : A) : K St A :=
  fun st => FSL.val (pairT st a).

(** I extend here [set] and [get] operators of monad, with a place
parameter.  Hence [set] and [get] are close to the usual constructions
of programming language, allowing to affect a reference or to derefence.
*)

Definition set (St:Type) (v:St) : K St unit :=
  fun st => FSL.val (pairT v tt).

Definition get (St:Type): K St St := 
  fun st => FSL.val (pairT st st).

Implicit Arguments get [].

(** The major change with respect to the usual monadic operators is [bind].
*)
Definition bind (A B St: Type) (m : K St A) (f : A -> K St B) : K St B :=
  fun st =>
  FSL.bind (m st)
    (fun p => let (st',a):=p in f a st').

Lemma bind_monotony :
 forall (A B St: Type) (sp1 sp2 : K St A) (f1 f2 : A -> K St B) (st : St),
 refInEnv st sp1 sp2 -> 
  (forall (st2 : St) (x : A), refInEnv st2 (f1 x) (f2 x)) -> 
  refInEnv st (bind sp1 f1) (bind sp2 f2).
Proof.
  unfold refInEnv, bind in |- *; intros.
  apply FSL.bind_monotony.
  auto with wprefines.
  intros p; case p; simpl in |- *;
  auto with wprefines.
Qed.

Hint Extern 0 (WP.refines _ _) => refine (FSL.bind_assoc_ref1 _ _ _): wprefines.
Hint Resolve refInEnv_refl bind_monotony: isl.

(** Here is a notation for IF-Then-Else *)
Notation "'If' pb 'Then' sp1 'Else' sp2" :=
  (bind pb (fun b => if b then sp1 else sp2))
  (at level 20, pb, sp1, sp2 at level 90): SSL_scope.

Lemma if_monotony :
 forall (A St: Type) (spb1 spb2 : K St bool) (sp1 sp2 sp3 sp4 : K St A) (st : St),
 refInEnv st spb1 spb2 -> 
 (forall st2 : St, refInEnv st2 sp1 sp2) -> 
 (forall st2 : St, refInEnv st2 sp3 sp4) -> 
 refInEnv st
   (If spb1 Then sp1 Else sp3)%SSL
   (If spb2 Then sp2 Else sp4)%SSL.
Proof.
  simpl in |- *; intros; apply bind_monotony; auto with isl.
  intros st0 x; case x; auto with isl.
Qed.

(** [bind] is associative. It is expressed by the two following lemmas *)
Lemma bind_assoc_ref1 :
 forall (A B C St: Type) (sp1 : K St A) (sp2 : A -> K St B) (sp3 : B -> K St C) (st : St),
 refInEnv st (bind (bind sp1 sp2) sp3) 
   (bind sp1 (fun a => bind (sp2 a) sp3)).
Proof.
  unfold bind, refInEnv in |- *.
  intros A B C St sp1 sp2 sp3 st.
  eapply WP.refines_trans.
  apply bind_assoc_ref1.
  apply FSL.bind_monotony; auto with wprefines.
  intros a; case a; case a; simpl; intros.
  intros; apply FSL.bind_monotony; auto with wprefines; 
  autorewrite with placeeq; auto with wprefines. 
Qed.

(* begin hide *)
Hint Extern 0 (WP.refines _ _) => refine (FSL.bind_assoc_ref2 _ _ _): wprefines.
(* end hide *)

Lemma bind_assoc_ref2 :
 forall (A B C St: Type) (sp1 : K St A) (sp2 : A -> K St B) (sp3 : B -> K St C) (st : St),
   refInEnv st (bind sp1 (fun a => bind (sp2 a) sp3))
               (bind (bind sp1 sp2) sp3).
Proof.
  unfold bind, refInEnv in |- *.
  intros A B C St sp1 sp2 sp3 st.
  eapply WP.refines_trans.
  2: apply bind_assoc_ref2.
  apply FSL.bind_monotony; auto with wprefines.
  intros a; case a; case a; simpl; intros.
  intros; apply FSL.bind_monotony; auto with wprefines; 
  autorewrite with placeeq; auto with wprefines. 
Qed.

Lemma bind_val_r_ref1:
  forall (A St: Type) (sp1 sp2: K St A) (sp3:A -> (K St A)) (st : St),
    (refInEnv st sp1 sp2)
      -> (forall (a:A) (st:St), (refInEnv st (sp3 a) (val a)))   
         ->  (refInEnv st (bind sp1 sp3) sp2).
Proof.
  intros A St sp1 sp2 sp3 st H1 H2; 
  apply refInEnv_trans with (sp2:=(bind sp1 (fun a => (val a)))).
  auto with isl.
  clear H2; generalize (refInEnv_from_wp H1); intros H2.
  apply refInEnv2wp; simpl.
  intros H3; eapply sdemon_weaken.
  apply H2; auto.
  intros b; case b; simpl; auto.
Qed. 

Lemma bind_val_r_ref2:
  forall (A St: Type) (sp1 sp2: K St A) (sp3:A -> (K St A)) (st : St),
    (refInEnv st sp2 sp1)
      -> (forall (a:A) (st:St), (refInEnv st (val a) (sp3 a)))   
         ->  (refInEnv st sp2 (bind sp1 sp3)).
Proof.
  intros A St sp1 sp2 sp3 st H1 H2; 
  apply refInEnv_trans with (sp2:=(bind sp1 (fun a => (val a)))).
  clear H2; generalize (refInEnv_from_wp H1); intros H2.
  apply refInEnv2wp; simpl.
  intros H3; eapply sdemon_weaken.
  apply H2.
  eauto with wipi.
  simpl; intros b H4.
  eapply angel_weaken.
  apply H4.
  intros b0 H5; subst.
  case b0; simpl; auto.  
  auto with isl.
Qed. 

Lemma bind_left_unassoc :
 forall (A B C St: Type) (sp1 : K St A) (sp2 : A -> K St B) (sp3 : B -> K St C) (sp4 : K St C) (st : St),
 refInEnv st (bind sp1 (fun a => bind (sp2 a) sp3)) sp4 -> 
 refInEnv st (bind (bind sp1 sp2) sp3) sp4.
Proof.
  intros; refine (refInEnv_trans (bind_assoc_ref1 _ _ _ _) _); auto.
Qed.

Lemma bind_left_assoc :
 forall (A B C St: Type) (sp1 : K St A) (sp2 : A -> K St B) (sp3 : B -> K St C) (sp4 : K St C) (st : St),
 refInEnv st (bind (bind sp1 sp2) sp3) sp4 -> 
 refInEnv st (bind sp1 (fun a => bind (sp2 a) sp3)) sp4.
Proof.
  intros;  refine (refInEnv_trans (bind_assoc_ref2 _ _ _ _) _); auto.
Qed.

Lemma bind_right_unassoc :
 forall (A B C St: Type) (sp1 : K St A) (sp2 : A -> K St B) (sp3 : B -> K St C) (sp4 : K St C) (st : St),
 refInEnv st sp4 (bind sp1 (fun a => bind (sp2 a) sp3)) -> 
 refInEnv st sp4 (bind (bind sp1 sp2) sp3).
Proof.
  intros; refine (refInEnv_trans _ (bind_assoc_ref2 _ _ _ _)); auto.
Qed.

Lemma bind_right_assoc :
 forall (A B C St: Type) (sp1 : K St A) (sp2 : A -> K St B) (sp3 : B -> K St C) (sp4 : K St C) (st : St),
 refInEnv st sp4 (bind (bind sp1 sp2) sp3) -> 
 refInEnv st sp4 (bind sp1 (fun a => bind (sp2 a) sp3)).
Proof.
  intros; refine (refInEnv_trans _ (bind_assoc_ref1 _ _ _ _)); auto.
Qed.

(** Now, let us see "Proof by step-by-step execution" lemmas 
*) 

Lemma bind_simpl_left :
 forall (A B St: Type) (sp1 : K St A) (sp2 : A -> K St B) (sp3 : K St B) (st : St),
 sdemon (sp1 st)
   (fun p => let (st',a):=p in refInEnv st (bind (set st') (fun _ => sp2 a)) sp3) 
    -> refInEnv st (bind sp1 sp2) sp3. 
Proof.
  simpl in |- *; unfold refInEnv, bind in |- *; intros.
  refine (FSL.bind_simpl_left _ _ _ _); auto.
  simpl in |- *; intros; eapply sdemon_weaken. eauto.
  simpl in |- *; intros p; case p; case p; simpl in |- *; auto with wprefines.
Qed.


Lemma bind_simpl_right :
 forall (A B St: Type) (sp1 : K St A) (sp2 : A -> K St B) (sp3 : K St B) (st : St),
 (pre (sp1 st) -> 
  angel (sp1 st)
    (fun p => let (st',a):=p in
         refInEnv st' (bind (set st) (fun _ => sp3)) (sp2 a)
     )) -> refInEnv st sp3 (bind sp1 sp2). 
Proof.
  simpl in |- *; unfold refInEnv, bind in |- *; intros.
  refine (FSL.bind_simpl_right _ _ _ _); auto.
  simpl in |- *; intros; eapply angel_weaken. eauto.
  simpl in |- *; intros p; case p; simpl in |- *; auto with wprefines.
Qed.

Lemma bind_simpl_both :
 forall (A B St: Type) (sp1 sp3 : K St A) (sp2 sp4 : A -> K St B) (st : St),
 refInEnv st sp1 sp3 -> 
  (pre (sp1 st) -> 
   sdemon (sp1 st)
     (fun p => let (st',a):=p in 
                 refInEnv st' (sp2 a) (sp4 a)))
   -> refInEnv st (bind sp1 sp2) (bind sp3 sp4).
Proof.
  simpl in |- *; unfold refInEnv, bind in |- *; intros.
  refine (FSL.bind_simpl_both _ _ _ _); auto.
  intros; eapply sdemon_weaken. eapply (H0 H1).
  simpl in |- *; intros p; case p; simpl in |- *; 
  intros; auto with wprefines.
Qed. 

(* begin hide *)
Lemma seq_left_assoc :
 forall (B C St: Type) (sp1 : K St unit) (sp2 : K St B) (sp3 : B -> K St C) (sp4 : K St C) (st : St),
 refInEnv st (bind (bind sp1 (fun _: unit => sp2)) sp3) sp4 -> 
 refInEnv st (bind sp1 (fun _: unit => bind sp2 sp3)) sp4.
Proof.
  intros; apply bind_left_assoc with (sp2:=(fun _:unit => sp2));
  auto.
Qed.

Lemma seq_right_assoc :
 forall (B C St: Type) (sp1 : K St unit) (sp2 : K St B) (sp3 : B -> K St C) (sp4 : K St C) (st : St),
 refInEnv st sp4 (bind (bind sp1 (fun _: unit => sp2)) sp3) -> 
 refInEnv st sp4 (bind sp1 (fun _: unit => bind sp2 sp3)).
Proof.
  intros; apply bind_right_assoc with (sp2:=(fun _:unit => sp2)); auto.
Qed.



Lemma bind_enter_left :
 forall (A B St: Type) (sp1 sp2:K St A) (sp3 : A -> K St B) (sp4: K St B) (st : St),
 refInEnv st sp1 sp2 -> 
 refInEnv st (bind sp2 sp3) sp4 -> refInEnv st (bind sp1 sp3) sp4.
Proof.
  intros; eapply refInEnv_trans.
  2: eauto.
  auto with isl.
Qed.

Lemma bind_enter_right :
 forall (A B St: Type) (sp1 sp2:K St A) (sp3 : A -> K St B) (sp4: K St B) (st : St),
 refInEnv st sp2 sp1 -> 
 refInEnv st sp4 (bind sp2 sp3) -> refInEnv st sp4 (bind sp1 sp3).
Proof.
  intros; eapply refInEnv_trans. 
  eauto.
  auto with isl.
Qed.

(** Here are a few lemmas to simplify "set". There are symmetric lemmas for the right side of refines relation.
 *)

Lemma set_skip_left :
 forall (A St: Type) (sp1 sp2 : K St A) (st x:St),
  x=st -> refInEnv st sp1 sp2 -> refInEnv st (bind (set x) (fun _ => sp1)) sp2.
Proof.
  intros A St sp1 sp2 st x H1 H; subst; apply refInEnv2wp.
   generalize (refInEnv_from_wp H); simpl in |- *; auto.
Qed. 

(* begin hide *)
Lemma set_skip_right :
 forall (A St: Type) (sp1 sp2 : K St A) (st x:St),
  x=st -> refInEnv st sp1 sp2 -> refInEnv st sp1 (bind (set x) (fun _ => sp2)).
Proof.
  intros A St sp1 sp2 st x H1 H; subst; apply refInEnv2wp.
   generalize (refInEnv_from_wp H); simpl in |- *;  auto.
Qed. 
(* end hide *)

Lemma set_swap_leftright:
 forall (A St: Type) (sp1 sp2 : K St A) (st x:St),
   refInEnv x sp1 (bind (set st) (fun _ => sp2)) 
     -> refInEnv st (bind (set x) (fun _ => sp1)) sp2.
Proof.
   intros A St sp1 sp2 st x H; apply refInEnv_trans 
    with (sp2:=(bind (set x) (fun _ => (bind (set st) (fun _ => sp2)))));
   auto with isl.
   apply seq_left_assoc.
   apply bind_simpl_left; simpl.
   apply set_skip_left; [ simpl; reflexivity | idtac ]. 
   auto with isl.
Qed.

(* begin hide *)
Lemma set_swap_rightleft:
 forall (A St: Type) (sp1 sp2 : K St A) (st x:St),
  refInEnv x (bind (set st) (fun _ => sp1)) sp2 
     -> refInEnv st sp1 (bind (set x) (fun _ => sp2)).
Proof.
  intros A St sp1 sp2 st x H; apply refInEnv_trans 
    with (sp2:=(bind (set x) (fun _ => (bind (set st) (fun _ => sp1)))));
   auto with isl.
   apply seq_right_assoc.
   apply bind_simpl_right; simpl.
   intros; apply set_skip_left; [ simpl; reflexivity | idtac ]. 
   auto with isl.
Qed.
(* end hide *)

Lemma set_simplify_left :
 forall (A St: Type) (sp1 sp2 : K St A) (st : St) (x:St),
 refInEnv st (bind (set x) (fun _ => sp1)) sp1 -> 
 refInEnv st sp1 sp2 -> 
 refInEnv st (bind (set x) (fun _ => sp1)) sp2.
Proof.
  intros; apply refInEnv_trans with (sp2 := sp1); auto.
Qed.

(* begin hide *)
Lemma set_simplify_right :
 forall (A St: Type) (sp1 sp2 : K St A) (st : St) (x:St),
 refInEnv st sp2 (bind (set x) (fun _ => sp2)) -> 
 refInEnv st sp1 sp2 -> 
 refInEnv st sp1 (bind (set x) (fun _ => sp2)).
Proof.
  intros; apply refInEnv_trans with (sp2 := sp2); auto.
Qed.


Lemma seq_monotony_left :
 forall (B St: Type) (sp1 : K St unit) (sp2 sp3 sp4 : K St B) (st : St),
 (forall st2 : St, refInEnv st2 sp2 sp3) -> 
 refInEnv st (bind sp1 (fun _ : unit => sp3)) sp4 -> 
 refInEnv st (bind sp1 (fun _ : unit => sp2)) sp4.
Proof.
  intros; apply refInEnv_trans with (sp2 := bind sp1 (fun _ : unit => sp3));
  auto with isl.  
Qed.

Lemma seq_monotony_right :
 forall (B St: Type) (sp1 : K St unit) (sp2 sp3 sp4 : K St B) (st : St),
 (forall st2 : St, refInEnv st2 sp2 sp3) -> 
 refInEnv st sp4 (bind sp1 (fun _ : unit => sp2)) -> 
 refInEnv st sp4 (bind sp1 (fun _ : unit => sp3)).
Proof.
  intros; apply refInEnv_trans with (sp2 := bind sp1 (fun _ : unit => sp2));
  auto with isl.  
Qed.


Hint Resolve bind_val_r_ref1 bind_val_r_ref2: isl.


(* end hide *)

(** Here is the lift operator from state transformer monad. It allows to define [any], [require], ...
directly from FSL. *)
Definition lift (A St: Type) (wp : WP.DS A) : K St A :=
  fun st => FSL.bind wp (fun a => FSL.val (pairT st a)). 

(** [lift] and [set] commutate. *)
Lemma set_permut_lift_left:
  forall (A B St: Type) (wp: WP.DS A) (sp1: A -> (K St B)) (sp2: (K St B)) (st x:St),
    (refInEnv st
      (bind (lift wp) (fun a => (bind (set x) (fun _ => (sp1 a))))) sp2)
   ->    
    (refInEnv st 
      (bind (set x) (fun _ => (bind (lift wp) sp1))) sp2).
Proof.
  intros; eapply refInEnv_trans. 
  2:eauto.
 apply refInEnv2wp; simpl.
 intros H1.
 generalize (refInEnv_from_wp (refInEnv_refl x (bind (lift wp) sp1))).
 simpl.
 auto.
Qed.

Definition abort (St A: Type) : K St A := lift (FSL.abort A).

Definition magic (St A: Type) : K St A := lift (FSL.magic A).

Definition require (St:Type) (p : Prop) : K St unit := lift (FSL.require p).

Definition ensure (St:Type) (p : Prop) : K St unit := lift (FSL.ensure p).

Definition any (St A: Type) : K St A := lift (FSL.any A).

Lemma any_permut_bind :
 forall (A B C St: Type) (sp : K St A) (f : A -> B -> K St C) (st : St),
 pre (sp st) \/ (NonEmpty B) -> 
 refInEnv st (bind sp (fun a => bind (any B) (fun b => f a b)))
             (bind (any B) (fun b => bind sp (fun a => f a b))).
Proof.
  simpl in |- *;
   unfold abort, any, refInEnv, bind, lift in |- *;
   intros A B C St sp f st H; apply WP.refines_fold; 
   simpl in |- *.
  intros H1; apply sdemon_fold.
  case H; auto.
  intros H2; case H2; intros x; generalize (H1 x); eauto with wipi.
  intros p; case p; simpl in |- *; clear p; simpl in |- *.
  intros st' a H2 b. lapply
  (WP.refines_unfold (refInEnv_refl st (bind sp (fun a : A => f a b)))).
  intros H3;
   lapply (sdemon_unfold (sp st) _ (pairT st' a) H3); 
   auto.
  simpl in |- *; intros; eapply sdemon_weaken. eauto. simpl in |- *; eauto.
  simpl in |- *; auto.
Qed.

(* begin hide *)

Definition or (St A:Type) (sp1 sp2:(K St A)) : (K St A)
 := (fun st => (or (sp1 st) (sp2 st))).

Lemma refInEnv_or_intro_l: forall (St A:Type) (sp1 sp2 sp:(K St A)) (st:St),
  (refInEnv st sp sp1) -> (refInEnv st sp (or sp1 sp2)).
Proof.
  unfold refInEnv, or; auto with wprefines.
Qed.

Lemma refInEnv_or_intro_r: forall (St A:Type) (sp1 sp2 sp:(K St A)) (st:St),
  (refInEnv st sp sp2) -> (refInEnv st sp (or sp1 sp2)).
Proof.
  unfold refInEnv, or; auto with wprefines.
Qed.

Lemma refInEnv_or_elim: forall (St A:Type) (sp1 sp2 sp:(K St A)) (st:St),
  (refInEnv st sp1 sp) -> (refInEnv st sp2 sp) -> 
    (refInEnv st (or sp1 sp2) sp).
Proof.
  unfold refInEnv, or; auto with wprefines.
Qed.

Hint Resolve refInEnv_or_intro_l refInEnv_or_intro_r refInEnv_or_elim
 set_skip_left set_skip_right: isl.

Definition or_x  (St A : Type) (sp1 sp2:K St A) :=
  (bind (any bool) (fun b => if b then sp1 else sp2)).
                                                                                                        
Lemma refInEnv_or_x: forall (St A : Type) (sp1 sp2:K St A) (st:St),
 (refInEnv st (or sp1 sp2) (or_x sp1 sp2)).
Proof.
  intros; unfold or_x; apply refInEnv_or_elim.
  apply bind_simpl_right; simpl.
  intros H; clear H.
  constructor 1 with (x:=true); auto with isl.
  apply bind_simpl_right; simpl.
  intros H; clear H.
  constructor 1 with (x:=false); auto with isl.
Qed.
                                                                                                        
Lemma refines_x_or: forall (St A : Type) (sp1 sp2:K St A) (st:St),
 (refInEnv st (or_x sp1 sp2) (or sp1 sp2)).
Proof.
  intros; unfold or_x; apply bind_simpl_left; simpl.
  intros b; case b; auto with isl.
Qed.

Hint Resolve refInEnv_or_x refines_x_or: isl.

Lemma refInEnv_or_bind_assoc2_1:
 forall (St A B:Type) (sp: K St A) (sp1 sp2: A -> (K St B)) (st:St),
   (refInEnv st (bind sp (fun a => (or (sp1 a) (sp2 a)))) (or (bind sp sp1) (bind sp sp2))).
Proof.
  intros; apply refInEnv_trans with (sp2:=(bind sp (fun a => (or_x (sp1 a) (sp2 a)))));
  auto with isl.
  unfold or_x; eapply refInEnv_trans.
  apply any_permut_bind with
    (f:=(fun a (b:bool) => if b then sp1 a else sp2 a)).
  auto with nonempty.
  apply bind_simpl_left; simpl ; intros b; case b; 
  auto with isl.
Qed.

Lemma if_or_ref1:
  forall (St A:Type) (spb: K St bool) (sp1 sp2: K St A) (st: St),
     (refInEnv st (If spb Then sp1 Else sp2)%SSL
                 (or (bind spb (fun b => bind (ensure (b=true)) (fun _ => sp1)))
                     (bind spb (fun b => bind (ensure (b=false)) (fun _ => sp2))))).
Proof.
  intros.
  apply refInEnv_trans with 
    (sp2:=(bind spb (fun b => (or (bind (ensure (b=true)) (fun _ => sp1))
                                  (bind (ensure (b=false)) (fun _ => sp2)))))).
  apply bind_monotony; simpl; auto with isl.
  intros st0 b; case b; simpl.
  apply refInEnv_or_intro_l; apply bind_simpl_right; simpl;
  intuition auto with isl.
  apply refInEnv_or_intro_r; apply bind_simpl_right; simpl;
  intuition auto with isl.
  refine (refInEnv_or_bind_assoc2_1 _ _ _ _).
Qed.

Lemma if_elim:
 forall (St A:Type) (spb: K St bool) (sp1 sp2 sp: K St A) (st: St),
   (refInEnv st (bind spb (fun b => (bind (ensure (b=true)) (fun _ => sp1)))) sp) -> 
   (refInEnv st (bind spb (fun b => (bind (ensure (b=false)) (fun _ => sp2)))) sp) -> 
   (refInEnv st (If spb Then sp1 Else sp2)%SSL sp).
Proof.
  intros; eapply refInEnv_trans.
  apply if_or_ref1.
  auto with isl.
Qed.

Definition sync (St A B:Type) (sp: A ->(K St B)) : (K St B) 
 := fun st => (FSL.sync (fun x:A => (sp x st))).
 
Lemma sync_elim: forall (A B St:Type) (sp1: A -> (K St B)) (sp2:(K St B)) (a:A) (st:St),
  (refInEnv st (sp1 a) sp2) -> (refInEnv st (sync sp1) sp2).
Proof.
  unfold refInEnv, sync; intros; eapply sync_elim; eauto.
Qed.

Lemma sync_intro: forall (A B St:Type) (sp1:(K St B)) (sp2: A -> (K St B)) (st:St),
  (forall (a:A), (refInEnv st sp1 (sp2 a))) -> (refInEnv st sp1 (sync sp2)).
Proof.
  unfold refInEnv, sync; auto with wprefines2.
Qed.

Hint Resolve sync_elim sync_intro: isl2.
                                                                                
Lemma sync_monotony :
 forall (St A B: Type) (s1 s2: A -> K St B) (st:St),
 (forall (a:A), refInEnv st (s1 a) (s2 a)) ->
    refInEnv st (sync s1) (sync s2).
Proof.
  intros; eauto with isl2.
Qed.
                                                                                
Hint Resolve sync_monotony: isl2.
 
Lemma bind_sync_distr_r_1:
  forall (A B C St:Type) (s1: A->K St B) (s2: B -> (K St C)) (st:St),
   (refInEnv st (bind (sync s1) s2) (sync (fun x => bind (s1 x) s2))).
Proof.
 intros; apply sync_intro.
 intros; apply bind_monotony; eauto with isl isl2.
Qed.
                                                                                
Lemma bind_sync_distr_l_1:
 forall (A B C St:Type) (s1: K St A) (s2: A -> B -> (K St C)) (st:St),
   (refInEnv st (bind s1 (fun a => sync (s2 a))) 
                (sync (fun b => (bind s1 (fun a => s2 a b))))).
Proof.
 intros; apply sync_intro.
 intros; apply bind_monotony; eauto with isl isl2.
Qed.

 
Unset Implicit Arguments.
(* end hide *)
