(* begin hide *)
Set Implicit Arguments.
Require Export SSLfriendly.

(* end hide *)

(** ** Module [AbsPost]: Hoare Specifications

*** Hoare specifications.

With respect to section %\ref{subsec:extdsl}%, I now consider that a
such Hoare specification may work only on a place of the current
state. 
*)

Section ABSPOST.

(** I assume [A] is the type of the returned result. [q] is the postcondition relating the final state and 
the result.
*)

Variable A St: Type.

Variable q: St -> A -> Prop.

(**  The abstract postcondition of [v] and [q] is defined as follows:
*)
Definition absPost: (K St A)
  := (Bind result <- (any A) In
         (Bind new_st <- (any St) In 
                   (set new_st) -;  
                   (ensure (q new_st result))) -;
                 (val result))%SSL.

Variable old_st: St.

Lemma abspost2wp:
  forall (pt:(K St A)),
  (sdemon (pt old_st)
   (fun p => let (new_st,a):=p in q new_st a))      
    -> (refInEnv old_st pt absPost).
Proof.
  intros sp H1; unfold absPost.
  apply refInEnv2wp; simpl.
  intros; eapply sdemon_weaken.
  apply H1.
  intros a; case a; simpl.
  intros st0 a0 H3. 
  constructor 1 with (x:=a0).
  constructor 1 with (x:=st0).
  intuition auto.
Qed.

(** This lemma corresponds to a ``refinement in a context'' rule. In a
goal [(refInEnv old_st sp1 sp2)] such [sp1] satifies a postcondition
[q], it is possible to strengthen [sp1] with a clause ensuring that
the result and the final state of [sp1] satisfy [q].  
*)
Lemma intro_abspost:
  forall (sp1 sp2:(K St A)),
   (refInEnv old_st sp1 absPost)
      -> (refInEnv old_st (Bind a <- sp1 In
                          (Bind new_st <- get St In 
                           ensure (q new_st a)) -; (val a))
                      sp2)%SSL       
  -> (refInEnv old_st sp1 sp2)%SSL.
Proof.
  intros sp1 sp2 H1 H. eapply refInEnv_trans.
  2:apply H.
  clear H; unfold refInEnv.
  apply refines_fold.
  simpl; intros H3.
  lapply (refines_unfold H1).
  intros H4; apply sdemon_fold.
  eapply sdemon_weaken.
  apply H4.
  auto.
  intros a H5.
  apply angel_fold with (a:=a); intuition auto.
  case (sdemon_unfold _ _ a H4 H5).
  simpl.
  intros x H6; decompose [ex Logic.and] H6.
  subst. simpl.  intuition auto.
  simpl; intuition auto. 
Qed.

(** This rule is a generalization of the preceding one.
*)
Lemma intro_bind_abspost:
  forall (B:Type) (sp1:(K St A)) (sp2: A -> (K St B)) (sp3: (K St B)),
    (refInEnv old_st sp1 absPost)
        -> (refInEnv old_st (Bind a <- sp1 In
                             (Bind new_st <- get St In
                               ensure (q new_st a)) -; 
                             (sp2 a))
                       sp3)%SSL
  -> (refInEnv old_st (bind sp1 sp2) sp3)%SSL.
Proof.
  intros B sp1 sp2 sp3 H1 H3.
  eapply refInEnv_trans.
  2:apply H3.
  apply refInEnv_trans with
   (sp2:=(Bind b <- (Bind a <- sp1 In (Bind new_st <- get St In 
                           ensure (q new_st a)) -; (val a)) In (sp2 b))%SSL).
   SSL_monotony.
   eapply intro_abspost.
   apply H1.
   auto with isl.
   SSL_unassoc_left.
   SSL_monotony.
   intros; SSL_simplify_left; simpl.
   intros; SSL_simplify_right; simpl.
   intuition auto.
   auto with isl.
Qed.

End ABSPOST.


(** *** Rules for Hoare specifications.
*)

Lemma abspost_monotony: forall (A St:Type) (q1 q2: St -> A -> Prop) (st:St),
   (forall (st2:St) (a:A), (q1 st2 a) -> (q2 st2 a)) -> 
    (refInEnv st (absPost q1) (absPost q2)).
Proof.
  intros; apply abspost2wp. 
  simpl.
  auto.
Qed.

Lemma abspost_split: forall (A St:Type) (q1 q2: St -> A -> Prop) 
  (st:St) (sp:(K St A)),
   (refInEnv st sp (absPost q1))
     -> (refInEnv st sp (absPost q2))
      ->
    (refInEnv st sp (absPost (fun (stf:St) (r:A) => (q1 stf r)  /\ (q2 stf r)))).
Proof.
  intros A St q1 q2 st sp H1 H2.
  eapply intro_abspost.
  apply H2.
  SSL_enter_left.
  apply H1.
  intros; eapply abspost2wp. 
  simpl.
  intuition auto.
Qed.

Lemma abspost_splitforall: forall (A B St:Type) (p: B -> Prop) (q: B -> St -> A -> Prop) 
  (st:St) (sp:(K St A)),
   (exists x:B, (p x)) -> 
   (forall (x:B), (p x) -> (refInEnv st sp (absPost (q x)))) ->
    (refInEnv st sp (absPost (fun (stf:St) (r:A) => forall (x:B), (p x) -> (q x stf r)))).
Proof.
  intros.
  apply abspost2wp. 
  case H; clear H.
  intros x H; apply sdemon_fold.  
  lapply (refines_unfold (H0 x H)); simpl; intros; eauto with wipi.
  intros a; case a; clear a.
  intros s a H1 x0 H2; intros.
  lapply (refines_unfold (H0 x0 H2)); simpl; auto.
  intros H3; lapply (sdemon_unfold _ _ (pairT s a) H3); auto.
  clear H3; intros H3.
  decompose [ex Logic.and] H3.
  inversion H6. 
  subst; auto.
Qed.

Unset Implicit Arguments.
(* end hide *)
