(* The game of Nim in the state DSM *)

Set Implicit Arguments.

Require Export SSLWhile.
Require Export Bool.

Definition eval_ (St A:Type) (f: St -> A) : K St A := 
  bind (get _) (fun st => val (f st)).
Implicit Arguments eval_ [St A].

Definition require_ (St:Type) (p: St -> Prop): K St unit
 := bind (get _) (fun st => require (p st)).
Implicit Arguments require_ [St].

Definition ensure_ (St:Type) (p: St -> Prop): K St unit
 := bind (get _) (fun st => ensure (p st)).
Implicit Arguments ensure_ [St].

Lemma pair_decompose: forall (A B:Type) (a1 a2:A) (b1 b2:B), a1=a2 -> b1=b2 -> (a1,b1)=(a2,b2).
Proof.
  intros A B a1 a2 b1 b2 H1 H2; rewrite H1; rewrite H2; auto.
Qed.

Hint Resolve pair_decompose: isl.

(*********************************************************)
(*          Parametric Module "TwoPlayersGame"           *)
(*********************************************************)

Module Type GameDataM.

Parameter GameData: Type.    (* see type "GameSt" below *)

End GameDataM.


Module TwoPlayersGame (GD: GameDataM).

Export GD.

(* Type of game states *)
Record GameSt : Type := mkGameSt {
  player:  bool ;       (* name of the next player        *)
  gdata:   GameData     (* data of the game *)
}.

Definition dualSt (st:GameSt) := (mkGameSt (negb (player st)) (gdata st)). 

Lemma dualSt_involutive: forall st, dualSt (dualSt st) = st.
Proof.
  intros st; case st; unfold dualSt; simpl; intros b gd.
  rewrite negb_involutive.
  auto.
Qed.

Hint Rewrite dualSt_involutive negb_involutive: game.

Definition change_player: K GameSt unit 
 := bind (get _) (fun st => set (dualSt st)).

(* execution of a two-player game *)
Definition game (not_gameover: GameSt -> bool) (pt pf: K GameSt unit): K GameSt unit  :=
    while (eval_ not_gameover)
          ((If eval_ player Then pt Else pf) -;
           change_player).

Lemma game_monotonic:
  forall (cond1 cond2: GameSt -> bool) (pt1 pt2 pf1 pf2: K GameSt unit),
    (forall st, (cond1 st)=(cond2 st))
   -> (forall st, (player st)=true -> (cond1 st)=true -> [< st >] pt1 >==> pt2)
   ->  (forall st, (player st)=false -> (cond1 st)=true -> [< st >] pf1 >==> pf2)
  -> forall st, [< st >] (game cond1 pt1 pf1) >==> (game cond2 pt2 pf2).
Proof.
  intros; unfold game.
  intros; apply while_monotonic_prec; auto with isl.
  intros; SSL_monotony.
  intros; SSL_simplify_all; auto with isl.
  intros; SSL_simplify_both.
  intros; SSL_simplify_all; auto with isl.
  intros; SSL_simplify_both.
  case st0; intros b; case b; simpl; intros; SSL_monotony;
  SSL_simplify_both; auto.
Qed.

(* a generalization of the previous lemma *)
Theorem game_monotonic_with_invariant:   forall cond1 cond2 pt1 pf1 pt2 pf2 (inv: GameSt -> Prop),
  (forall st, (inv st) -> (cond1 st)=(cond2 st)) 
 -> (forall st, (player st)=true -> (cond1 st)=true -> (inv st) ->
       [< st >] pt1 >==> (pt2 -; ensure_ (fun st => inv (dualSt st))))
 -> (forall st, (player st)=false -> (cond1 st)=true -> (inv st) ->
       [< st >] pf1 >==> (pf2 -; ensure_ (fun st => inv (dualSt st))))
 -> forall st, [< st >] (game cond1 pt1 pf1) >==> (require_ (fun st => inv st) -; game cond2 pt2 pf2).
Proof.
  unfold game, dualSt; intros cond1 cond2 pt1 pf1 pt2 pf2 inv Hcond Hpt Hpf.
  apply while_morph with (G:= fun (k:K GameSt unit) => (require_ (fun st => inv st) -; k)).
  intros; SSL_monotony.
  intros; SSL_simplify_right.
  intuition; auto with isl.
  intros; SSL_monotony.
  intros; SSL_simplify_all; auto with isl.
  intros k st; case st; clear st; simpl; intros pl gd; SSL_simplify_right.
  intros H; case H; clear H; intros H H0; clear H0.
  SSL_simplify_both.
  intros; SSL_simplify_all; auto with isl.
  intros D1; SSL_simplify_both.
  intros D2 X; clear D1 D2; SSL_unassoc_left.
  SSL_unassoc_left.
  SSL_unassoc_right.
  SSL_unassoc_right. 
  SSL_simplify_both.
  intros.
  eapply bind_enter_left with (sp2:=(if pl then pt2 else pf2) -; ensure_ (fun st => inv (mkGameSt (negb (player st)) (gdata st)))).
  generalize H X; case pl; simpl; clear H X pl; intuition.
  SSL_unassoc_left. 
  SSL_monotony.
  intros st2; case st2; clear st2; intros pl' gd'; intros; SSL_simplify_left.
  intuition subst.
  SSL_simplify_left.
  SSL_simplify_both.
  intros; SSL_simplify_left.
  intuition.
Qed.

Lemma two_change_player_refines_skip: forall st, [< st >] (change_player -; change_player) >==> skip.
Proof.
  intros; SSL_simplify_all.
  case st; intros b; case b; simpl; auto.
Qed.

Lemma two_change_player_refined_skip: forall st, [< st >] skip >==> (change_player -; change_player).
Proof.
  intros; SSL_simplify_all.
  case st; intros b; case b; simpl; auto.
Qed.

Hint Resolve two_change_player_refines_skip two_change_player_refined_skip: game.

Definition dualPlayer (p: K GameSt unit) :=
  (change_player -; p -; change_player).

Lemma dualPlayer_monotonic: 
  forall (p1 p2: K GameSt unit) st, 
    ([< dualSt st >] p1 >==> p2)
    -> ([< st >] (dualPlayer p1) >==> (dualPlayer p2)).
Proof.
  intros; unfold dualPlayer; SSL_simplify_both.
Qed.

Lemma dualPlayer_invol_1: 
  forall (p: K GameSt unit) st, 
    [< st >] (dualPlayer (dualPlayer p)) >==> p.
Proof.
  intros; unfold dualPlayer.
  eapply seq_monotony_left.
  intros; SSL_unassoc_left.
  auto with isl.
  SSL_assoc_left.
  SSL_enter_left.
  apply two_change_player_refines_skip.  
  SSL_simplify_left.
  SSL_unassoc_left.
  eapply seq_monotony_left.
  intros; apply two_change_player_refines_skip.   
  apply bind_val_r_ref1; auto with isl;
  intros a; case a; simpl; auto with isl.
Qed.

Lemma dualPlayer_invol_2: 
  forall (p: K GameSt unit) st, 
    [< st >] p >==> (dualPlayer (dualPlayer p)).
Proof.
  intros; unfold dualPlayer.
  eapply seq_monotony_right.
  intros; SSL_unassoc_right.
  auto with isl.
  SSL_assoc_right.
  SSL_enter_right.
  apply two_change_player_refined_skip.  
  SSL_simplify_right.
  intros; SSL_unassoc_right.
  eapply seq_monotony_right.
  intros; apply two_change_player_refined_skip.   
  apply bind_val_r_ref2; auto with isl;
  intros a; case a; simpl; auto with isl.
Qed.

Hint Resolve dualPlayer_invol_2 dualPlayer_invol_1 dualPlayer_monotonic: game.

Lemma dualPlayer_inversion: 
  forall (p1 p2: K GameSt unit) st, 
    ([< dualSt st >] (dualPlayer p1) >==> (dualPlayer p2))
    -> ([< st >] p1 >==> p2).
Proof.
  intros; SSL_trans.
  apply dualPlayer_invol_2.
  SSL_trans.
  2: apply dualPlayer_invol_1.
  auto with game.
Qed.

Lemma dualPlayer_left_right: 
  forall (p1 p2: K GameSt unit) st, 
    ([< dualSt st >] p1 >==> (dualPlayer p2))
      -> ([< st >] (dualPlayer p1) >==> p2).
Proof.
  intros; apply dualPlayer_inversion.
  SSL_trans.
  2: apply H.
  auto with game.
Qed.

Lemma dualPlayer_right_left: 
  forall (p1 p2: K GameSt unit) st, 
    ([< dualSt st >] (dualPlayer p1) >==> p2)
      -> ([< st >] p1 >==> (dualPlayer p2)).
Proof.
  intros; apply dualPlayer_inversion.
  SSL_trans.
  eauto.
  auto with game.
Qed.

Lemma game_symmetry:
  forall cond pt pf st, 
   [< st >] (game cond pt pf) >==> (dualPlayer (game (fun st => cond (dualSt st)) (dualPlayer pf) (dualPlayer pt))).
Proof.
  intros cond pt pf; unfold dualSt, game.
  apply while_morph with (G:=dualPlayer).
  intros k1 k2 H st; case st; simpl; auto with game.
  intros st; case st; simpl; intros b gd; SSL_simplify_all.
  case b; simpl; intuition. 
  intros k st; case st; simpl; intros b gd.
  SSL_simplify_left.
  SSL_simplify_left.
  intros H; SSL_simplify_left.
  SSL_unassoc_left.
  SSL_unassoc_left.
  SSL_simplify_left.
  apply dualPlayer_right_left.
  SSL_simplify_right; simpl.
  intros; SSL_simplify_right; simpl.
  constructor 1.
  generalize H; case b; simpl; auto.
  SSL_simplify_left.
  SSL_unassoc_right.
  SSL_unassoc_right.
  SSL_simplify_right.
  unfold dualPlayer.
  intros; eapply seq_monotony_left.
  intros; SSL_unassoc_left.
  intros; eapply seq_monotony_left.
  intros; SSL_unassoc_left.
  auto with isl.
  SSL_assoc_left; auto with isl.
  SSL_assoc_left. 
  SSL_monotony.
  case b; simpl; auto with isl.
  intros; SSL_unassoc_left.
  SSL_monotony.
  intros; SSL_unassoc_left.
  intros; eapply seq_monotony_left.
  intros; apply two_change_player_refines_skip.
  apply bind_val_r_ref1; auto with isl;
  intros a; case a; simpl; auto with isl.
Qed.

Definition absPlayer (gameRule: GameSt -> GameSt -> Prop): K GameSt unit :=
    bind (get _) (fun st => absPost (fun st' _ => gameRule st st' /\ (player st')=(player st))).

Lemma absPlayer_autodual:
  forall gameRule st, [< st >] (dualPlayer (absPlayer (fun st st' => gameRule (dualSt st) (dualSt st')))) >==> (absPlayer gameRule).
Proof.
  intros; unfold absPlayer.
  SSL_simplify_right.
  intuition; apply abspost2wp; simpl.
  intros xx a0; case st; case a0; simpl; intuition (subst; autorewrite with game in * |- *; auto).
Qed.

Lemma absPlayer_monotonic:
  forall (rule1 rule2: GameSt -> GameSt -> Prop) st, 
    (forall st st', (rule1 st st') -> (rule2 st st')) 
    -> [< st >] (absPlayer rule1) >==> (absPlayer rule2).
Proof.
  intros; unfold absPlayer.
  SSL_simplify_right.
  intuition; apply abspost2wp; simpl.
  intuition.
Qed.

Definition invPlayer (p: K GameSt unit) (inv: GameSt -> Prop) (gameRule: GameSt -> GameSt -> Prop) :=
    require_ (fun st => inv st) -; p -; ensure_ (fun st => forall st', (gameRule (dualSt st) (dualSt st')) -> (player st')=(player st) -> inv st').

Lemma invPlayer_monotonic: 
  forall p1 p2 (inv: GameSt -> Prop) rule st,
    ((inv st) ->  [< st >] p1 >==> p2) 
      -> [< st >] (invPlayer p1 inv rule) >==> (invPlayer p2 inv rule).
Proof.
  intros p1 p2 inv rule st H; unfold invPlayer.
  SSL_simplify_both.
  intuition; intros; SSL_monotony.
Qed.

Lemma invPlayer_trivial_trivok: 
  forall p rule st, [< st >] p >==>  (invPlayer p (fun _ => True) rule).
Proof.
  intros p rule st; unfold invPlayer.
  SSL_simplify_right.
  intuition; intros.
  SSL_trans.
  eapply bind_val_r_ref2 with (sp3:=fun _:unit => val (St:=GameSt) tt).
  auto with isl.
  intros a; case a; simpl; auto with isl.
  SSL_monotony.
  intros; SSL_Trivial.
Qed.

Definition gameInv (invpt invpf: GameSt -> Prop) (rule: GameSt -> GameSt -> Prop) (st: GameSt) 
  :=  if (player st) then (invpt st) /\ forall st', (rule st (dualSt st')) -> (player st')=false -> (invpf st') 
                     else (invpf st) /\ forall st', (rule st (dualSt st')) -> (player st')=true  -> (invpt st').

(* a decomposition (e.g. "rely-guarantee") theorem: useful to study abstract strategies and their implementations *)
Theorem game_PlayerRefinement:   forall cond pt1 pf1 pt2 pf2 invpt invpf rule,
      (forall st, (player st)=true  -> (cond st)=true ->  [< st >] pt1 >==> invPlayer pt1 invpt rule)
   -> (forall st, (player st)=false -> (cond st)=true ->  [< st >] pf1 >==> invPlayer pf1 invpf rule)
   -> (forall st, (player st)=true  -> (cond st)=true -> (invpt st) -> [< st >] pt1 >==> pt2)
   -> (forall st, (player st)=false -> (cond st)=true -> (invpf st) -> [< st >] pf1 >==> pf2)
   -> (forall st, (player st)=true  -> (cond st)=true -> (invpt st) -> [< st >] pt2 >==> absPlayer rule)
   -> (forall st, (player st)=false -> (cond st)=true -> (invpf st) -> [< st >] pf2 >==> absPlayer rule)
   ->  forall st, (gameInv invpt invpf rule st) 
       -> [< st >] (game cond pt1 pf1) >==> (game cond pt2 pf2).
Proof.
  intros cond pt1 pf1 pt2 pf2 invpt invpf rule H H0 H1 H2 H3 H4 st H5; SSL_trans; [
     apply game_monotonic_with_invariant with (cond2:=cond) (pt2:=pt2) (pf2:=pf2) (inv:=gameInv invpt invpf rule)
    | SSL_simplify_left; intuition];
  unfold gameInv, dualSt; intros st0; case st0; clear st0; simpl; intros b gd X; subst; intuition;
  (SSL_trans; [ (apply H || apply H0); simpl; auto | ]);
  unfold invPlayer; SSL_simplify_left;
  intuition; SSL_simplify_left;
  (SSL_enter_left; [ (apply H1 || apply H2); simpl; auto | ]);
  (eapply intro_bind_abspost; [ 
    SSL_trans; [ auto |]; simpl; auto; unfold absPlayer; SSL_simplify_left; auto with isl | ]);
  SSL_monotony;
  simpl; SSL_simplify_all; simpl;
  case st2; simpl; intuition (subst; simpl in * |- *); intuition (subst; simpl in * |- *);
  intuition.
Qed.

Hint Resolve absPlayer_autodual absPlayer_monotonic: game.

End TwoPlayersGame.

(*********************************************************)
(*             Parametric Module "NimGame"               *)
(*      a first "abstract" view of the game              *)
(*      without rule on memory sharing                   *)
(*********************************************************)

Require Export ZmodAux.

Lemma Zmod_discr1: forall (a b c:Z),
   c > 0 -> a - c < b < a -> b mod c <> a mod c.
Proof.
   intros a b c X H H1.
   absurd ((a-b) mod c = 0).
   rewrite <- Z_mod_unicity; omega.
   rewrite Z_mod_minus_mod; auto.
   rewrite H1.
   cutrewrite (a mod c - a mod c =0); try ring.
   unfold Zmod; simpl; auto.
Qed.

Open Scope SSL_scope.


Module Type NimDataM.

Parameter GameData: Type.
Parameter nb_matches: GameData -> Z.

End NimDataM.


Module NimGame(ND: NimDataM).

Module Export G:=TwoPlayersGame(ND).

Export ND.

Notation "'nbm' st" := (nb_matches (gdata st)) (at level 50).

Definition nimgame (pt pf: K GameSt unit)
 := game (fun st => Zlt_bool 0 (nbm st)) pt pf.

Definition nimRule (st st': GameSt): Prop :=
     let nm:=(nbm st) in
     let nm':=(nbm st') in
          0 <= nm'
      /\  nm-3 <= nm' <  nm.

Definition is_winner (p: bool) (st': GameSt): Prop 
 := (player st')=negb p.

Ltac st_split := 
  let x := fresh in
    intro x; case x; clear x; simpl.

Lemma nimgame_end: forall st, (nbm st) >=0 ->
        [< st >] (nimgame (absPlayer nimRule) (absPlayer nimRule)) 
                  >==> (absPost (fun st _ => nbm st=0)).
Proof.
   unfold nimgame, game; intros st H1. 
   SSL_trans.
   apply while_refines_while_abs with (inv:=(fun st => nbm st >= 0)).
   (* preservation of invariant *)
   unfold inv_preservation.
   unfold nimRule; st_split. intros b gd H. 
   SSL_simplify_left.
   zlt_case 0 (nb_matches gd).
   case b; intros; apply abspost2wp; simpl; intuition.
   intros; apply abspost2wp; simpl; intuition.
   (* cas final *)
   apply abspost2wp; simpl; intuition.
   generalize (Zlt_bool_ge _ _ H2).
   omega.
Qed.

Definition nimStrategy (st st': GameSt): Prop :=
     let nm:=(nbm st) in
     let r:=(nm mod 4) in
     if (Zeq_bool r 0) then
        nimRule st st'
     else
        (nbm st')=nm - r.

Definition expertPlayer: K GameSt unit := 
    absPlayer nimStrategy.

Lemma expertPlayer_isNimPlayer:
  forall st, 0<(nbm st) -> [< st >] expertPlayer >==> (absPlayer nimRule).
Proof.
  unfold absPlayer. st_split.
  intros b gd H.
  SSL_simplify_right; intros.
  apply abspost2wp; simpl.
  unfold nimStrategy; simpl.
  intros d; st_split; intros b0 gd0.
  zeq_case ((nb_matches gd) mod 4) 0; intuition subst.
  unfold nimRule; simpl.
  set (n:=(nb_matches gd)) in * |- *.
  rewrite H3; clear H3.
  generalize (Z_mod_le (a:=n) (b:=4)).
  generalize (Z_mod_lt n 4). 
  omega.
Qed.

Hint Resolve expertPlayer_isNimPlayer: nim.

Lemma expertPlayer_autodual:
  forall st, [< st >] (dualPlayer expertPlayer) >==> expertPlayer.
Proof.
  unfold expertPlayer; intros; SSL_trans.
  2: apply absPlayer_autodual.
  case st. intros; apply dualPlayer_monotonic.
  unfold dualSt; simpl; auto with game.
Qed.

Hint Resolve expertPlayer_autodual: nim.

Definition false_can_win : K GameSt unit
  := bind (get _) (fun st =>
        require ((nbm st) >= 0) -;
        absPost (fun st' _ =>  (Zeq_bool ((nbm st) mod 4) 0)=(player st) -> is_winner false st')).

Theorem expertPlayer_nimgame:
  forall st, [< st >] (nimgame (absPlayer nimRule) expertPlayer) >==> false_can_win.
Proof.
  unfold nimgame, game; apply while_refines_invspec.
  st_split.
  intros b gd.
   SSL_simplify_left.
   unfold false_can_win, is_winner; SSL_simplify_right; simpl.
   intros D; clear D; SSL_simplify_right; intuition.
   zlt_case 0 (nb_matches gd).
   (* 1 cas recursif *)
   intros;  SSL_unassoc_left.
   SSL_unassoc_left.
   SSL_simplify_left.
   case b; simpl.
   (* 1.1 absPlayer joue *)
   apply abspost2wp; simpl.
   unfold nimRule; intros a0; st_split.
   intros b' gd'; set (nm:=(nb_matches gd)) in *|- * ; generalize (nb_matches gd'); intros nm' ; 
   intuition subst.
   omega.
   apply H5; clear H5 a0 a; simpl.
   apply Zdiff_eq_bool.
   generalize (Zeq_cases (nm mod 4) 0).
   rewrite H7; simpl; clear H7.
   intros X1 X2.
   generalize (Zmod_discr1 (a:=nm) (b:=nm') (c:=4)); intuition.
   (* 1.2 expertPlayer joue *)
   eapply intro_bind_abspost.
   SSL_trans.
   apply expertPlayer_isNimPlayer; auto with zarith.
   unfold absPlayer.
   SSL_simplify_left.
   auto with isl.
   apply abspost2wp; simpl; unfold nimStrategy, nimRule; simpl.
   intros a0. st_split.
   intros b' gd'; set (nm:=(nb_matches gd)) in *|- * ; generalize (nb_matches gd'); intros nm'; 
   intuition subst.
   omega.
   apply H7; clear H7 a0.
   apply Zeq_eq_bool.
   generalize (Zeq_cases (nm mod 4) 0).
   rewrite H9 in * |- *; simpl; clear H9; intro. subst.
   cutrewrite ((nm - nm mod 4)=0 + ((nm/4) * 4)).
   rewrite Z_mod_plus; auto.
   omega.
   generalize (Z_div_mod_eq nm 4); omega.
  (* 2 cas final *)
   intros; cutrewrite ((nb_matches gd)=0); simpl.
   apply abspost2wp; vm_compute; auto.
   omega.
Qed.

Hint Resolve (* Zlt_lt_bool *) Zlt_bool_lt: zarith.

Lemma expertPlayer_as_player_true:
  forall st, (nbm st) >= 0 -> (Zeq_bool ((nbm st) mod 4) 0) = negb (player st) -> 
    [< st >] (nimgame expertPlayer (absPlayer nimRule)) >==> (absPost (fun st' _ => is_winner true st')).
Proof.
 st_split; intros b gd; simpl; intros.
  (SSL_trans; [  | apply dualPlayer_left_right; (SSL_trans; [ apply expertPlayer_nimgame | ]) ]).
  unfold nimgame; SSL_trans; [ apply game_symmetry | ] .
  apply dualPlayer_monotonic; apply game_monotonic; auto with nim zarith.
  intros; SSL_trans; [ | apply absPlayer_autodual ].
  apply dualPlayer_monotonic.
  unfold dualSt; simpl; auto with game.
  apply dualPlayer_right_left; simpl;
  apply abspost2wp; simpl.
  rewrite H0; simpl.
  constructor 1; auto.
  intros x; unfold is_winner, dualSt; st_split; simpl.
  intros b'; rewrite negb_involutive. 
  intuition subst; auto.
Qed.

End NimGame.

(*********************************************************)
(*             Parametric Module "NimWGame"              *)
(*      a refinement of NimGame                          *)
(*      with "write-private" data                        *)
(*********************************************************)

Module Type NimWDataM.

Parameter WDatat WDataf: Type.
(* WData1: write-private data of player true  *)
(* WData2: write-private data of player false *)

End NimWDataM.


Module NimWGame(NWD:NimWDataM).

Export NWD.

Module Export ND <: NimDataM.

Record _GameData: Type := mkGameData {
  nb_matches: Z;
  writet: WDatat ; (* write-private data of player true  *)
  writef: WDataf   (* write-private data of player false  *)
}.

Definition GameData := _GameData.

End ND.

Module Export NG:=NimGame(ND).

Notation "'wrt' st" := (writet (gdata st)) (at level 50).
Notation "'wrf' st" := (writef (gdata st)) (at level 50).

Definition WRule (gameRule: GameSt-> GameSt -> Prop) (st st': GameSt): Prop :=
  gameRule st st' /\ if (player st) then (wrf st')=(wrf st) else (wrt st')=(wrt st).
     
Definition absWPlayer (gameRule: GameSt -> GameSt -> Prop) :=
  absPlayer (WRule gameRule).

Lemma absWPlayer_absPlayer:
  forall gameRule st, [< st >] absWPlayer gameRule >==> absPlayer gameRule.
Proof.
  unfold absWPlayer, WRule; intros; SSL_Trivial.
Qed.

Lemma absWPlayer_decompose:
  forall p gameRule st,
    ([< st >] p >==> absPlayer gameRule)
    -> ([< st >] p >==>  absPost (fun st' _ => if (player st) then (wrf st')=(wrf st)
                                                              else (wrt st')=(wrt st)))
      -> ([< st >] p >==> absWPlayer gameRule).
Proof.
  unfold absWPlayer, WRule, absPlayer; intros p gameRule st H1 H2.
  SSL_simplify_right; intros.
  eapply abspost_split.
  eapply abspost_split; auto.
  SSL_trans; [ apply H1 | ].
  SSL_Trivial.
  SSL_trans; [ apply H1 | ].
  SSL_Trivial.
Qed.
 

Ltac st_split := 
  let x := fresh in
    intro x; case x; clear x; simpl ;
    let y:= fresh "player" in
    let gd:= fresh in
      intros y gd; case gd; simpl; generalize y; clear gd y.

Definition expertWPlayer: K GameSt unit := absWPlayer nimStrategy.

Lemma expertWPlayer_expertPlayer:
  forall st, [< st >] expertWPlayer >==> expertPlayer.
Proof.
  unfold expertWPlayer, expertPlayer.
  intros; refine (absWPlayer_absPlayer _ _).
Qed.

Lemma expertWPlayer_isNimWPlayer:
  forall st, 0 < (nbm st) -> [< st >] expertWPlayer >==> (absWPlayer nimRule).
Proof.
  st_split.
  intros b n st1 st2 H; refine (absWPlayer_decompose _ _); simpl.
  SSL_trans; [ apply expertWPlayer_expertPlayer | ].
  apply expertPlayer_isNimPlayer; simpl; auto.
  unfold expertWPlayer, absWPlayer, WRule, absPlayer.
  SSL_Trivial.
Qed.

Require Import Zwf.
Import WfT.

Definition Rgame (st1 st2: GameSt): Prop := Zwf 0 (nbm st1) (nbm st2).

Lemma Rgame_well_founded: (well_founded Rgame).
Proof.
   unfold Rgame.
   apply wf_inverse_image with (f:=fun st => (nbm st)).
   apply Zwf_well_founded.
Qed.

Hint Resolve Rgame_well_founded: nim.

Definition nimgamewf (pt pf: K GameSt unit) (invpt invpf: GameSt -> Prop) : K _ unit :=
       require (
          (forall st, (player st)=true  -> 0 < (nbm st) -> [< st >] pt >==> invPlayer pt invpt (WRule nimRule))
       /\ (forall st, (player st)=false -> 0 < (nbm st) -> [< st >] pf >==> invPlayer pf invpf (WRule nimRule))
       /\ (forall st, (player st)=true  -> 0 < (nbm st) -> (invpt st) -> [< st >] pt >==> absWPlayer nimRule)
       /\ (forall st, (player st)=false -> 0 < (nbm st) -> (invpf st) -> [< st >] pf >==> absWPlayer nimRule) )-;
       (whileWF (* cond *) (eval_ (fun st => Zlt_bool 0 (nbm st)))
                (* body *) ((If eval_ player Then pt Else pf) -; change_player)
                (* inv. *) (gameInv invpt invpf (WRule nimRule))
                (* var. *) Rgame).

Lemma nimgamewf_feasible: 
  forall pt pf (invpt invpf: GameSt -> Prop), 
      (forall st, (player st)=true  -> 0 < (nbm st) -> invpt st -> feasible pt st)
   -> (forall st, (player st)=false -> 0 < (nbm st) -> invpf st -> feasible pf st)
   ->  forall st, feasible (nimgamewf pt pf invpt invpf) st.
Proof.
  intros; unfold nimgamewf, feasible.
 SSL_unassoc_right.
 SSL_simplify_right.
 intros D.
 apply feasible_def.
 apply whileWF_feasible.
 unfold gameInv; clear D st.
 st_split; intros b nm w1 w2 H1; simpl.
 apply feasible_bind_simplify; simpl.
 intro D; clear D.
 zlt_case 0 nm.
 intros;  apply feasible_bind_split.
 apply feasible_bind_simplify; simpl.
 intros D; clear D. 
 generalize H1; clear H1; case b; simpl; intuition.
 intros; apply feasible_wp; simpl; auto.
 intros; apply feasible_wp; simpl; auto.
Qed.

Lemma nimgame_refines_nimgamewf: forall pt pf invpt invpf st,
      [< st >] (nimgame pt pf) >==> (nimgamewf pt pf invpt invpf).
Proof.
  unfold nimgamewf, nimgame, game, whileWF.
  intros; SSL_simplify_right.
  intuition.
  intros; SSL_simplify_right.
  intuition.
Qed.

Hint Resolve Zlt_bool_lt: zarith.

Lemma nimgamewf_refines_nimgame:
  forall pt pf invpt invpf,
          (forall st, (player st)=true  -> 0 < (nbm st) -> [< st >] pt >==> invPlayer pt invpt (WRule nimRule))
       -> (forall st, (player st)=false -> 0 < (nbm st) -> [< st >] pf >==> invPlayer pf invpf (WRule nimRule))
       -> (forall st, (player st)=true  -> 0 < (nbm st) -> (invpt st) -> [< st >] pt >==> absWPlayer nimRule)
       -> (forall st, (player st)=false -> 0 < (nbm st) -> (invpf st) -> [< st >] pf >==> absWPlayer nimRule)
       -> forall st, 
             (gameInv invpt invpf (WRule nimRule) st)
              -> [< st >] (nimgamewf pt pf invpt invpf) >==> (nimgame pt pf).
Proof.
  unfold nimgamewf, nimgame, game, whileWF.
  intros; SSL_simplify_left; intuition.
  SSL_simplify_left.
  SSL_simplify_left.
  intuition.
  (* invariant preservation *)
  clear H3 st; unfold inv_preservation.
  st_split.
  intros b nm w1 w2 H3; SSL_simplify_left.
  zlt_case 0 nm.
  intro X; 
  apply bind_enter_left with (sp2:=(absWPlayer nimRule) -; 
                                     ensure_ (fun st' => gameInv invpt invpf (WRule nimRule) (dualSt st'))).
  SSL_simplify_left; generalize H3; unfold gameInv; clear H3; case b; simpl; intro H3;
  (SSL_trans; [ (apply H || apply H0); auto | ]) ;
  unfold invPlayer;
  SSL_simplify_left; intuition;
  (SSL_simplify_left; SSL_enter_left; [ (apply H1 || apply H2); auto | ]);
  SSL_simplify_both;
  intros D xx; clear D xx; st_split;
  simpl; intuition subst;
  SSL_Trivial.
  apply abspost2wp; simpl; intuition.
  intros; apply abspost2wp; simpl; intuition.
  (* variant decreasing *)
  clear H3 st; unfold variant_decreasing.
  st_split.
  intros b nm w1 w2 H3; SSL_simplify_left.
  SSL_simplify_left.
  intros; SSL_simplify_left.
  apply bind_enter_left with (sp2:=(absWPlayer nimRule)).
  SSL_simplify_left; generalize H3; unfold gameInv; case b; simpl; intuition auto with zarith.
  apply abspost2wp; simpl; unfold Rgame, WRule, nimRule, Zwf; simpl.
  intuition.
Qed.

Lemma expertWPlayer_as_player_false:
  forall st, (nbm st) >= 0 -> 
               (Zeq_bool ((nbm st) mod 4) 0) = (player st) -> 
    [< st >] (nimgame (absWPlayer nimRule) expertWPlayer) 
                 >==> (absPost (fun st' _ => is_winner false st')).
Proof.
  intros; SSL_trans.
  2: SSL_trans; [ apply expertPlayer_nimgame; eauto | ].
  unfold nimgame; apply game_monotonic; auto with zarith isl.  
  intros; apply absWPlayer_absPlayer; auto with zarith.
  intros; apply expertWPlayer_expertPlayer; auto with zarith.
  apply abspost2wp; simpl.
  intuition.
Qed.

Lemma expertWPlayer_as_player_true:
  forall st, (nbm st) >= 0 -> 
               (Zeq_bool ((nbm st) mod 4) 0) = negb (player st) -> 
    [< st >] (nimgame expertWPlayer (absWPlayer nimRule)) 
                 >==> (absPost (fun st' _ => is_winner true st')).
Proof.
  intros; SSL_trans.
  2: apply expertPlayer_as_player_true; eauto.
  unfold nimgame; apply game_monotonic; auto with zarith isl.  
  intros; apply expertWPlayer_expertPlayer; auto with zarith.
  intros; apply absWPlayer_absPlayer; auto with zarith.
Qed.
 

Theorem expertWPlayer_optimal_as_player_false:
  forall pf invpf,
     (forall st, (player st)=false -> 0 < (nbm st) -> [< st >] pf >==> invPlayer pf invpf (WRule nimRule))
     -> (forall st, (player st)=false -> 0 < (nbm st) -> (invpf st) -> [< st >] pf >==> absWPlayer nimRule)
     -> (forall st, (player st)=false -> 0 < (nbm st) -> (invpf st) -> feasible pf st)
     -> forall st, (nbm st) >= 0
      -> (if (player st) then forall st', (WRule nimRule st (dualSt st')) -> (player st')=false -> (invpf st')
                              else (invpf st))
       -> ([< st >] (nimgame (absWPlayer nimRule) pf) >==> absPost (fun st' _ => is_winner false st'))
        -> (player st)=(Zeq_bool ((nbm st) mod 4) 0).
Proof.
 intros pf invpf H1 H2 H3. st_split; intros b nm w1 w2; simpl; intros.
 assert (H5:forall (b':bool), b'=b \/  b' = (negb b)).
 intros b'; case b; case b'; auto.
 case (H5 (Zeq_bool (nm mod 4) 0)); auto.
 clear H5; intros H5.
 absurd ([< mkGameSt b (mkGameData nm w1 w2) >]
     nimgame expertWPlayer pf >==> 
     absPost (fun _ _ => False)).
 (* cas 1: "expertWPlayer vs pf" can not refine magic *)
 intros H6.
 cut (feasible (absPost (fun _ (_:unit) => False)) (mkGameSt b (mkGameData nm w1 w2))).
 (* cas 1.1: magic is not feasible *)
 intros H7; generalize (feasible_unfold H7).
 simpl.
 intros H8; case H8; auto; clear H8.
 intros a H8; decompose [ex Logic.and] H8; auto.
 (* cas 1.2: hypothesis implies magic is feasible *)
 unfold feasible.
 SSL_enter_right.
 apply H6.
 SSL_enter_right.
 apply nimgamewf_refines_nimgame.
 intros; apply invPlayer_trivial_trivok.
 apply H1.
 intros; apply expertWPlayer_isNimWPlayer; auto.
 auto.
 generalize H0; unfold gameInv; case b; simpl; intuition.
 apply feasible_def.
 apply nimgamewf_feasible; auto.
 clear H6 H5 H4 H0 H b nm w1 w2.
 st_split; intros b nm w1 w2 H; simpl; subst.
 intros; apply feasible_wp.
 simpl.
 intros X; clear X.
 constructor 1 with (x:=tt).
 unfold nimStrategy, WRule, nimRule; simpl.
 zeq_case (nm mod 4) 0.
 intros; constructor 1 with (x:=(mkGameSt true (mkGameData (nm-1) w1 w2))).
 simpl; intuition (try omega).
 intros; constructor 1 with (x:=(mkGameSt true (mkGameData (nm-(nm mod 4)) w1 w2))).
 intuition.
 (* cas 2: hypothesis implies "expertPlayer vs pf" refines magic *)
 SSL_trans.
 apply abspost_split with (q1:=fun (st:GameSt) (_:unit) => is_winner false st)
                                     (q2:=fun (st:GameSt) (_:unit) => is_winner true st).
 SSL_trans.
 2:apply H4.
 unfold nimgame; apply game_monotonic; auto with isl. 
 intros; apply expertWPlayer_isNimWPlayer; auto with zarith.
 SSL_trans.
 2:apply expertWPlayer_as_player_true; simpl; auto.
 unfold nimgame;
 eapply game_PlayerRefinement.
 intros; apply invPlayer_trivial_trivok.
 intros; apply H1; auto with zarith.
 auto with isl.
 intros; apply H2; auto with zarith.
 intros; fold (absWPlayer nimRule).
 apply expertWPlayer_isNimWPlayer; auto with zarith.
 intros; auto with isl.
 generalize H0; unfold gameInv; case b; simpl; intuition.
 apply abspost_monotony.
 unfold is_winner; st_split.
 intuition subst.
 discriminate H8.
Qed. 

End NimWGame.

(*********************************************************)
(*             Parametric Module "NimLazyPlayer"            *)
(*      an instance of NimWGame                                *)
(*      for "LazyPlayer" (a refinement of ExpertPlayer)  *)
(*********************************************************)

Module Type PlayerTrueDataM.

Parameter WDatat: Type.
(* WData1: wr-private data of player true  *)

End PlayerTrueDataM.

Module NimLazyPlayer(PTD: PlayerTrueDataM).

Export PTD.

Module NWD <: NimWDataM.

Definition WDatat := WDatat.

Definition WDataf : Set := option (bool*Z).
(* None: represents initialization case 
 * Some (b,v): has already played with last value "v". 
 *    if b then the player will win else the player may lose.
 *)

End NWD.

Module Export NWG := NimWGame(NWD).

Definition runLazyPlayer (nm:Z) (ist:WDataf) : bool*Z :=
  match ist with
  | None => let x := nm mod 4 in if (Zeq_bool x 0) then (false,nm-1) else (true,nm-x)
  | Some (true,old) => (true,old-4)
  | Some (false,old) => if (Zeq_bool (old-nm) 3) then (false,old-4) else (true,old-3)
  end.  

Definition lazyPlayer: K GameSt unit 
 := bind (get _) (fun st => 
        let r := (runLazyPlayer (nbm st) (wrf st)) in
        set (mkGameSt (player st) (mkGameData (snd r) (wrt st) (Some r)))).

Definition invLazyPlayer (st:GameSt): Prop :=
  match (wrf st) with
  | None => True
  | Some (b,old) => old - 3 <= (nbm st) < old
             /\ if b then old mod 4=0 else old mod 4 = 3
  end.

Ltac omegabs := abstract (omega).

Lemma lazyPlayer_preserves_invLazyPlayer:
  forall st, (player st)=false -> 0 < (nbm st) ->
    [< st >] lazyPlayer >==> (invPlayer lazyPlayer invLazyPlayer (WRule nimRule)).
Proof.
  unfold invPlayer, WRule, nimRule; st_split; intros b nm e1 e2 H H'; subst; SSL_simplify_all.
  intros H; case H; intros H1 H2; clear H H2; constructor 1; auto.
  st_split.
  intros b' nm' e'1 e'2 H0; case H0; clear H0; intros H H3.
  generalize H; clear H; subst.
  generalize H1; case e2; clear e2 H1; simpl; unfold invLazyPlayer, nimRule; simpl.
  (* 1. deja joue en position *)
  intros p; case p; clear p; intros b0 old; case b0; clear b0; simpl.
  (* 1.1 cas gagnant *)
  intuition.
  rewrite Z_mod_minus_mod; try omegabs.
  rewrite H3; auto.
  (* 1.2. cas perdant *)
  zeq_case (old-nm) 3;
  intuition;
  rewrite Z_mod_minus_mod; try omegabs;
  rewrite H4;  auto.
  (* 2. premier coup *)
  zeq_case (nm mod 4) 0; intuition;
  rewrite Z_mod_minus_mod; try omegabs.
  rewrite H; auto.
  rewrite Z_mod_mod_id; try omegabs.
  cutrewrite (nm mod 4 - nm mod 4 = 0); auto with zarith.
Qed.

Lemma lazyPlayer_refines_expertWPlayer_aux:
  forall st, (player st)=false -> 0 < (nbm st) -> (invLazyPlayer st) -> 
    [< st >] lazyPlayer >==> expertWPlayer.
Proof.
  st_split; intros b nm e1 e2 XX1 XX2; subst.
  intros H; unfold lazyPlayer, expertWPlayer, absWPlayer, absPlayer. 
  SSL_simplify_both.
  intros D; clear D.
  apply abspost2wp; simpl.
  unfold WRule; constructor 1; auto.
  generalize H; clear H; unfold invLazyPlayer, nimStrategy, nimRule; case e2; simpl.
  (* 1. deja joue en position *)
  intros p; case p; clear p; intros b0 old; case b0; clear b0;  simpl; intuition.
  (* 1.1. cas gagnant *)
  intuition. 
  zeq_case (nm mod 4) 0.
  intros X; case (Zmod_discr1 (a:=old) (b:=nm) (c:=4)); intuition. 
  intros X; cutrewrite (nm mod 4=(4 - (old - nm))).
  ring.
  rewrite (Z_mod_unicity (a:=4 - (old - nm)) (b:=4)); try omegabs.
  apply Z_mod_eq_intro; try omegabs.
  cutrewrite (nm - (4 - (old - nm)) = old - 4); try ring.
  rewrite Z_mod_minus_mod; try omegabs.
  rewrite H1; auto.
  (* 1.2. cas perdant *)
  zeq_case (old - nm) 3; intros.
  (* 1.2.1  on reste perdant *)
  assert (X: nm = old - 3).
  omegabs.
  subst; rewrite Z_mod_minus_mod; try omegabs.
  rewrite H1.
  unfold Zmod; simpl.
  omegabs.
  (* 1.2.2 simu de expertPlayer *)
  assert (X: old - 2 <= nm <= old - 1).
  omegabs.
  cutrewrite (nm mod 4=(3 - (old - nm))).
  assert (X0:(3 - (old - nm))<>0).
  omegabs.
  rewrite (Zdiff_eq_bool _ _ X0).
  ring.
  rewrite (Z_mod_unicity (a:=3 - (old - nm)) (b:=4)); try omegabs.
  apply Z_mod_eq_intro; try omegabs.
  cutrewrite (nm - (3 - (old - nm)) = old - 3); try omegabs.
  rewrite Z_mod_minus_mod; try omegabs.
  rewrite H1.
  auto.
   (* 2. premier coup *)
  intros; zeq_case (nm mod 4) 0; intuition.
Qed.

Hint Resolve Zlt_bool_lt: zarith.

Theorem lazyPlayer_refines_expertWPlayer: forall pt invpt, 
      (forall st, (player st)=true  -> 0<(nbm st) ->  [< st >] pt >==> invPlayer pt invpt (WRule nimRule))
   -> (forall st, (player st)=true  ->  0<(nbm st) -> (invpt st) -> [< st >] pt >==> (absWPlayer nimRule))
   ->  forall st, (wrf st)=None
      -> (if (player st) then (invpt st) 
                              else forall st', (WRule nimRule st (dualSt st')) -> (player st')=true -> (invpt st'))
      ->   [< st >] (nimgame pt lazyPlayer) >==> (nimgame pt expertWPlayer).
Proof.
  unfold nimgame, absWPlayer; intros pt invpt H1 H2; st_split.
  intros b nm w1 w2 H3 H4.
  eapply game_PlayerRefinement;  simpl; auto with zarith.
  intros; apply lazyPlayer_preserves_invLazyPlayer; auto with isl zarith.
  intros; auto with isl. 
  intros; apply lazyPlayer_refines_expertWPlayer_aux; auto with zarith.
  simpl; st_split; intros; subst. SSL_trans. apply expertWPlayer_isNimWPlayer; auto with zarith.
  unfold absWPlayer; auto with isl.
  generalize H4; clear H4; unfold invLazyPlayer, WRule; simpl.
  subst. case b; simpl; auto.
  intros H5; constructor 1; auto.
  st_split; intuition subst; simpl; auto.
  unfold gameInv; simpl; intuition.
Qed.

End NimLazyPlayer.


