(* Basic definitions of weakest-preconditions 
 * S. Boulme ---  coq 8.0 
 *)

(* begin hide *)
Set Implicit Arguments.
Require Export NonEmpty.
(* end hide *)

(** ** Module [WP]: Definitions and properties of weakest-preconditions. 

*** Definition of weakest-preconditions.

On functional expressions, weakest-preconditions are of type [(A -> Prop) -> Prop] 
where [A] is the type of the result returned by the
expression. Weakest preconditions are kind of continuations semantics:
given a post-condition on the result, they returned a precondition
ensuring that the post-condition hold on some result of the
expression. In presence, of non-determinism, there are two families of
preconditions: demonic that ensure the post-condition hold for all
possible results of the expression, and angelic that ensure there
exists a result such the post-condition holds. A pair of demonic and
angelic weakest-precondition satisfies the following [is_wp_pair]
predicate.
*)
Definition is_wp_pair (A : Type) (sdemon angel : (A -> Prop) -> Prop) :=
  forall R : A -> Prop,
  (sdemon R <-> sdemon (fun _ => True) /\ (forall a : A, angel (eq a) -> R a))
  /\ (angel R <-> (exists a : A, angel (eq a) /\ R a)).

(** A functional Dijkstra specification is simply such a pair. *)
Record DS (A : Type) : Type := 
  {
   sdemon : (A -> Prop) -> Prop;
   angel : (A -> Prop) -> Prop;
   DScorrect : is_wp_pair sdemon angel
  }.

(** In the following, I introduce syntactic macros.  If [s] is a
specification, [(sdemon s (fun _ => True))] is called the precondition
of [s] and is noted [(pre s)].  For a given [a], [(angel s (eq a))] is
called the postcondition that [s] returns [a], and is noted 
[(post s a)].  
*)
Notation "'pre' s" := (sdemon s (fun _ => True)) (at level 70, s at level 0).
Notation "'post' s a":= (angel s (eq a)) (at level 70, s at level 0, a at level 0).

(** Here are some direct reformulations of the definitions. These
lemmas shows that a specification is uniquely determined by a its [pre]/[post] couple.
Such a couple is traditionally called the ``normal form'' of the specification 
in refinement calculus.
*)

Lemma sdemon_fold :
 forall (A : Type) (s : DS A) (R : A -> Prop),
 pre s -> 
 (forall a : A, post s a -> R a) -> 
 sdemon s R.
Proof.
  simpl in |- *; intros A s R; case (DScorrect s R); intuition auto.
Qed.

Lemma sdemon_unfold :
 forall (A : Type) (s : DS A) (R : A -> Prop) (a : A),
 sdemon s R -> (post s a) -> R a.
Proof.
  simpl in |- *; intros A s R; case (DScorrect s R); intuition auto.
Qed.

Lemma angel_fold :
 forall (A : Type) (s : DS A) (R : A -> Prop) (a : A),
 post s a -> R a -> angel s R.
Proof.
  simpl in |- *; intros A s R a H1 H2; case (DScorrect s R);
   intuition auto.
  assert
   (EXT1 :
    exists a : _, (post s a) /\ R a);
   auto.
  constructor 1 with (x := a); intuition auto.
Qed.
 
Lemma angel_unfold :
 forall (A : Type) (s : DS A) (R : A -> Prop),
 angel s R -> 
 exists a : A, (post s a) /\ R a.
Proof.
  simpl in |- *; intros A s R; case (DScorrect s R); intuition auto.
Qed.

(* begin hide *)
Hint Resolve sdemon_fold angel_fold angel_unfold: wipi.
(* end hide *)


(** *** Basic properties of weakest-preconditions

The monotony of [sdemon] and [angel] with respect to logical
implication is directly derived from these properties.  *)
Lemma sdemon_weaken :
 forall (A : Type) (s : DS A) (R1 R2 : A -> Prop),
 sdemon s R1 -> (forall a : A, R1 a -> R2 a) -> sdemon s R2.
Proof.
  intros A s R1 R2 H; case (DScorrect s R1); case (DScorrect s R2);
   intuition eauto.
Qed.

Lemma angel_weaken :
 forall (A : Type) (s : DS A) (R1 R2 : A -> Prop),
 angel s R1 -> (forall a : A, R1 a -> R2 a) -> angel s R2.
Proof.
  intros A s R1 R2 H H0; case (angel_unfold _ _ H); intuition eauto with wipi.
Qed.

(** Sometimes, I need a finner weakenning property for [angel], 
using [pre] and [post].
*)
Lemma angel_weaken_pre :
 forall (A : Type) (s : DS A) (R1 R2 : A -> Prop),
 pre s
 -> angel s R1
  -> (forall a:A, (post s a) -> R1 a -> R2 a)
   -> angel s R2.
Proof.
  intros A s R1 R2 H H0 H1; case (angel_unfold _ _ H0);
   intuition eauto with wipi.
Qed.

(* begin hide *)
Hint Resolve sdemon_weaken angel_weaken angel_weaken_pre: wipi.

(* Others properties of wp *)
Lemma sdemon_eta :
 forall (A : Type) (s : DS A) (R : A -> Prop),
 sdemon s (fun a => R a) -> sdemon s R.
Proof.
  eauto with wipi.
Qed.

Lemma angel_eta :
 forall (A : Type) (s : DS A) (R : A -> Prop),
 angel s (fun a => R a) -> angel s R.
Proof.
  eauto with wipi.
Qed.
(* end hide *)

(** I directly derive from [sdemon_weaken] the following properties on [sdemon]. 
Of course, these properties are also be proved for [angel] using [angel_weaken].
*) 
Lemma sdemon_and_distr: forall (A:Type) (s:(DS A)) (R1 R2:A -> Prop),
   (sdemon s (fun a:A => (R1 a) /\ (R2 a))) -> (sdemon s R1) /\ (sdemon s R2).
Proof.
  intros A s R1 R2 H; constructor 1; eapply sdemon_weaken; eauto.
  simpl; tauto.
  simpl; tauto.
Qed. 

Lemma sdemon_or_lift: forall (A:Type) (s:(DS A)) (R1 R2:A -> Prop),
  (sdemon s R1) \/ (sdemon s R2) -> (sdemon s (fun a:A => (R1 a) \/ (R2 a))). 
Proof.
  intros A s R1 R2 H; case H; eauto with wipi.
Qed. 

Lemma sdemon_forall_distr:forall (A B:Type) (s:(DS A)) (R:B -> A -> Prop),
  (sdemon s (fun a:A => forall (b:B),(R b a))) -> forall (b:B), (sdemon s (R b)).
Proof.
  intros; eauto with wipi.
Qed. 

Lemma sdemon_exists_lift:forall (A B:Type) (s:(DS A)) (R:B -> A -> Prop) (b:B),
  (sdemon s (R b)) -> (sdemon s (fun a:A => (exists b:B, (R b a)))).
Proof.
  intros; eauto with wipi.
Qed.

(* begin hide *)
Lemma angel_and_distr: forall (A:Type) (s:(DS A)) (R1 R2:A -> Prop),
   (angel s (fun a:A => (R1 a) /\ (R2 a))) -> (angel s R1) /\ (angel s R2).
Proof.
  intros A s R1 R2 H; constructor 1; eapply angel_weaken; eauto.
  simpl; tauto.
  simpl; tauto.
Qed. 

Lemma angel_or_lift: forall (A:Type) (s:(DS A)) (R1 R2:A -> Prop),
  (angel s R1) \/ (angel s R2) -> (angel s (fun a:A => (R1 a) \/ (R2 a))). 
Proof.
  intros A s R1 R2 H; case H; eauto with wipi.
Qed. 

Lemma angel_forall_distr:forall (A B:Type) (s:(DS A)) (R:B -> A -> Prop),
  (angel s (fun a:A => forall (b:B),(R b a))) -> forall (b:B), (angel s (R b)).
Proof.
  intros; eauto with wipi.
Qed. 

Lemma angel_exists_lift:forall (A B:Type) (s:(DS A)) (R:B -> A -> Prop) (b:B),
  (angel s (R b)) -> (angel s (fun a:A => (exists b:B, (R b a)))).
Proof.
  intros; eauto with wipi.
Qed.
(* end hide *)

(** Using the definition of [sdemon] and [angel], it is possible to
prove the converse of the properties [sdemon_and_distr], [angel_or_lift]
and  [angel_exists_lift]. *)
Lemma sdemon_and_lift :
 forall (A : Type) (s : DS A) (R1 R2 : A -> Prop),
 sdemon s R1 -> sdemon s R2 -> sdemon s (fun a : A => R1 a /\ R2 a).
Proof.
  intros A s R1 R2 H1 H2; apply sdemon_fold; eauto with wipi.
  intros a H3; lapply (sdemon_unfold _ _ a H1); auto. lapply (sdemon_unfold _ _ a H2); auto.
Qed. 

Lemma angel_or_distr :
 forall (A : Type) (s : DS A) (R1 R2 : A -> Prop),
 angel s (fun a : A => R1 a \/ R2 a) -> angel s R1 \/ angel s R2.
Proof.
  intros A s R1 R2 H; case (angel_unfold _ _ H); intuition eauto with wipi.
Qed. 

Lemma angel_exists_distr :
 forall (A B : Type) (s : DS A) (R : B -> A -> Prop),
 angel s (fun a : A => exists b : B, R b a) -> exists b : B, angel s (R b).
Proof.
  intros A B s R H; case (angel_unfold _ _ H); intros a H1;
   decompose [and ex] H1; eauto with wipi.
Qed. 

(** The converse of [sdemon_forall_distr] do not hold. Indeed, assuming [forall (b:B), sdemon s (R b)], 
then [sdemon (fun a : A => forall b : B, R b a)] may not be true. Because [B] could be empty and [pre s] 
be [False]. By excluding this bad case, I can however prove two weaker versions of this property.
*)
Lemma sdemon_forall_lift1 :
 forall (A B : Type) (s : DS A) (R : B -> A -> Prop),
 pre s -> 
 (forall b : B, sdemon s (R b)) -> 
 sdemon s (fun a : A => forall b : B, R b a).
Proof.
  intros A B s R H H0; apply sdemon_fold; eauto with wipi.
  intros a H1 b; lapply (sdemon_unfold _ _ a (H0 b)); auto.
Qed. 

Lemma sdemon_forall_lift2 :
 forall (A B : Type) (s : DS A) (R : B -> A -> Prop),
 (NonEmpty B)
  -> (forall b : B, sdemon s (R b)) -> 
 sdemon s (fun a : A => forall b : B, R b a).
Proof.
  intros A B s R H1 H2; case H1; intros x. apply sdemon_fold.
  generalize (H2 x); eauto with wipi.
  intros a H b; lapply (sdemon_unfold _ _ a (H2 b)); auto.
Qed. 

(** *** Refinement in the functional specification language.
I now define the refinement relation between specifications. There are at least
three equivalent definitions. I choose one, and shows the equivalence with 
the two others.

Here is the first definition of the refinement. Let me note that in
intuitionnistic logic, both sides of the conjonction are necessary,
whereas in classical logic the left part is sufficient (in classical
logic, [angel] can be expressed in terms of [sdemon] using a double
negation).
*)
Definition refines (A : Type) (s1 s2 : DS A) :=
  forall R : A -> Prop,
  (sdemon s2 R -> sdemon s1 R) /\ (pre s2 -> angel s1 R -> angel s2 R).    

(** With this definition, it is almost trivial to prove that refinement is a pre-order. 
*)
Lemma refines_refl : forall (A : Type) (s : DS A), refines s s.
Proof.
  firstorder auto.
Qed.

Lemma refines_trans :
 forall (A : Type) (s1 s2 s3 : DS A),
 refines s1 s2 -> refines s2 s3 -> refines s1 s3.
Proof.
  firstorder auto.
Qed.

(* begin hide *)
Hint Resolve refines_refl refines_trans: wprefines. 
(* end hide *)

(** I now show that this definition is equivalent to the traditional one, expressed in terms of
[pre]/[post]. The proof only applies the relation between [sdemon]/[angel] and [pre]/[post].
*)
Lemma refines_prepost :
 forall (A : Type) (s1 s2 : DS A),
 refines s1 s2 <-> 
 pre s2
  -> pre s1 /\ (forall a : A, post s1 a -> post s2 a).
Proof.
  unfold refines in |- *; intros A s1 s2; constructor 1.
  firstorder eauto.
  intros H R; case (DScorrect s1 R); case (DScorrect s2 R);
   case (DScorrect s2 (fun _ : A => True)); intuition eauto 20.
  assert
   (EXT1 :
    exists a : _, post s1 a /\ R a);
   auto.
  decompose [and ex] EXT1; intuition eauto 20.
Qed.

(** I now prove the equivalence with a third definition. For more
convenience, the equivalence is split into two lemmas. This definition
is actually only a reformulation of the preceding ones, using [sdemon]
and [angel] instead of [pre]/[post]. With respect with the first
definition, this definition avoids a quantification over all
postconditions.  In particular, the lemma [refines_fold] can be
interpreted as a proof rule allowing to simplify refinement proofs
using wp-calculus.  
*)
Lemma refines_unfold :
 forall (A : Type) (s1 s2 : DS A),
 refines s1 s2 -> 
   pre s2 -> sdemon s1 (fun a => post s2 a).
Proof. 
  intros A s1 s2; case (refines_prepost s1 s2);
   case (DScorrect s1 (fun a => post s2 a));
   intuition auto.
Qed.

Lemma refines_fold :
 forall (A : Type) (s1 s2 : DS A),
 (pre s2 -> sdemon s1 (fun a => post s2 a)) -> 
 refines s1 s2.
Proof. 
  intros A s1 s2; case (refines_prepost s1 s2);
   case (DScorrect s1 (fun a => post s2 a));
   intuition eauto.
Qed.

(* begin hide *)
Hint Resolve refines_fold refines_unfold: wprefines.
(* end hide *)

(** This lemma is a little proof rule (in refinement proofs) that 
allows to introduce in the context the precondition of the specification being 
refined.
*)
Lemma refines_pre_intro :
 forall (A : Type) (s1 s2 : DS A),
 (pre s2 -> refines s1 s2) -> refines s1 s2.
Proof.
  auto with wprefines.
Qed.

(* begin hide *)
(* Deterministic (and valid) specifications *)

Record extractInfo (A:Type) (s:DS A): Type := 
  { extract: A ; 
    extract_proof: forall R : A -> Prop, angel s R <-> R extract }.

Definition impl (A : Type) (s : DS A) := pre s -> extractInfo s.

Lemma impl_angel_unfold :
 forall (A : Type) (s : DS A) (H : impl s) (H1 : pre s) (R : A -> Prop),
 angel s R -> R (extract (H H1)). 
Proof.
  intros A s H H1 R H0; case (extract_proof (H H1) R); tauto.
Qed.

Lemma impl_angel_fold :
 forall (A : Type) (s : DS A) (H : impl s) (H1 : pre s) (R : A -> Prop),
 R (extract (H H1)) -> angel s R. 
Proof.
  intros A s H H1 R H0; case (extract_proof (H H1) R); tauto.
Qed.

(*
Theorem impl_valid :
 forall (A : Type) (s : DS A) (H : impl s) (H1 : pre s),
  (post s (extract (H H1))).
Proof.
  simpl in |- *; intros; eapply impl_angel_fold; eauto.
Qed.

Lemma impl_proof_irrel :
 forall (A : Type) (s : DS A) (H H' : impl s) (H1 H1' : pre s),
 extract (H H1) = extract (H' H1').
Proof.
  intros; simpl in |- *;
   apply (sdemon_unfold s (fun a => extract (H H1) = a)).
  apply sdemon_fold; auto.
  intros; symmetry  in |- *; apply impl_angel_unfold; eauto with wipi.
  apply impl_valid.
Qed.
*)

Lemma refines_weakpre: 
 forall (A : Type) (s1 s2: DS A),
   (refines s1 s2) -> (pre s2) -> pre s1.
Proof.
  intros; case (refines_prepost s1 s2).
  intuition.
Qed.

Theorem impl_valid:
 forall (A : Type) (s1 s2: DS A) (H:impl s1),
   (refines s1 s2) -> 
      (pre s2) -> 
        forall (H1:pre s1), (post s2 (extract (H H1))).
Proof.
  intros A s1 s2 H H2 H3 H1; case (refines_prepost s1 s2).
  intuition.
  apply H6.
  eapply impl_angel_fold; eauto.
Qed.


Lemma impl_compat: forall (A : Type) (s1 s2: DS A),
  (refines s1 s2) -> (refines s2 s1) -> (impl s1) -> (impl s2).
Proof.
  intros A s1 s2 H1 H2 H3 H4; assert (H: pre s1).
  case (refines_prepost s1 s2); firstorder auto.
  case (H3 H); intros a H5; constructor 1 with (extract:=a).
  unfold refines in H1, H2. firstorder auto.
Qed.

Theorem extract_unicity:
    forall (A : Type) (s1 s2: DS A),
    (refines s1 s2) -> 
      forall (H1:impl s1) (H1p: pre s1) (H2:impl s2) (H2p: pre s2),
         (extract (H1 H1p))=(extract (H2 H2p)).
Proof.
   intros.
   apply (impl_angel_unfold H2 H2p _ (impl_valid H1 H H2p H1p)).
Qed.

Lemma impl_refinement:
  forall (A : Type) (s1 s2: DS A),
    (refines s1 s2) -> (impl s2) -> (impl s1) -> ((pre s1) -> (pre s2)) -> (refines s2 s1).
Proof.
  intros A s1 s2 H H2 H1 X; case (refines_prepost s2 s1).
  intros X0 X1.
  apply X1; clear X1 X0; intuition.
  eapply (impl_angel_fold H1 H0).
  rewrite (extract_unicity H H1 H0 H2 H3).
  apply (impl_angel_unfold H2 H3 _ H4).
Qed.

(* OLD VERSION: not enough general for wellfounded iteration or "and" *)
 Record DSsig (T : Type) (P : T -> Prop) : Type := DSexist
  {DSwit : T; DSsigP : P DSwit}.  

Definition strdet (A : Type) (s : DS A) :=
  pre s -> DSsig (fun a : A => forall R : A -> Prop, angel s R = R a).

Theorem strdet_valid :
 forall (A : Type) (s : DS A) (H : strdet s) (H1 : pre s),
   (post s (DSwit (H H1))).
Proof.
  simpl in |- *; intros; rewrite (DSsigP (H H1)); auto.
Qed.

Lemma strdet_proof_irrel :
 forall (A : Type) (s : DS A) (H H' : strdet s) (H1 H1' : pre s),
 DSwit (H H1) = DSwit (H' H1').
Proof.
  intros; simpl in |- *;
   apply (sdemon_unfold s (fun a => DSwit (H H1) = a)).
  apply sdemon_fold; auto.
  intros a H2; rewrite (DSsigP (H H1)) in H2; eauto with wipi.
  apply strdet_valid.
Qed.

Hint Resolve  (* sdemon_and_distr *) angel_or_distr
  sdemon_and_lift (* angel_or_lift sdemon_forall_distr *) 
  sdemon_forall_lift1 angel_exists_distr (* angel_exists_lift *)  sdemon_eta
  angel_eta: wipi.

Unset Implicit Arguments.
(* end hide *)
