Set Implicit Arguments.

Require Import ZArith.

Definition AbsArray : Type -> Type :=
  fun A => Z -> A.

Definition eqext
 := fun (A:Type) (t1 t2:(AbsArray A)) (a b:Z) =>
   forall i:Z, (a <= i <= b)%Z -> (t1 i)=(t2 i).
 
Lemma eqext_refl:
  forall (A:Type) (t:(AbsArray A)) (a b:Z), (eqext t t a b).
Proof.
  unfold eqext; intros; auto.   
Qed.

Lemma eqext_sym:
  forall (A:Type) (t1 t2:(AbsArray A)) (a b:Z), 
     (eqext t1 t2 a b) -> (eqext t2 t1 a b).
Proof.
  unfold eqext; intros; symmetry; auto.   
Qed.

Lemma eqext_trans:
  forall (A:Type) (t1 t2 t3:(AbsArray A)) (a b:Z), 
     (eqext t1 t2 a b) -> (eqext t2 t3 a b)
        -> (eqext t1 t3 a b).
Proof.
  unfold eqext; intros A t1 t2 t3 a b H1 H2; 
  intros; eapply trans_eq.
  eapply H1; eauto.
  eauto.
Qed.

(* A permutation is defined as a bijection on indexes.
 * This bijection is defined as a pair of a function and its inverse.
 *)
Definition permutation (A:Type) (t1 t2:(AbsArray A)) (a b:Z) : Prop 
  := exists p1_2: Z->Z, exists p2_1: Z->Z, forall i:Z, 
      (a <= i <= b)%Z -> 
       (  (a <= (p2_1 i) <= b)%Z 
          /\ (a <= (p1_2 i) <= b)%Z 
          /\ (p1_2 (p2_1 i))=i 
          /\ (p2_1 (p1_2 i))=i 
          /\ (t2 i)=(t1 (p1_2 i)) ).

(* a technical lemma to help in reasoning with permutation *)
Lemma permutation_ind:
 forall (A:Type) (t1 t2:(AbsArray A)) (a b:Z),
 (permutation t1 t2 a b) -> forall (P:Prop),
  (forall (p1_2 p2_1: Z -> Z),
          (forall i:Z, (a <= i <= b)%Z -> (a <= (p2_1 i) <= b)%Z)
       -> (forall i:Z, (a <= i <= b)%Z -> (a <= (p1_2 i) <= b)%Z)
       -> (forall i:Z, (a <= i <= b)%Z -> (p1_2 (p2_1 i))=i)
       -> (forall i:Z, (a <= i <= b)%Z -> (p2_1 (p1_2 i))=i) 
       -> (forall i:Z, (a <= i <= b)%Z -> (t2 i)=(t1 (p1_2 i)))
       -> (forall i:Z, (a <= i <= b)%Z -> (t1 i)=(t2 (p2_1 i)))
       -> P)
    ->  P.
Proof.
 intros A t1 t2 a b H2 P H1; unfold permutation in H2.
 case H2; clear H2; intros p1_2 H2; case H2; clear H2; intros p2_1 H2.
 apply H1 with (p1_2:=p1_2) (p2_1:=p2_1);
 try (intros i H3; case (H2 i); clear H2; intuition auto; fail).
 intros i H3; lapply (H2 i); intuition auto.
 assert (X: (t2 (p2_1 i) = t1 (p1_2 (p2_1 i)))).
 lapply (H2 (p2_1 i)); intuition auto.  
 rewrite X.
 rewrite H3.
 auto.
Qed.

Lemma permutation_eqext:
 forall (A:Type) (t1 t2:(AbsArray A)) (a b:Z),
    (eqext t1 t2 a b) -> (permutation t1 t2 a b).
Proof.
  unfold eqext; intros; unfold permutation.
  constructor 1 with (x:=fun (x:Z) => x).
  constructor 1 with (x:=fun (x:Z) => x).
  firstorder auto.
Qed.

Lemma permutation_sym:
  forall (A:Type) (t1 t2:(AbsArray A)) (a b:Z), 
     (permutation t1 t2 a b) -> (permutation t2 t1 a b).
Proof.
  intros A t1 t2 a b H; apply (permutation_ind H).
  intros p1_2 p2_1 H1 H2 H3 H4 H5 H6.
  unfold permutation.
  constructor 1 with (x:=p2_1).
  constructor 1 with (x:=p1_2).
  firstorder auto.
Qed.

Lemma permutation_trans:
  forall (A:Type) (t1 t2 t3:(AbsArray A)) (a b:Z), 
     (permutation t1 t2 a b) -> (permutation t2 t3 a b)
        -> (permutation t1 t3 a b).
Proof.
  intros A t1 t2 t3 a b H1 H2; apply (permutation_ind H2); apply (permutation_ind H1); clear H1 H2.
  intros p1_2 p2_1 H1 H2 H3 H4 H5 H6.
  intros p2_3 p3_2 H7 H8 H9 H10 H11 H12.
  unfold permutation.
  constructor 1 with (x:=fun x => p1_2 (p2_3 x)).
  constructor 1 with (x:=fun x => p3_2 (p2_1 x)).
  intros i; repeat (constructor 1).
  case (H7 (p2_1 i)); auto.
  case (H7 (p2_1 i)); auto.
  case (H2 (p2_3 i)); auto.
  case (H2 (p2_3 i)); auto.
  rewrite H9; auto.
  rewrite H4; auto.
  rewrite H11; auto.  
Qed.

Require Import ZArith_dec.

Inductive Exchange (A:Type) (t1 t2:(AbsArray A)) (a b:Z) : Prop :=
  Exchange_proof: 
  (t1 a)=(t2 b) -> 
    (t1 b)=(t2 a) -> 
       (forall (i:Z), ((i<>a) -> (i <> b) -> (t1 i)=(t2 i))) 
          -> (Exchange t1 t2 a b).

Lemma permutation_invol:
 forall (A:Type) (t1 t2:(AbsArray A)) (a b:Z),
    (exists p: Z->Z, forall i:Z, 
      (a <= i <= b)%Z -> 
       (  (a <= (p i) <= b)%Z 
          /\ (p (p i))=i 
          /\ (t2 i)=(t1 (p i)) ) 
     )
  -> (permutation t1 t2 a b).
Proof.
 intros A t1 t2 a b H; case H; clear H.
 intros p H.
 unfold permutation.
 constructor 1 with (x:=p);
 constructor 1 with (x:=p).
 firstorder auto.
Qed.

Lemma Exchange_Permutation:
  forall (A:Type) (t1 t2:(AbsArray A)) (a b c d:Z),
    (c <= a <= d)%Z -> (c <= b <= d)%Z
       -> (Exchange t1 t2 a b) -> (permutation t1 t2 c d).
Proof.
  intros A t1 t2 a b c d H1 H2 H3; case H3; clear H3; intros H3 H4 H5.
  apply permutation_invol.
  constructor 1 with (x:= fun i => 
     if (Z_eq_dec i a) 
     then b 
     else (if (Z_eq_dec i b) then a else i)).
  intros i H.
  case (Z_eq_dec i a).
  intros; subst; intuition auto.
  case (Z_eq_dec b a); auto.
  case (Z_eq_dec b b); auto.
  intros X; case X; auto.
  case (Z_eq_dec i b).
  intros; subst; intuition auto.
  case (Z_eq_dec a a); auto.
  intros X; case X; auto.
  firstorder auto.
  case (Z_eq_dec i a).
  intros; subst.
  absurd (a<>a); auto.
  case (Z_eq_dec i b).
  intros; subst.
  absurd (b<>b); auto.
  auto.
Qed.


Inductive Maximize (t:(AbsArray Z)) (m:Z) (a:Z): Z -> Prop :=
  Maxim_cons : forall (b:Z), 
     ((a <= b)%Z -> ((t b)<=m)%Z) -> 
     ((a < b)%Z -> (Maximize t m a (b-1)%Z)) -> (Maximize t m a b).

(* Require Omega. *)

(* Signification "extensionnelle" de ce prdicat: *)
Lemma Maximize_ext1 : forall (t:(AbsArray Z)) (m a b i:Z),
  (Maximize t m a b) -> (a <= i <= b)%Z -> ((t i) <= m)%Z.
Proof.
  intros t m a b i H1; elim H1; auto.
  intros b0 H2 H3 HR H4; case (Z_eq_dec b0 i).
   intros H; rewrite <- H; apply H2; abstract omega.
   intros; apply HR; abstract omega.
Qed.

Require Import Zwf.

Lemma Maximize_ext2 : forall (t:(AbsArray Z)) (m a b:Z),
  (forall i, (a <= i <= b)%Z -> ((t i) <= m)%Z) -> (Maximize t m a b). 
Proof.
  intros t m a b; pattern b; apply well_founded_ind with (R:=(Zwf a)).
  apply Zwf_well_founded.
  clear b; intros b HR H; constructor 1.
  intros; apply H; abstract omega.
  intros; apply HR.
  unfold Zwf; abstract omega.
  intros; apply H; abstract omega.
Qed.

Lemma Maximize_Zle: forall (t:(AbsArray Z)) (m1 m2 a b:Z),
  (Maximize t m1 a b) -> (m1 <= m2)%Z -> (Maximize t m2 a b).
Proof.
  intros t m1 m2 a b H0; elim H0.
  intros b0 H1 H2 H3 H4; constructor 1; auto.
  intros; abstract omega. 
Qed.

Inductive Sorted (t:(AbsArray Z)) (a:Z): Z -> Prop :=
  sorted_maximize: forall (b:Z), 
     (Maximize t (t b) a (b-1)%Z) -> 
     ((a < b)%Z ->  (Sorted t a (b-1)%Z)) -> (Sorted t a b).

(* Signification extensionnelle de ce predicat *)
Lemma Sorted_ext1 : forall (t:(AbsArray Z)) (a b i j:Z),
  (Sorted t a b) ->  (a <= i <= j)%Z -> (j <= b)%Z -> ((t i) <= (t j))%Z.
Proof.
  intros t a b i j H; elim H; clear H b; auto.
  intros b H1 H2 H3 H4 H5; case (Z_eq_dec b j).
  intros; subst.
  case (Z_eq_dec i j).
  intros; subst; abstract omega.
  intros; eapply Maximize_ext1; eauto.
  abstract omega.
  intros; apply H3; abstract omega.
Qed.

Lemma Sorted_ext2 : forall (t:(AbsArray Z)) (a b:Z),
  (forall (i:Z), (a <= i < b)%Z -> ((t i) <= (t (i+1))))%Z -> (Sorted t a b).
Proof.
  intros t a b; pattern b; apply well_founded_ind with (R:=(Zwf a)).
  apply Zwf_well_founded.
  clear b; intros b HR H; constructor 1.
  apply Maximize_ext2.
  intros i H1.
  assert ((t i) <= (t (b-1)))%Z.
  eapply Sorted_ext1.
  eapply HR with (y:=(b-1)%Z).
  unfold Zwf; abstract omega.
  intros; apply H; abstract omega.
  abstract omega.
  abstract omega.
  assert ((t (b-1)) <= (t b))%Z.
  generalize (H (b-1)%Z).
  ring_simplify (b-1+1)%Z.
  abstract omega.
  abstract omega.
  intros; apply HR.
  unfold Zwf; abstract omega.
  intros; eapply H; eauto; abstract omega.
Qed.


Lemma Sorted_union: forall (t:(AbsArray Z)) (a b c:Z),
  (Sorted t a b) -> (Sorted t b c) -> (Sorted t a c).
Proof.
   intros t a b c H1 H2; apply Sorted_ext2.
   intros i H3.
   case (Z_lt_dec i b).
   intros; apply Sorted_ext1 with (a:=a) (b:=b); auto.
   abstract omega.
   abstract omega.
   intros; apply Sorted_ext1 with (a:=b) (b:=c); auto; abstract omega.
Qed.
   
