(* ATTENTION,
   Les commentaires de ce fichier sont interpretes par "coqdoc"
   pour generer un fichier tex, inclus dans le rapport.
*)
(* begin hide *)
Set Implicit Arguments.

Require Import Relation_Definitions.
Require Import Iter.

Implicit Arguments preorder [A].
Implicit Arguments preord_trans [A R].
Implicit Arguments preord_refl [A R].

(* end hide *)
(** ** Tarski's theorem in constructive type theory 

In set theory, Tarski's theorem applies on complete lattice: preorders
such that each subset %\(E\)% of the lattice admits a greatest lower bound
and a least upper bound%\footnote{These bound may not be in
the subset \(E\)}%. 

Here, we must adapt this notion of complete lattice to type
theory. The notion of subset does not exist, but we can translate it
by considering a predicate %\(P\)% instead, where %\(P\)%
represents the subset of elements satisfying %\(P\)%.

Hence, if [A] is a type, [R] a relation on [A], and [glb] of type [(A->Prop) -> A], 
the following structure represents a predicate defining [glb] has a greatest lower bound 
operator with respect to [R].
*)
Record is_glb (A:Type) (R: A -> A -> Prop) (glb: (A->Prop) -> A) : Prop := {
  glb_lb: forall (P:A->Prop) (x:A), (P x) -> (R (glb P) x) ;
  glb_glb: forall (P:A->Prop) (x:A), 
     (forall (y:A), (P y) -> (R x y)) -> (R x (glb P))
}.

(* begin hide *)
Lemma is_glb_weaken: forall (A:Type) (R: A -> A -> Prop) (glb: (A->Prop) -> A),
  (is_glb R glb) ->  
    forall (P Q: A -> Prop), 
      (forall x:A, P x -> Q x) -> (R (glb Q) (glb P)).
Proof.
  intros A R glb H P Q H0.
  apply (glb_glb H).
  intros y H1; apply (glb_lb H); auto.
Qed. 
(* end hide *)

(** A lower semi-lattice is a preorder [R] with a greatest lower bound operator [glb].
By definition, [glb] is unique up to the equivalence associated to the preorder.
*)
Record inf_semilattice (A:Type) (R: A -> A -> Prop) (glb: (A->Prop) -> A) : Prop := {
  isemlat_preord:> (preorder R) ; 
  isemlat_glb:> (is_glb R glb)
}.

(** On lower semi-lattice, Tarski's theorem ensures the existence of smallest fixpoints.
  I define below Tarski's smallest fixpoint operator [sfix] and its main
  properties in a %\coq{}% section. 
*) 
Section Smallest_Fixpoints.

Variable A: Type.
Variable R: A -> A -> Prop.
Variable glb: (A -> Prop) -> A.

Hypothesis semi_lattice: (inf_semilattice R glb).

Section SFix_Def.

(** I denote [F] the function whose fixpoint is being computed.
    This function is assumed to be monotonic. 
*)
Variable F: A -> A.
Hypothesis F_monotony: forall (x y:A), (R x y) -> (R (F x) (F y)).

(** Below, I introduce the predicate [prefix] defining [x] as a prefixpoint. 
*)
Definition prefix (x:A) : Prop := (R (F x) x).

(** The smallest fixpoint of [F], denoted here [sfix], is the greatest
lower bound of the subset of prefixpoints.
*)

Definition sfix:= glb prefix.

(** By the [glb_lb] property of [glb], [sfix] is lower than any prefixpoint [x].
*)
Lemma sfix_smallest: forall (x:A), (R (F x) x) -> (R sfix x).
Proof.
  unfold sfix. intros; apply (glb_lb semi_lattice).
  auto.
Qed.
                                                                                
(** As a consequence, if [x] is a prefixpoint, then by [F_monotony], [(F sfix)] is lower than [F x].
By transitivity, [(F sfix)] is lower than [x]. As this last proposition holds for any prefixpoint [x], 
by the [glb_glb] property of [glb], [(F sfix)] is lower than [sfix]. Thus, [sfix] is a prefixpoint.
*)
Lemma sfix_prefix: (R (F sfix) sfix).
Proof.
  unfold sfix at 2.
  apply (glb_glb semi_lattice).
  intros y H; eapply (preord_trans semi_lattice).
  2:apply H.
  apply F_monotony.
  apply sfix_smallest.
  auto.
Qed.

(** Applying [F_monotony] to [sfix_prefix], we have that [(F sfix)] is a prefixpoint. 
Thus, by property [sfix_smallest], [sfix] is lower than [F sfix]. Thus, [sfix] is a postfixpoint.
*)
Lemma sfix_postfix: (R sfix (F sfix)).
Proof.
  apply sfix_smallest.
  apply F_monotony.
  apply sfix_prefix.
Qed.

(* begin hide *)
Lemma iterFsfix_R_sfix:
  forall (n:nat), (R (iter n F sfix) sfix).
Proof.
  induction n; simpl.
  eapply (preord_refl semi_lattice).
  eapply (preord_trans semi_lattice).
  2: apply sfix_prefix.
  auto.
Qed.

Lemma sfix_R_iterFsfix:
  forall (n:nat), (R sfix (iter n F sfix)).
Proof.
  induction n; simpl.
  eapply (preord_refl semi_lattice).
  eapply (preord_trans semi_lattice).
  apply sfix_postfix.
  auto.
Qed.

(* end hide *)
End SFix_Def.

(** The fixpoint operator [sfix] is itself monotonic.
*)
Section SFix_monotony.

Variable F1 F2: A -> A.
(*Hypothesis F1_monotony: forall (x y:A), (R x y) -> (R (F1 x) (F1 y)).*)
Hypothesis F2_monotony: forall (x y:A), (R x y) -> (R (F2 x) (F2 y)).

Lemma sfix_monotony: (forall (x:A), (R (F1 x) (F2 x))) -> (R (sfix F1) (sfix F2)).
Proof.
  intros; apply sfix_smallest.
  eapply (preord_trans semi_lattice).
  2: apply sfix_prefix; auto.
  auto. 
Qed.

End SFix_monotony.

End Smallest_Fixpoints.

(**
I give a last property allowing to relate two fixpoints, when their equations are themselves related.
Hence, given the following diagram:
%
\begin{center}
\begin{tabular}{c@{\hspace*{5em}}c}
& \\[0.5em]
\rnode{XA1}{{\tt A1}} & \rnode{XA2}{{\tt A2}} \\[3em]
\rnode{YA1}{{\tt A1}} & \rnode{YA2}{{\tt A2}} \\[0.5em]
\ncline{->}{XA1}{YA1}\Bput{{\tt F1}}
\ncline{->}{XA2}{YA2}\Aput{{\tt F2}}
\ncline{->}{XA1}{XA2}\Aput{{\tt G}}
\ncline{->}{YA1}{YA2}\Bput{{\tt G}}
\end{tabular}
\end{center}
%
 
%\noindent% If the right path %{\tt F2}\(\circ\){\tt G}%
is pointwise lower than the left path %{\tt G}\(\circ\){\tt F1}%, 
then [(sfix F2)] is lower than [(G (sfix F1))]. 
Section %\ref{subsec:recuspecex}% give examples showing that this property is useful in proofs.
An example also shows that when the left path is pointwise lower than the right path, then 
[(G (sfix F1))] may not be lower than [(sfix F2)].
*)

Section SFix_Morph.

Variable A1 A2: Type.
Variable R1: A1 -> A1 -> Prop.
Variable R2: A2 -> A2 -> Prop.
Variable glb1: (A1 -> Prop) -> A1.
Variable glb2: (A2 -> Prop) -> A2.

Hypothesis semi_lattice1: (inf_semilattice R1 glb1).
Hypothesis semi_lattice2: (inf_semilattice R2 glb2).

Variable F1: A1 -> A1.
Hypothesis F1_monotony: forall (x y:A1), (R1 x y) -> (R1 (F1 x) (F1 y)).

Variable F2: A2 -> A2.
(* Hypothesis F2_monotony: forall (x y:A2), (R2 x y) -> (R2 (F2 x) (F2 y)). *)

Variable G: A1 -> A2.
Hypothesis G_monotony: forall (x y:A1), (R1 x y) -> (R2 (G x) (G y)).

Lemma sfix2_R_G_sfix1:
 (forall (x: A1), (R2 (F2 (G x)) (G (F1 x))))
    -> (R2 (sfix R2 glb2 F2) (G (sfix R1 glb1 F1))).
Proof.
 intros F2G_R_GF1; apply sfix_smallest; auto.
  eapply (preord_trans semi_lattice2).
  apply F2G_R_GF1.
  apply G_monotony.
  apply sfix_prefix; auto.
Qed.

Hypothesis G_preserves_glb: forall (P: A2->Prop), 
 (R2 (G (glb1 (fun x:A1 => (P (G x))))) (glb2 P)).

Hypothesis G_injection: forall (x y:A1), (R2 (G x) (G y)) -> (R1 x y).

Lemma G_sfix1_refines_sfix2:
 (forall (x: A1), (R2 (G (F1 x)) (F2 (G x))))
    -> (R2 (G (sfix R1 glb1 F1)) (sfix R2 glb2 F2)).
Proof.
  intros; unfold sfix.
  eapply (preord_trans semi_lattice2).
  2: apply G_preserves_glb.
  apply G_monotony.
  apply (is_glb_weaken semi_lattice1).
  unfold prefix.
  intros; apply G_injection.
  eapply (preord_trans semi_lattice2).
  eapply H.
  auto.  
Qed.


End SFix_Morph.

(** I now introduce biggest fixpoints using a dual principle. Given a preorder [R], its dual is the preorder
[(fun x y => R y x)].
*)
Lemma dual_preorder:
  forall (A:Type) (R:A->A->Prop),
    (preorder (fun x y => R y x)) -> (preorder R).
Proof.
  intros A R H; case H; constructor 1; auto.
  unfold transitive.
  intros; eapply preord_trans; eauto.
Qed.

(** An upper semi-lattice is a preorder [R] with a least upper bound operator [lub], that is
a great lower bound operator for the dual of [R].
*)
Record sup_semilattice (A:Type) (R:A -> A -> Prop) (lub: (A->Prop) -> A) : Prop := {
  ssemlat_preord:> (preorder R) ; 
  ssemlat_lub:> (is_glb (fun x y => R y x) lub)
}.


(** Hence, an upper semi-lattice is a lower semi-lattice on the dual preorder.
*)
Lemma dual_ssemilat:
  forall (A:Type) (R:A->A->Prop) (lub: (A->Prop) -> A),
    (sup_semilattice R lub) -> (inf_semilattice (fun x y => R y x) lub).
Proof.
  intros A R lub H; case H; constructor 1; auto.
  apply dual_preorder; case ssemlat_preord0; intros; constructor 1; auto.
Qed.

(** Thus, given an upper semi-lattice for a preorder [R] and a least upper bound [lub], 
then the biggest fixpoint [bfix] is the smallest fixpoint for the dual: 
[sfix (fun x y => R y x) lub].
*)

(* begin hide *)
Hint Resolve dual_ssemilat: lattice.

Section Biggest_Fixpoints.

Variable A: Type.
Variable R: relation A.
Variable lub: (A -> Prop) -> A.

Hypothesis semi_lattice: (sup_semilattice R lub).


Section BFix_Def.

Variable F: A -> A.
Hypothesis F_monotony: forall (x y:A), (R x y) -> (R (F x) (F y)).

Definition bfix:= sfix (fun x y => R y x) lub F.

Lemma bfix_biggest:
  forall (x:A), (R x (F x)) -> (R x bfix).
Proof.
  unfold bfix. intros; apply sfix_smallest with (R:=(fun x y => R y x));
  auto with lattice.
Qed.
                                                                                
Lemma bfix_prefix: (R (F bfix) bfix).
Proof.
  unfold bfix. intros; apply sfix_postfix with (R:=(fun x y => R y x));
  auto with lattice.
Qed.

Lemma bfix_postfix: (R bfix (F bfix)).
Proof.
  unfold bfix. intros; apply sfix_prefix with (R:=(fun x y => R y x));
  auto with lattice.
Qed.

Lemma iterFbfix_R_bfix:
  forall (n:nat), (R (iter n F bfix) bfix).
Proof.
  unfold bfix. intros; apply sfix_R_iterFsfix with (R:=(fun x y => R y x));
  auto with lattice.
Qed.

Lemma bfix_R_iterFbfix:
  forall (n:nat), (R bfix (iter n F bfix)).
Proof.
  unfold bfix. intros; apply iterFsfix_R_sfix with (R:=(fun x y => R y x));
  auto with lattice.
Qed.

End BFix_Def.

Section BFix_monotony.

Variable F1 F2: A -> A.
Hypothesis F1_monotony: forall (x y:A), (R x y) -> (R (F1 x) (F1 y)).
(* Hypothesis F2_monotony: forall (x y:A), (R x y) -> (R (F2 x) (F2 y)). *)

Lemma bfix_monotony: 
  (forall (x:A), (R (F1 x) (F2 x))) -> (R (bfix F1) (bfix F2)).
Proof.
  unfold bfix. intros; apply sfix_monotony with (R:=(fun x y => R y x));
  auto with lattice.
Qed.

End BFix_monotony.

End Biggest_Fixpoints.

Section GFix_Morph.

Variable A1 A2: Type.
Variable R1: relation A1.
Variable R2: relation A2.
Variable lub1: (A1 -> Prop) -> A1.
Variable lub2: (A2 -> Prop) -> A2.

Hypothesis semi_lattice1: (sup_semilattice R1 lub1).
Hypothesis semi_lattice2: (sup_semilattice R2 lub2).

Variable F1: A1 -> A1.
Hypothesis F1_monotony: forall (x y:A1), (R1 x y) -> (R1 (F1 x) (F1 y)). 

Variable F2: A2 -> A2.
(* Hypothesis F2_monotony: forall (x y:A2), (R2 x y) -> (R2 (F2 x) (F2 y)). *)

Variable G: A1 -> A2.
Hypothesis G_monotony: forall (x y:A1), (R1 x y) -> (R2 (G x) (G y)).

Lemma G_bfix1_R_bfix2:
 (forall (x: A1), (R2 (G (F1 x)) (F2 (G x))))
    -> (R2 (G (bfix R1 lub1 F1)) (bfix R2 lub2 F2)).
Proof.
  unfold bfix. intros; 
  apply sfix2_R_G_sfix1 with (R1:=(fun x y => R1 y x))
                             (R2:=(fun x y => R2 y x));
  auto with lattice.
Qed.

End GFix_Morph.

Unset Implicit Arguments.
(* end hide *)
